@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -21,9 +21,9 @@
 */
 
 #include "bdConverterMain.h"
-#include "dbGDS2Writer.h"
+#include "bdWriterOptions.h"
 
 BD_PUBLIC int strm2gds (int argc, char *argv[])
 {
-  return bd::converter_main (argc, argv, db::GDS2WriterOptions ().format_name ());
+  return bd::converter_main (argc, argv, bd::GenericWriterOptions::gds2_format_name);
 }
