@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -70,7 +70,11 @@ struct basic_array_iterator
 
   virtual long index_a () const { return -1; }
   virtual long index_b () const { return -1; }
-  
+
+  virtual size_t quad_id () const { return 0; }
+  virtual box_type quad_box () const { return box_type::world (); }
+  virtual void skip_quad () { }
+
   virtual disp_type get () const = 0;
   
   virtual basic_array_iterator<Coord> *clone () const = 0;
@@ -111,8 +115,12 @@ struct ArrayBase
 
   virtual bool equal (const ArrayBase *) const = 0;
 
+  virtual bool fuzzy_equal (const ArrayBase *) const = 0;
+
   virtual bool less (const ArrayBase *) const = 0;
 
+  virtual bool fuzzy_less (const ArrayBase *) const = 0;
+
   virtual void mem_stat (MemStatistics *stat, MemStatistics::purpose_t purpose, int cat, bool no_self = false, void *parent = 0) const = 0;
 
   bool in_repository;
@@ -417,6 +425,19 @@ struct regular_array
         ab_coord (b.p2 ())
       };
 
+      //  NOTE: we need to take some care we don't overiterate in case of vanishing row
+      //  or column vectors. Hence eff_amax and eff_bmax which are 1 in this case.
+
+      unsigned long eff_amax = m_amax;
+      if (m_a.equal (vector_type ())) {
+        eff_amax = 1;
+      }
+
+      unsigned long eff_bmax = m_bmax;
+      if (m_b.equal (vector_type ())) {
+        eff_bmax = 1;
+      }
+
       double amin = ab [0].first;
       double amax = ab [0].first;
       double bmin = ab [0].second;
@@ -436,8 +457,8 @@ struct regular_array
         } else {
           amini = (unsigned long) (amin + 1.0 - epsilon);
         }
-        if (amini > m_amax) {
-          amini = m_amax;
+        if (amini > eff_amax) {
+          amini = eff_amax;
         }
       }
          
@@ -448,8 +469,8 @@ struct regular_array
         } else {
           amaxi = (unsigned long) (amax + epsilon) + 1;
         }
-        if (amaxi > m_amax) {
-          amaxi = m_amax;
+        if (amaxi > eff_amax) {
+          amaxi = eff_amax;
         }
       }
 
@@ -460,8 +481,8 @@ struct regular_array
         } else {
           bmini = (unsigned long) (bmin + 1.0 - epsilon);
         }
-        if (bmini > m_bmax) {
-          bmini = m_bmax;
+        if (bmini > eff_bmax) {
+          bmini = eff_bmax;
         }
       }
          
@@ -472,8 +493,8 @@ struct regular_array
         } else {
           bmaxi = (unsigned long) (bmax + epsilon) + 1;
         }
-        if (bmaxi > m_bmax) {
-          bmaxi = m_bmax;
+        if (bmaxi > eff_bmax) {
+          bmaxi = eff_bmax;
         }
       }
       
@@ -541,6 +562,12 @@ struct regular_array
     return (m_a == d->m_a && m_b == d->m_b && m_amax == d->m_amax && m_bmax == d->m_bmax);
   }
 
+  virtual bool fuzzy_equal (const ArrayBase *b) const
+  {
+    const regular_array<Coord> *d = static_cast<const regular_array<Coord> *> (b);
+    return (m_a.equal (d->m_a) && m_b.equal (d->m_b) && m_amax == d->m_amax && m_bmax == d->m_bmax);
+  }
+
   virtual bool less (const ArrayBase *b) const
   {
     const regular_array<Coord> *d = static_cast<const regular_array<Coord> *> (b);
@@ -549,6 +576,14 @@ struct regular_array
            m_amax < d->m_amax || (m_amax == d->m_amax && m_bmax < d->m_bmax)))));
   }
 
+  virtual bool fuzzy_less (const ArrayBase *b) const
+  {
+    const regular_array<Coord> *d = static_cast<const regular_array<Coord> *> (b);
+    return m_a.less (d->m_a) || (m_a.equal (d->m_a) && (
+           m_b.less (d->m_b) || (m_b.equal (d->m_b) && (
+           m_amax < d->m_amax || (m_amax == d->m_amax && m_bmax < d->m_bmax)))));
+  }
+
   virtual bool is_regular_array (vector_type &a, vector_type &b, unsigned long &amax, unsigned long &bmax) const
   {
     a = m_a;
@@ -590,16 +625,44 @@ protected:
   unsigned long m_amax, m_bmax;
   double m_det;
 
+  inline vector_type eff_a () const
+  {
+    if (m_a.equal (vector_type ())) {
+      if (m_b.equal (vector_type ())) {
+        return vector_type (1, 0);
+      } else {
+        return vector_type (m_b.y (), -m_b.x ());
+      }
+    } else {
+      return m_a;
+    }
+  }
+
+  inline vector_type eff_b () const
+  {
+    if (m_b.equal (vector_type ())) {
+      if (m_a.equal (vector_type ())) {
+        return vector_type (0, 1);
+      } else {
+        return vector_type (-m_a.y (), m_a.x ());
+      }
+    } else {
+      return m_b;
+    }
+  }
+
   std::pair <double, double> ab_coord (const point_type &p) const
   {
-    double a = (double (p.x ()) * double (m_b.y ()) - double (p.y ()) * double (m_b.x ())) / m_det;
-    double b = (double (m_a.x ()) * double (p.y ()) - double (m_a.y ()) * double (p.x ())) / m_det;
-    return std::make_pair (a, b);
+    vector_type a = eff_a (), b = eff_b ();
+    double ia = (double (p.x ()) * double (b.y ()) - double (p.y ()) * double (b.x ())) / m_det;
+    double ib = (double (a.x ()) * double (p.y ()) - double (a.y ()) * double (p.x ())) / m_det;
+    return std::make_pair (ia, ib);
   }
 
   void compute_det () 
   {
-    m_det = double (m_a.x ()) * double (m_b.y ()) - double (m_a.y ()) * double (m_b.x ());
+    vector_type a = eff_a (), b = eff_b ();
+    m_det = double (a.x ()) * double (b.y ()) - double (a.y ()) * double (b.x ());
   }
 };
 
@@ -661,6 +724,18 @@ struct regular_complex_array
     return regular_array<Coord>::equal (b);
   }
 
+  virtual bool fuzzy_equal (const ArrayBase *b) const
+  {
+    const regular_complex_array<Coord> *d = static_cast<const regular_complex_array<Coord> *> (b);
+    if (fabs (m_acos - d->m_acos) > epsilon) {
+      return false;
+    }
+    if (fabs (m_mag - d->m_mag) > epsilon) {
+      return false;
+    }
+    return regular_array<Coord>::fuzzy_equal (b);
+  }
+
   virtual bool less (const ArrayBase *b) const
   {
     const regular_complex_array<Coord> *d = static_cast<const regular_complex_array<Coord> *> (b);
@@ -673,6 +748,18 @@ struct regular_complex_array
     return regular_array<Coord>::less (b);
   }
 
+  virtual bool fuzzy_less (const ArrayBase *b) const
+  {
+    const regular_complex_array<Coord> *d = static_cast<const regular_complex_array<Coord> *> (b);
+    if (fabs (m_acos - d->m_acos) > epsilon) {
+      return m_acos < d->m_acos;
+    }
+    if (fabs (m_mag - d->m_mag) > epsilon) {
+      return m_mag < d->m_mag;
+    }
+    return regular_array<Coord>::fuzzy_less (b);
+  }
+
   virtual void mem_stat (MemStatistics *stat, MemStatistics::purpose_t purpose, int cat, bool no_self, void *parent) const
   {
     if (!no_self) {
@@ -765,6 +852,21 @@ struct iterated_array_iterator
     return new iterated_array_iterator <Coord> (*this);
   }
 
+  virtual size_t quad_id () const
+  {
+    return m_t.quad_id ();
+  }
+
+  virtual box_type quad_box () const
+  {
+    return m_t.quad_box ();
+  }
+
+  virtual void skip_quad ()
+  {
+    m_t.skip_quad ();
+  }
+
 private:
   box_tree_const_iterator m_b, m_e;
   box_tree_touching_iterator m_t;
@@ -862,9 +964,7 @@ struct iterated_array
   virtual std::pair <basic_array_iterator <Coord> *, bool>
   begin_touching (const box_type &b) const
   {
-    if (b.empty ()) {
-      return std::make_pair (new iterated_array_iterator <Coord> (m_v.begin (), m_v.end ()), false);
-    } else if (! b.touches (m_box)) {
+    if (b.empty () || ! b.touches (m_box)) {
       return std::make_pair (new iterated_array_iterator <Coord> (m_v.end (), m_v.end ()), false);
     } else {
       box_convert_type bc;
@@ -923,6 +1023,20 @@ struct iterated_array
     return true;
   }
 
+  virtual bool fuzzy_equal (const ArrayBase *b) const
+  {
+    const iterated_array<Coord> *d = static_cast<const iterated_array<Coord> *> (b);
+    if (m_v.size () != d->m_v.size ()) {
+      return false;
+    }
+    for (const_iterator p1 = m_v.begin (), p2 = d->m_v.begin (); p1 != m_v.end (); ++p1, ++p2) {
+      if (! p1->equal (*p2)) {
+        return false;
+      }
+    }
+    return true;
+  }
+
   virtual bool less (const ArrayBase *b) const
   {
     const iterated_array<Coord> *d = static_cast<const iterated_array<Coord> *> (b);
@@ -937,6 +1051,20 @@ struct iterated_array
     return false;
   }
 
+  virtual bool fuzzy_less (const ArrayBase *b) const
+  {
+    const iterated_array<Coord> *d = static_cast<const iterated_array<Coord> *> (b);
+    if (m_v.size () != d->m_v.size ()) {
+      return (m_v.size () < d->m_v.size ());
+    }
+    for (const_iterator p1 = m_v.begin (), p2 = d->m_v.begin (); p1 != m_v.end (); ++p1, ++p2) {
+      if (! p1->equal (*p2)) {
+        return (p1->less (*p2));
+      }
+    }
+    return false;
+  }
+
   virtual void mem_stat (MemStatistics *stat, MemStatistics::purpose_t purpose, int cat, bool no_self, void *parent) const
   {
     if (!no_self) {
@@ -1044,6 +1172,18 @@ struct iterated_complex_array
     return iterated_array<Coord>::equal (b);
   }
 
+  virtual bool fuzzy_equal (const ArrayBase *b) const
+  {
+    const iterated_complex_array<Coord> *d = static_cast<const iterated_complex_array<Coord> *> (b);
+    if (fabs (m_acos - d->m_acos) > epsilon) {
+      return false;
+    }
+    if (fabs (m_mag - d->m_mag) > epsilon) {
+      return false;
+    }
+    return iterated_array<Coord>::fuzzy_equal (b);
+  }
+
   virtual bool less (const ArrayBase *b) const
   {
     const iterated_complex_array<Coord> *d = static_cast<const iterated_complex_array<Coord> *> (b);
@@ -1056,6 +1196,18 @@ struct iterated_complex_array
     return iterated_array<Coord>::less (b);
   }
 
+  virtual bool fuzzy_less (const ArrayBase *b) const
+  {
+    const iterated_complex_array<Coord> *d = static_cast<const iterated_complex_array<Coord> *> (b);
+    if (fabs (m_acos - d->m_acos) > epsilon) {
+      return m_acos < d->m_acos;
+    }
+    if (fabs (m_mag - d->m_mag) > epsilon) {
+      return m_mag < d->m_mag;
+    }
+    return iterated_array<Coord>::fuzzy_less (b);
+  }
+
   virtual complex_trans_type complex_trans (const simple_trans_type &s) const
   {
     return complex_trans_type (s, m_acos, m_mag);
@@ -1146,7 +1298,12 @@ struct single_complex_inst
     return true;
   }
 
-  virtual bool less (const ArrayBase *b) const 
+  virtual bool fuzzy_equal (const ArrayBase *b) const
+  {
+    return equal (b);
+  }
+
+  virtual bool less (const ArrayBase *b) const
   {
     const double epsilon = 1e-10;
     const single_complex_inst<Coord> *d = static_cast<const single_complex_inst<Coord> *> (b);
@@ -1159,6 +1316,11 @@ struct single_complex_inst
     return false;
   }
 
+  virtual bool fuzzy_less (const ArrayBase *b) const
+  {
+    return less (b);
+  }
+
   virtual void mem_stat (MemStatistics *stat, MemStatistics::purpose_t purpose, int cat, bool no_self, void *parent) const
   {
     if (!no_self) {
@@ -1379,6 +1541,47 @@ struct array_iterator
     return mp_base ? mp_base->index_b () : -1;
   }
 
+  /**
+   *  @brief For iterators supporting quads (iterated arrays), this method will return the quad ID
+   */
+  size_t quad_id () const
+  {
+    return mp_base ? mp_base->quad_id () : 0;
+  }
+
+  /**
+   *  @brief For iterators supporting quads (iterated arrays), this method will return the quad bounding box
+   *
+   *  Note that this method will only return a valid quad box is the quad_id is non-null.
+   *
+   *  This method will return the bounding box of all array offsets in the quad.
+   */
+  db::box<Coord> quad_box () const
+  {
+    return mp_base ? mp_base->quad_box () : db::box<Coord>::world ();
+  }
+
+  /**
+   *  @brief For iterators supporting quads (iterated arrays), this method will skip the current quad
+   */
+  void skip_quad ()
+  {
+    if (mp_base) {
+      mp_base->skip_quad ();
+    }
+  }
+
+  /**
+   *  @brief Gets a value indicating whether the iterator is a synthetic one
+   *
+   *  "is_singular" is true, if the iterator was default-created or with a single
+   *  transformation.
+   */
+  bool is_singular () const
+  {
+    return mp_base == 0;
+  }
+
 private:
   trans_type m_trans;
   basic_array_iterator <Coord> *mp_base;
@@ -1505,7 +1708,7 @@ struct array
    *  an appropriate basic_array object using a complex transformation.
    */
   array (const Obj &obj, const complex_trans_type &ct, const vector_type &a, const vector_type &b, unsigned long amax, unsigned long bmax)
-    : m_obj (obj), m_trans (ct), mp_base (new regular_complex_array <coord_type> (ct.rcos (), ct.mag (), a, b, amax, bmax))
+    : m_obj (obj), m_trans (ct), mp_base (ct.is_complex () ? new regular_complex_array <coord_type> (ct.rcos (), ct.mag (), a, b, amax, bmax) : new regular_array <coord_type> (a, b, amax, bmax))
   {
     //  .. nothing yet ..
   }
@@ -1533,7 +1736,7 @@ struct array
    *  it's own storage.
    */
   array (const Obj &obj, const complex_trans_type &ct, ArrayRepository &rep, const vector_type &a, const vector_type &b, unsigned long amax, unsigned long bmax)
-    : m_obj (obj), m_trans (ct), mp_base (rep.insert (regular_complex_array <coord_type> (ct.rcos (), ct.mag (), a, b, amax, bmax)))
+    : m_obj (obj), m_trans (ct), mp_base (ct.is_complex () ? rep.insert (regular_complex_array <coord_type> (ct.rcos (), ct.mag (), a, b, amax, bmax)) : rep.insert (regular_array <coord_type> (a, b, amax, bmax)))
   {
     //  .. nothing yet ..
   }
@@ -1551,6 +1754,32 @@ struct array
     //  .. nothing yet ..
   }
 
+  /**
+   *  @brief The iterated array constructor
+   *
+   *  This is basically a convenience function that creates
+   *  an appropriate basic_array object.
+   */
+  template <class Iter>
+  array (const Obj &obj, const trans_type &trans, Iter from, Iter to)
+    : m_obj (obj), m_trans (trans), mp_base (new iterated_array <coord_type> (from, to))
+  {
+    //  .. nothing yet ..
+  }
+
+  /**
+   *  @brief The complex iterated array constructor
+   *
+   *  This is basically a convenience function that creates
+   *  an appropriate basic_array object.
+   */
+  template <class Iter>
+  array (const Obj &obj, const complex_trans_type &ct, Iter from, Iter to)
+    : m_obj (obj), m_trans (ct), mp_base (ct.is_complex () ? new iterated_complex_array <coord_type> (ct.rcos (), ct.mag (), from, to) : new iterated_array <coord_type> (from, to))
+  {
+    //  .. nothing yet ..
+  }
+
   /**
    *  @brief The singular complex instance constructor
    *
@@ -1559,7 +1788,7 @@ struct array
    */
   explicit 
   array (const Obj &obj, const complex_trans_type &ct)
-    : m_obj (obj), m_trans (ct), mp_base (new single_complex_inst <coord_type> (ct.rcos (), ct.mag ()))
+    : m_obj (obj), m_trans (ct), mp_base (ct.is_complex () ? new single_complex_inst <coord_type> (ct.rcos (), ct.mag ()) : 0)
   {
     //  .. nothing yet ..
   }
@@ -1589,7 +1818,7 @@ struct array
    */
   explicit 
   array (const Obj &obj, const complex_trans_type &ct, ArrayRepository &rep)
-    : m_obj (obj), m_trans (ct), mp_base (rep.insert (single_complex_inst <coord_type> (ct.rcos (), ct.mag ())))
+    : m_obj (obj), m_trans (ct), mp_base (ct.is_complex () ? rep.insert (single_complex_inst <coord_type> (ct.rcos (), ct.mag ())) : 0)
   {
     //  .. nothing yet ..
   }
@@ -1670,17 +1899,13 @@ struct array
   array_iterator <coord_type, Trans> begin_touching (const box_type &b, const BoxConv &bc) const 
   {
     if (b.empty ()) {
-      if (mp_base) {
-        return array_iterator <coord_type, Trans> (m_trans, mp_base->begin_touching (box_type ()));
-      } else {
-        return array_iterator <coord_type, Trans> (m_trans, true); 
-      }
+      return array_iterator <coord_type, Trans> (m_trans, true);
     } else if (b == box_type::world ()) {
       return begin ();
     } else if (mp_base) {
       box_type ob (bc (m_obj));
       if (ob.empty ()) {
-        return array_iterator <coord_type, Trans> (m_trans, mp_base->begin_touching (box_type ()));
+        return array_iterator <coord_type, Trans> (m_trans, true);
       } else {
         if (mp_base->is_complex ()) {
           complex_trans_type ct = mp_base->complex_trans (simple_trans_type (m_trans));
@@ -1762,7 +1987,7 @@ struct array
    *  The raw bounding boxes can be accumulated for all arrays having the same
    *  orientation, magnification and object. 
    *  bbox_from_raw_bbox can be used to compute the total bbox from such an
-   *  accumulated raw bounding box. This is for example exploitet in the 
+   *  accumulated raw bounding box. This is for example exploited in the 
    *  update_bbox method of db::cell.
    */
   box_type raw_bbox () const
@@ -1794,6 +2019,32 @@ struct array
     }
   }
 
+  /**
+   *  @brief Gets the bounding box from the iterator's current quad
+   *
+   *  The bounding box is that of all objects in the current quad and
+   *  is confined to the array's total bounding box.
+   */
+  template <class Iter, class BoxConv>
+  box_type quad_box (const Iter &iter, const BoxConv &bc) const
+  {
+    box_type bb;
+    if (mp_base) {
+      bb = mp_base->bbox (box_type (0, 0, 0, 0));
+    }
+    bb &= iter.quad_box ();
+
+    if (mp_base) {
+      if (mp_base->is_complex ()) {
+        return bb * box_type (mp_base->complex_trans (simple_trans_type (m_trans)) * bc (m_obj));
+      } else {
+        return bb * (m_trans * bc (m_obj));
+      }
+    } else {
+      return bb * (m_trans * bc (m_obj));
+    }
+  }
+
   /**
    *  @brief The number of single instances in the array
    */
@@ -1877,6 +2128,21 @@ struct array
     }
   }
 
+  /**
+   *  @brief Compare operator for equality (fuzzy version)
+   */
+  bool equal (const array<Obj, Trans> &d) const
+  {
+    if (! mp_base) {
+      return (m_trans.equal (d.m_trans) && m_obj == d.m_obj && ! d.mp_base);
+    } else {
+      if (! m_trans.equal (d.m_trans) || ! (m_obj == d.m_obj) || type () != d.type ()) {
+        return false;
+      }
+      return mp_base && mp_base->fuzzy_equal (d.mp_base);
+    }
+  }
+
   /**
    *  @brief A sorting order criterion
    */
@@ -1902,6 +2168,31 @@ struct array
     }
   }
 
+  /**
+   *  @brief A fuzzy sorting order criterion
+   */
+  bool less (const array<Obj, Trans> &d) const
+  {
+    if (! (m_obj == d.m_obj)) {
+      return (m_obj < d.m_obj);
+    }
+    if (! m_trans.equal (d.m_trans)) {
+      return m_trans.less (d.m_trans);
+    }
+    if (type () != d.type ()) {
+      return (type () < d.type ());
+    }
+    if (mp_base == d.mp_base) {
+      return false;
+    } else if (! mp_base) {
+      return true;
+    } else if (! d.mp_base) {
+      return false;
+    } else {
+      return mp_base->fuzzy_less (d.mp_base);
+    }
+  }
+
   /**
    *  @brief Compare operator
    *
@@ -2202,6 +2493,14 @@ struct array
     return a;
   }
 
+  /**
+   *  @brief A dummy implementation of the "scaled" function for API compatibility
+   */
+  array scaled (double f) const
+  {
+    return transformed (db::complex_trans<coord_type, coord_type> (f));
+  }
+
   /**
    *  @brief Transformation of an array into a new system
    *
@@ -2222,7 +2521,7 @@ struct array
     }
     db::mem_stat (stat, purpose, cat, m_obj, true, (void *) this);
     if (mp_base) {
-      db::mem_stat (stat, purpose, cat, *mp_base, false, (void *) this);
+      mp_base->mem_stat (stat, purpose, cat, false, (void *) this);
     }
   }
 
