@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -55,15 +55,18 @@ template <class Coord> class generic_repository;
 class Layout;
 class Library;
 class ImportLayerMapping;
+class CellMapping;
+class LayerMapping;
+class PCellDeclaration;
 
 /**
  *  @brief The cell object
  * 
  *  A cell object consists of a set of shape containers (called layers),
- *  a set of child cell instances and auxiliary informations such as
+ *  a set of child cell instances and auxiliary information such as
  *  the parent instance list.
  *  A cell is identified through an index given to the cell upon instantiation.
- *  The cell index is valid in the context of a cell graph object which
+ *  The cell index is valid in the context of a layout object which
  *  must issue the cell index.
  */
 
@@ -126,7 +129,7 @@ public:
   /**
    *  @brief Return the const shapes list of the given layer
    *
-   *  This method allows to access the shapes list on a certain layer.
+   *  This method allows one to access the shapes list on a certain layer.
    *  If the layer does not exist yet, a reference to an empty list is
    *  returned.
    *
@@ -139,7 +142,7 @@ public:
   /**
    *  @brief Return the shapes list of the given layer
    *
-   *  This method allows to access the shapes list on a certain layer.
+   *  This method allows one to access the shapes list on a certain layer.
    *  If the layer does not exist yet, it is created.
    *
    *  @param index The layer index of the shapes list to retrieve
@@ -180,6 +183,13 @@ public:
    */
   void copy (unsigned int src, unsigned int dest);
 
+  /**
+   *  @brief Copy the shapes from layer src to dest (only shapes from given classes)
+   *
+   *  The target layer is not overwritten. Instead, the shapes are added to the target layer's shapes.
+   */
+  void copy (unsigned int src, unsigned int dest, unsigned int types);
+
   /**
    *  @brief Move the shapes from layer src to dest
    *
@@ -187,6 +197,13 @@ public:
    */
   void move (unsigned int src, unsigned int dest);
 
+  /**
+   *  @brief Move the shapes from layer src to dest (only shapes from given classes)
+   *
+   *  The target layer is not overwritten. Instead, the shapes are added to the target layer's shapes.
+   */
+  void move (unsigned int src, unsigned int dest, unsigned int types);
+
   /**
    *  @brief Swap the layers given
    */
@@ -197,7 +214,12 @@ public:
    */
   void clear (unsigned int index);
 
-  /** 
+  /**
+   *  @brief Clear the shapes on the given layer (only the shapes from the given classes)
+   */
+  void clear (unsigned int index, unsigned int types);
+
+  /**
    *  @brief Erase a cell instance given by a instance proxy
    *
    *  Erasing a cell instance will destroy the sorting order and invalidate
@@ -301,6 +323,30 @@ public:
     return m_instances.transform_into (ref, t);
   }
 
+  /**
+   *  @brief Transforms the cell by the given transformation.
+   *
+   *  The transformation is applied to all instances and shapes. Magnified transformations will
+   *  render magnified instances. See \transform_into for a version which avoids this.
+   *
+   *  @param t The transformation to apply
+   */
+  template <class Trans>
+  void transform (const Trans &t)
+  {
+    m_instances.transform (t);
+    for (typename shapes_map::iterator s = m_shapes_map.begin (); s != m_shapes_map.end (); ++s) {
+      if (! s->second.empty ()) {
+        //  Note: don't use the copy ctor here - it will copy the attachment to the manager
+        //  and create problems when destroyed. Plus: swap would be more efficient. But by using
+        //  assign_transformed we get undo support for free.
+        shapes_type d;
+        d = s->second;
+        s->second.assign_transformed (d, t);
+      }
+    }
+  }
+
   /**
    *  @brief Transforms the cell into a new coordinate system.
    *
@@ -327,18 +373,17 @@ public:
   }
 
   /**
-   *  @brief Insert an instance given by a instance reference with a different cell index and property ID
+   *  @brief Insert an instance given by a instance reference with a different cell index
    *
    *  This member may be used to map an instance to another layout object. 
    *
    *  @param ref The instance reference of which to insert a copy
    *  @param im The mapper to new cell index to use (for mapping to a different layout for example)
-   *  @param pm The mapper to new cell property ID use (for mapping to a different layout for example)
    */
-  template <class IndexMap, class PropIdMap>
-  instance_type insert (const instance_type &ref, IndexMap &im, PropIdMap &pm)
+  template <class IndexMap>
+  instance_type insert (const instance_type &ref, IndexMap &im)
   {
-    return (m_instances.insert (ref, im, pm));
+    return (m_instances.insert (ref, im));
   }
 
   /**
@@ -429,6 +474,27 @@ public:
    */
   instance_type change_pcell_parameters (const instance_type &ref, const std::vector<tl::Variant> &new_parameters);
 
+  /**
+   *  @brief Changes the PCell parameters of a PCell instance using a dict
+   *
+   *  @return A reference to the new instance. The original reference may be invalid.
+   */
+  instance_type change_pcell_parameters (const instance_type &ref, const std::map<std::string, tl::Variant> &new_parameters);
+
+  /**
+   *  @brief Gets the PCellDeclaration object of the instance if the instance is a PCell instance
+   *
+   *  If the instance is not a PCell instance, 0 is returned.
+   */
+  const db::PCellDeclaration *pcell_declaration_of_inst (const db::Cell::instance_type &ref) const;
+
+  /**
+   *  @brief Gets the PCellDeclaration object of the cell is the cell is a PCell variant
+   *
+   *  If the cell is not a PCell variant, 0 is returned.
+   */
+  const db::PCellDeclaration *pcell_declaration () const;
+
   /**
    *  @brief The cell index accessor method
    *
@@ -456,22 +522,8 @@ public:
    */
   bool is_shape_bbox_dirty () const;
 
-  /** 
-   *  @brief Update the bbox 
-   *
-   *  This will update the bbox from the shapes and instances.
-   *  This requires the bboxes of the child cells to be computed
-   *  before. Practically this will be done by computing the
-   *  bboxes bottom-up in the hierarchy.
-   *  In addition, the number of hierarchy leves below is also
-   *  updated.
-   *
-   *  @param layers The max. number of layers in the child cells
-   *  @return true, if the bounding box has changed.
-   */
-  bool update_bbox (unsigned int layers);
   /**
-   *  @brief Sort the shapes lists 
+   *  @brief Sorts the shapes lists
    *
    *  This will sort the shapes lists for query of regions 
    *  on a per-shape basis. Since sorting of the shapes is
@@ -485,18 +537,29 @@ public:
    *
    *  Before the bounding box can be retrieved, it must have
    *  been computed using update_bbox. This is performed by 
-   *  requesting an update from the graph.
+   *  requesting an update from the layout.
    *
    *  @return The bounding box that was computed by update_bbox
    */
   const box_type &bbox () const;
 
+  /**
+   *  @brief Retrieve the bounding box of the cell, including empty cells
+   *
+   *  This method behaves like "bbox", but includes empty cells as single-point
+   *  boxes (0,0;0,0). This bounding box is used for drawing and allows
+   *  including empty cells.
+   *
+   *  @return The bounding box that was computed by update_bbox
+   */
+  const box_type &bbox_with_empty () const;
+
   /**
    *  @brief Retrieve the per-layer bounding box of the cell
    *
    *  Before the bounding box can be retrieved, it must have
    *  been computed using update_bbox. This is performed by 
-   *  requesting an update from the graph.
+   *  requesting an update from the layout.
    *
    *  @return The bounding box that was computed by update_bbox
    */
@@ -601,7 +664,7 @@ public:
   /**
    *  @brief The parent instance list begin iterator
    *
-   *  The begin_parent_insts() allows to access to the parent instance list.
+   *  The begin_parent_insts() allows one to access to the parent instance list.
    */
   parent_inst_iterator begin_parent_insts () const;
 
@@ -695,6 +758,11 @@ public:
     return shapes (layer).begin_touching (box, flags, prop_sel, inv_prop_sel);
   }
 
+  /**
+   *  @brief A quick, recursive test whether the cell has shapes touching the given box on the given layer
+   */
+  bool has_shapes_touching (unsigned int layer, const db::Box &box) const;
+
   /**
    *  @brief Collect all calling cells (either calling this cell directly or indirectly)
    *
@@ -725,7 +793,7 @@ public:
    *  call this cell to the set given. It is assumed that the if the set contains a cell, it
    *  will also contain all called cells, so it may act as a cache.
    *
-   *  This version allows to restrict the search to a cone of the hierarchy tree, that is 
+   *  This version allows one to restrict the search to a cone of the hierarchy tree, that is
    *  a set of cells which are collected from another call of "collect_called_cells" with another initial cell.
    *  
    *  @param callers The set of called cells (used as cache) 
@@ -781,6 +849,13 @@ public:
    */
   virtual void update (ImportLayerMapping * /*layer_mapping*/ = 0) { }
 
+  /**
+   *  @brief Checks if the cell is locked
+   *
+   *  This method throws an exception if the cell is locked.
+   */
+  void check_locked () const;
+
   /**
    *  @brief Tell, if this cell is a proxy cell
    *
@@ -800,7 +875,7 @@ public:
   void set_name (const std::string &name);
 
   /**
-   *  @brief Get the basic name
+   *  @brief Gets the basic name
    *
    *  The basic name of the cell is either the cell name or the cell name in the
    *  target library (for library proxies) or the PCell name (for PCell proxies).
@@ -808,6 +883,14 @@ public:
    */
   virtual std::string get_basic_name () const;
 
+  /**
+   *  @brief Gets the variant name
+   *
+   *  The variant name is the PCell's "cell_name" - which may encode the PCell parameters
+   *  into a formal name. Usually that is identical to the PCell name.
+   */
+  virtual std::string get_variant_name () const;
+
   /**
    *  @brief Gets the display name
    *
@@ -852,6 +935,27 @@ public:
     m_ghost_cell = g;
   }
 
+  /**
+   *  @brief Gets a value indicating whether the cell is locked
+   *
+   *  A locked cell cannot be modified in terms of instances or shapes.
+   *  The name of a locked cell can be changed though.
+   */
+  bool is_locked () const
+  {
+    return m_locked;
+  }
+
+  /**
+   *  @brief Sets the locked state of the cell
+   *
+   *  See \is_locked for details about locked state.
+   */
+  void set_locked (bool f)
+  {
+    m_locked = f;
+  }
+
   /**
    *  @brief Returns a value indicating whether the cell is empty
    *
@@ -913,14 +1017,109 @@ public:
     return mp_layout;
   }
 
+  /**
+   *  @brief Copies the shapes from the source cell's tree to this cell
+   *
+   *  This variant uses the given cell mapping and layer mapping.
+   */
+  void copy_tree_shapes (const db::Cell &source_cell, const db::CellMapping &cm);
+
+  /**
+   *  @brief Copies the shapes from the source cell's tree to this cell
+   *
+   *  This variant uses the given cell mapping and layer mapping.
+   */
+  void copy_tree_shapes (const db::Cell &source_cell, const db::CellMapping &cm, const db::LayerMapping &lm);
+
+  /**
+   *  @brief Copies instances and shapes from the source cell to this cell
+   *
+   *  If the source and target layout are different ones, the whole cell tree of the source cell
+   *  will be copied.
+   *  This will create new cells in the target layout to accommodate the source cell tree.
+   *  Returns an array with the freshly created cells which accommodate the cell tree.
+   */
+  std::vector<db::cell_index_type> copy_tree (const db::Cell &source_cell);
+
+  /**
+   *  @brief Copies the instances from the source to this cell.
+   */
+  void copy_instances (const db::Cell &source_cell);
+
+  /**
+   *  @brief Copies all shapes from the source cell to this cell
+   */
+  void copy_shapes (const db::Cell &source_cell);
+
+  /**
+   *  @brief Copies all shapes from the source cell to this cell using the given layer mapping
+   */
+  void copy_shapes (const db::Cell &source_cell, const db::LayerMapping &layer_mapping);
+
+  /**
+   *  @brief Moves the shapes from the source cell's tree to this cell
+   *
+   *  This variant uses the given cell mapping and layer mapping.
+   */
+  void move_tree_shapes (db::Cell &source_cell, const db::CellMapping &cm);
+
+  /**
+   *  @brief Moves the shapes from the source cell's tree to this cell
+   *
+   *  This variant uses the given cell mapping and layer mapping.
+   */
+  void move_tree_shapes (db::Cell &source_cell, const db::CellMapping &cm, const db::LayerMapping &lm);
+
+  /**
+   *  @brief Moves instances and shapes from the source cell to this cell
+   *
+   *  If the source and target layout are different ones, the whole cell tree of the source cell
+   *  will be copied.
+   *  This will create new cells in the target layout to accommodate the source cell tree.
+   *  Returns an array with the freshly created cells which accommodate the cell tree.
+   */
+  std::vector<db::cell_index_type> move_tree (db::Cell &source_cell);
+
+  /**
+   *  @brief Moves the instances from the source to this cell.
+   */
+  void move_instances (db::Cell &source_cell);
+
+  /**
+   *  @brief Moves all shapes from the source cell to this cell
+   */
+  void move_shapes (db::Cell &source_cell);
+
+  /**
+   *  @brief Moves all shapes from the source cell to this cell using the given layer mapping
+   */
+  void move_shapes (db::Cell &source_cell, const db::LayerMapping &layer_mapping);
+
+  /**
+   *  @brief Gets the current bounding box (with empty) without calling Layout::update
+   *
+   *  This method is intended for internal purposes only.
+   */
+  const box_type &bbox_with_empty_no_update () const
+  {
+    return m_bbox_with_empty;
+  }
+
+  /**
+   *  @brief Gets the current per-layer bounding box without calling Layout::update
+   *
+   *  This method is intended for internal purposes only.
+   */
+  const box_type &bbox_no_update (unsigned int l) const;
+
 protected:
   /**
    *  @brief Standard constructor: create an empty cell object
    * 
-   *  Takes the manager object from the graph object.
+   *  Takes the manager object from the layout object.
    *
    *  @param ci The index of the cell
-   *  @param g A reference to the graph object that owns the cell
+   *  @param g A reference to the layout object that owns the cell
    */
   Cell (cell_index_type ci, db::Layout &g);
 
@@ -938,17 +1137,19 @@ protected:
   virtual Cell *clone (db::Layout &layout) const;
 
 private:
+  friend class db::Layout;
   cell_index_type m_cell_index;
   mutable db::Layout *mp_layout;
   shapes_map m_shapes_map;
   instances_type m_instances;
-  box_type m_bbox;
+  box_type m_bbox, m_bbox_with_empty;
   box_map m_bboxes;
   db::properties_id_type m_prop_id;
 
   // packed fields
   unsigned int m_hier_levels : 29;
   bool m_bbox_needs_update : 1;
+  bool m_locked : 1;
   bool m_ghost_cell : 1;
 
   static box_type ms_empty_box;
@@ -956,7 +1157,7 @@ private:
   //  linked list, used by Layout
   Cell *mp_last, *mp_next;
 
-  //  clear the shapes without telling the graph
+  //  clear the shapes without telling the layout
   void clear_shapes_no_invalidate ();
 
   //  helper function for computing the number of hierarchy levels
@@ -972,9 +1173,9 @@ private:
   }
 
   /**
-   *  @brief Return a reference to the instances object 
+   *  @brief Return a reference to the instances object
    */
-  instances_type &instances ()
+  const instances_type &instances () const
   {
     return m_instances;
   }
@@ -982,7 +1183,7 @@ private:
   /**
    *  @brief Return a reference to the instances object 
    */
-  const instances_type &instances () const
+  instances_type &instances ()
   {
     return m_instances;
   }
@@ -1007,7 +1208,7 @@ private:
    *  @brief Sort the child instance list
    * 
    *  The child instances are first sorted by cell index, then by raw transformation
-   *  (excluding displacement). This allows to simplify the bbox computation by 
+   *  (excluding displacement). This allows one to simplify the bbox computation by
    *  convolution of the displacements bboxes with the object bboxes.
    */
   void sort_child_insts ();
@@ -1018,9 +1219,24 @@ private:
    *  This will sort the cell instance list. As a prerequesite
    *  the cell's bounding boxes must have been computed.
    *
-   *  @param layers The maximum number of layers in the child cells
+   *  @param force Force sorting, even if not strictly needed
    */
-  void sort_inst_tree ();
+  void sort_inst_tree (bool force);
+
+  /**
+   *  @brief Updates the bbox
+   *
+   *  This will update the bbox from the shapes and instances.
+   *  This requires the bboxes of the child cells to be computed
+   *  before. Practically this will be done by computing the
+   *  bboxes bottom-up in the hierarchy.
+   *  In addition, the number of hierarchy levels below is also
+   *  updated.
+   *
+   *  @param layers The max. number of layers in the child cells
+   *  @return true, if the bounding box has changed.
+   */
+  bool update_bbox (unsigned int layers);
 };
 
 /**
@@ -1034,16 +1250,5 @@ mem_stat (MemStatistics *stat, MemStatistics::purpose_t purpose, int cat, const
 
 } // namespace db
 
-namespace tl
-{
-  /**
-   *  @brief Type traits 
-   */
-  template <> struct type_traits <db::Cell> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-}
-
 #endif
 
