@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -147,7 +147,7 @@ InstOp<Inst, ET>::erase (Instances *insts)
     std::vector<bool> done;
     done.resize (m_insts.size (), false);
 
-    //  This is not quite effective but seems to be the simpliest way
+    //  This is not quite effective but seems to be the simplest way
     //  of implementing this: search for each element and erase these.
     //  The alternative would be to store the iterator along with the object.
     std::vector<typename tree_type::const_iterator> to_erase;
@@ -241,20 +241,32 @@ Instance::to_string (bool resolve_cell_name) const
     r = "cell_index=" + tl::to_string (ci.object ().cell_index ());
   }
 
-  if (ci.is_complex ()) {
-    r += " " + ci.complex_trans ().to_string ();
-  } else {
-    r += " " + (*ci.begin ()).to_string ();
-  }
-
   db::vector<coord_type> a, b;
   unsigned long amax, bmax;
   if (ci.is_regular_array (a, b, amax, bmax)) {
+
+    if (ci.is_complex ()) {
+      r += " " + ci.complex_trans ().to_string ();
+    } else {
+      r += " " + (*ci.begin ()).to_string ();
+    }
+
     r += " array=(" + a.to_string () + "," + b.to_string () + " " + tl::to_string (amax) + "x" + tl::to_string (bmax) + ")";
+
+  } else {
+
+    for (db::CellInstArray::iterator i = ci.begin (); ! i.at_end (); ++i) {
+      if (ci.is_complex ()) {
+        r += " " + ci.complex_trans (*i).to_string ();
+      } else {
+        r += " " + (*i).to_string ();
+      }
+    }
+
   }
 
   if (has_prop_id ()) {
-    r += " prop_id=" + tl::to_string (prop_id ());
+    r += std::string (" props=") + db::properties (prop_id ()).to_dict_var ().to_string ();
   }
 
   return r;
@@ -268,12 +280,18 @@ void
 instance_iterator<Traits>::release_iter ()
 { 
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()).~stable_iter_wp_type ();
       } else {
         basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()).~stable_iter_type ();
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()).~stable_unsorted_iter_wp_type ();
+      } else {
+        basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()).~stable_unsorted_iter_type ();
+      }
     } else {
       if (m_with_props) {
         basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()).~iter_wp_type ();
@@ -289,12 +307,18 @@ void
 instance_iterator<Traits>::make_iter ()
 { 
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         new (&basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ())) stable_iter_wp_type ();
       } else {
         new (&basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ())) stable_iter_type ();
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        new (&basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ())) stable_unsorted_iter_wp_type ();
+      } else {
+        new (&basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ())) stable_unsorted_iter_type ();
+      }
     } else {
       if (m_with_props) {
         new (&basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ())) iter_wp_type ();
@@ -310,18 +334,24 @@ template <class Traits>
 bool
 instance_iterator<Traits>::operator== (const instance_iterator<Traits> &d) const
 {
-  if (! (m_type == d.m_type && m_stable == d.m_stable && m_with_props == d.m_with_props)) {
+  if (! (m_type == d.m_type && m_stable == d.m_stable && m_with_props == d.m_with_props && m_unsorted == d.m_unsorted)) {
     return false;
   }
   if (m_type == TNull) {
     return true;
   } else {
     if (m_stable) {
-      if (m_with_props) {
+      if (m_with_props && ! m_unsorted) {
         return (basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()) == d.basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
       } else {
         return (basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()) == d.basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        return (basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()) == d.basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+      } else {
+        return (basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()) == d.basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+      }
     } else {
       if (m_with_props) {
         return (basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()) == d.basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -343,16 +373,23 @@ instance_iterator<Traits>::operator= (const instance_iterator<Traits> &iter)
     m_type = iter.m_type;
     m_stable = iter.m_stable;
     m_with_props = iter.m_with_props;
+    m_unsorted = iter.m_unsorted;
     m_traits = iter.m_traits;
 
     if (m_type == TInstance) {
 
-      if (m_stable) {
+      if (m_stable && ! m_unsorted) {
         if (m_with_props) {
           new (&basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ())) stable_iter_wp_type (iter.basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
         } else {
           new (&basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ())) stable_iter_type (iter.basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
         }
+      } else if (m_stable) {
+        if (m_with_props) {
+          new (&basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ())) stable_unsorted_iter_wp_type (iter.basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+        } else {
+          new (&basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ())) stable_unsorted_iter_type (iter.basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+        }
       } else {
         if (m_with_props) {
           new (&basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ())) iter_wp_type (iter.basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -375,12 +412,18 @@ db::Box
 instance_iterator<Traits>::quad_box () const
 {
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         return m_traits.quad_box (basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
       } else {
         return m_traits.quad_box (basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        return m_traits.quad_box (basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+      } else {
+        return m_traits.quad_box (basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+      }
     } else {
       if (m_with_props) {
         return m_traits.quad_box (basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -397,12 +440,18 @@ size_t
 instance_iterator<Traits>::quad_id () const
 {
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         return m_traits.quad_id (basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
       } else {
         return m_traits.quad_id (basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        return m_traits.quad_id (basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+      } else {
+        return m_traits.quad_id (basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+      }
     } else {
       if (m_with_props) {
         return m_traits.quad_id (basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -419,12 +468,18 @@ void
 instance_iterator<Traits>::skip_quad () 
 {
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         m_traits.skip_quad (basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
       } else {
         m_traits.skip_quad (basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        m_traits.skip_quad (basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+      } else {
+        m_traits.skip_quad (basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+      }
     } else {
       if (m_with_props) {
         m_traits.skip_quad (basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -442,12 +497,18 @@ instance_iterator<Traits> &
 instance_iterator<Traits>::operator++() 
 {
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         ++basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ());
       } else {
         ++basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ());
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        ++basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ());
+      } else {
+        ++basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ());
+      }
     } else {
       if (m_with_props) {
         ++basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ());
@@ -466,7 +527,7 @@ void
 instance_iterator<Traits>::make_next () 
 {
   while (true) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         if (! basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()).at_end ()) {
           return;
@@ -476,6 +537,16 @@ instance_iterator<Traits>::make_next ()
           return;
         }
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        if (! basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()).at_end ()) {
+          return;
+        }
+      } else {
+        if (! basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()).at_end ()) {
+          return;
+        }
+      }
     } else {
       if (m_with_props) {
         if (! basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()).at_end ()) {
@@ -502,12 +573,18 @@ void
 instance_iterator<Traits>::update_ref ()
 {
   if (m_type == TInstance) {
-    if (m_stable) {
+    if (m_stable && ! m_unsorted) {
       if (m_with_props) {
         m_ref = m_traits.instance_from_stable_iter (basic_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
       } else {
         m_ref = m_traits.instance_from_stable_iter (basic_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
       }
+    } else if (m_stable) {
+      if (m_with_props) {
+        m_ref = m_traits.instance_from_stable_iter (basic_unsorted_iter (cell_inst_wp_array_type::tag (), InstancesEditableTag ()));
+      } else {
+        m_ref = m_traits.instance_from_stable_iter (basic_unsorted_iter (cell_inst_array_type::tag (), InstancesEditableTag ()));
+      }
     } else {
       if (m_with_props) {
         m_ref = value_type (m_traits.instances (), *basic_iter (cell_inst_wp_array_type::tag (), InstancesNonEditableTag ()));
@@ -539,7 +616,7 @@ void
 NormalInstanceIteratorTraits::init (instance_iterator<NormalInstanceIteratorTraits> *iter) const
 {
   tl_assert (mp_insts != 0);
-  if (iter->m_stable) {
+  if (iter->m_stable && ! iter->m_unsorted) {
     if (iter->m_with_props) {
       cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
       iter->basic_iter (tag, InstancesEditableTag ()) = mp_insts->inst_tree (tag, InstancesEditableTag ()).begin_flat ();
@@ -547,6 +624,14 @@ NormalInstanceIteratorTraits::init (instance_iterator<NormalInstanceIteratorTrai
       cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
       iter->basic_iter (tag, InstancesEditableTag ()) = mp_insts->inst_tree (tag, InstancesEditableTag ()).begin_flat ();
     }
+  } else if (iter->m_stable) {
+    if (iter->m_with_props) {
+      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
+      iter->basic_unsorted_iter (tag, InstancesEditableTag ()) = stable_unsorted_iter_wp_type (mp_insts->inst_tree (tag, InstancesEditableTag ()).begin (), mp_insts->inst_tree (tag, InstancesEditableTag ()).end ());
+    } else {
+      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
+      iter->basic_unsorted_iter (tag, InstancesEditableTag ()) = stable_unsorted_iter_type (mp_insts->inst_tree (tag, InstancesEditableTag ()).begin (), mp_insts->inst_tree (tag, InstancesEditableTag ()).end ());
+    }
   } else {
     if (iter->m_with_props) {
       cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
@@ -640,13 +725,12 @@ OverlappingInstanceIteratorTraits::init (instance_iterator<OverlappingInstanceIt
 //  ChildCellIterator implementation
 
 ChildCellIterator::ChildCellIterator ()
-  : m_iter (), m_end (), mp_insts (0)
+  : m_iter (), m_end ()
 { }
 
 ChildCellIterator::ChildCellIterator (const instances_type *insts)
   : m_iter (insts->begin_sorted_insts ()),
-    m_end (insts->end_sorted_insts ()),
-    mp_insts (insts)
+    m_end (insts->end_sorted_insts ())
 { }
 
 cell_index_type 
@@ -823,9 +907,30 @@ Instance::bbox () const
   }
 }
 
+Instance::box_type
+Instance::bbox_with_empty () const
+{
+  const db::Instances *i = instances ();
+  const db::Cell *c = i ? i->cell () : 0;
+  const db::Layout *g = c ? c->layout () : 0;
+  if (g) {
+    return bbox (db::box_convert<cell_inst_type, false> (*g));
+  } else {
+    return db::Instance::box_type ();
+  }
+}
+
 // -------------------------------------------------------------------------------------
 //  Instances implementation
 
+static void
+check_is_editable_for_undo_redo (const Instances *instances)
+{
+  if (! instances->is_editable ()) {
+    throw tl::Exception (tl::to_string (tr ("No undo/redo support on non-editable instance lists")));
+  }
+}
+
 Instances::Instances (cell_type *cell)
   : mp_cell (cell)
 {
@@ -921,6 +1026,9 @@ Instances::operator= (const Instances &d)
 
     m_parent_insts = d.m_parent_insts;
 
+    set_instance_by_cell_index_needs_made (true);
+    set_instance_tree_needs_sort (true);
+
   }
   return *this;
 }
@@ -928,13 +1036,36 @@ Instances::operator= (const Instances &d)
 bool 
 Instances::is_editable () const
 {
-  return mp_cell && mp_cell->layout () ? mp_cell->layout ()->is_editable () : true;
+  return cell () && cell ()->layout () ? cell ()->layout ()->is_editable () : true;
 }
 
 db::Layout *
 Instances::layout () const
 {
-  return mp_cell ? mp_cell->layout () : 0;
+  return cell () ? cell ()->layout () : 0;
+}
+
+void
+Instances::invalidate_insts ()
+{
+  db::Cell *cp = cell ();
+  if (cp) {
+    cp->check_locked ();
+    cp->invalidate_insts ();
+  }
+
+  set_instance_by_cell_index_needs_made (true);
+  set_instance_tree_needs_sort (true);
+
+  invalidate_prop_ids ();
+}
+
+void
+Instances::invalidate_prop_ids ()
+{
+  if (layout ()) {
+    layout ()->invalidate_prop_ids ();
+  }
 }
 
 template <class Tag, class ET, class I>
@@ -943,10 +1074,15 @@ Instances::erase_positions (Tag tag, ET editable_tag, I first, I last)
 {
   typedef instances_editable_traits<ET> editable_traits;
 
-  if (mp_cell) { 
-    mp_cell->invalidate_insts ();  //  HINT: must come before the change is done!
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, first, last, true /*dummy*/));
+  invalidate_insts ();  //  HINT: must come before the change is done!
+
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
+      if (! is_editable ()) {
+        throw tl::Exception (tl::to_string (tr ("No undo/redo support for non-editable instance lists in 'erase_positions'")));
+      }
+      cell ()->manager ()->queue (cell (), new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, first, last, true /*dummy*/));
     }
   }
 
@@ -960,17 +1096,19 @@ Instances::insert (const InstArray &inst)
 {
   bool editable = is_editable ();
 
-  if (mp_cell) {
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
       if (editable) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesEditableTag> (true /*insert*/, inst));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesEditableTag> (true /*insert*/, inst));
       } else {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesNonEditableTag> (true /*insert*/, inst));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesNonEditableTag> (true /*insert*/, inst));
       }
     }
-    mp_cell->invalidate_insts ();
   }
 
+  invalidate_insts ();
+
   // TODO: simplify this, i.e. through instance_from_pointer
   if (editable) {
     return instance_type (this, inst_tree (typename InstArray::tag (), InstancesEditableTag ()).insert (inst));
@@ -986,13 +1124,14 @@ Instances::insert (I from, I to)
   typedef std::iterator_traits<I> it_traits;
   typedef typename it_traits::value_type value_type;
 
-  if (mp_cell) {
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename it_traits::value_type, ET> (true /*insert*/, from, to));
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
+      cell ()->manager ()->queue (cell (), new db::InstOp<typename it_traits::value_type, ET> (true /*insert*/, from, to));
     }
-    mp_cell->invalidate_insts ();
   }
 
+  invalidate_insts ();
   inst_tree (typename value_type::tag (), ET ()).insert (from, to);
 }
 
@@ -1036,19 +1175,21 @@ template <class InstArray>
 void 
 Instances::replace (const InstArray *replace, const InstArray &with)
 {
-  if (mp_cell) {
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
       if (is_editable ()) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesEditableTag> (false /*not insert*/, *replace));
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesEditableTag> (true /*insert*/, with));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesEditableTag> (false /*not insert*/, *replace));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesEditableTag> (true /*insert*/, with));
       } else {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesNonEditableTag> (false /*not insert*/, *replace));
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray, InstancesNonEditableTag> (true /*insert*/, with));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesNonEditableTag> (false /*not insert*/, *replace));
+        cell ()->manager ()->queue (cell (), new db::InstOp<InstArray, InstancesNonEditableTag> (true /*insert*/, with));
       }
     }
-    mp_cell->invalidate_insts ();
   }
 
+  invalidate_insts ();
+
   //  HINT: this only works because we know our box trees well:
   *((InstArray *)replace) = with;
 }
@@ -1057,7 +1198,7 @@ Instances::instance_type
 Instances::replace (const instance_type &ref, const cell_inst_wp_array_type &inst)
 {
   if (ref.instances () != this) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Trying to replace an object in a list that it does not belong to")));
+    throw tl::Exception (tl::to_string (tr ("Trying to replace an object in a list that it does not belong to")));
   }
 
   const cell_inst_wp_array_type *cp = ref.basic_ptr (cell_inst_wp_array_type::tag ());
@@ -1080,7 +1221,7 @@ Instances::instance_type
 Instances::replace (const instance_type &ref, const cell_inst_array_type &inst)
 {
   if (ref.instances () != this) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Trying to replace an object in a list that it does not belong to")));
+    throw tl::Exception (tl::to_string (tr ("Trying to replace an object in a list that it does not belong to")));
   }
 
   const cell_inst_array_type *cp = ref.basic_ptr (cell_inst_array_type::tag ());
@@ -1121,13 +1262,15 @@ void
 Instances::erase_inst_by_iter (Tag tag, ET editable_tag, I iter)
 {
   if (iter.vector () != &inst_tree (tag, editable_tag).objects ()) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Trying to erase an object from a list that it does not belong to")));
+    throw tl::Exception (tl::to_string (tr ("Trying to erase an object from a list that it does not belong to")));
   }
 
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, *iter));
+  invalidate_insts ();
+
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
+      cell ()->manager ()->queue (cell (), new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, *iter));
     }
   }
 
@@ -1138,10 +1281,12 @@ template <class Tag, class ET>
 void 
 Instances::erase_inst_by_tag (Tag tag, ET editable_tag, const typename Tag::object_type &obj)
 {
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, obj));
+  invalidate_insts ();
+
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
+      cell ()->manager ()->queue (cell (), new db::InstOp<typename Tag::object_type, ET> (false /*not insert*/, obj));
     }
   }
 
@@ -1169,15 +1314,17 @@ template <class ET>
 void
 Instances::clear_insts (ET editable_tag)
 {
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
+  invalidate_insts ();
+
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
       const Instances *const_this = this;
       if (! const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).empty ()) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
+        cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
       }
       if (! const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).empty ()) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_wp_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
+        cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_wp_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
       }
     }
   }
@@ -1198,9 +1345,7 @@ Instances::clear_insts ()
 void 
 Instances::clear (Instances::cell_inst_array_type::tag) 
 {
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-  }
+  invalidate_insts ();
 
   if (m_generic.any) {
     if (is_editable ()) {
@@ -1215,9 +1360,7 @@ Instances::clear (Instances::cell_inst_array_type::tag)
 void 
 Instances::clear (Instances::cell_inst_wp_array_type::tag) 
 {
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-  }
+  invalidate_insts ();
 
   if (m_generic_wp.any) {
     if (is_editable ()) {
@@ -1342,8 +1485,13 @@ struct cell_inst_compare_f
 };
 
 void 
-Instances::sort_child_insts () 
+Instances::sort_child_insts (bool force)
 {
+  if (! force && ! instance_by_cell_index_needs_made ()) {
+    return;
+  }
+  set_instance_by_cell_index_needs_made (false);
+
   m_insts_by_cell_index = sorted_inst_vector ();
   m_insts_by_cell_index.reserve (cell_instances ());
 
@@ -1376,25 +1524,108 @@ Instances::sort_child_insts ()
   std::sort (m_insts_by_cell_index.begin (), m_insts_by_cell_index.end (), cell_inst_compare_f<basic_inst_type> ());
 }
 
-void 
-Instances::sort_inst_tree (const Layout *g)
+namespace {
+
+/**
+ *  @brief An alternative box converter for CellInst which does not call layout.update
+ *
+ *  Using this converter in cell.update_bbox() prevents recursive calls to layout.update.
+ */
+class InternalCellInstBoxConverter
 {
+public:
+  typedef db::Cell::box_type box_type;
+  typedef db::simple_bbox_tag complexity;
+
+  InternalCellInstBoxConverter (const db::Layout *layout)
+    : mp_layout (layout)
+  {
+    //  .. nothing yet ..
+  }
+
+  db::Cell::box_type operator() (const db::Cell::cell_inst_type &cell_inst) const
+  {
+    return mp_layout->cell (cell_inst.cell_index ()).bbox_with_empty_no_update ();
+  }
+
+private:
+  const db::Layout *mp_layout;
+};
+
+/**
+ *  @brief An alternative box converter for CellInstArray with properties which does not call layout.update
+ *
+ *  Using this converter in cell.update_bbox() prevents recursive calls to layout.update.
+ */
+struct InternalCellInstArrayWithPropertiesBoxConverter
+{
+  typedef db::Cell::cell_inst_array_type cell_inst_array;
+  typedef db::Cell::box_type box_type;
+  typedef db::complex_bbox_tag complexity;
+
+  InternalCellInstArrayWithPropertiesBoxConverter (const db::Layout *layout)
+    : m_bc (layout)
+  { }
+
+  box_type operator() (const db::object_with_properties<cell_inst_array> &array) const
+  {
+    return array.bbox (m_bc);
+  }
+
+private:
+  InternalCellInstBoxConverter m_bc;
+};
+
+/**
+ *  @brief An alternative box converter for CellInstArray which does not call layout.update
+ *
+ *  Using this converter in cell.update_bbox() prevents recursive calls to layout.update.
+ */
+struct InternalCellInstArrayBoxConverter
+{
+  typedef db::Cell::cell_inst_array_type cell_inst_array;
+  typedef db::Cell::box_type box_type;
+  typedef db::complex_bbox_tag complexity;
+
+  InternalCellInstArrayBoxConverter (const db::Layout *layout)
+    : m_bc (layout)
+  { }
+
+  box_type operator() (const cell_inst_array &array) const
+  {
+    return array.bbox (m_bc);
+  }
+
+private:
+  InternalCellInstBoxConverter m_bc;
+};
+
+}
+
+void 
+Instances::sort_inst_tree (const Layout *g, bool force)
+{
+  if (! force && ! instance_tree_needs_sort ()) {
+    return;
+  }
+  set_instance_tree_needs_sort (false);
+
   if (m_generic.any) {
     if (is_editable ()) {
-      m_generic.stable_tree->sort (cell_inst_array_box_converter (*g));
+      m_generic.stable_tree->sort (InternalCellInstArrayBoxConverter (g));
     } else {
-      m_generic.unstable_tree->sort (cell_inst_array_box_converter (*g));
+      m_generic.unstable_tree->sort (InternalCellInstArrayBoxConverter (g));
       //  since we use unstable instance trees in non-editable mode, we need to resort the child instances in this case
-      sort_child_insts ();
+      sort_child_insts (true);
     }
   }
   if (m_generic_wp.any) {
     if (is_editable ()) {
-      m_generic_wp.stable_tree->sort (cell_inst_wp_array_box_converter (*g));
+      m_generic_wp.stable_tree->sort (InternalCellInstArrayWithPropertiesBoxConverter (g));
     } else {
-      m_generic_wp.unstable_tree->sort (cell_inst_wp_array_box_converter (*g));
+      m_generic_wp.unstable_tree->sort (InternalCellInstArrayWithPropertiesBoxConverter (g));
       //  since we use unstable instance trees in non-editable mode, we need to resort the child instances in this case
-      sort_child_insts ();
+      sort_child_insts (true);
     }
   }
 
@@ -1468,10 +1699,11 @@ Instances::instance_type
 Instances::replace_prop_id (const instance_type &ref, db::properties_id_type prop_id)
 {
   if (ref.instances () != this) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Trying to replace an object in a list that it does not belong to")));
+    throw tl::Exception (tl::to_string (tr ("Trying to replace an object in a list that it does not belong to")));
   }
 
-  if (! ref.is_null ()) {
+  if (! ref.is_null () && (!ref.has_prop_id () || ref.prop_id () != prop_id)) {
+    invalidate_prop_ids ();
     cell_inst_wp_array_type new_inst (ref.cell_inst (), prop_id);
     return replace (ref, new_inst);
   } else {
@@ -1479,7 +1711,29 @@ Instances::replace_prop_id (const instance_type &ref, db::properties_id_type pro
   }
 }
 
-void 
+Instances::instance_type
+Instances::clear_properties (const instance_type &ref)
+{
+  if (ref.instances () != this) {
+    throw tl::Exception (tl::to_string (tr ("Trying to replace an object in a list that it does not belong to")));
+  }
+
+  if (! ref.is_null () && ref.has_prop_id ()) {
+
+    invalidate_prop_ids ();
+
+    cell_inst_array_type new_inst (ref.cell_inst ());
+    erase (ref);
+    return insert (new_inst);
+
+  } else {
+
+    return ref;
+
+  }
+}
+
+void
 Instances::do_clear_insts ()
 {
   if (m_generic.any) {
@@ -1523,8 +1777,7 @@ Instances::redo (db::Op *op)
 
 Instances::instance_type 
 Instances::do_insert (const Instances::instance_type &ref, 
-                      tl::func_delegate_base <db::cell_index_type> &im,
-                      tl::func_delegate_base <db::properties_id_type> &pm)
+                      tl::func_delegate_base <db::cell_index_type> &im)
 {
   if (ref.instances () == this) {
 
@@ -1539,7 +1792,7 @@ Instances::do_insert (const Instances::instance_type &ref,
 
       cell_inst_wp_array_type inst_wp = *ref.basic_ptr (cell_inst_wp_array_type::tag ());
       inst_wp.object () = cell_inst_type (im (ref.cell_index ()));
-      inst_wp.properties_id (pm (ref.prop_id ()));
+      inst_wp.properties_id (ref.prop_id ());
 
       return insert (inst_wp);
 
@@ -1559,7 +1812,7 @@ Instances::do_insert (const Instances::instance_type &ref,
       cell_inst_array_type inst (*ref.basic_ptr (cell_inst_wp_array_type::tag ()), layout () ? &layout ()->array_repository () : 0);
       inst.object () = cell_inst_type (im (ref.cell_index ()));
 
-      return insert (cell_inst_wp_array_type (inst, pm (ref.prop_id ())));
+      return insert (cell_inst_wp_array_type (inst, ref.prop_id ()));
 
     }
 
@@ -1574,15 +1827,17 @@ void Instances::apply_op (const Op &op, ET editable_tag)
   bool has_insts = ! const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).empty ();
   bool has_wp_insts = ! const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).empty ();
 
-  if (mp_cell) {
-    mp_cell->invalidate_insts ();
-    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
+  invalidate_insts ();
+
+  if (cell ()) {
+    if (cell ()->manager () && cell ()->manager ()->transacting ()) {
+      check_is_editable_for_undo_redo (this);
       transacting = true;
       if (has_insts) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
+        cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
       }
       if (has_wp_insts) {
-        mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_wp_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
+        cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_wp_array_type, ET> (false /*not insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
       }
     }
   }
@@ -1607,10 +1862,10 @@ void Instances::apply_op (const Op &op, ET editable_tag)
 
   if (transacting) {
     if (has_insts) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_array_type, ET> (true /*insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
+      cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_array_type, ET> (true /*insert*/, const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_array_type::tag (), editable_tag).end ()));
     }
     if (has_wp_insts) {
-      mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_wp_array_type, ET> (true /*insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
+      cell ()->manager ()->queue (cell (), new db::InstOp<cell_inst_wp_array_type, ET> (true /*insert*/, const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).begin (), const_this->inst_tree (cell_inst_wp_array_type::tag (), editable_tag).end ()));
     }
   }
 }
