@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@ namespace db
 //  Recursive shape iterator implementation
 
 RecursiveShapeIterator::RecursiveShapeIterator (const RecursiveShapeIterator &d)
+  : gsi::ObjectBase (d)
 {
   operator= (d);
 }
@@ -48,6 +49,8 @@ RecursiveShapeIterator &RecursiveShapeIterator::operator= (const RecursiveShapeI
     mp_shape_prop_sel = d.mp_shape_prop_sel;
     m_shape_inv_prop_sel = d.m_shape_inv_prop_sel;
     m_overlapping = d.m_overlapping;
+    m_for_merged_input = d.m_for_merged_input;
+
     m_start = d.m_start;
     m_stop = d.m_stop;
 
@@ -64,20 +67,27 @@ RecursiveShapeIterator &RecursiveShapeIterator::operator= (const RecursiveShapeI
 
     m_box_convert = d.m_box_convert;
 
+    m_locker = d.m_locker;
     m_inst = d.m_inst;
     m_inst_array = d.m_inst_array;
     m_empty_cells_cache = d.m_empty_cells_cache;
     m_layer = d.m_layer;
     mp_cell = d.mp_cell;
     m_current_layer = d.m_current_layer;
+    m_skip_shapes = d.m_skip_shapes;
+    m_skip_shapes_member = d.m_skip_shapes_member;
     m_shape = d.m_shape;
     m_trans = d.m_trans;
+    m_global_trans = d.m_global_trans;
+    m_property_translator = d.m_property_translator;
     m_trans_stack = d.m_trans_stack;
     m_inst_iterators = d.m_inst_iterators;
     m_inst_array_iterators = d.m_inst_array_iterators;
     m_cells = d.m_cells;
     m_local_complex_region_stack = d.m_local_complex_region_stack;
     m_local_region_stack = d.m_local_region_stack;
+    m_skip_shapes_stack = d.m_skip_shapes_stack;
+    m_skip_shapes_member_stack = d.m_skip_shapes_member_stack;
     m_needs_reinit = d.m_needs_reinit;
     m_inst_quad_id = d.m_inst_quad_id;
     m_inst_quad_id_stack = d.m_inst_quad_id_stack;
@@ -92,12 +102,12 @@ RecursiveShapeIterator::RecursiveShapeIterator ()
   //  anything. Not necessary reasonable.
   m_layer = 0;
   m_has_layers = false;
-  mp_layout = 0;
   mp_shapes = 0;
   mp_top_cell = 0;
   mp_cell = 0;
   m_current_layer = 0;
   m_overlapping = false;
+  m_for_merged_input = false;
   m_max_depth = std::numeric_limits<int>::max (); // all
   m_min_depth = 0;
   m_shape_flags = shape_iterator::All;
@@ -112,10 +122,10 @@ RecursiveShapeIterator::RecursiveShapeIterator (const shapes_type &shapes)
 {
   m_layer = 0;
   m_has_layers = false;
-  mp_layout = 0;
   mp_shapes = &shapes;
   mp_top_cell = 0;
   m_overlapping = false;
+  m_for_merged_input = false;
   init ();
   init_region (box_type::world ());
 }
@@ -124,10 +134,10 @@ RecursiveShapeIterator::RecursiveShapeIterator (const shapes_type &shapes, const
 {
   m_layer = 0;
   m_has_layers = false;
-  mp_layout = 0;
   mp_shapes = &shapes;
   mp_top_cell = 0;
   m_overlapping = overlapping;
+  m_for_merged_input = false;
   init ();
   init_region (region);
 }
@@ -136,137 +146,152 @@ RecursiveShapeIterator::RecursiveShapeIterator (const shapes_type &shapes, const
 {
   m_layer = 0;
   m_has_layers = false;
-  mp_layout = 0;
   mp_shapes = &shapes;
   mp_top_cell = 0;
   m_overlapping = overlapping;
+  m_for_merged_input = false;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, const box_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, const box_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout, layer)
 {
   m_layer = layer;
   m_has_layers = false;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, const region_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, const region_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout, layer)
 {
   m_layer = layer;
   m_has_layers = false;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, bool for_merged_input)
   : m_box_convert (layout, layer)
 {
   m_layer = layer;
   m_has_layers = false;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = false;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (box_type::world ());
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers, const box_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers, const box_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers = layers;
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers, const region_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers, const region_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers = layers;
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::vector<unsigned int> &layers, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers = layers;
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = false;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (box_type::world ());
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers, const box_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers, const box_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers.insert (m_layers.end (), layers.begin (), layers.end ());
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers, const region_type &region, bool overlapping)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers, const region_type &region, bool overlapping, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers.insert (m_layers.end (), layers.begin (), layers.end ());
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = overlapping;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (region);
 }
 
-RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers)
+RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, const std::set<unsigned int> &layers, bool for_merged_input)
   : m_box_convert (layout)
 {
   m_layer = 0;
   m_layers.insert (m_layers.end (), layers.begin (), layers.end ());
   m_has_layers = true;
-  mp_layout = &layout;
+  mp_layout.reset (const_cast<db::Layout *> (&layout));
   mp_shapes = 0;
   mp_top_cell = &cell;
   m_overlapping = false;
+  m_for_merged_input = for_merged_input;
   init ();
   init_region (box_type::world ());
 }
 
+RecursiveShapeIterator::~RecursiveShapeIterator ()
+{
+    //  .. nothing yet ..
+}
+
+
 void 
 RecursiveShapeIterator::init ()
 {
@@ -280,6 +305,8 @@ RecursiveShapeIterator::init ()
   m_shape_quad_id = 0;
   mp_cell = 0;
   m_current_layer = 0;
+  m_global_trans = cplx_trans_type ();
+  m_property_translator = db::PropertiesTranslator ();
 }
 
 void
@@ -312,12 +339,32 @@ RecursiveShapeIterator::init_region (const RecursiveShapeIterator::region_type &
   }
 }
 
+void
+RecursiveShapeIterator::set_global_trans (const cplx_trans_type &tr)
+{
+  if (m_global_trans != tr) {
+    m_global_trans = tr;
+    reset ();
+  }
+}
+
+const db::RecursiveShapeIterator::cplx_trans_type &
+RecursiveShapeIterator::always_apply () const
+{
+  if (m_trans_stack.empty ()) {
+    return m_global_trans;
+  } else {
+    static cplx_trans_type unity;
+    return unity;
+  }
+}
+
 void
 RecursiveShapeIterator::set_region (const box_type &region)
 {
   if (m_region != region || mp_complex_region.get () != 0) {
     init_region (region);
-    m_needs_reinit = true;
+    reset ();
   }
 }
 
@@ -325,7 +372,7 @@ void
 RecursiveShapeIterator::set_region (const region_type &region)
 {
   init_region (region);
-  m_needs_reinit = true;
+  reset ();
 }
 
 void
@@ -338,7 +385,7 @@ RecursiveShapeIterator::confine_region (const box_type &region)
   } else {
     init_region (m_region & region);
   }
-  m_needs_reinit = true;
+  reset ();
 }
 
 void
@@ -351,7 +398,29 @@ RecursiveShapeIterator::confine_region (const region_type &region)
   } else {
     init_region (region & region_type (m_region));
   }
-  m_needs_reinit = true;
+  reset ();
+}
+
+void
+RecursiveShapeIterator::set_layer (unsigned int layer)
+{
+  if (m_has_layers || m_layer != layer) {
+    m_has_layers = false;
+    m_layers.clear ();
+    m_layer = layer;
+    reset ();
+  }
+}
+
+void
+RecursiveShapeIterator::set_layers (const std::vector<unsigned int> &layers)
+{
+  if (! m_has_layers || m_layers != layers) {
+    m_has_layers = true;
+    m_layers = layers;
+    m_layer = 0;
+    reset ();
+  }
 }
 
 namespace {
@@ -377,7 +446,7 @@ private:
 }
 
 void
-RecursiveShapeIterator::validate () const
+RecursiveShapeIterator::validate (RecursiveShapeReceiver *receiver) const
 {
   if (! m_needs_reinit) {
     return;
@@ -386,19 +455,25 @@ RecursiveShapeIterator::validate () const
   m_needs_reinit = false;
 
   //  re-initialize
+  m_locker = db::LayoutLocker ();
+
   mp_cell = mp_top_cell;
   m_trans_stack.clear ();
   m_inst_iterators.clear ();
   m_inst_quad_id_stack.clear ();
   m_inst_array_iterators.clear ();
   m_cells.clear ();
-  m_trans = cplx_trans_type ();
+  m_trans = m_global_trans;
   m_current_layer = 0;
   m_shape = shape_iterator ();
   m_shape_quad_id = 0;
 
   m_local_region_stack.clear ();
-  m_local_region_stack.push_back (m_region);
+  m_local_region_stack.push_back (m_global_trans.inverted () * m_region);
+  m_skip_shapes_stack.clear ();
+  m_skip_shapes_stack.push_back (false);
+  m_skip_shapes_member_stack.clear ();
+  m_skip_shapes_member_stack.push_back (false);
 
   m_local_complex_region_stack.clear ();
   if (mp_complex_region.get ()) {
@@ -425,19 +500,36 @@ RecursiveShapeIterator::validate () const
   }
 
   if (mp_shapes) {
+
     //  Ensures the trees are built properly - this is important in MT contexts (i.e. TilingProcessor)
     //  TODO: get rid of that const cast
     (const_cast <db::Shapes *> (mp_shapes))->update ();
+
     start_shapes ();
+
   } else if (mp_layout && (! m_has_layers || m_current_layer < m_layers.size ())) {
+
     //  Ensures the trees are built properly - this is important in MT contexts (i.e. TilingProcessor)
     mp_layout->update ();
-    new_cell ();
-    next_shape ();
+
+    new_cell (receiver);
+    next_shape (receiver);
+
+  }
+
+  if (mp_layout && ! at_end ()) {
+    m_locker = db::LayoutLocker (const_cast<db::Layout *> (mp_layout.get ()), true);
   }
 }
 
-void 
+void
+RecursiveShapeIterator::reset ()
+{
+  m_needs_reinit = true;
+  m_locker = db::LayoutLocker ();
+}
+
+void
 RecursiveShapeIterator::reset_selection ()
 {
   if (mp_layout) {
@@ -445,7 +537,7 @@ RecursiveShapeIterator::reset_selection ()
     m_start.clear ();
     m_stop.clear ();
 
-    m_needs_reinit = true;
+    reset ();
 
   }
 }
@@ -460,7 +552,7 @@ RecursiveShapeIterator::unselect_cells (const std::set<db::cell_index_type> &cel
       m_start.erase (*c);
     }
 
-    m_needs_reinit = true;
+    reset ();
 
   }
 }
@@ -475,7 +567,7 @@ RecursiveShapeIterator::unselect_all_cells ()
       m_stop.insert (c->cell_index ());
     }
 
-    m_needs_reinit = true;
+    reset ();
 
   }
 }
@@ -490,7 +582,7 @@ RecursiveShapeIterator::select_cells (const std::set<db::cell_index_type> &cells
       m_stop.erase (*c);
     }
 
-    m_needs_reinit = true;
+    reset ();
 
   }
 }
@@ -505,7 +597,7 @@ RecursiveShapeIterator::select_all_cells ()
       m_start.insert (c->cell_index ());
     }
 
-    m_needs_reinit = true;
+    reset ();
 
   }
 }
@@ -513,10 +605,17 @@ RecursiveShapeIterator::select_all_cells ()
 bool
 RecursiveShapeIterator::at_end () const
 {
-  validate ();
+  validate (0);
   return m_shape.at_end () || is_inactive ();
 }
 
+bool
+RecursiveShapeIterator::at_end_no_lock () const
+{
+  RecursiveShapeIterator copy (*this);
+  return copy.at_end ();
+}
+
 std::vector<db::InstElement>
 RecursiveShapeIterator::path () const
 {
@@ -543,6 +642,8 @@ RecursiveShapeIterator::bbox () const
     }
   }
 
+  box = box.transformed (m_global_trans);
+
   if (m_region != box_type::world ()) {
     box &= m_region;
   }
@@ -594,7 +695,7 @@ RecursiveShapeIterator::skip_inst_iter_for_complex_region () const
 
     //  skip insts outside the complex region
     if (! m_inst.at_end ()) {
-      if (! is_outside_complex_region (m_inst->bbox ())) {
+      if (! is_outside_complex_region (m_inst->bbox (m_box_convert))) {
         break;
       } else {
         ++m_inst;
@@ -605,7 +706,7 @@ RecursiveShapeIterator::skip_inst_iter_for_complex_region () const
 }
 
 void
-RecursiveShapeIterator::next ()
+RecursiveShapeIterator::next (RecursiveShapeReceiver *receiver)
 {
   if (! at_end ()) {
 
@@ -616,14 +717,20 @@ RecursiveShapeIterator::next ()
     }
 
     if (! mp_shapes && m_shape.at_end ()) {
-      next_shape ();
+      next_shape (receiver);
+    }
+
+    if (at_end ()) {
+      //  Take this opportunity the release the layout lock.
+      //  This way, the shape iterator can be held further, without blocking the layout.
+      m_locker = db::LayoutLocker ();
     }
 
   }
 }
 
 void
-RecursiveShapeIterator::next_shape () const
+RecursiveShapeIterator::next_shape (RecursiveShapeReceiver *receiver) const
 {
   while (at_end ()) {
 
@@ -639,9 +746,15 @@ RecursiveShapeIterator::next_shape () const
       
       if (! m_inst.at_end () && int (m_inst_iterators.size ()) < m_max_depth) {
 
+        tl_assert (mp_layout);
+
         //  determine whether the cell is empty with respect to the layers specified
         bool is_empty = false;
-        if (! m_has_layers) {
+        if (receiver && receiver->wants_all_cells ()) {
+
+          //  don't skip empty cells in that case
+
+        } else if (! m_has_layers) {
 
           is_empty = mp_layout->cell (m_inst->cell_index ()).bbox (m_layer).empty ();
 
@@ -667,11 +780,21 @@ RecursiveShapeIterator::next_shape () const
 
         if (is_empty) {
 
+          //  skip entire cell
           ++m_inst;
-          new_inst ();
+          new_inst (receiver);
+
+        } else if (!down (receiver)) {
+
+          //  skip this instance array member
+          ++m_inst_array;
+          new_inst_member (receiver);
+
+          if (m_inst_array.at_end ()) {
+            ++m_inst;
+            new_inst (receiver);
+          }
 
-        } else {
-          down ();
         }
 
       } else {
@@ -682,12 +805,14 @@ RecursiveShapeIterator::next_shape () const
         }
 
         //  no more instances: up and next instance
-        up ();
+        up (receiver);
 
         ++m_inst_array;
+        new_inst_member (receiver);
+
         if (m_inst_array.at_end ()) {
           ++m_inst;
-          new_inst ();
+          new_inst (receiver);
         }
 
       }
@@ -697,9 +822,11 @@ RecursiveShapeIterator::next_shape () const
   }
 }
 
-void
-RecursiveShapeIterator::down () const
+bool
+RecursiveShapeIterator::down (RecursiveShapeReceiver *receiver) const
 {
+  tl_assert (mp_layout);
+
   m_trans_stack.push_back (m_trans);
   m_cells.push_back (mp_cell);
 
@@ -708,8 +835,10 @@ RecursiveShapeIterator::down () const
   m_inst_quad_id_stack.push_back (m_inst_quad_id);
 
   bool ia = is_inactive ();
+  bool aoi = is_all_of_instance ();
   mp_cell = &mp_layout->cell (m_inst->cell_index ());
   set_inactive (ia);
+  set_all_of_instance (aoi);
 
   m_trans = m_trans * m_inst->complex_trans (*m_inst_array);
 
@@ -717,11 +846,14 @@ RecursiveShapeIterator::down () const
   box_type new_region = box_type::world ();
 
   //  compute the region inside the new cell
-  if (new_region != m_local_region_stack.front ()) {
-    new_region = m_trans.inverted () * m_local_region_stack.front ();
-    new_region &= cell ()->bbox ();
+  if (new_region != m_local_region_stack.back ()) {
+    new_region = m_inst->complex_trans (*m_inst_array).inverted () * m_local_region_stack.back ();
+    new_region &= cell_bbox (cell_index ());
   }
+
   m_local_region_stack.push_back (new_region);
+  m_skip_shapes_stack.push_back (m_skip_shapes);
+  m_skip_shapes_member_stack.push_back (m_skip_shapes_member);
 
   if (! m_local_complex_region_stack.empty ()) {
 
@@ -755,11 +887,39 @@ RecursiveShapeIterator::down () const
 
   }
 
-  new_cell ();
+  //  do not descend if the box is empty
+
+  if (m_local_region_stack.back ().empty ()) {
+
+    pop ();
+
+    return false;
+
+  } else {
+
+    if (receiver) {
+      receiver->enter_cell (this, cell (), m_local_region_stack.back (), m_local_complex_region_stack.empty () ? 0 : &m_local_complex_region_stack.back ());
+    }
+
+    new_cell (receiver);
+
+    return true;
+
+  }
 }
 
 void
-RecursiveShapeIterator::up () const
+RecursiveShapeIterator::up (RecursiveShapeReceiver *receiver) const
+{
+  if (receiver) {
+    receiver->leave_cell (this, cell ());
+  }
+
+  pop ();
+}
+
+void
+RecursiveShapeIterator::pop () const
 {
   m_shape = shape_iterator ();
   m_shape_quad_id = 0;
@@ -767,6 +927,8 @@ RecursiveShapeIterator::up () const
   m_inst = m_inst_iterators.back ();
   m_inst_array = m_inst_array_iterators.back ();
   m_inst_quad_id = m_inst_quad_id_stack.back ();
+  m_skip_shapes = m_skip_shapes_stack.back ();
+  m_skip_shapes_member = m_skip_shapes_member_stack.back ();
   m_inst_iterators.pop_back ();
   m_inst_array_iterators.pop_back ();
   m_inst_quad_id_stack.pop_back ();
@@ -776,6 +938,8 @@ RecursiveShapeIterator::up () const
   mp_cell = m_cells.back ();
   m_cells.pop_back ();
   m_local_region_stack.pop_back ();
+  m_skip_shapes_stack.pop_back ();
+  m_skip_shapes_member_stack.pop_back ();
   if (! m_local_complex_region_stack.empty ()) {
     m_local_complex_region_stack.pop_back ();
   }
@@ -800,7 +964,7 @@ RecursiveShapeIterator::start_shapes () const
 void
 RecursiveShapeIterator::new_layer () const
 {
-  if (int (m_trans_stack.size ()) < m_min_depth || int (m_trans_stack.size ()) > m_max_depth) {
+  if (skip_shapes () || int (m_trans_stack.size ()) < m_min_depth || int (m_trans_stack.size ()) > m_max_depth) {
     m_shape = shape_iterator ();
   } else if (! m_overlapping) {
     m_shape = cell ()->shapes (m_layer).begin_touching (m_local_region_stack.back (), m_shape_flags, mp_shape_prop_sel, m_shape_inv_prop_sel);
@@ -817,35 +981,64 @@ RecursiveShapeIterator::new_layer () const
 }
 
 void 
-RecursiveShapeIterator::new_cell () const
+RecursiveShapeIterator::new_cell (RecursiveShapeReceiver *receiver) const
 {
   if (m_has_layers) {
     m_current_layer = 0;
     m_layer = m_layers.front ();
   }
 
-  if (! m_start.empty () && m_start.find (cell_index ()) != m_start.end ()) {
-    set_inactive (false);
-  } else if (! m_stop.empty () && m_stop.find (cell_index ()) != m_stop.end ()) {
-    set_inactive (true);
+  bool new_cell_inactive = is_child_inactive (cell_index ());
+  if (is_inactive () != new_cell_inactive) {
+    set_inactive (new_cell_inactive);
   }
 
   new_layer ();
 
-  m_inst = cell ()->begin_touching (m_local_region_stack.back ());
+  if (m_overlapping) {
+    m_inst = cell ()->begin_touching (m_local_region_stack.back ().enlarged (box_type::vector_type (-1, -1)));
+  } else {
+    m_inst = cell ()->begin_touching (m_local_region_stack.back ());
+  }
 
   m_inst_quad_id = 0;
 
   //  skip instance quad if possible
-  if (! m_local_complex_region_stack.empty ()) {
+  if (! m_local_complex_region_stack.empty () && (! receiver || ! receiver->wants_all_cells ())) {
     skip_inst_iter_for_complex_region ();
   }
 
-  new_inst ();
+  new_inst (receiver);
+}
+
+bool
+RecursiveShapeIterator::instance_is_covered (const box_type &inst_bx, unsigned int layer) const
+{
+  //  Try some optimization: if the instance we're looking at is entirely covered
+  //  by a rectangle (other objects are too expensive to check), then we skip it
+  //
+  //  We check 10 shapes max.
+
+  auto si = cell ()->shapes (layer).begin_overlapping (inst_bx, m_shape_flags, mp_shape_prop_sel, m_shape_inv_prop_sel);
+  size_t nmax = 10;
+  while (! si.at_end () && nmax-- > 0) {
+    if (inst_bx.inside (si->rectangle ())) {
+      return true;
+    }
+    ++si;
+  }
+
+  return false;
+}
+
+bool
+RecursiveShapeIterator::skip_shapes () const
+{
+  return m_skip_shapes_stack.back () || m_skip_shapes_member_stack.back ();
 }
 
 void 
-RecursiveShapeIterator::new_inst () const
+RecursiveShapeIterator::new_inst (RecursiveShapeReceiver *receiver) const
 {
   //  look for the next instance with a non-empty array iterator. The latter can be 
   //  empty because we use a per-layer box converter for that case what we don't for the
@@ -853,19 +1046,57 @@ RecursiveShapeIterator::new_inst () const
   while (! m_inst.at_end ()) {
 
     //  skip instance quad if possible
-    if (! m_local_complex_region_stack.empty ()) {
+    if (! m_local_complex_region_stack.empty () && (! receiver || ! receiver->wants_all_cells ())) {
       skip_inst_iter_for_complex_region ();
       if (m_inst.at_end ()) {
         break;
       }
     }
 
-    if (m_local_region_stack.back () != box_type::world ()) {
-      m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region_stack.back (), m_box_convert);
+    bool all_of_instance = false;
+    bool with_region = false;
+
+    if (m_local_region_stack.back () != box_type::world () && ! m_inst->cell_inst ().bbox (m_box_convert).inside (m_local_region_stack.back ())) {
+      with_region = true;
     } else {
-      m_inst_array = m_inst->cell_inst ().begin (); 
+      //  TODO: optimization potential: only report all_of_instance == false, if not entirely within the complex region
+      all_of_instance = m_local_complex_region_stack.empty ();
     }
 
+    m_skip_shapes = skip_shapes ();
+    m_skip_shapes_member = false;
+
+    if (m_for_merged_input && ! m_skip_shapes && (! m_has_layers || m_layers.size () == 1)) {
+      box_type inst_bx = m_inst->bbox (m_box_convert);
+      m_skip_shapes = instance_is_covered (inst_bx, m_has_layers ? m_layers.front () : m_layer);
+    }
+
+    RecursiveShapeReceiver::new_inst_mode ni = RecursiveShapeReceiver::NI_all;
+    if (receiver) {
+      ni = receiver->new_inst (this, m_inst->cell_inst (), always_apply (), m_local_region_stack.back (), m_local_complex_region_stack.empty () ? 0 : &m_local_complex_region_stack.back (), all_of_instance, m_skip_shapes);
+    } else if (m_skip_shapes) {
+      ni = RecursiveShapeReceiver::NI_skip;
+    }
+
+    if (ni == RecursiveShapeReceiver::NI_skip) {
+      m_inst_array = inst_array_iterator ();
+    } else if (ni == RecursiveShapeReceiver::NI_single) {
+      //  a singular iterator
+      m_inst_array = db::CellInstArray::iterator (m_inst->cell_inst ().front (), false);
+    } else if (with_region) {
+      if (m_overlapping) {
+        m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region_stack.back ().enlarged (box_type::vector_type (-1, -1)), m_box_convert);
+      } else {
+        m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region_stack.back (), m_box_convert);
+      }
+    } else {
+      m_inst_array = m_inst->cell_inst ().begin ();
+    }
+
+    set_all_of_instance (all_of_instance);
+
+    new_inst_member (receiver);
+
     if (! m_inst_array.at_end ()) {
       break;
     } else {
@@ -875,6 +1106,51 @@ RecursiveShapeIterator::new_inst () const
   }
 }
 
+void
+RecursiveShapeIterator::new_inst_member (RecursiveShapeReceiver *receiver) const
+{
+  if (! m_local_complex_region_stack.empty ()) {
+
+    //  skip instance array members not part of the complex region
+    while (! m_inst_array.at_end ()) {
+      box_type ia_box = m_inst->complex_trans (*m_inst_array) * cell_bbox (m_inst->cell_index ());
+      if (! is_outside_complex_region (ia_box)) {
+        break;
+      } else {
+        ++m_inst_array;
+      }
+    }
+
+  }
+
+  m_skip_shapes_member = false;
+
+  while (! m_inst_array.at_end () && (m_for_merged_input || receiver)) {
+
+    m_skip_shapes_member = m_skip_shapes;
+    if (m_for_merged_input && ! m_inst_array.is_singular () && ! m_skip_shapes && (! m_has_layers || m_layers.size () == 1)) {
+
+      box_type ia_box = m_inst->complex_trans (*m_inst_array) * cell_bbox (m_inst->cell_index ());
+      m_skip_shapes_member = instance_is_covered (ia_box, m_has_layers ? m_layers.front () : m_layer);
+
+    }
+
+    bool skip = false;
+    if (receiver) {
+      skip = ! receiver->new_inst_member (this, m_inst->cell_inst (), always_apply (), m_inst->complex_trans (*m_inst_array), m_local_region_stack.back (), m_local_complex_region_stack.empty () ? 0 : &m_local_complex_region_stack.back (), is_all_of_instance (), m_skip_shapes_member);
+    } else {
+      skip = m_skip_shapes_member;
+    }
+
+    if (skip) {
+      ++m_inst_array;
+    } else {
+      break;
+    }
+
+  }
+}
+
 bool
 RecursiveShapeIterator::is_outside_complex_region (const db::Box &box) const
 {
@@ -885,5 +1161,44 @@ RecursiveShapeIterator::is_outside_complex_region (const db::Box &box) const
   }
 }
 
+bool
+RecursiveShapeIterator::is_child_inactive (db::cell_index_type new_child) const
+{
+  bool inactive = is_inactive ();
+  if (! m_start.empty () && m_start.find (new_child) != m_start.end ()) {
+    inactive = false;
+  } else if (! m_stop.empty () && m_stop.find (new_child) != m_stop.end ()) {
+    inactive = true;
+  }
+  return inactive;
+}
+
+void
+RecursiveShapeIterator::push (RecursiveShapeReceiver *receiver)
+{
+  //  force reset so we can validate with a receiver
+  reset ();
+
+  receiver->begin (this);
+
+  try {
+
+    validate (receiver);
+
+    while (! at_end ()) {
+      receiver->shape (this, *m_shape, always_apply (), m_trans, m_local_region_stack.back (), m_local_complex_region_stack.empty () ? 0 : &m_local_complex_region_stack.back ());
+      next (receiver);
+    }
+
+    receiver->end (this);
+
+  } catch (...) {
+
+    receiver->end (this);
+    throw;
+
+  }
+}
+
 }
 
