@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@
 #include "dbPolygon.h"
 #include "dbPath.h"
 #include "dbEdge.h"
+#include "dbEdgePair.h"
 #include "dbText.h"
 #include "dbBox.h"
 #include "dbBoxConvert.h"
@@ -616,6 +617,7 @@ public:
   typedef db::path_ref<path_type, unit_trans_type> path_ptr_type;
   typedef db::array<path_ptr_type, disp_type> path_ptr_array_type;
   typedef db::edge<coord_type> edge_type;
+  typedef db::edge_pair<coord_type> edge_pair_type;
   typedef db::text<coord_type> text_type;
   typedef db::text_ref<text_type, disp_type> text_ref_type;
   typedef db::text_ref<text_type, unit_trans_type> text_ptr_type;
@@ -642,6 +644,8 @@ public:
   typedef tl::reuse_vector<db::path_ref<path_type, unit_trans_type> >::const_iterator path_ptr_iter_type;
   typedef tl::reuse_vector<db::array<path_ptr_type, disp_type> >::const_iterator path_ptr_array_iter_type;
   typedef tl::reuse_vector<db::edge<coord_type> >::const_iterator edge_iter_type;
+  typedef tl::reuse_vector<db::edge_pair<coord_type> >::const_iterator edge_pair_iter_type;
+  typedef tl::reuse_vector<db::point<coord_type> >::const_iterator point_iter_type;
   typedef tl::reuse_vector<db::text<coord_type> >::const_iterator text_iter_type;
   typedef tl::reuse_vector<db::text_ref<text_type, disp_type> >::const_iterator text_ref_iter_type;
   typedef tl::reuse_vector<db::text_ref<text_type, unit_trans_type> >::const_iterator text_ptr_iter_type;
@@ -665,6 +669,8 @@ public:
   typedef tl::reuse_vector<db::object_with_properties<db::path_ref<path_type, unit_trans_type> > >::const_iterator ppath_ptr_iter_type;
   typedef tl::reuse_vector<db::object_with_properties<db::array<path_ptr_type, disp_type> > >::const_iterator ppath_ptr_array_iter_type;
   typedef tl::reuse_vector<db::object_with_properties<db::edge<coord_type> > >::const_iterator pedge_iter_type;
+  typedef tl::reuse_vector<db::object_with_properties<db::edge_pair<coord_type> > >::const_iterator pedge_pair_iter_type;
+  typedef tl::reuse_vector<db::object_with_properties<db::point<coord_type> > >::const_iterator ppoint_iter_type;
   typedef tl::reuse_vector<db::object_with_properties<db::text<coord_type> > >::const_iterator ptext_iter_type;
   typedef tl::reuse_vector<db::object_with_properties<db::text_ref<text_type, disp_type> > >::const_iterator ptext_ref_iter_type;
   typedef tl::reuse_vector<db::object_with_properties<db::text_ref<text_type, unit_trans_type> > >::const_iterator ptext_ptr_iter_type;
@@ -690,6 +696,7 @@ public:
     SimplePolygonPtrArray,
     SimplePolygonPtrArrayMember,
     Edge,
+    EdgePair,
     Path,
     PathRef,
     PathPtrArray,
@@ -704,6 +711,7 @@ public:
     TextRef,
     TextPtrArray,
     TextPtrArrayMember,
+    Point,
     UserObject
   };
     
@@ -781,8 +789,8 @@ public:
    *  we would probably create too much constness. Hence we use const_cast to convert it
    *  to a non-const one.
    */
-  template <class Obj>
-  Shape (const db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj> &tree_iter)
+  template <class Obj, bool R>
+  Shape (const db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj, R> &tree_iter)
     : mp_shapes (const_cast<db::Shapes *> (shapes)), m_with_props (false), m_stable (true)
   {
     typename Obj::tag tag = typename Obj::tag ();
@@ -795,8 +803,8 @@ public:
   /**
    *  @brief Construct as a proxy to a certain object given by an iterator
    */
-  template <class Obj>
-  Shape (db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj> &tree_iter)
+  template <class Obj, bool R>
+  Shape (db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj, R> &tree_iter)
     : mp_shapes (shapes), m_with_props (false), m_stable (true)
   {
     typename Obj::tag tag = typename Obj::tag ();
@@ -813,8 +821,8 @@ public:
    *  we would probably create too much constness. Hence we use const_cast to convert it
    *  to a non-const one.
    */
-  template <class Obj, class Trans>
-  Shape (const db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj> &tree_iter, const Trans &trans)
+  template <class Obj, bool R, class Trans>
+  Shape (const db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj, R> &tree_iter, const Trans &trans)
     : mp_shapes (const_cast<db::Shapes *> (shapes)), m_with_props (false), m_stable (true)
   {
     typename Obj::tag tag = typename Obj::tag ();
@@ -827,8 +835,8 @@ public:
   /**
    *  @brief Construct as a proxy to a certain object given by an iterator
    */
-  template <class Obj, class Trans>
-  Shape (db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj> &tree_iter, const Trans &trans)
+  template <class Obj, bool R, class Trans>
+  Shape (db::Shapes *shapes, const tl::reuse_vector_const_iterator<Obj, R> &tree_iter, const Trans &trans)
     : mp_shapes (shapes), m_with_props (false), m_stable (true)
   {
     typename Obj::tag tag = typename Obj::tag ();
@@ -910,7 +918,7 @@ public:
   }
 
   /** 
-   *  @brief Construct a shape proxy as a reference to a simple olygon
+   *  @brief Construct a shape proxy as a reference to a simple polygon
    */
   void init (simple_polygon_type::tag)
   {
@@ -1016,7 +1024,23 @@ public:
     m_type = Edge;
   }
 
-  /** 
+  /**
+   *  @brief Construct a shape proxy as a reference to a edge pair
+   */
+  void init (edge_pair_type::tag)
+  {
+    m_type = EdgePair;
+  }
+
+  /**
+   *  @brief Construct a shape proxy as a reference to a point
+   */
+  void init (point_type::tag)
+  {
+    m_type = Point;
+  }
+
+  /**
    *  @brief Construct a shape proxy as a reference to a box
    */
   void init (box_type::tag)
@@ -1075,16 +1099,24 @@ public:
   }
 
   /**
-   *  @brief Determine, if the shape is of "object_with_properties" type.
+   *  @brief Gets the index of layer the shape sits in
    *
-   *  Usually, the properties id (see prop_id()) should be used.
-   *
-   *  @return true, if the shape is of "object_with_properties<Sh> type".
+   *  This getter only applies to shapes that are members of a cell
+   *  and a layout.
+   *  For other shapes this method returns -1.
    */
-  bool with_props () const
-  {
-    return m_with_props;
-  }
+  int layer () const;
+
+  /**
+   *  @brief Changes the layer the shape sits in
+   *
+   *  This setter applies to shapes are are members of a cell
+   *  and a layout.
+   *  Changing the layer effectively moves the shape to a different
+   *  container. For invalid layer indexes or if the shape is
+   *  not member of a cell and layout, this method does nothing.
+   */
+  void set_layer (unsigned int layer_index);
 
   /**
    *  @brief Get the properties Id associated with the shape
@@ -1186,7 +1218,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const polygon_type *basic_ptr (polygon_type::tag) const
@@ -1202,7 +1234,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const simple_polygon_type *basic_ptr (simple_polygon_type::tag) const
@@ -1218,7 +1250,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const polygon_ref_type *basic_ptr (polygon_ref_type::tag) const
@@ -1234,7 +1266,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const polygon_ptr_array_type *basic_ptr (polygon_ptr_array_type::tag) const
@@ -1250,7 +1282,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const simple_polygon_ref_type *basic_ptr (simple_polygon_ref_type::tag) const
@@ -1266,7 +1298,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const simple_polygon_ptr_array_type *basic_ptr (simple_polygon_ptr_array_type::tag) const
@@ -1282,7 +1314,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const path_type *basic_ptr (path_type::tag) const
@@ -1298,7 +1330,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const path_ref_type *basic_ptr (path_ref_type::tag) const
@@ -1315,7 +1347,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const path_ptr_array_type *basic_ptr (path_ptr_array_type::tag) const
@@ -1331,7 +1363,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const edge_type *basic_ptr (edge_type::tag) const
@@ -1344,10 +1376,42 @@ public:
     }
   }
 
-  /** 
+  /**
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
+   *  target object.
+   */
+  const edge_pair_type *basic_ptr (edge_pair_type::tag) const
+  {
+    tl_assert (m_type == EdgePair);
+    if (m_stable) {
+      return m_with_props ? &**(((pedge_pair_iter_type *) m_generic.iter)) : &**(((edge_pair_iter_type *) m_generic.iter));
+    } else {
+      return m_with_props ? m_generic.pedge_pair : m_generic.edge_pair;
+    }
+  }
+
+  /**
+   *  @brief Return the actual object that this shape reference is pointing to
+   *
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
+   *  target object.
+   */
+  const point_type *basic_ptr (point_type::tag) const
+  {
+    tl_assert (m_type == Point);
+    if (m_stable) {
+      return m_with_props ? &**(((ppoint_iter_type *) m_generic.iter)) : &**(((point_iter_type *) m_generic.iter));
+    } else {
+      return m_with_props ? m_generic.ppoint : m_generic.point;
+    }
+  }
+
+  /**
+   *  @brief Return the actual object that this shape reference is pointing to
+   *
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const text_type *basic_ptr (text_type::tag) const
@@ -1363,7 +1427,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const text_ref_type *basic_ptr (text_ref_type::tag) const
@@ -1379,7 +1443,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const text_ptr_array_type *basic_ptr (text_ptr_array_type::tag) const
@@ -1395,7 +1459,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const user_object_type *basic_ptr (user_object_type::tag) const
@@ -1411,7 +1475,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const box_type *basic_ptr (box_type::tag) const
@@ -1427,7 +1491,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const box_array_type *basic_ptr (box_array_type::tag) const
@@ -1443,7 +1507,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const short_box_type *basic_ptr (short_box_type::tag) const
@@ -1459,7 +1523,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const short_box_array_type *basic_ptr (short_box_array_type::tag) const
@@ -1475,7 +1539,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<polygon_type> *basic_ptr (db::object_with_properties<polygon_type>::tag) const
@@ -1492,7 +1556,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<simple_polygon_type> *basic_ptr (db::object_with_properties<simple_polygon_type>::tag) const
@@ -1509,7 +1573,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<polygon_ref_type> *basic_ptr (db::object_with_properties<polygon_ref_type>::tag) const
@@ -1526,7 +1590,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<polygon_ptr_array_type> *basic_ptr (db::object_with_properties<polygon_ptr_array_type>::tag) const
@@ -1543,7 +1607,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<simple_polygon_ref_type> *basic_ptr (db::object_with_properties<simple_polygon_ref_type>::tag) const
@@ -1571,7 +1635,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<path_type> *basic_ptr (db::object_with_properties<path_type>::tag) const
@@ -1588,7 +1652,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<path_ref_type> *basic_ptr (db::object_with_properties<path_ref_type>::tag) const
@@ -1605,7 +1669,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<path_ptr_array_type> *basic_ptr (db::object_with_properties<path_ptr_array_type>::tag) const
@@ -1622,7 +1686,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<edge_type> *basic_ptr (db::object_with_properties<edge_type>::tag) const
@@ -1636,10 +1700,44 @@ public:
     }
   }
 
-  /** 
+  /**
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
+   *  target object.
+   */
+  const db::object_with_properties<edge_pair_type> *basic_ptr (db::object_with_properties<edge_pair_type>::tag) const
+  {
+    tl_assert (m_type == EdgePair);
+    tl_assert (m_with_props);
+    if (m_stable) {
+      return &**(((pedge_pair_iter_type *) m_generic.iter));
+    } else {
+      return m_generic.pedge_pair;
+    }
+  }
+
+  /**
+   *  @brief Return the actual object that this shape reference is pointing to for objects with properties
+   *
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
+   *  target object.
+   */
+  const db::object_with_properties<point_type> *basic_ptr (db::object_with_properties<point_type>::tag) const
+  {
+    tl_assert (m_type == Point);
+    tl_assert (m_with_props);
+    if (m_stable) {
+      return &**(((ppoint_iter_type *) m_generic.iter));
+    } else {
+      return m_generic.ppoint;
+    }
+  }
+
+  /**
+   *  @brief Return the actual object that this shape reference is pointing to for objects with properties
+   *
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<text_type> *basic_ptr (db::object_with_properties<text_type>::tag) const
@@ -1656,7 +1754,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<text_ref_type> *basic_ptr (db::object_with_properties<text_ref_type>::tag) const
@@ -1673,7 +1771,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<text_ptr_array_type> *basic_ptr (db::object_with_properties<text_ptr_array_type>::tag) const
@@ -1690,7 +1788,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<user_object_type> *basic_ptr (db::object_with_properties<user_object_type>::tag) const
@@ -1707,7 +1805,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<box_type> *basic_ptr (db::object_with_properties<box_type>::tag) const
@@ -1724,7 +1822,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<box_array_type> *basic_ptr (db::object_with_properties<box_array_type>::tag) const
@@ -1741,7 +1839,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<short_box_type> *basic_ptr (db::object_with_properties<short_box_type>::tag) const
@@ -1758,7 +1856,7 @@ public:
   /** 
    *  @brief Return the actual object that this shape reference is pointing to for objects with properties
    *
-   *  This is a generalisation of the polygon (), etc. methods using a tag to identify the
+   *  This is a generalization of the polygon (), etc. methods using a tag to identify the
    *  target object.
    */
   const db::object_with_properties<short_box_array_type> *basic_ptr (db::object_with_properties<short_box_array_type>::tag) const
@@ -1862,7 +1960,25 @@ public:
     return *(((edge_iter_type *) m_generic.iter));
   }
 
-  /** 
+  /**
+   *  @brief Return the iterator (in stable reference mode) by tag
+   */
+  edge_pair_iter_type basic_iter (edge_pair_type::tag) const
+  {
+    tl_assert (m_type == EdgePair && ! m_with_props);
+    return *(((edge_pair_iter_type *) m_generic.iter));
+  }
+
+  /**
+   *  @brief Return the iterator (in stable reference mode) by tag
+   */
+  point_iter_type basic_iter (point_type::tag) const
+  {
+    tl_assert (m_type == Point && ! m_with_props);
+    return *(((point_iter_type *) m_generic.iter));
+  }
+
+  /**
    *  @brief Return the iterator (in stable reference mode) by tag 
    */
   text_iter_type basic_iter (text_type::tag) const
@@ -2024,7 +2140,25 @@ public:
     return *(((pedge_iter_type *) m_generic.iter));
   }
 
-  /** 
+  /**
+   *  @brief Return the iterator (in stable reference mode) by tag for objects with properties
+   */
+  pedge_pair_iter_type basic_iter (db::object_with_properties<edge_pair_type>::tag) const
+  {
+    tl_assert (m_type == EdgePair && m_with_props);
+    return *(((pedge_pair_iter_type *) m_generic.iter));
+  }
+
+  /**
+   *  @brief Return the iterator (in stable reference mode) by tag for objects with properties
+   */
+  ppoint_iter_type basic_iter (db::object_with_properties<point_type>::tag) const
+  {
+    tl_assert (m_type == Point && m_with_props);
+    return *(((ppoint_iter_type *) m_generic.iter));
+  }
+
+  /**
    *  @brief Return the iterator (in stable reference mode) by tag for objects with properties
    */
   ptext_iter_type basic_iter (db::object_with_properties<text_type>::tag) const
@@ -2288,7 +2422,93 @@ public:
     return edge (p);
   }
 
-  /** 
+  /**
+   *  @brief Return a reference to the edge pair if one is referenced
+   */
+  const edge_pair_type &edge_pair () const
+  {
+    tl_assert (m_type == EdgePair);
+    return *basic_ptr (edge_pair_type::tag ());
+  }
+
+  /**
+   *  @brief Test if the shape proxy points to a edge pair
+   */
+  bool is_edge_pair () const
+  {
+    return (m_type == EdgePair);
+  }
+
+  /**
+   *  @brief Instantiate the edge pair object
+   *
+   *  If an edge pair is referenced, this object is instantiated
+   *  by this method.
+   *  Returns true, if the conversion was successful.
+   */
+  bool edge_pair (edge_pair_type &e) const
+  {
+    if (is_edge_pair ()) {
+      e = edge_pair ();
+      return true;
+    } else {
+      return false;
+    }
+  }
+
+  /**
+   *  @brief Alias for polymorphic expansion
+   *  Returns true, if the conversion was successful.
+   */
+  bool instantiate (edge_pair_type &p) const
+  {
+    return edge_pair (p);
+  }
+
+  /**
+   *  @brief Return a reference to the point if one is referenced
+   */
+  const point_type &point () const
+  {
+    tl_assert (m_type == Point);
+    return *basic_ptr (point_type::tag ());
+  }
+
+  /**
+   *  @brief Test if the shape proxy points to a point
+   */
+  bool is_point () const
+  {
+    return (m_type == Point);
+  }
+
+  /**
+   *  @brief Instantiate the edge object
+   *
+   *  If a edge is referenced, this object is instantiated
+   *  by this method.
+   *  Returns true, if the conversion was successful.
+   */
+  bool point (point_type &p) const
+  {
+    if (is_point ()) {
+      p = point ();
+      return true;
+    } else {
+      return false;
+    }
+  }
+
+  /**
+   *  @brief Alias for polymorphic expansion
+   *  Returns true, if the conversion was successful.
+   */
+  bool instantiate (point_type &p) const
+  {
+    return point (p);
+  }
+
+  /**
    *  @brief Return a reference to the text if one is referenced
    */
   const text_type &text () const
@@ -2439,6 +2659,16 @@ public:
    */
   box_type bbox () const;
 
+  /**
+   *  @brief Returns the box if the object represents a rectangle or an empty box if not
+   *
+   *  This method returns the rectangle (aka box) the shape represents a polygon
+   *  that is a rectangle, a path with two points and no rounded ends or an actual box.
+   *
+   *  If not, an empty box is returned.
+   */
+  box_type rectangle () const;
+
   /** 
    *  @brief Compute the area of the shape
    */
@@ -2523,7 +2753,7 @@ public:
    *
    *  Ordering of shapes is not specified by the identity of the objects but by the
    *  order of the pointers.
-   *  Since pointers are volatile objects, the ordering is not strictly reproducable!
+   *  Since pointers are volatile objects, the ordering is not strictly reproducible!
    */
   bool operator< (const Shape &d) const
   {
@@ -2547,6 +2777,11 @@ public:
     return m_trans < d.m_trans;
   }
 
+  /**
+   *  @brief Hash value
+   */
+  size_t hash_value () const;
+
   /** 
    *  @brief Convert to a string
    */
@@ -2567,6 +2802,8 @@ public:
     const text_ref_type *text_ref;
     const text_ptr_array_type *text_aref;
     const edge_type *edge;
+    const edge_pair_type *edge_pair;
+    const point_type *point;
     const path_type *path;
     const path_ref_type *path_ref;
     const path_ptr_array_type *path_aref;
@@ -2586,6 +2823,8 @@ public:
     const db::object_with_properties<text_ref_type> *ptext_ref;
     const db::object_with_properties<text_ptr_array_type> *ptext_aref;
     const db::object_with_properties<edge_type> *pedge;
+    const db::object_with_properties<edge_pair_type> *pedge_pair;
+    const db::object_with_properties<point_type> *ppoint;
     const db::object_with_properties<path_type> *ppath;
     const db::object_with_properties<path_ref_type> *ppath_ref;
     const db::object_with_properties<path_ptr_array_type> *ppath_aref;
@@ -2611,6 +2850,21 @@ public:
 };
 
 }  // namespace db
+
+namespace std
+{
+
+  //  provide a template specialization for std::hash<T>
+  template <>
+  struct hash <db::Shape>
+  {
+    size_t operator() (const db::Shape &s) const
+    {
+      return s.hash_value ();
+    }
+  };
+
+}  // namespace std
   
 #endif
 
