@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include "tlExpression.h"
 #include "tlProgress.h"
 #include "tlThreadedWorkers.h"
+#include "tlThreads.h"
 #include "gsiDecl.h"
 
 #include <cmath>
@@ -51,6 +52,31 @@ public:
     mp_shapes->insert (t.transformed (m_trans));
   }
 
+  template <class P>
+  void insert_polygon (const P &p)
+  {
+    if (p.is_box () && ! m_trans.is_complex ()) {
+      mp_shapes->insert (p.box ().transformed (m_trans));
+    } else {
+      if (mp_shapes->cell () && mp_shapes->cell ()->layout ()) {
+        db::polygon_ref<P, db::Disp> pr (p.transformed (m_trans), mp_shapes->cell ()->layout ()->shape_repository ());
+        mp_shapes->insert (pr);
+      } else {
+        mp_shapes->insert (p.transformed (m_trans));
+      }
+    }
+  }
+
+  void operator() (const db::Polygon &p)
+  {
+    insert_polygon (p);
+  }
+
+  void operator() (const db::SimplePolygon &p)
+  {
+    insert_polygon (p);
+  }
+
   void operator() (const db::EdgePair &ep)
   {
     mp_shapes->insert (ep.normalized ().to_polygon (m_ep_sizing).transformed (m_trans));
@@ -163,6 +189,34 @@ private:
   const db::ICplxTrans m_trans;
 };
 
+/**
+ *  @brief A helper class for the generic implementation of the text collection insert functionality
+ */
+class TextsInserter
+{
+public:
+  TextsInserter (db::Texts *texts, const db::ICplxTrans &trans)
+    : mp_texts (texts), m_trans (trans)
+  {
+    //  .. nothing yet ..
+  }
+
+  template <class T>
+  void operator() (const T &)
+  {
+    //  .. discard anything except Texts ..
+  }
+
+  void operator() (const db::Text &t)
+  {
+    mp_texts->insert (t.transformed (m_trans));
+  }
+
+private:
+  db::Texts *mp_texts;
+  const db::ICplxTrans m_trans;
+};
+
 class TileLayoutOutputReceiver
   : public db::TileOutputReceiver
 {
@@ -270,6 +324,26 @@ private:
   db::EdgePairs *mp_edge_pairs;
 };
 
+class TileTextsOutputReceiver
+  : public db::TileOutputReceiver
+{
+public:
+  TileTextsOutputReceiver (db::Texts *texts)
+    : mp_texts (texts)
+  {
+    //  .. nothing yet ..
+  }
+
+  void put (size_t /*ix*/, size_t /*iy*/, const db::Box &tile, size_t /*id*/, const tl::Variant &obj, double /*dbu*/, const db::ICplxTrans &trans, bool clip)
+  {
+    TextsInserter inserter (mp_texts, trans);
+    insert_var (inserter, obj, tile, clip);
+  }
+
+private:
+  db::Texts *mp_texts;
+};
+
 class TilingProcessorJob
   : public tl::JobBase
 {
@@ -278,11 +352,14 @@ public:
     : tl::JobBase (nworkers),
       mp_proc (proc),
       m_has_tiles (has_tiles),
-      m_progress (0)
+      m_progress_count (0),
+      m_progress (std::string ())
   {
     //  .. nothing yet ..
   }
 
+  void start (const std::string &job_description);
+
   bool has_tiles () const
   {
     return m_has_tiles;
@@ -290,19 +367,19 @@ public:
 
   void next_progress () 
   {
-    QMutexLocker locker (&m_mutex);
-    ++m_progress;
+    tl::MutexLocker locker (&m_mutex);
+    ++m_progress_count;
   }
 
-  void update_progress (tl::RelativeProgress &progress) 
+  void update_progress ()
   {
     unsigned int p;
     {
-      QMutexLocker locker (&m_mutex);
-      p = m_progress;
+      tl::MutexLocker locker (&m_mutex);
+      p = m_progress_count;
     }
 
-    progress.set (p, true /*force yield*/);
+    m_progress.set (p, true /*force yield*/);
   }
 
   TilingProcessor *processor () const
@@ -312,11 +389,14 @@ public:
 
   virtual tl::Worker *create_worker ();
 
+  virtual void after_sync_task (tl::Task *task);
+
 private:
   TilingProcessor *mp_proc;
   bool m_has_tiles;
-  unsigned int m_progress;
-  QMutex m_mutex;
+  unsigned int m_progress_count;
+  tl::Mutex m_mutex;
+  tl::RelativeProgress m_progress;
 };
 
 class TilingProcessorTask
@@ -394,6 +474,7 @@ private:
   TilingProcessorJob *mp_job;
 
   void do_perform (const TilingProcessorTask *task);
+  void make_input_var (const TilingProcessor::InputSpec &is, const db::RecursiveShapeIterator *iter, tl::Eval &eval, double sf);
 };
 
 class TilingProcessorReceiverFunction
@@ -406,7 +487,7 @@ public:
     //  .. nothing yet ..
   }
 
-  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant &out, const std::vector<tl::Variant> &args) const
+  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant &out, const std::vector<tl::Variant> &args, const std::map<std::string, tl::Variant> * /*kwargs*/) const
   {
     out = mp_proc->receiver (args);
   }
@@ -425,7 +506,7 @@ public:
     //  .. nothing yet ..
   }
 
-  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant & /*out*/, const std::vector<tl::Variant> &args) const 
+  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant & /*out*/, const std::vector<tl::Variant> &args, const std::map<std::string, tl::Variant> * /*kwargs*/) const
   {
     mp_proc->put (m_ix, m_iy, m_tile_box, args);
   }
@@ -445,12 +526,30 @@ public:
     //  .. nothing yet ..
   }
 
-  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant & /*out*/, const std::vector<tl::Variant> & /*args*/) const 
+  void execute (const tl::ExpressionParserContext & /*context*/, tl::Variant & /*out*/, const std::vector<tl::Variant> & /*args*/, const std::map<std::string, tl::Variant> * /*kwargs*/) const
   {
     // TODO: ... implement ..
   }
 };
 
+void
+TilingProcessorWorker::make_input_var (const TilingProcessor::InputSpec &is, const db::RecursiveShapeIterator *iter, tl::Eval &eval, double sf)
+{
+  if (! iter) {
+    iter = &is.iter;
+  }
+
+  if (is.type == TilingProcessor::TypeRegion) {
+    eval.set_var (is.name, tl::Variant (db::Region (*iter, db::ICplxTrans (sf) * is.trans, is.merged_semantics)));
+  } else if (is.type == TilingProcessor::TypeEdges) {
+    eval.set_var (is.name, tl::Variant (db::Edges (*iter, db::ICplxTrans (sf) * is.trans, is.merged_semantics)));
+  } else if (is.type == TilingProcessor::TypeEdgePairs) {
+    eval.set_var (is.name, tl::Variant (db::EdgePairs (*iter, db::ICplxTrans (sf) * is.trans)));
+  } else if (is.type == TilingProcessor::TypeTexts) {
+    eval.set_var (is.name, tl::Variant (db::Texts (*iter, db::ICplxTrans (sf) * is.trans)));
+  }
+}
+
 void
 TilingProcessorWorker::do_perform (const TilingProcessorTask *tile_task)
 {
@@ -491,11 +590,7 @@ TilingProcessorWorker::do_perform (const TilingProcessorTask *tile_task)
 
     if (! mp_job->has_tiles ()) { 
 
-      if (i->region) {
-        eval.set_var (i->name, tl::Variant (db::Region (i->iter, db::ICplxTrans (sf) * i->trans, i->merged_semantics)));
-      } else {
-        eval.set_var (i->name, tl::Variant (db::Edges (i->iter, db::ICplxTrans (sf) * i->trans, i->merged_semantics)));
-      }
+      make_input_var (*i, 0, eval, sf);
 
     } else {
 
@@ -508,11 +603,7 @@ TilingProcessorWorker::do_perform (const TilingProcessorTask *tile_task)
         iter.confine_region (region_dbu);
       }
 
-      if (i->region) {
-        eval.set_var (i->name, tl::Variant (db::Region (iter, db::ICplxTrans (sf) * i->trans, i->merged_semantics)));
-      } else {
-        eval.set_var (i->name, tl::Variant (db::Edges (iter, db::ICplxTrans (sf) * i->trans, i->merged_semantics)));
-      }
+      make_input_var (*i, &iter, eval, sf);
 
     }
 
@@ -541,9 +632,28 @@ TilingProcessorJob::create_worker ()
   return new TilingProcessorWorker (this);
 }
 
+void
+TilingProcessorJob::after_sync_task (tl::Task * /*task*/)
+{
+  //  This needs to be done here as there is no external loop to do this
+  update_progress ();
+}
+
+void
+TilingProcessorJob::start (const std::string &job_description)
+{
+  m_progress = tl::RelativeProgress (job_description, tasks (), 1);
+  //  prevents child progress objects from showing
+  m_progress.set_final (true);
+
+  tl::JobBase::start ();
+}
+
 // ----------------------------------------------------------------------------------
 //  The tiling processor implementation
 
+tl::Mutex TilingProcessor::s_output_lock;
+
 TilingProcessor::TilingProcessor ()
   : m_tile_width (0.0), m_tile_height (0.0),
     m_ntiles_w (0), m_ntiles_h (0), 
@@ -558,7 +668,7 @@ TilingProcessor::TilingProcessor ()
 }
 
 void 
-TilingProcessor::input (const std::string &name, const db::RecursiveShapeIterator &iter, const db::ICplxTrans &trans, bool as_region, bool merged_semantics)
+TilingProcessor::input (const std::string &name, const db::RecursiveShapeIterator &iter, const db::ICplxTrans &trans, Type type, bool merged_semantics)
 {
   if (m_inputs.empty () && iter.layout ()) {
     m_dbu = iter.layout ()->dbu ();
@@ -567,7 +677,7 @@ TilingProcessor::input (const std::string &name, const db::RecursiveShapeIterato
   m_inputs.back ().name = name;
   m_inputs.back ().iter = iter;
   m_inputs.back ().trans = trans;
-  m_inputs.back ().region = as_region;
+  m_inputs.back ().type = type;
   m_inputs.back ().merged_semantics = merged_semantics;
 }
 
@@ -698,7 +808,17 @@ TilingProcessor::output (const std::string &name, db::EdgePairs &edge_pairs)
   m_outputs.back ().receiver = new TileEdgePairsOutputReceiver (&edge_pairs);
 }
 
-void 
+void
+TilingProcessor::output (const std::string &name, db::Texts &texts)
+{
+  m_top_eval.set_var (name, m_outputs.size ());
+  m_outputs.push_back (OutputSpec ());
+  m_outputs.back ().name = name;
+  m_outputs.back ().id = 0;
+  m_outputs.back ().receiver = new TileTextsOutputReceiver (&texts);
+}
+
+void
 TilingProcessor::output (const std::string &name, db::Edges &edges)
 {
   m_top_eval.set_var (name, m_outputs.size ());
@@ -711,15 +831,15 @@ TilingProcessor::output (const std::string &name, db::Edges &edges)
 tl::Variant
 TilingProcessor::receiver (const std::vector<tl::Variant> &args)
 {
-  QMutexLocker locker (&m_output_mutex);
+  tl::MutexLocker locker (&s_output_lock);
 
   if (args.size () != 1) {
-    throw tl::Exception (tl::to_string (QObject::tr ("_rec function requires one argument: the handle of the output channel")));
+    throw tl::Exception (tl::to_string (tr ("_rec function requires one argument: the handle of the output channel")));
   }
 
   size_t index = args[0].to<size_t> ();
   if (index >= m_outputs.size ()) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Invalid handle in _rec function call")));
+    throw tl::Exception (tl::to_string (tr ("Invalid handle in _rec function call")));
   }
 
   gsi::Proxy *proxy = new gsi::Proxy (gsi::cls_decl<TileOutputReceiver> ());
@@ -733,17 +853,17 @@ TilingProcessor::receiver (const std::vector<tl::Variant> &args)
 void 
 TilingProcessor::put (size_t ix, size_t iy, const db::Box &tile, const std::vector<tl::Variant> &args)
 {
-  QMutexLocker locker (&m_output_mutex);
+  tl::MutexLocker locker (&s_output_lock);
 
   if (args.size () < 2 || args.size () > 3) {
-    throw tl::Exception (tl::to_string (QObject::tr ("_output function requires two or three arguments: handle and object and a clip flag (optional)")));
+    throw tl::Exception (tl::to_string (tr ("_output function requires two or three arguments: handle and object and a clip flag (optional)")));
   }
 
   bool clip = ((args.size () <= 2 || args [2].to_bool ()) && ! tile.empty ());
 
   size_t index = args[0].to<size_t> ();
   if (index >= m_outputs.size ()) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Invalid handle (first argument) in _output function call")));
+    throw tl::Exception (tl::to_string (tr ("Invalid handle (first argument) in _output function call")));
   }
 
   m_outputs[index].receiver->put (ix, iy, tile, m_outputs[index].id, args[1], dbu (), m_outputs[index].trans, clip);
@@ -756,7 +876,7 @@ TilingProcessor::execute (const std::string &desc)
 
   if (tot_box.empty ()) {
     for (std::vector<InputSpec>::const_iterator i = m_inputs.begin (); i != m_inputs.end (); ++i) {
-      if (! i->iter.at_end ()) {
+      if (! i->iter.at_end_no_lock ()) {
         if (scale_to_dbu ()) {
           double dbu_value = i->iter.layout () ? i->iter.layout ()->dbu () : dbu ();
           tot_box += i->iter.bbox ().transformed (db::CplxTrans (dbu_value) * db::CplxTrans (i->trans));
@@ -813,7 +933,7 @@ TilingProcessor::execute (const std::string &desc)
   //  is just a single tile.
   bool has_tiles = (ntiles_w > 1 || ntiles_h > 1 || ! m_frame.empty ());
 
-  TilingProcessorJob job (this, m_threads, has_tiles);
+  TilingProcessorJob job (this, int (m_threads), has_tiles);
 
   double l = 0.0, b = 0.0;
 
@@ -860,11 +980,6 @@ TilingProcessor::execute (const std::string &desc)
 
   }
 
-  //  TODO: there should be a general scheme of how thread-specific progress is merged
-  //  into a global one ..
-  size_t todo_count = ntiles_w * ntiles_h * m_scripts.size ();
-  tl::RelativeProgress progress (desc, todo_count, 1);
-
   try {
 
     try {
@@ -876,10 +991,10 @@ TilingProcessor::execute (const std::string &desc)
         }
       }
 
-      job.start ();
+      job.start (desc);
       while (job.is_running ()) {
         //  This may throw an exception, if the cancel button has been pressed.
-        job.update_progress (progress);
+        job.update_progress ();
         job.wait (100);
       }
 
@@ -909,7 +1024,7 @@ TilingProcessor::execute (const std::string &desc)
   }
 
   if (job.has_error ()) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Errors occured during processing. First error message says:\n")) + job.error_messages ().front ());
+    throw tl::Exception (tl::to_string (tr ("Errors occurred during processing. First error message says:\n")) + job.error_messages ().front ());
   }
 }
 
