@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@
 
 #include "tlExpression.h"
 #include "tlTypeTraits.h"
+#include "tlThreads.h"
 
 namespace db
 {
@@ -106,7 +107,7 @@ public:
 
   /**
    *  @brief Indicate the end of the execution
-   *  @param sucess Will be true if all tiles executed successfully
+   *  @param success Will be true if all tiles executed successfully
    */
   virtual void finish (bool /*success*/) { }
 
@@ -134,6 +135,27 @@ private:
   TilingProcessor *mp_proc;
 };
 
+/**
+ *  @brief Delivery of tiling processor output
+ *  This utility is put between the container and the receiver.
+ *  The inserter is an object having an operator() that takes the object.
+ *  This function is responsible for preparing (i.e. clipping) and delivering the output.
+ */
+template <class X>
+void insert (X &inserter, const db::Box &o, const db::Box &tile, bool clip)
+{
+  if (clip) {
+    //  clipping
+    db::Box oc = o & tile;
+    if (! oc.empty () && oc.width () > 0 && oc.height () > 0) {
+      inserter (oc);
+    }
+  } else {
+    //  no clipping
+    inserter (o);
+  }
+}
+
 /**
  *  @brief Delivery of tiling processor output
  *  This utility is put between the container and the receiver.
@@ -143,7 +165,10 @@ private:
 template <class X>
 void insert (X &inserter, const db::Polygon &o, const db::Box &tile, bool clip)
 {
-  if (clip && ! o.box ().inside (tile)) {
+  if (o.is_box ()) {
+    //  simple case: box
+    insert (inserter, o.box (), tile, clip);
+  } else if (clip && ! o.box ().inside (tile)) {
     //  apply clipping
     if (o.box ().touches (tile)) {
       std::vector <db::Polygon> clipped_poly;
@@ -167,7 +192,10 @@ void insert (X &inserter, const db::Polygon &o, const db::Box &tile, bool clip)
 template <class X>
 void insert (X &inserter, const db::SimplePolygon &o, const db::Box &tile, bool clip)
 {
-  if (clip && ! o.box ().inside (tile)) {
+  if (o.is_box ()) {
+    //  simple case: box
+    insert (inserter, o.box (), tile, clip);
+  } else if (clip && ! o.box ().inside (tile)) {
     //  apply clipping
     if (o.box ().touches (tile)) {
       std::vector <db::SimplePolygon> clipped_poly;
@@ -241,27 +269,6 @@ void insert (X &inserter, const db::Text &o, const db::Box &tile, bool clip)
   }
 }
 
-/**
- *  @brief Delivery of tiling processor output
- *  This utility is put between the container and the receiver.
- *  The inserter is an object having an operator() that takes the object.
- *  This function is responsible for preparing (i.e. clipping) and delivering the output.
- */
-template <class X>
-void insert (X &inserter, const db::Box &o, const db::Box &tile, bool clip)
-{
-  if (clip) {
-    //  clipping
-    db::Box oc = o & tile;
-    if (! oc.empty ()) {
-      inserter (oc);
-    }
-  } else {
-    //  no clipping
-    inserter (o);
-  }
-}
-
 /**
  *  @brief Delivery of tiling processor output
  *  This utility is put between the container and the receiver.
@@ -340,7 +347,21 @@ void insert (X &inserter, const db::Edges &data, const db::Box &tile, bool clip)
 template <class X>
 void insert (X &inserter, const db::EdgePairs &data, const db::Box &tile, bool clip)
 {
-  for (db::EdgePairs::const_iterator o = data.begin (); o != data.end (); ++o) {
+  for (db::EdgePairs::const_iterator o = data.begin (); ! o.at_end (); ++o) {
+    insert (inserter, *o, tile, clip);
+  }
+}
+
+/**
+ *  @brief Delivery of tiling processor output
+ *  This utility is put between the container and the receiver.
+ *  The inserter is an object having an operator() that takes the object.
+ *  This function is responsible for preparing (i.e. clipping) and delivering the output.
+ */
+template <class X>
+void insert (X &inserter, const db::Texts &data, const db::Box &tile, bool clip)
+{
+  for (db::Texts::const_iterator o = data.begin (); ! o.at_end (); ++o) {
     insert (inserter, *o, tile, clip);
   }
 }
@@ -363,6 +384,9 @@ bool insert_var (X &inserter, const tl::Variant &obj, const db::Box &tile, bool
   } else if (obj.is_user<db::Edges> ()) {
     insert (inserter, obj.to_user<db::Edges> (), tile, clip);
     return true;
+  } else if (obj.is_user<db::Texts> ()) {
+    insert (inserter, obj.to_user<db::Texts> (), tile, clip);
+    return true;
   } else if (obj.is_user<db::Box> ()) {
     insert (inserter, obj.to_user<db::Box> (), tile, clip);
     return true;
@@ -416,6 +440,8 @@ bool insert_var (X &inserter, const tl::Variant &obj, const db::Box &tile, bool
 class DB_PUBLIC TilingProcessor
 {
 public:
+  enum Type { TypeRegion, TypeEdges, TypeEdgePairs, TypeTexts };
+
   /**
    *  @brief Constructor
    */
@@ -440,7 +466,7 @@ public:
    *  The transformation can be used to convert between database units.  
    *  If "as_region" is false, the input is taken as edge input.
    */
-  void input (const std::string &name, const db::RecursiveShapeIterator &iter, const db::ICplxTrans &trans = db::ICplxTrans (), bool as_region = true, bool merged_semantics = true);
+  void input (const std::string &name, const db::RecursiveShapeIterator &iter, const db::ICplxTrans &trans = db::ICplxTrans (), Type type = TypeRegion, bool merged_semantics = true);
 
   /**
    *  @brief Specifies the output 
@@ -488,6 +514,14 @@ public:
    */
   void output (const std::string &name, db::EdgePairs &edge_pairs);
 
+  /**
+   *  @brief Specifies output to an text collection
+   *
+   *  This version will specify output to a Texts object.
+   *  Only texts will be stored.
+   */
+  void output (const std::string &name, db::Texts &texts);
+
   /**
    *  @brief Specifies output to an edge collection
    *
@@ -616,6 +650,14 @@ public:
    */
   void execute (const std::string &desc);
 
+  /**
+   *  @brief Gets the output mutex for operations not using the output method
+   */
+  static tl::Mutex &output_lock ()
+  {
+    return s_output_lock;
+  }
+
 private:
   friend class TilingProcessorWorker;
   friend class TilingProcessorOutputFunction;
@@ -623,11 +665,11 @@ private:
 
   struct InputSpec
   {
-    InputSpec () : region (false), merged_semantics (false) { }
+    InputSpec () : type (TilingProcessor::TypeRegion), merged_semantics (false) { }
     std::string name;
     db::RecursiveShapeIterator iter;
     db::ICplxTrans trans;
-    bool region;
+    TilingProcessor::Type type;
     bool merged_semantics;
   };
 
@@ -661,21 +703,11 @@ private:
   bool m_dbu_specific_set;
   bool m_scale_to_dbu;
   std::vector<std::string> m_scripts;
-  QMutex m_output_mutex;
   tl::Eval m_top_eval;
+  static tl::Mutex s_output_lock;
 };
 
 }
 
-namespace tl
-{
-  template <>
-  struct type_traits<db::TilingProcessor> : public type_traits<void>
-  {
-    typedef tl::true_tag has_default_constructor;
-    typedef tl::false_tag has_copy_constructor;
-  };
-}
-
 #endif
 
