@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -150,11 +150,19 @@ struct unit_trans
     // .. nothing else ..
   }
 
+  /**
+   *  @brief Copy ctor (which basically does nothing)
+   */
+  unit_trans (const unit_trans<C> &)
+  {
+    // .. nothing else ..
+  }
+
   /**
    *  @brief Copy ctor (which basically does nothing)
    */
   template <class D>
-  explicit unit_trans (const unit_trans<D> &)
+  unit_trans (const unit_trans<D> &)
   {
     // .. nothing else ..
   }
@@ -240,6 +248,15 @@ struct unit_trans
     return *this;
   }
 
+  /**
+   *  @brief Assignment (which basically does nothing)
+   */
+  template <class D>
+  unit_trans &operator= (const unit_trans<D> &) 
+  {
+    return *this;
+  }
+
   /**
    *  @brief Inversion
    */
@@ -359,6 +376,32 @@ public:
     // .. nothing else ..
   }
 
+  /**
+   *  @brief Reduction
+   */
+  template <class T>
+  explicit fixpoint_trans (const T &t)
+    : m_f (0)
+  {
+    *this = t.fp_trans ();
+  }
+
+  /**
+   *  @brief Reduction from a matrix2d
+   */
+  explicit fixpoint_trans (const db::matrix_2d<coord_type> &t)
+  {
+    m_f = ((int (floor (t.angle () / 90.0 + 0.5) + 4)) % 4) + (t.is_mirror () ? 4 : 0);
+  }
+
+  /**
+   *  @brief Reduction from a matrix3d
+   */
+  explicit fixpoint_trans (const db::matrix_3d<coord_type> &t)
+  {
+    m_f = ((int (floor (t.angle () / 90.0 + 0.5) + 4)) % 4) + (t.is_mirror () ? 4 : 0);
+  }
+
   /**
    *  @brief Returns true, if the transformation is unity
    */
@@ -403,7 +446,18 @@ public:
     // .. nothing else ..
   }
 
-  /** 
+  /**
+   *  @brief The standard constructor using a code rather than angle and mirror and no displacement
+   *
+   *  @param f The rotation/mirror code (r0 .. m135 constants)
+   */
+  explicit fixpoint_trans (unsigned int f)
+    : m_f (f)
+  {
+    // .. nothing else ..
+  }
+
+  /**
    *  @brief The rotation/mirror codes
    */
   static const int r0   = 0;  //  No rotation
@@ -730,22 +784,31 @@ public:
   /**
    *  @brief The "conversion" from the unit transformation to a displacement
    */
-  explicit disp_trans (unit_trans<C>)
+  disp_trans (unit_trans<C>)
     : m_u ()
   {
     // .. nothing else ..
   }
 
+  /**
+   *  @brief The copy constructor 
+   *
+   *  @param d The source from which to copy
+   */
+  disp_trans (const disp_trans<C> &d)
+    : m_u (d.disp ())
+  { }
+
   /**
    *  @brief The copy constructor that converts also
    *
-   *  The copy constructor allows to convert between different
+   *  The copy constructor allows converting between different
    *  coordinate types, if possible.
    *
    *  @param d The source from which to copy
    */
   template <class D>
-  explicit disp_trans (const disp_trans<D> &d)
+  disp_trans (const disp_trans<D> &d)
     : m_u (d.disp ())
   { }
 
@@ -779,6 +842,25 @@ public:
     // .. nothing else ..
   }
 
+  /**
+   *  @brief Assignment
+   */
+  disp_trans &operator= (const disp_trans<C> &d)
+  {
+    m_u = d.disp ();
+    return *this;
+  }
+
+  /**
+   *  @brief Assignment with type conversion
+   */
+  template <class D>
+  disp_trans &operator= (const disp_trans<D> &d)
+  {
+    m_u = d.disp ();
+    return *this;
+  }
+
   /**
    *  @brief Returns true, if the transformation is unity
    */
@@ -1090,10 +1172,22 @@ public:
     : fixpoint_trans<C> (d.rot ()), m_u (d.disp ())
   { }
 
+  /**
+   *  @brief Assignment
+   *
+   *  @param d The source from which to take the data
+   */
+  simple_trans &operator= (const simple_trans<C> &d)
+  { 
+    fixpoint_trans<C>::operator= (d);
+    m_u = d.disp ();
+    return *this;
+  }
+
   /**
    *  @brief The copy constructor that converts to a different coordinate type also
    *
-   *  The copy constructor allows to convert between different
+   *  The copy constructor allows converting between different
    *  coordinate types, if possible.
    *
    *  @param d The source from which to copy
@@ -1103,6 +1197,22 @@ public:
     : fixpoint_trans<C> (d.rot ()), m_u (d.disp ())
   { }
 
+  /**
+   *  @brief Assignment which also converts
+   *
+   *  This assignment implementation will also convert
+   *  between different coordinate types if possible.
+   *
+   *  @param d The source from which to take the data
+   */
+  template <class D>
+  simple_trans &operator= (const simple_trans<D> &d)
+  { 
+    fixpoint_trans<C>::operator= (d);
+    m_u = d.disp ();
+    return *this;
+  }
+
   /**
    *  @brief The standard constructor using angle and mirror flag
    *
@@ -1362,10 +1472,10 @@ public:
   /**
    *  @brief String conversion
    */
-  std::string to_string () const
+  std::string to_string (double dbu = 0.0) const
   {
     std::string s1 = fixpoint_trans<C>::to_string ();
-    std::string s2 = m_u.to_string ();
+    std::string s2 = m_u.to_string (dbu);
     if (! s1.empty () && ! s2.empty ()) {
       return s1 + " " + s2;
     } else {
@@ -1433,7 +1543,7 @@ operator<< (std::ostream &os, const simple_trans<C> &t)
 /**
  *  @brief A complex transformation
  *
- *  A complex transformation provides magnification, mirroring at the x-axis, rotation by an arbitary
+ *  A complex transformation provides magnification, mirroring at the x-axis, rotation by an arbitrary
  *  angle and a displacement. The template parameters for this transformation are
  *  I (the input coordinate type), F (the output coordinate type) and R (the representation
  *  type used internally for representing the floating-point members).
@@ -1533,6 +1643,13 @@ public:
   complex_trans (const simple_trans<I> &s, double acos, double mag)
     : m_u (s.disp ())
   {
+    //  This prevents rounding issues (like acos = 1.0000000000002):
+    if (acos > 1.0) {
+      acos = 1.0;
+    } else if (acos < -1.0) {
+      acos = -1.0;
+    }
+
     db::point<R> t (1.0, 0.0);
     t = fixpoint_trans<R> (s.fp_trans ()) (t);
     double asin = sqrt (1.0 - acos * acos); // we may to this since we know that the angle is between 0 and 90 degree
@@ -1571,7 +1688,7 @@ public:
   {
     tl_assert (! m.has_shear ());
     tl_assert (! m.has_perspective ());
-    std::pair<double, double> mag = m.mag ();
+    std::pair<double, double> mag = m.mag2 ();
     tl_assert (fabs (mag.first - mag.second) < 1e-10);
     double rot = m.angle () * M_PI / 180.0;
     m_mag = m.is_mirror () ? -mag.first : mag.first;
@@ -1594,7 +1711,7 @@ public:
     : m_u (u)
   {
     tl_assert (! m.has_shear ());
-    std::pair<double, double> mag = m.mag ();
+    std::pair<double, double> mag = m.mag2 ();
     tl_assert (fabs (mag.first - mag.second) < 1e-10);
     double rot = m.angle () * M_PI / 180.0;
     m_mag = m.is_mirror () ? -mag.first : mag.first;
@@ -1618,7 +1735,7 @@ public:
   /**
    *  @brief The copy constructor that converts also
    *
-   *  The copy constructor allows to convert between different
+   *  The copy constructor allows converting between different
    *  coordinate types, if possible.
    *
    *  @param d The source from which to copy
@@ -1894,6 +2011,15 @@ public:
     m_mag = m_mag < 0.0 ? -m : m;
   }
 
+  /**
+   *  @brief Returns a value indicating whether the transformation is a complex one
+   *  The transformation can safely be converted to a simple transformation if this value is false.
+   */
+  bool is_complex () const
+  {
+    return is_mag () || ! is_ortho ();
+  }
+
   /** 
    *  @brief Test, if the transformation is mirroring
    */
@@ -2116,22 +2242,14 @@ public:
     return ! equal (t);
   }
 
-  /**
-   *  @brief Default string conversion
-   */
-  std::string to_string () const
-  {
-    return to_string (false);
-  }
-
   /**
    *  @brief String conversion
    *
-   *  The lazy and micron flags allow to customize the output somewhat.
+   *  The lazy and micron flags allow customization of the output to some degree.
    *  When lazy is set to true, output that is not required (i.e. magnification when 1)
    *  is dropped. If dbu is set, the coordinates are multiplied with this factor to render micron units.
    */
-  std::string to_string (bool lazy, double dbu = 0.0) const
+  std::string to_string (bool lazy = false, double dbu = 0.0) const
   {
     std::string s;
     if (is_mirror ()) {
@@ -2444,7 +2562,7 @@ typedef complex_trans<db::Coord, db::DCoord> CplxTrans;
  *  @brief Specialization: concatenation of CplxTrans
  *
  *  The combination of two of these objects is basically not allowed, since the
- *  output and input types in not compatible. For sake of similicity however, we
+ *  output and input types in not compatible. For sake of simplicity however, we
  *  allow this now.
  */
 inline CplxTrans operator* (const CplxTrans &a, const CplxTrans &b)
@@ -2461,7 +2579,7 @@ typedef complex_trans<db::DCoord, db::Coord> VCplxTrans;
  *  @brief Specialization: concatenation of VCplxTrans
  *
  *  The combination of two of these objects is basically not allowed, since the
- *  output and input types in not compatible. For sake of similicity however, we
+ *  output and input types in not compatible. For sake of simplicity however, we
  *  allow this now.
  */
 inline VCplxTrans operator* (const VCplxTrans &a, const VCplxTrans &b)
@@ -2553,6 +2671,7 @@ namespace tl
   template<> DB_PUBLIC void extractor_impl (tl::Extractor &ex, db::CplxTrans &t);
   template<> DB_PUBLIC void extractor_impl (tl::Extractor &ex, db::VCplxTrans &t);
   template<> DB_PUBLIC void extractor_impl (tl::Extractor &ex, db::DCplxTrans &t);
+  template<> DB_PUBLIC void extractor_impl (tl::Extractor &ex, db::ICplxTrans &t);
 
   template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::UnitTrans &t);
   template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::DUnitTrans &t);
@@ -2565,6 +2684,7 @@ namespace tl
   template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::CplxTrans &t);
   template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::VCplxTrans &t);
   template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::DCplxTrans &t);
+  template<> DB_PUBLIC bool test_extractor_impl (tl::Extractor &ex, db::ICplxTrans &t);
 
 } // namespace tl
 
