@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -24,14 +24,6 @@
 #include "dbUserObject.h"
 #include "tlClassRegistry.h"
 
-namespace tl
-{
-  template<> DB_PUBLIC tl::Registrar<db::user_object_factory_base<db::Coord> > *tl::Registrar<db::user_object_factory_base<db::Coord> >::instance = 0;
-  template class DB_PUBLIC tl::RegisteredClass<db::user_object_factory_base<db::Coord> >;
-  template<> DB_PUBLIC tl::Registrar<db::user_object_factory_base<db::DCoord> > *tl::Registrar<db::user_object_factory_base<db::DCoord> >::instance = 0;
-  template class DB_PUBLIC tl::RegisteredClass<db::user_object_factory_base<db::DCoord> >;
-}
-
 namespace db
 {
   static unsigned int s_class_id = 1;
