@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -33,7 +33,7 @@
 namespace drc
 {
   DRC_PUBLIC int _force_link_f ();
-  static int _force_link_target = _force_link_f ();
+  int _force_link_target = _force_link_f ();
 }
 
 #endif
