@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -24,11 +24,13 @@
 #ifndef _HDR_gsiClass
 #define _HDR_gsiClass
 
+#include "gsiClassBase.h"
 #include "gsiObject.h"
 
 #include "tlTypeTraits.h"
 #include "tlHeap.h"
 #include "tlUtils.h"
+#include "tlVariant.h"
 
 #include <memory>
 
@@ -43,74 +45,104 @@ namespace gsi
 /**
  *  @brief A helper function to implement equal as efficiently as possible
  */
-template<class T, class I>
-bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, I);
+template<class T, bool> struct _var_user_equal_impl;
 
 template<class T>
-bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl * /*delegate*/, tl::true_tag)
+struct _var_user_equal_impl<T, true>
 {
-  return *a == *b;
-}
+  static bool call (const T *a, const T *b, const VariantUserClassImpl * /*delegate*/) { return *a == *b; }
+};
 
 template<class T>
-bool _var_user_equal_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::false_tag)
+struct _var_user_equal_impl<T, false>
 {
-  return delegate->equal_impl ((void *) a, (void *) b);
-}
+  static bool call (const T *a, const T *b, const VariantUserClassImpl *delegate) { return delegate->equal_impl ((void *) a, (void *) b); }
+};
 
 /**
- *  @brief A helper function to implement less as efficiently as possible
+ *  @brief A helper function to implement equal as efficiently as possible
  */
-template<class T, class I>
-bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, I);
+template<class T, bool> struct _var_user_less_impl;
 
 template<class T>
-bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl * /*delegate*/, tl::true_tag)
+struct _var_user_less_impl<T, true>
 {
-  return *a < *b;
-}
+  static bool call (const T *a, const T *b, const VariantUserClassImpl * /*delegate*/) { return *a < *b; }
+};
 
 template<class T>
-bool _var_user_less_impl (const T *a, const T *b, const VariantUserClassImpl *delegate, tl::false_tag)
+struct _var_user_less_impl<T, false>
 {
-  return delegate->less_impl ((void *) a, (void *) b);
-}
+  static bool call (const T *a, const T *b, const VariantUserClassImpl *delegate) { return delegate->less_impl ((void *) a, (void *) b); }
+};
 
 /**
  *  @brief A helper function to implement to_string as efficiently as possible
  */
-template<class T, class I>
-std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl *delegate, I);
+template<class T, bool> struct _var_user_to_string_impl;
 
 template<class T>
-std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl * /*delegate*/, tl::true_tag)
+struct _var_user_to_string_impl<T, true>
 {
-  return a->to_string ();
-}
+  static std::string call (const T *a, const VariantUserClassImpl * /*delegate*/) { return a ? a->to_string () : std::string (); }
+};
 
 template<class T>
-std::string _var_user_to_string_impl (const T *a, const VariantUserClassImpl *delegate, tl::false_tag)
+struct _var_user_to_string_impl<T, false>
 {
-  return delegate->to_string_impl ((void *) a);
-}
+  static std::string call (const T *a, const VariantUserClassImpl *delegate) { return delegate->to_string_impl ((void *) a); }
+};
 
 /**
- *  @brief A helper function to implement read as efficiently as possible
+ *  @brief A helper function to implement to_variant as efficiently as possible
  */
-template<class T, class I>
-void _var_user_read_impl (T *a, tl::Extractor &ex, I);
+template<class T, bool> struct _var_user_to_variant_impl;
 
 template<class T>
-void _var_user_read_impl (T *a, tl::Extractor &ex, tl::true_tag)
+struct _var_user_to_variant_impl<T, true>
 {
-  ex.read (*a);
-}
+  static tl::Variant call (const T *a, const VariantUserClassImpl * /*delegate*/) { return a->to_variant (); }
+};
 
 template<class T>
-void _var_user_read_impl (T * /*a*/, tl::Extractor & /*ex*/, tl::false_tag)
+struct _var_user_to_variant_impl<T, false>
 {
-  tl_assert (false);
-}
+  static tl::Variant call (const T *a, const VariantUserClassImpl *delegate) { return delegate->to_variant_impl ((void *) a); }
+};
+
+/**
+ *  @brief A helper function to implement to_int as efficiently as possible
+ */
+template<class T, bool> struct _var_user_to_int_impl;
+
+template<class T>
+struct _var_user_to_int_impl<T, true>
+{
+  static int call (const T *a, const VariantUserClassImpl * /*delegate*/) { return a->to_int (); }
+};
+
+template<class T>
+struct _var_user_to_int_impl<T, false>
+{
+  static int call (const T *a, const VariantUserClassImpl *delegate) { return delegate->to_int_impl ((void *) a); }
+};
+
+/**
+ *  @brief A helper function to implement to_double as efficiently as possible
+ */
+template<class T, bool> struct _var_user_to_double_impl;
+
+template<class T>
+struct _var_user_to_double_impl<T, true>
+{
+  static double call (const T *a, const VariantUserClassImpl * /*delegate*/) { return a->to_double (); }
+};
+
+template<class T>
+struct _var_user_to_double_impl<T, false>
+{
+  static double call (const T *a, const VariantUserClassImpl *delegate) { return delegate->to_double_impl ((void *) a); }
+};
 
 /**
  *  @brief A VariantUserClassBase specialization that links GSI classes and Variant classes
@@ -164,20 +196,32 @@ public:
 
   virtual bool equal (const void *a, const void *b) const
   {
-    typename tl::type_traits<T>::has_equal_operator f;
-    return gsi::_var_user_equal_impl ((const T *) a, (const T *) b, this, f);
+    return gsi::_var_user_equal_impl<T, tl::has_equal_operator<T>::value>::call ((const T *) a, (const T *) b, this);
   }
 
   virtual bool less (const void *a, const void *b) const
   {
-    typename tl::type_traits<T>::has_less_operator f;
-    return gsi::_var_user_less_impl ((const T *) a, (const T *) b, this, f);
+    return gsi::_var_user_less_impl<T, tl::has_less_operator<T>::value>::call ((const T *) a, (const T *) b, this);
+  }
+
+  virtual void to_variant (const void *a, tl::Variant &var) const
+  {
+    var = gsi::_var_user_to_variant_impl<T, tl::has_to_variant<T>::value>::call ((const T *) a, this);
   }
 
   virtual std::string to_string (const void *a) const
   {
-    typename tl::type_traits<T>::supports_to_string f;
-    return gsi::_var_user_to_string_impl ((const T *) a, this, f);
+    return gsi::_var_user_to_string_impl<T, tl::has_to_string<T>::value>::call ((const T *) a, this);
+  }
+
+  virtual int to_int (const void *a) const
+  {
+    return gsi::_var_user_to_int_impl<T, tl::has_to_int<T>::value>::call ((const T *) a, this);
+  }
+
+  virtual double to_double (const void *a) const
+  {
+    return gsi::_var_user_to_double_impl<T, tl::has_to_double<T>::value>::call ((const T *) a, this);
   }
 
   void *clone (const void *obj) const
@@ -211,8 +255,8 @@ public:
 
   void read (void *a, tl::Extractor &ex) const
   {
-    typename tl::type_traits<T>::supports_extractor f;
-    gsi::_var_user_read_impl ((T *) a, ex, f);
+    T *t = (T *) a;
+    ex.read (*t);
   }
 
   const gsi::ClassBase *gsi_cls () const
@@ -243,256 +287,21 @@ private:
 // -----------------------------------------------------------------------------
 //  GSI implementation
 
-template <class X> 
-void *_get_vector_of (SerialArgs & /*from*/, const ArgType & /*a*/, void * /*data*/, void (* /*cb*/) (void * /*data*/, void * /*obj*/), tl::false_tag /*has_copy_ctor*/) 
-{
-  tl_assert (false); // cannot copy object of this type
-  return 0;
-}
+template <class X, bool> struct _destroy;
+template <class X> struct _destroy<X, false> { static void call (X *) { tl_assert (false); } };
+template <class X> struct _destroy<X, true> { static void call (X *x) { delete x; } };
 
-template <class X> 
-void _get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj), tl::true_tag /*has_copy_ctor*/) 
-{
-  std::vector<X> vv;
-  const std::vector<X> *v = &vv;
-  if (a.is_cref ()) {
-    v = &from.template read<const std::vector<X> &> ();
-  } else if (a.is_cptr ()) {
-    v = from.template read<const std::vector<X> *> ();
-  } else if (a.is_ref ()) {
-    v = &from.template read<std::vector<X> &> ();
-  } else if (a.is_ptr ()) {
-    v = from.template read<std::vector<X> *> ();
-  } else {
-    vv = from.template read< std::vector<X> > ();
-  }
-  for (typename std::vector<X>::const_iterator o = v->begin (); o != v->end (); ++o) {
-    (*cb) (data, new X (*o));
-  }
-}
+template <class X, bool> struct _create;
+template <class X> struct _create<X, false> { static void *call () { throw tl::Exception (tl::to_string (tr ("Object cannot be created here"))); } };
+template <class X> struct _create<X, true> { static void *call () { return new X (); } };
 
-template <class X> 
-void _get_cptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
-{
-  std::vector<const X *> vv;
-  const std::vector<const X *> *v = &vv;
-  if (a.is_cref ()) {
-    v = &from.template read<const std::vector<const X *> &> ();
-  } else if (a.is_cptr ()) {
-    v = from.template read<const std::vector<const X *> *> ();
-  } else if (a.is_ref ()) {
-    v = &from.template read<std::vector<const X *> &> ();
-  } else if (a.is_ptr ()) {
-    v = from.template read<std::vector<const X *> *> ();
-  } else {
-    vv = from.template read< std::vector<const X *> > ();
-  }
-  for (typename std::vector<const X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
-    (*cb) (data, (void *) *o);
-  }
-}
+template <class X, bool> struct _clone;
+template <class X> struct _clone<X, false> { static void *call (const void *) { throw tl::Exception (tl::to_string (tr ("Object cannot be copied here"))); } };
+template <class X> struct _clone<X, true> { static void *call (const void *other) { return new X (*(const X *)other); } };
 
-template <class X> 
-void _get_ptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
-{
-  std::vector<X *> vv;
-  const std::vector<X *> *v = &vv;
-  if (a.is_cref ()) {
-    v = &from.template read<const std::vector<X *> &> ();
-  } else if (a.is_cptr ()) {
-    v = from.template read<const std::vector<X *> *> ();
-  } else if (a.is_ref ()) {
-    v = &from.template read<std::vector<X *> &> ();
-  } else if (a.is_ptr ()) {
-    v = from.template read<std::vector<X *> *> ();
-  } else {
-    vv = from.template read< std::vector<X *> > ();
-  }
-  for (typename std::vector<X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
-    (*cb) (data, *o);
-  }
-}
-
-template <class X> 
-void _destroy (X *, tl::false_tag /*has public dtor*/)
-{
-  tl_assert (false); // cannot delete object of this type
-}
-
-template <class X> 
-void _destroy (X *x, tl::true_tag /*has public dtor*/)
-{
-  delete x;
-}
-
-template <class X> 
-void _push_vector_of (SerialArgs & /*to*/, const ArgType & /*a*/, tl::Heap & /*heap*/, const std::vector<void *> & /*objects*/, tl::false_tag /*has_copy_ctor*/) 
-{
-  tl_assert (false); // cannot copy object of this type
-}
-
-template <class X> 
-void _push_vector_of (SerialArgs &to, const ArgType &a, tl::Heap &heap, const std::vector<void *> &objects, tl::true_tag /*has_copy_ctor*/) 
-{
-  tl_assert (a.inner () != 0);
-
-  std::vector<X> vv;
-  std::vector<X> *v;
-  if (a.is_ref () || a.is_cref () || a.is_ptr () || a.is_cptr ()) {
-    v = new std::vector<X> ();
-    heap.push (v);
-  } else {
-    v = &vv;
-  }
-
-  v->reserve (objects.size ());
-  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
-    v->push_back (*(X *)*o);
-  }
-
-  if (a.is_cref ()) {
-    to.write<const std::vector<X> &> (*v);
-  } else if (a.is_cptr ()) {
-    to.write<const std::vector<X> *> (v);
-  } else if (a.is_ref ()) {
-    to.write<std::vector<X> &> (*v);
-  } else if (a.is_ptr ()) {
-    to.write<std::vector<X> *> (v);
-  } else {
-    to.write<std::vector<X> > (vv);
-  }
-}
-
-template <class X> 
-void _push_cptr_vector_of (SerialArgs &to, const ArgType &a, tl::Heap &heap, const std::vector<void *> &objects)
-{
-  tl_assert (a.inner () != 0);
-
-  std::vector<const X *> vv;
-  std::vector<const X *> *v;
-  if (a.is_ref () || a.is_cref () || a.is_ptr () || a.is_cptr ()) {
-    v = new std::vector<const X *> ();
-    heap.push (v);
-  } else {
-    v = &vv;
-  }
-
-  v->reserve (objects.size ());
-  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
-    v->push_back ((const X *)*o);
-  }
-
-  if (a.is_cref ()) {
-    to.write<const std::vector<const X *> &> (*v);
-  } else if (a.is_cptr ()) {
-    to.write<const std::vector<const X *> *> (v);
-  } else if (a.is_ref ()) {
-    to.write<std::vector<const X *> &> (*v);
-  } else if (a.is_ptr ()) {
-    to.write<std::vector<const X *> *> (v);
-  } else {
-    to.write<std::vector<const X *> > (vv);
-  }
-}
-
-template <class X> 
-void _push_ptr_vector_of (SerialArgs &to, const ArgType &a, tl::Heap &heap, const std::vector<void *> &objects)
-{
-  tl_assert (a.inner () != 0);
-
-  std::vector<X *> vv;
-  std::vector<X *> *v;
-  if (a.is_ref () || a.is_cref () || a.is_ptr () || a.is_cptr ()) {
-    v = new std::vector<X *> ();
-    heap.push (v);
-  } else {
-    v = &vv;
-  }
-
-  v->reserve (objects.size ());
-  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
-    v->push_back ((X *)*o);
-  }
-
-  if (a.is_cref ()) {
-    to.write<const std::vector<X *> &> (*v);
-  } else if (a.is_cptr ()) {
-    to.write<const std::vector<X *> *> (v);
-  } else if (a.is_ref ()) {
-    to.write<std::vector<X *> &> (*v);
-  } else if (a.is_ptr ()) {
-    to.write<std::vector<X *> *> (v);
-  } else {
-    to.write<std::vector<X *> > (vv);
-  }
-}
-
-template <class X> 
-void *_create (tl::false_tag)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Object cannot be created here")));
-  return 0;
-}
-
-template <class X> 
-void *_create (tl::true_tag)
-{
-  return new X ();
-}
-
-template <class X> 
-void *_clone (tl::false_tag, const void * /*other*/)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Object cannot be copied here")));
-  return 0;
-}
-
-template <class X> 
-void *_clone (tl::true_tag, const void *other)
-{
-  return new X (*(const X *)other);
-}
-
-template <class X> 
-void _assign (tl::false_tag /*has_copy_ctor*/, void *, const void *)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Object cannot be copied here")));
-}
-
-template <class X> 
-void _assign (tl::true_tag /*has_copy_ctor*/, void *dest, const void *src)
-{
-  *(X *)dest = *(const X *)src;
-}
-
-/**
- *  @brief A class predicate telling us whether X is polymorphic
- *
- *  The trick was taken from boost - if a class is made polymorphic and it
- *  has not been before, the size of the object changes since a virtual 
- *  function table will be added.
- */
-template <class X>
-struct is_polymorphic
-{
-private:
-  struct P1 : public X
-  {
-    P1 ();
-    ~P1 ();
-    int something;
-  };
-
-  struct P2 : public X
-  {
-    P2 ();
-    virtual ~P2 ();
-    int something;
-  };
-
-public:
-  typedef typename tl::boolean_value<sizeof (P1) == sizeof (P2)>::value value;
-};
+template <class X, bool> struct _assign;
+template <class X> struct _assign<X, false> { static void call (void *, const void *) { throw tl::Exception (tl::to_string (tr ("Object cannot be copied here"))); } };
+template <class X> struct _assign<X, true> { static void call (void *dest, const void *src) { *(X *)dest = *(const X *)src; } };
 
 /**
  *  @brief A helper class which tests whether a given object can be upcast 
@@ -512,14 +321,14 @@ public:
  *  The can_upcast method will return true, if the object (which has at least to 
  *  be of type B) can be upcast to X.
  */
-template <class X, class B, class B_IS_POLYMORPHIC>
+template <class X, class B, bool B_IS_POLYMORPHIC>
 class SubClassTester;
 
 /**
  *  @brief Specialization for polymorphic types - we can use dynamic_cast to tests whether B object can be cast to X
  */
 template <class X, class B>
-class SubClassTester<X, B, tl::true_tag>
+class SubClassTester<X, B, true>
   : public SubClassTesterBase 
 {
 public:
@@ -533,7 +342,7 @@ public:
  *  @brief Specialization for non-polymorphic types
  */
 template <class X, class B>
-class SubClassTester<X, B, tl::false_tag>
+class SubClassTester<X, B, false>
   : public SubClassTesterBase 
 {
 public:
@@ -555,10 +364,6 @@ class GSI_PUBLIC_TEMPLATE ClassExt
   : public ClassBase
 {
 public:
-  typedef typename tl::type_traits<X>::has_copy_constructor has_copy_ctor;
-  typedef typename tl::type_traits<X>::has_default_constructor has_default_ctor;
-  typedef typename tl::type_traits<X>::has_public_destructor has_public_dtor;
-
   ClassExt (const Methods &mm, const std::string &doc = std::string ())
     : ClassBase (doc, mm), mp_declaration (0)
   {
@@ -578,15 +383,15 @@ public:
    *  This feature is not quite useful usually and is reserved for special use cases
    *  such as including enums into a declaration namespace.
    */
-  ClassExt (const ClassBase &import, const std::string &name, const std::string &doc = std::string ())
+  ClassExt (const ClassBase &import, const std::string &name = std::string (), const std::string &doc = std::string ())
     : ClassBase (doc, Methods ()), mp_declaration (&import)
   {
     set_name (name);
   }
 
-  virtual bool is_of_type (const std::type_info &ti) const 
+  virtual bool binds () const
   {
-    return (ti == typeid (X));
+    return false;
   }
 
   virtual const std::type_info &type () const
@@ -640,7 +445,9 @@ struct NoAdaptorTag { };
 template <class X, class Adapted>
 struct adaptor_type_info
 {
-  static const std::type_info *type_info () 
+  typedef Adapted final_type;
+
+  static const std::type_info *type_info ()
   {
     return &typeid (Adapted);
   }
@@ -671,6 +478,8 @@ struct adaptor_type_info
 template <class X>
 struct adaptor_type_info<X, NoAdaptorTag>
 {
+  typedef X final_type;
+
   static const std::type_info *type_info () 
   {
     return 0;
@@ -712,34 +521,36 @@ class GSI_PUBLIC_TEMPLATE Class
   : public ClassBase
 {
 public:
-  typedef typename tl::type_traits<X>::has_copy_constructor has_copy_ctor;
-  typedef typename tl::type_traits<X>::has_default_constructor has_default_ctor;
-  typedef typename tl::type_traits<X>::has_public_destructor has_public_dtor;
+  typedef typename adaptor_type_info<X, Adapted>::final_type final_type;
 
-  Class (const std::string &name, const Methods &mm, const std::string &doc = std::string (), bool do_register = true)
+  Class (const std::string &module, const std::string &name, const Methods &mm, const std::string &doc = std::string (), bool do_register = true)
     : ClassBase (doc, mm, do_register)
   {
+    set_module (module);
     set_name (name);
   }
 
   template <class B>
-  Class (const Class<B> &base, const std::string &name, const Methods &mm, const std::string &doc = std::string (), bool do_register = true)
-    : ClassBase (doc, mm, do_register), m_subclass_tester (new SubClassTester<X, B, typename is_polymorphic<B>::value> ())
+  Class (const Class<B> &base, const std::string &module, const std::string &name, const Methods &mm, const std::string &doc = std::string (), bool do_register = true)
+    : ClassBase (doc, mm, do_register), m_subclass_tester (new SubClassTester<X, B, std::is_polymorphic<B>::value> ())
   {
+    set_module (module);
     set_name (name);
     set_base (&base);
   }
 
-  Class (const std::string &name, const std::string &doc = std::string (), bool do_register = true)
+  Class (const std::string &module, const std::string &name, const std::string &doc = std::string (), bool do_register = true)
     : ClassBase (doc, Methods (), do_register)
   {
+    set_module (module);
     set_name (name);
   }
 
   template <class B>
-  Class (const Class<B> &base, const std::string &name, const std::string &doc = std::string (), bool do_register = true)
-    : ClassBase (doc, Methods (), do_register), m_subclass_tester (new SubClassTester<X, B, typename is_polymorphic<B>::value> ())
+  Class (const Class<B> &base, const std::string &module, const std::string &name, const std::string &doc = std::string (), bool do_register = true)
+    : ClassBase (doc, Methods (), do_register), m_subclass_tester (new SubClassTester<X, B, std::is_polymorphic<B>::value> ())
   {
+    set_module (module);
     set_name (name);
     set_base (&base);
   }
@@ -780,15 +591,12 @@ public:
   virtual void destroy (void *p) const
   {
     X *x = (X *)p;
-    has_public_dtor hpd;
-    _destroy (x, hpd);
+    _destroy<X, std::is_destructible<X>::value>::call (x);
   }
 
   virtual void *create () const
   {
-    has_default_ctor cst;
-    void *r = _create<X> (cst);
-    return r;
+    return _create<X, std::is_default_constructible<X>::value>::call ();
   }
 
   virtual void *create_from_adapted (const void *x) const
@@ -813,33 +621,27 @@ public:
 
   virtual void *clone (const void *other) const
   {
-    has_copy_ctor cst;
-    void *r = _clone<X> (cst, other);
-    return r;
+    return _clone<X, std::is_copy_constructible<X>::value>::call (other);
   }
 
   virtual void assign (void *dest, const void *src) const
   {
-    has_copy_ctor cst;
-    _assign<X> (cst, dest, src);
+    _assign<X, std::is_copy_assignable<X>::value>::call (dest, src);
   }
 
   virtual bool can_destroy () const
   {
-    has_public_dtor hpd;
-    return tl::value_of (hpd);
+    return std::is_destructible<X>::value;
   }
 
   virtual bool can_copy () const
   {
-    has_copy_ctor cpt;
-    return tl::value_of (cpt);
+    return std::is_copy_constructible<X>::value;
   }
 
   virtual bool can_default_create () const
   {
-    has_default_ctor cpt;
-    return tl::value_of (cpt);
+    return std::is_default_constructible<X>::value;
   }
 
   virtual const ClassBase *subclass_decl (const void *p) const 
@@ -860,15 +662,9 @@ public:
     return m_subclass_tester.get () && m_subclass_tester->can_upcast (p);
   }
 
-  virtual bool is_of_type (const std::type_info &ti) const 
+  virtual bool binds () const
   {
-    if (adapted_type_info ()) {
-      //  A class matches the typeinfo of the adapted type. We'll sort this out later
-      //  on the client side. 
-      return (ti == *adapted_type_info ());
-    } else {
-      return (ti == typeid (X));
-    }
+    return true;
   }
 
   virtual const std::type_info &type () const
@@ -891,10 +687,10 @@ public:
   }
 
 private:
-  gsi::VariantUserClass<X> m_var_cls;
-  gsi::VariantUserClass<X> m_var_cls_c;
-  gsi::VariantUserClass<X> m_var_cls_cls;
-  std::auto_ptr<SubClassTesterBase> m_subclass_tester;
+  gsi::VariantUserClass<final_type> m_var_cls;
+  gsi::VariantUserClass<final_type> m_var_cls_c;
+  gsi::VariantUserClass<final_type> m_var_cls_cls;
+  std::unique_ptr<SubClassTesterBase> m_subclass_tester;
 };
 
 /**
@@ -909,14 +705,14 @@ class SubClass
   : public Class<X, Adapted>
 {
 public:
-  SubClass (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
-    : Class<X, Adapted> (name, mm, doc)
+  SubClass (const std::string &module, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
+    : Class<X, Adapted> (module, name, mm, doc)
   {
     //  .. nothing yet ..
   }
 
-  SubClass (const std::string &name, const std::string &doc = std::string ())
-    : Class<X, Adapted> (name, doc)
+  SubClass (const std::string &module, const std::string &name, const std::string &doc = std::string ())
+    : Class<X, Adapted> (module, name, doc)
   {
     //  .. nothing yet ..
   }
@@ -941,14 +737,14 @@ class GSI_PUBLIC_TEMPLATE ChildClass
   : public Class<X, Adapted>
 {
 public:
-  ChildClass (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
-    : Class<X, Adapted> (name, mm, doc)
+  ChildClass (const std::string &module, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
+    : Class<X, Adapted> (module, name, mm, doc)
   {
     //  .. nothing yet ..
   }
 
-  ChildClass (const std::string &name, const std::string &doc = std::string ())
-    : Class<X, Adapted> (name, doc)
+  ChildClass (const std::string &module, const std::string &name, const std::string &doc = std::string ())
+    : Class<X, Adapted> (module, name, doc)
   {
     //  .. nothing yet ..
   }
@@ -959,7 +755,8 @@ public:
     ClassBase *non_const_pcls = const_cast<ClassBase *> (cls_decl<P> ());
     non_const_pcls->add_child_class (this);
 
-    return Class<X, Adapted>::consolidate ();
+    //  no longer required as it is a child now.
+    return false;
   }
 };
 
@@ -974,14 +771,14 @@ class GSI_PUBLIC_TEMPLATE ChildSubClass
   : public SubClass<X, B, Adapted>
 {
 public:
-  ChildSubClass (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
-    : SubClass<X, B, Adapted> (name, mm, doc)
+  ChildSubClass (const std::string &module, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
+    : SubClass<X, B, Adapted> (module, name, mm, doc)
   {
     //  .. nothing yet ..
   }
 
-  ChildSubClass (const std::string &name, const std::string &doc = std::string ())
-    : SubClass<X, B, Adapted> (name, doc)
+  ChildSubClass (const std::string &module, const std::string &name, const std::string &doc = std::string ())
+    : SubClass<X, B, Adapted> (module, name, doc)
   {
     //  .. nothing yet ..
   }
