@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -21,8 +21,8 @@
 */
 
 
-#ifndef _HDR_gsiSerialisation
-#define _HDR_gsiSerialisation
+#ifndef _HDR_gsiSerialization
+#define _HDR_gsiSerialization
 
 #include "gsiTypes.h"
 #include "tlHeap.h"
@@ -30,6 +30,7 @@
 
 #include <list>
 #include <memory>
+#include <cstring>
 
 namespace gsi
 {
@@ -78,7 +79,18 @@ struct GSI_PUBLIC ArglistUnderflowException
   : public tl::Exception
 {
   ArglistUnderflowException ()
-    : tl::Exception (tl::to_string (QObject::tr ("Too few arguments or no return value supplied")))
+    : tl::Exception (tl::to_string (tr ("Too few arguments or no return value supplied")))
+  { }
+};
+
+/**
+ *  @brief An exception thrown if there are not enough arguments on the serialization buffer
+ */
+struct GSI_PUBLIC ArglistUnderflowExceptionWithType
+  : public tl::Exception
+{
+  ArglistUnderflowExceptionWithType (const ArgSpecBase &as)
+    : tl::Exception (tl::to_string (tr ("Too few arguments - missing '%s'")), as.name ())
   { }
 };
 
@@ -89,7 +101,18 @@ struct GSI_PUBLIC NilPointerToReference
   : public tl::Exception
 {
   NilPointerToReference ()
-    : tl::Exception (tl::to_string (QObject::tr ("nil object passed to a reference")))
+    : tl::Exception (tl::to_string (tr ("nil object passed to a reference")))
+  { }
+};
+
+/**
+ *  @brief An exception thrown if a reference is null (nil)
+ */
+struct GSI_PUBLIC NilPointerToReferenceWithType
+  : public tl::Exception
+{
+  NilPointerToReferenceWithType (const ArgSpecBase &as)
+    : tl::Exception (tl::to_string (tr ("nil object passed to a reference for '%s'")), as.name ())
   { }
 };
 
@@ -207,7 +230,16 @@ public:
   template <class X>
   inline X read (tl::Heap &heap)
   {
-    return this->read_impl<X> (typename type_traits<X>::tag (), heap);
+    return this->read_impl<X> (typename type_traits<X>::tag (), heap, 0);
+  }
+
+  /**
+   *  @brief Reads a value from the buffer
+   */
+  template <class X>
+  inline X read (tl::Heap &heap, const ArgSpecBase *as)
+  {
+    return this->read_impl<X> (typename type_traits<X>::tag (), heap, as);
   }
 
   /**
@@ -230,6 +262,26 @@ private:
     }
   }
 
+  inline void check_data (const ArgSpecBase *as) const
+  {
+    if (! *this) {
+      if (as) {
+        throw ArglistUnderflowExceptionWithType (*as);
+      } else {
+        throw ArglistUnderflowException ();
+      }
+    }
+  }
+
+  inline void throw_nil_for_reference (const ArgSpecBase *as) const
+  {
+    if (as) {
+      throw NilPointerToReferenceWithType (*as);
+    } else {
+      throw NilPointerToReference ();
+    }
+  }
+
   // -----------------------------------------------------------
   //  reader implementations
 
@@ -368,18 +420,18 @@ private:
   //  reader implementations
 
   template <class X>
-  X read_impl (const pod_direct_tag &, tl::Heap &)
+  X read_impl (const pod_direct_tag &, tl::Heap &, const ArgSpecBase *as)
   {
-    check_data ();
+    check_data (as);
     X r = *((X *)mp_read);
     mp_read += item_size<X> ();
     return r;
   }
 
   template <class X>
-  X read_impl (const x_tag &, tl::Heap &)
+  X read_impl (const x_tag &, tl::Heap &, const ArgSpecBase *as)
   {
-    check_data ();
+    check_data (as);
     X *xp = *(X **)mp_read;
     X x = *xp;
     delete xp;
@@ -388,7 +440,7 @@ private:
   }
 
   template <class X>
-  X read_impl (const vptr_tag &, tl::Heap &)
+  X read_impl (const vptr_tag &, tl::Heap &, const ArgSpecBase *)
   {
     void *r = *((void **)mp_read);
     mp_read += item_size<void *> ();
@@ -397,73 +449,73 @@ private:
   }
 
   template <class X>
-  X read_impl (const ref_tag &, tl::Heap &)
+  X read_impl (const ref_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     typedef typename type_traits<X>::value_type value_type;
-    check_data ();
+    check_data (as);
     value_type *r = *((value_type **)mp_read);
     mp_read += item_size<value_type *> ();
     if (! r) {
-      throw NilPointerToReference ();
+      throw_nil_for_reference (as);
     }
     return *r;
   }
 
   template <class X>
-  X read_impl (const pod_cref_tag &, tl::Heap &)
+  X read_impl (const pod_cref_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X &)
     typedef typename type_traits<X>::value_type value_type;
-    check_data ();
+    check_data (as);
     const value_type *r = ((const value_type *)mp_read);
     mp_read += item_size<value_type> ();
     return *r;
   }
 
   template <class X>
-  X read_impl (const npod_cref_tag &, tl::Heap &)
+  X read_impl (const npod_cref_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X &)
     typedef typename type_traits<X>::value_type value_type;
-    check_data ();
+    check_data (as);
     const value_type *r = *((const value_type **)mp_read);
     mp_read += item_size<const value_type *> ();
     if (! r) {
-      throw NilPointerToReference ();
+      throw_nil_for_reference (as);
     }
     return *r;
   }
 
   template <class X>
-  X read_impl (const x_cref_tag &, tl::Heap &)
+  X read_impl (const x_cref_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X &)
     typedef typename type_traits<X>::value_type value_type;
-    check_data ();
+    check_data (as);
     const value_type *r = *((const value_type **)mp_read);
     mp_read += item_size<const value_type *> ();
     if (! r) {
-      throw NilPointerToReference ();
+      throw_nil_for_reference (as);
     }
     return *r;
   }
 
   template <class X>
-  X read_impl (const ptr_tag &, tl::Heap &)
+  X read_impl (const ptr_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (X *)
     typedef typename type_traits<X>::value_type value_type;
-    check_data ();
+    check_data (as);
     value_type * const &r = *((value_type **)mp_read);
     mp_read += item_size<value_type *> ();
     return r;
   }
 
   template <class X>
-  X read_impl (const pod_cptr_tag &, tl::Heap &)
+  X read_impl (const pod_cptr_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X *)
-    check_data ();
+    check_data (as);
     bool h = *(bool *)mp_read;
     mp_read += item_size<bool> ();
     X r = h ? (X)mp_read : (X)0;
@@ -473,33 +525,36 @@ private:
 
   //  see notes on the serialization for this type:
   template <class X>
-  X read_impl (const npod_cptr_tag &, tl::Heap &)
+  X read_impl (const npod_cptr_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X *)
-    check_data ();
+    check_data (as);
     X r = *((X *)mp_read);
     mp_read += item_size<X> ();
     return r;
   }
 
   template <class X>
-  X read_impl (const x_cptr_tag &, tl::Heap &)
+  X read_impl (const x_cptr_tag &, tl::Heap &, const ArgSpecBase *as)
   {
     //  X is actually an (const X *)
-    check_data ();
+    check_data (as);
     X r = *((X *)mp_read);
     mp_read += item_size<X> ();
     return r;
   }
 
   template <class X>
-  X read_impl (const adaptor_direct_tag &, tl::Heap &heap)
+  X read_impl (const adaptor_direct_tag &, tl::Heap &heap, const ArgSpecBase *as)
   {
-    check_data ();
+    check_data (as);
 
-    std::auto_ptr<AdaptorBase> p (*(AdaptorBase **)mp_read);
+    AdaptorBase *p = *(AdaptorBase **)mp_read;
     mp_read += item_size<AdaptorBase *> ();
-    tl_assert (p.get () != 0);
+
+    tl_assert (p != 0);
+    //  late-destroy the adaptor since the new X object may still need data from there (e.g. QLatin1String)
+    heap.push (p);
 
     X x = X ();
     copy_to<X> (*p, x, heap);
@@ -507,15 +562,18 @@ private:
   }
 
   template <class X>
-  X read_impl (const adaptor_cref_tag &, tl::Heap &heap)
+  X read_impl (const adaptor_cref_tag &, tl::Heap &heap, const ArgSpecBase *as)
   {
     typedef typename tl::get_inner_type<X>::result x_type;
 
-    check_data ();
+    check_data (as);
 
-    std::auto_ptr<AdaptorBase> p (*(AdaptorBase **)mp_read);
+    AdaptorBase *p = *(AdaptorBase **)mp_read;
     mp_read += item_size<AdaptorBase *> ();
-    tl_assert (p.get () != 0);
+
+    tl_assert (p != 0);
+    //  late-destroy the adaptor since the new X object may still need data from there (e.g. QLatin1String)
+    heap.push (p);
 
     x_type *x = new x_type ();
     heap.push (x);
@@ -525,11 +583,11 @@ private:
   }
 
   template <class X>
-  X read_impl (const adaptor_ref_tag &, tl::Heap &heap)
+  X read_impl (const adaptor_ref_tag &, tl::Heap &heap, const ArgSpecBase *as)
   {
     typedef typename tl::get_inner_type<X>::result x_type;
 
-    check_data ();
+    check_data (as);
 
     AdaptorBase *p = *(AdaptorBase **)mp_read;
     mp_read += item_size<AdaptorBase *> ();
@@ -543,31 +601,36 @@ private:
   }
 
   template <class X>
-  X read_impl (const adaptor_cptr_tag &, tl::Heap &heap)
+  X read_impl (const adaptor_cptr_tag &, tl::Heap &heap, const ArgSpecBase *as)
   {
     typedef typename tl::get_inner_type<X>::result x_type;
 
-    check_data ();
+    check_data (as);
 
-    std::auto_ptr<AdaptorBase> p (*(AdaptorBase **)mp_read);
+    AdaptorBase *p = *(AdaptorBase **) mp_read;
     mp_read += item_size<AdaptorBase *> ();
 
     x_type *x = 0;
-    if (p.get () != 0) {
+    if (p != 0) {
+
+      //  late-destroy the adaptor since the new X object may still need data from there (e.g. QLatin1String)
+      heap.push (p);
+
       x = new x_type ();
       heap.push (x);
       copy_to<x_type> (*p, *x, heap);
+
     }
 
     return x;
   }
 
   template <class X>
-  X read_impl (const adaptor_ptr_tag &, tl::Heap &heap)
+  X read_impl (const adaptor_ptr_tag &, tl::Heap &heap, const ArgSpecBase *as)
   {
     typedef typename tl::get_inner_type<X>::result x_type;
 
-    check_data ();
+    check_data (as);
 
     AdaptorBase *p = *(AdaptorBase **)mp_read;
     mp_read += item_size<AdaptorBase *> ();
@@ -585,6 +648,46 @@ private:
   }
 };
 
+// ------------------------------------------------------------
+//  A rebinding of SerialArgs::read to a functor
+
+template <class X>
+struct GSI_PUBLIC_TEMPLATE arg_reader
+{
+  inline X operator() (gsi::SerialArgs &args, tl::Heap &heap) { return args.read<X> (heap); }
+};
+
+// ------------------------------------------------------------
+//  Provides a function analog to arg_reader, but for a static value
+//  The reasoning behind this functor is to support default arguments.
+//  Specifically in the case of (const X &) arguments, this must not
+//  create references to temporaries.
+
+template <class X>
+struct GSI_PUBLIC_TEMPLATE arg_maker
+{
+  inline X operator() (const X &x, tl::Heap &) { return x; }
+};
+
+template <class X>
+struct GSI_PUBLIC_TEMPLATE arg_maker<X &>
+{
+  inline X &operator() (X &x, tl::Heap &) { return x; }
+};
+
+template <class X>
+struct GSI_PUBLIC_TEMPLATE arg_maker<const X &>
+{
+  inline const X &operator() (const X &x, tl::Heap &heap)
+  { 
+    //  avoid references to temp. With this copy we can create a const
+    //  reference from a static value.
+    X *copy = new X (x);
+    heap.push (copy);
+    return *copy; 
+  }
+};
+
 // ------------------------------------------------------------
 //  Basic adaptor 
 
@@ -655,75 +758,7 @@ class GSI_PUBLIC_TEMPLATE StringAdaptorImpl
 {
 };
 
-/**
- *  @brief Specialization for QByteArray
- */
-template <>
-class GSI_PUBLIC StringAdaptorImpl<QByteArray>
-  : public StringAdaptor
-{
-public:
-  StringAdaptorImpl (QByteArray *s) 
-    : mp_s (s), m_is_const (false) 
-  { 
-    //  .. nothing yet ..
-  }
-
-  StringAdaptorImpl (const QByteArray *s) 
-    : mp_s (const_cast<QByteArray *> (s)), m_is_const (true) 
-  { 
-    //  .. nothing yet ..
-  }
-
-  StringAdaptorImpl (const QByteArray &s) 
-    : m_is_const (false), m_s (s) 
-  { 
-    mp_s = &m_s; 
-  }
-
-  StringAdaptorImpl () 
-    : m_is_const (false)
-  { 
-    mp_s = &m_s; 
-  }
-
-  virtual ~StringAdaptorImpl () 
-  { 
-    //  .. nothing yet ..
-  }
-
-  virtual size_t size () const 
-  { 
-    return mp_s->size (); 
-  }
-
-  virtual const char *c_str () const
-  {
-    return mp_s->constData (); 
-  }
-
-  virtual void set (const char *c_str, size_t s, tl::Heap &) 
-  {
-    if (! m_is_const) {
-      *mp_s = QByteArray (c_str, s);
-    }
-  }
-
-  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
-  {
-    StringAdaptorImpl<QByteArray> *s = dynamic_cast<StringAdaptorImpl<QByteArray> *>(target);
-    if (s) {
-      *s->mp_s = *mp_s;
-    } else {
-      StringAdaptor::copy_to (target, heap);
-    }
-  }
-   
-private:
-  QByteArray *mp_s;
-  bool m_is_const;
-  QByteArray m_s;
-};
+#if defined(HAVE_QT)
 
 /**
  *  @brief Specialization for QString
@@ -776,7 +811,7 @@ public:
   virtual void set (const char *c_str, size_t s, tl::Heap &) 
   {
     if (! m_is_const) {
-      *mp_s = QString::fromUtf8 (c_str, s);
+      *mp_s = QString::fromUtf8 (c_str, int (s));
     }
   }
 
@@ -848,7 +883,7 @@ public:
   virtual void set (const char *c_str, size_t s, tl::Heap &heap) 
   {
     if (! m_is_const) {
-      QString *qstr = new QString (QString::fromUtf8 (c_str, s));
+      QString *qstr = new QString (QString::fromUtf8 (c_str, int (s)));
       heap.push (qstr);
       *mp_s = QStringRef (qstr);
     }
@@ -871,6 +906,156 @@ private:
   mutable QByteArray m_s_utf8;
 };
 
+#if QT_VERSION >= 0x60000
+
+/**
+ *  @brief Specialization for QString
+ */
+template <>
+class GSI_PUBLIC StringAdaptorImpl<QStringView>
+  : public StringAdaptor
+{
+public:
+  StringAdaptorImpl (QStringView *s)
+    : mp_s (s), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  StringAdaptorImpl (const QStringView *s)
+    : mp_s (const_cast<QStringView *> (s)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  StringAdaptorImpl (const QStringView &s)
+    : m_is_const (false), m_s (s)
+  {
+    mp_s = &m_s;
+  }
+
+  StringAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_s = &m_s;
+  }
+
+  virtual ~StringAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual size_t size () const
+  {
+    return mp_s->toUtf8 ().size ();
+  }
+
+  virtual const char *c_str () const
+  {
+    m_s_utf8 = mp_s->toUtf8 ();
+    return m_s_utf8.constData ();
+  }
+
+  virtual void set (const char *c_str, size_t s, tl::Heap &heap)
+  {
+    if (! m_is_const) {
+      QString *hstr = heap.create<QString> ();
+      *hstr = QString::fromUtf8 (c_str, int (s));
+      *mp_s = QStringView (hstr->constData (), hstr->size ());
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    StringAdaptorImpl<QStringView> *s = dynamic_cast<StringAdaptorImpl<QStringView> *>(target);
+    if (s) {
+      QString *hstr = heap.create<QString> ();
+      *hstr = mp_s->toString ();
+      *s->mp_s = QStringView (hstr->constData (), hstr->size ());
+    } else {
+      StringAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  QStringView *mp_s;
+  bool m_is_const;
+  QStringView m_s;
+  mutable QByteArray m_s_utf8;
+};
+
+#endif
+
+#if QT_VERSION >= 0x50000
+
+/**
+ *  @brief Specialization for QLatin1String
+ */
+template <>
+class GSI_PUBLIC StringAdaptorImpl<QLatin1String>
+  : public StringAdaptor
+{
+public:
+  StringAdaptorImpl (QLatin1String *s)
+    : mp_s (s), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  StringAdaptorImpl (const QLatin1String *s)
+    : mp_s (const_cast<QLatin1String *> (s)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  StringAdaptorImpl (const QLatin1String &s)
+    : m_is_const (false), m_s (s)
+  {
+    mp_s = &m_s;
+  }
+
+  StringAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_s = &m_s;
+  }
+
+  virtual ~StringAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual size_t size () const
+  {
+    return QString::fromLatin1 (mp_s->data (), mp_s->size ()).toUtf8 ().size ();
+  }
+
+  virtual const char *c_str () const
+  {
+    m_s_utf8 = QString::fromLatin1 (mp_s->data (), mp_s->size ()).toUtf8 ();
+    return m_s_utf8.constData ();
+  }
+
+  virtual void set (const char *c_str, size_t s, tl::Heap &heap)
+  {
+    if (! m_is_const) {
+      QByteArray *latin1_holder = new QByteArray (QString::fromUtf8 (c_str, int (s)).toLatin1 ());
+      heap.push (latin1_holder);
+      *mp_s = QLatin1String (latin1_holder->constData (), latin1_holder->size ());
+    }
+  }
+
+private:
+  QLatin1String *mp_s;
+  bool m_is_const;
+  QLatin1String m_s;
+  mutable QByteArray m_s_utf8;
+};
+
+#endif
+
+#endif
+
 /**
  *  @brief Specialization for std::string
  */
@@ -1058,12 +1243,290 @@ public:
 };
 
 // ------------------------------------------------------------
-//  Variant adaptor framework
+//  ByteArray adaptor framework
 
 /**
  *  @brief A generic adaptor for strings
  *  This is the base class for implementing generic access to strings
  */
+class GSI_PUBLIC ByteArrayAdaptor
+  : public AdaptorBase
+{
+public:
+  /**
+   *  @brief Default constructor
+   */
+  ByteArrayAdaptor () { }
+
+  /**
+   *  @brief Destructor
+   */
+  virtual ~ByteArrayAdaptor () { }
+
+  /**
+   *  @brief Returns the size of the string
+   */
+  virtual size_t size () const = 0;
+
+  /**
+   *  @brief Returns a pointer to a UTF8 encoded character array with size size()
+   */
+  virtual const char *c_str () const = 0;
+
+  /**
+   *  @brief Sets the string to the given UTF8 string with length s
+   */
+  virtual void set (const char *c_str, size_t s, tl::Heap &heap) = 0;
+
+  /**
+   *  @brief copy_to implementation
+   */
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    ByteArrayAdaptor *s = dynamic_cast<ByteArrayAdaptor *>(target);
+    tl_assert (s);
+    s->set (c_str (), size (), heap);
+  }
+};
+
+/**
+ *  @brief Generic string adaptor implementation
+ */
+template <class X>
+class GSI_PUBLIC_TEMPLATE ByteArrayAdaptorImpl
+  : public ByteArrayAdaptor
+{
+};
+
+#if defined(HAVE_QT)
+
+/**
+ *  @brief Specialization for QByteArray
+ */
+template <>
+class GSI_PUBLIC ByteArrayAdaptorImpl<QByteArray>
+  : public ByteArrayAdaptor
+{
+public:
+  ByteArrayAdaptorImpl (QByteArray *s)
+    : mp_s (s), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const QByteArray *s)
+    : mp_s (const_cast<QByteArray *> (s)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const QByteArray &s)
+    : m_is_const (false), m_s (s)
+  {
+    mp_s = &m_s;
+  }
+
+  ByteArrayAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_s = &m_s;
+  }
+
+  virtual ~ByteArrayAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual size_t size () const
+  {
+    return mp_s->size ();
+  }
+
+  virtual const char *c_str () const
+  {
+    return mp_s->constData ();
+  }
+
+  virtual void set (const char *c_str, size_t s, tl::Heap &)
+  {
+    if (! m_is_const) {
+      *mp_s = QByteArray (c_str, int (s));
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    ByteArrayAdaptorImpl<QByteArray> *s = dynamic_cast<ByteArrayAdaptorImpl<QByteArray> *>(target);
+    if (s) {
+      *s->mp_s = *mp_s;
+    } else {
+      ByteArrayAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  QByteArray *mp_s;
+  bool m_is_const;
+  QByteArray m_s;
+};
+
+#if QT_VERSION > 0x60000
+
+/**
+ *  @brief Specialization for QByteArray
+ */
+template <>
+class GSI_PUBLIC ByteArrayAdaptorImpl<QByteArrayView>
+  : public ByteArrayAdaptor
+{
+public:
+  ByteArrayAdaptorImpl (QByteArrayView *s)
+    : mp_s (s), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const QByteArrayView *s)
+    : mp_s (const_cast<QByteArrayView *> (s)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const QByteArrayView &s)
+    : m_is_const (false), m_s (s)
+  {
+    mp_s = &m_s;
+  }
+
+  ByteArrayAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_s = &m_s;
+  }
+
+  virtual ~ByteArrayAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual size_t size () const
+  {
+    return mp_s->size ();
+  }
+
+  virtual const char *c_str () const
+  {
+    return mp_s->constData ();
+  }
+
+  virtual void set (const char *c_str, size_t s, tl::Heap &heap)
+  {
+    if (! m_is_const) {
+      QByteArray *str = heap.create<QByteArray> ();
+      *str = QByteArray (c_str, s);
+      *mp_s = QByteArrayView (str->constData (), str->size ());
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    ByteArrayAdaptorImpl<QByteArrayView> *s = dynamic_cast<ByteArrayAdaptorImpl<QByteArrayView> *>(target);
+    if (s) {
+      QByteArray *str = heap.create<QByteArray> ();
+      *str = QByteArray (mp_s->constData (), mp_s->size ());
+      *s->mp_s = *str;
+    } else {
+      ByteArrayAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  QByteArrayView *mp_s;
+  bool m_is_const;
+  QByteArrayView m_s;
+};
+
+#endif
+
+#endif
+
+/**
+ *  @brief Specialization for std::string
+ */
+template <>
+class GSI_PUBLIC ByteArrayAdaptorImpl<std::vector<char> >
+  : public ByteArrayAdaptor
+{
+public:
+  ByteArrayAdaptorImpl (std::vector<char> *s)
+    : mp_s (s), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const std::vector<char> *s)
+    : mp_s (const_cast<std::vector<char> *> (s)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  ByteArrayAdaptorImpl (const std::vector<char> &s)
+    : m_is_const (false), m_s (s)
+  {
+    mp_s = &m_s;
+  }
+
+  ByteArrayAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_s = &m_s;
+  }
+
+  virtual ~ByteArrayAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual size_t size () const
+  {
+    return mp_s->size ();
+  }
+
+  virtual const char *c_str () const
+  {
+    return &mp_s->front ();
+  }
+
+  virtual void set (const char *c_str, size_t s, tl::Heap &)
+  {
+    if (! m_is_const) {
+      *mp_s = std::vector<char> (c_str, c_str + s);
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    ByteArrayAdaptorImpl<std::vector<char> > *s = dynamic_cast<ByteArrayAdaptorImpl<std::vector<char> > *>(target);
+    if (s) {
+      *s->mp_s = *mp_s;
+    } else {
+      ByteArrayAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  std::vector<char> *mp_s;
+  bool m_is_const;
+  std::vector<char> m_s;
+};
+
+// ------------------------------------------------------------
+//  Variant adaptor framework
+
+/**
+ *  @brief A generic adaptor for variants
+ *  This is the base class for implementing generic access to variants
+ */
 class GSI_PUBLIC VariantAdaptor
   : public AdaptorBase
 {
@@ -1086,21 +1549,21 @@ public:
   /**
    *  @brief Sets the variant's value
    */
-  virtual void set (const tl::Variant &v) = 0;
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/) = 0;
 
   /**
    *  @brief Implementation of copy_to
    */
-  virtual void copy_to (AdaptorBase *target, tl::Heap & /*heap*/) const
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
   {
     VariantAdaptor *v = dynamic_cast<VariantAdaptor *>(target);
     tl_assert (v);
-    v->set (var ());
+    v->set (var (), heap);
   }
 };
 
 /**
- *  @brief Generic string adaptor implementation
+ *  @brief Generic variant adaptor implementation
  */
 template <class X> 
 class GSI_PUBLIC_TEMPLATE VariantAdaptorImpl
@@ -1108,6 +1571,8 @@ class GSI_PUBLIC_TEMPLATE VariantAdaptorImpl
 {
 };
 
+#if defined(HAVE_QT)
+
 /**
  *  @brief Specialization for QVariant
  */
@@ -1155,7 +1620,7 @@ public:
     return *mp_v;
   }
 
-  virtual void set (const tl::Variant &v) 
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/)
   {
     if (! m_is_const) {
       *mp_v = v.to_qvariant ();
@@ -1178,6 +1643,95 @@ private:
   QVariant m_v;
 };
 
+#if QT_VERSION >= 0x50000
+
+/**
+ *  @brief Specialization for QVariant
+ */
+template <typename T>
+class GSI_PUBLIC_TEMPLATE VariantAdaptorImpl<QPointer<T> >
+  : public VariantAdaptor
+{
+public:
+  VariantAdaptorImpl (QPointer<T> *v)
+    : mp_v (v), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  VariantAdaptorImpl (const QPointer<T> *v)
+    : mp_v (const_cast<QPointer<T> *> (v)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  VariantAdaptorImpl (const QPointer<T> &v)
+    : m_is_const (true), m_v (v)
+  {
+    mp_v = &m_v;
+  }
+
+  VariantAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_v = &m_v;
+  }
+
+  virtual ~VariantAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual tl::Variant var () const
+  {
+    return tl::Variant::make_variant_ref (mp_v->get ());
+  }
+
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/)
+  {
+    if (m_is_const) {
+      //  .. can't change
+    } else if (v.is_nil ()) {
+      mp_v->clear ();
+    } else if (v.is_user<T> ()) {
+      if (v.user_is_ref ()) {
+        *mp_v = (T *) v.to_user ();
+      } else {
+        //  basically should not happen as the QPointer cannot hold anything other than a reference
+        //  (a tl::Variant terminology)
+        tl_assert (false);
+      }
+    } else {
+      //  basically should not happen as the QPointer cannot hold anything other than a user object
+      //  (a tl::Variant terminology)
+      tl_assert (false);
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    VariantAdaptorImpl<QPointer<T>> *v = dynamic_cast<VariantAdaptorImpl<QPointer<T>> *>(target);
+    if (v) {
+      if (mp_v->isNull ()) {
+        v->mp_v->clear ();
+      } else {
+        *v->mp_v = QPointer<T> (*mp_v);
+      }
+    } else {
+      VariantAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  QPointer<T> *mp_v;
+  bool m_is_const;
+  QPointer<T> m_v;
+};
+
+#endif
+
+#endif
+
 /**
  *  @brief Specialization for tl::Variant
  */
@@ -1225,7 +1779,12 @@ public:
     return *mp_v;
   }
 
-  virtual void set (const tl::Variant &v) 
+  tl::Variant &var_ref_nc ()
+  {
+    return *mp_v;
+  }
+
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/)
   {
     if (! m_is_const) {
       *mp_v = v;
@@ -1248,6 +1807,85 @@ private:
   tl::Variant m_v;
 };
 
+#if __cplusplus >= 201703L
+
+/**
+ *  @brief Specialization for std::optional
+ */
+template <typename T>
+class GSI_PUBLIC_TEMPLATE VariantAdaptorImpl<std::optional<T> >
+  : public VariantAdaptor
+{
+public:
+  VariantAdaptorImpl (std::optional<T> *v)
+    : mp_v (v), m_is_const (false)
+  {
+    //  .. nothing yet ..
+  }
+
+  VariantAdaptorImpl (const std::optional<T> *v)
+    : mp_v (const_cast<std::optional<T> *> (v)), m_is_const (true)
+  {
+    //  .. nothing yet ..
+  }
+
+  VariantAdaptorImpl (const std::optional<T> &v)
+    : m_is_const (true), m_v (v)
+  {
+    mp_v = &m_v;
+  }
+
+  VariantAdaptorImpl ()
+    : m_is_const (false)
+  {
+    mp_v = &m_v;
+  }
+
+  virtual ~VariantAdaptorImpl ()
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual tl::Variant var () const
+  {
+    return *mp_v ? tl::Variant (mp_v->value ()) : tl::Variant ();
+  }
+
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/)
+  {
+    if (m_is_const) {
+      //  .. can't change
+    } else if (v.is_nil ()) {
+      mp_v->reset ();
+    } else if (v.is_user<T> ()) {
+      *mp_v = v.to_user<T> ();
+    } else {
+      *mp_v = v.to<T> ();
+    }
+  }
+
+  virtual void copy_to (AdaptorBase *target, tl::Heap &heap) const
+  {
+    VariantAdaptorImpl<std::optional<T>> *v = dynamic_cast<VariantAdaptorImpl<std::optional<T>> *>(target);
+    if (v) {
+      if (*mp_v) {
+        v->mp_v->reset ();
+      } else {
+        *v->mp_v = mp_v->value ();
+      }
+    } else {
+      VariantAdaptor::copy_to (target, heap);
+    }
+  }
+
+private:
+  std::optional<T> *mp_v;
+  bool m_is_const;
+  std::optional<T> m_v;
+};
+
+#endif
+
 // ------------------------------------------------------------
 //  Vector adaptor framework
 
@@ -1345,7 +1983,7 @@ public:
     gsi::SerialArgs rr (serial_size ());
     tl_assert (v->serial_size () == serial_size ());
 
-    std::auto_ptr<VectorAdaptorIterator> i (create_iterator ());
+    std::unique_ptr<VectorAdaptorIterator> i (create_iterator ());
     while (! i->at_end ()) {
       rr.reset ();
       i->get (rr, heap);
@@ -1407,11 +2045,15 @@ void push_vector (std::set<X> &v, const X &x)
   v.insert (x);
 }
 
+#if defined(HAVE_QT)
+
+#if QT_VERSION < 0x60000
 template <class X>
 void push_vector (QVector<X> &v, const X &x)
 {
   v.push_back (x);
 }
+#endif
 
 inline void push_vector (QStringList &v, const QString &x)
 {
@@ -1430,6 +2072,8 @@ void push_vector (QSet<X> &v, const X &x)
   v.insert (x);
 }
 
+#endif
+
 /**
  *  @brief Implementation of the generic adaptor for a specific container
  */
@@ -1614,7 +2258,7 @@ public:
     gsi::SerialArgs rr (serial_size ());
     tl_assert (v->serial_size () == serial_size ());
 
-    std::auto_ptr<MapAdaptorIterator> i (create_iterator ());
+    std::unique_ptr<MapAdaptorIterator> i (create_iterator ());
     while (! i->at_end ()) {
       rr.reset ();
       i->get (rr, heap);
@@ -1651,6 +2295,8 @@ struct map_access
   }
 };
 
+#if defined(HAVE_QT)
+
 /**
  *  @brief Specialization for QMap
  */
@@ -1705,6 +2351,8 @@ struct map_access<QHash<X, Y> >
   }
 };
 
+#endif
+
 /**
  *  @brief Implementation of the generic iterator adaptor for a specific container
  */
@@ -1852,6 +2500,12 @@ inline AdaptorBase *create_adaptor_by_category(const string_adaptor_tag & /*tag*
   return new StringAdaptorImpl<X> (v);
 }
 
+template <class X, class V>
+inline AdaptorBase *create_adaptor_by_category(const byte_array_adaptor_tag & /*tag*/, V v)
+{
+  return new ByteArrayAdaptorImpl<X> (v);
+}
+
 template <class X, class V>
 inline AdaptorBase *create_adaptor_by_category(const variant_adaptor_tag & /*tag*/, const V &v)
 {
@@ -1901,7 +2555,7 @@ inline AdaptorBase *adaptor_factory<Tag, X>::get (const X &v)
 template <class X>
 inline void copy_to (AdaptorBase &a, X &x, tl::Heap &heap)
 {
-  std::auto_ptr<AdaptorBase> t (adaptor_factory<typename gsi::type_traits<X &>::tag, X &>::get (x));
+  std::unique_ptr<AdaptorBase> t (adaptor_factory<typename gsi::type_traits<X &>::tag, X &>::get (x));
   a.copy_to (t.get (), heap);
 }
 
