@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@
 
 */
 
+#if defined(HAVE_QT)
 
 #ifndef HDR_imgNavigator
 #define HDR_imgNavigator
@@ -28,7 +29,8 @@
 
 namespace lay
 {
-  class PluginRoot;
+  class Dispatcher;
+  class LayoutViewWidget;
   class LayoutView;
   class ZoomService;
   class ViewService;
@@ -52,17 +54,14 @@ public:
   ~Navigator ();
 
   void background_color (QColor c);
-  img::Object *setup (lay::PluginRoot *root, img::Object *img);
+  img::Object *setup (lay::Dispatcher *root, img::Object *img);
 
-  lay::LayoutView *view () 
-  {
-    return mp_view;
-  }
+  lay::LayoutView *view ();
 
   void activate_service (lay::ViewService *service);
 
 private:
-  lay::LayoutView *mp_view;  
+  lay::LayoutViewWidget *mp_view;
   lay::ZoomService *mp_zoom_service;
 };
 
@@ -70,3 +69,4 @@ private:
 
 #endif
 
+#endif
