@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -29,18 +29,17 @@
 namespace img
 {
 
+extern const std::string cfg_images_visible;
+
 class PluginDeclaration
   : public lay::PluginDeclaration
 {
 public:
-  virtual void get_options (std::vector < std::pair<std::string, std::string> > & /*options*/) const
-  {
-    // .. nothing yet ..
-  }
-
   virtual void get_menu_entries (std::vector<lay::MenuEntry> &menu_entries) const;
-  virtual lay::Plugin *create_plugin (db::Manager *manager, lay::PluginRoot *, lay::LayoutView *view) const;
+  virtual lay::Plugin *create_plugin (db::Manager *manager, lay::Dispatcher *, lay::LayoutViewBase *view) const;
   virtual bool implements_editable (std::string &title) const;
+  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const;
+  virtual std::vector<std::string> additional_editor_options_pages () const;
 };
 
 }
