@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -20,12 +20,16 @@
 
 */
 
+#if defined(HAVE_QT)
 
 #include "imgPropertiesPage.h"
 #include "imgLandmarksDialog.h"
+#include "imgStream.h"
 #include "layLayoutView.h"
 #include "layFileDialog.h"
+#include "layQtTools.h"
 #include "tlExceptions.h"
+#include "tlFileUtils.h"
 
 namespace img
 {
@@ -36,11 +40,11 @@ const double max_gamma = 3.0;
 // -------------------------------------------------------------------------
 //  PropertiesPage implementation
 
-PropertiesPage::PropertiesPage (img::Service *service, QWidget *parent)
-  : lay::PropertiesPage (parent, service), mp_service (service), mp_direct_image (0)
+PropertiesPage::PropertiesPage (img::Service *service, db::Manager *manager, QWidget *parent)
+  : lay::PropertiesPage (parent, manager, service), mp_service (service), mp_direct_image (0)
 {
   mp_service->get_selection (m_selection);
-  m_pos = m_selection.begin ();
+  m_index = 0;
 
   mp_service->clear_highlights ();
 
@@ -48,7 +52,7 @@ PropertiesPage::PropertiesPage (img::Service *service, QWidget *parent)
 }
 
 PropertiesPage::PropertiesPage (QWidget *parent)
-  : lay::PropertiesPage (parent, 0), mp_service (0), mp_direct_image (0)
+  : lay::PropertiesPage (parent, 0, 0), mp_service (0), mp_direct_image (0)
 {
   init ();
 }
@@ -71,6 +75,7 @@ void
 PropertiesPage::init ()
 {
   m_no_signals = false;
+  m_in_color_mapping_signal = false;
 
   setupUi (this);
 
@@ -100,11 +105,17 @@ PropertiesPage::init ()
   connect (action, SIGNAL (triggered ()), this, SLOT (reverse_color_order ()));
   false_color_control->addAction (action);
 
-  color_pb->set_color (QColor ());
+  colors->set_color (std::make_pair (QColor (), QColor ()));
+  colors->setEnabled (false);
+  value_le->setEnabled (false);
+
+  attach_service (mp_service);
 
   connect (browse_pb, SIGNAL (clicked ()), this, SLOT (browse ()));
-  connect (color_pb, SIGNAL (color_changed (QColor)), false_color_control, SLOT (set_current_color (QColor)));
-  connect (false_color_control, SIGNAL (selection_changed (QColor)), color_pb, SLOT (set_color (QColor)));
+  connect (colors, SIGNAL (color_changed (std::pair<QColor, QColor>)), false_color_control, SLOT (set_current_color (std::pair<QColor, QColor>)));
+  connect (false_color_control, SIGNAL (selection_changed (std::pair<QColor, QColor>)), colors, SLOT (set_color (std::pair<QColor, QColor>)));
+  connect (false_color_control, SIGNAL (selection_changed (std::pair<QColor, QColor>)), this, SLOT (color_mapping_changed ()));
+  connect (false_color_control, SIGNAL (color_mapping_changed ()), this, SLOT (color_mapping_changed ()));
 
   connect (brightness_slider, SIGNAL (valueChanged (int)), this, SLOT (brightness_slider_changed (int)));
   connect (brightness_sb, SIGNAL (valueChanged (int)), this, SLOT (brightness_spinbox_changed (int)));
@@ -119,17 +130,39 @@ PropertiesPage::init ()
   connect (b_slider, SIGNAL (valueChanged (int)), this, SLOT (blue_slider_changed (int)));
   connect (b_sb, SIGNAL (valueChanged (double)), this, SLOT (blue_spinbox_changed (double)));
 
-  connect (false_color_control, SIGNAL (color_mapping_changed ()), this, SLOT (color_mapping_changed ()));
-  connect (false_color_control, SIGNAL (selection_changed ()), this, SLOT (color_mapping_changed ()));
-  connect (from_le, SIGNAL (returnPressed ()), this, SLOT (min_max_return_pressed ()));
-  connect (to_le, SIGNAL (returnPressed ()), this, SLOT (min_max_return_pressed ()));
-  connect (value_le, SIGNAL (returnPressed ()), this, SLOT (value_return_pressed ()));
+  connect (from_le, SIGNAL (editingFinished ()), this, SLOT (min_max_value_changed ()));
+  connect (to_le, SIGNAL (editingFinished ()), this, SLOT (min_max_value_changed ()));
+  connect (value_le, SIGNAL (editingFinished ()), this, SLOT (value_changed ()));
 
+  connect (width_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (height_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (x_offset_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (y_offset_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (angle_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (shear_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (persp_tx_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+  connect (persp_ty_le, SIGNAL (editingFinished ()), this, SIGNAL (edited ()));
+
+  connect (mirror_cbx, SIGNAL (clicked ()), this, SIGNAL (edited ()));
   connect (reset_pb, SIGNAL (clicked ()), this, SLOT (reset_pressed ()));
-  connect (preview_cbx, SIGNAL (clicked ()), this, SLOT (preview_checked ()));
+  connect (save_pb, SIGNAL (clicked ()), this, SLOT (save_pressed ()));
   connect (define_landmarks_pb, SIGNAL (clicked ()), this, SLOT (define_landmarks_pressed ()));
 }
 
+void
+PropertiesPage::attach_service (img::Service *service)
+{
+  layer_binding_cbx->set_new_layer_enabled (false);
+  layer_binding_cbx->set_no_layer_available (true);
+
+  if (service && service->view ()) {
+    int cv_index = service->view ()->active_cellview_index ();
+    layer_binding_cbx->set_view (service->view (), cv_index, true);
+  } else {
+    layer_binding_cbx->set_view (0, -1);
+  }
+}
+
 void
 PropertiesPage::invalidate ()
 {
@@ -139,44 +172,58 @@ PropertiesPage::invalidate ()
   }
 }
 
-void 
-PropertiesPage::back ()
+size_t
+PropertiesPage::count () const
 {
-  m_pos = m_selection.end ();
-  invalidate ();
+  return m_selection.size ();
 }
 
-void 
-PropertiesPage::front ()
+void
+PropertiesPage::select_entries (const std::vector<size_t> &entries)
 {
-  m_pos = m_selection.begin ();
-  invalidate ();
+  if (! entries.empty ()) {
+    m_index = entries.front ();
+    invalidate ();
+  }
 }
 
-bool 
-PropertiesPage::at_begin () const
+std::string
+PropertiesPage::description (size_t entry) const
 {
-  return (m_pos == m_selection.begin ());
+  const img::Object *obj = dynamic_cast <const img::Object *> (m_selection [entry]->ptr ());
+  if (! obj) {
+    return std::string ("nil");
+  }
+
+  std::string d = tl::to_string (tr ("Image"));
+  if (! obj->filename ().empty ()) {
+    d += "[" + tl::filename (obj->filename ()) + "]";
+  }
+  d += tl::sprintf ("(%dx%d)", obj->width (), obj->height ());
+  return d;
 }
 
-bool 
-PropertiesPage::at_end () const
+std::string
+PropertiesPage::description () const
 {
-  return (m_pos == m_selection.end ());
+  return tl::to_string (tr ("Images"));
 }
 
-void 
-PropertiesPage::operator-- ()
+void
+PropertiesPage::confine_selection (const std::vector<size_t> &remaining_entries)
 {
-  --m_pos;
-  invalidate ();
-}
+  if (! mp_service) {
+    return;
+  }
 
-void 
-PropertiesPage::operator++ ()
-{
-  ++m_pos;
-  invalidate ();
+  std::vector <img::Service::obj_iterator> org_selection;
+  m_selection.swap (org_selection);
+  for (auto i = remaining_entries.begin (); i != remaining_entries.end (); ++i) {
+    m_selection.push_back (org_selection [*i]);
+  }
+
+  mp_service->set_selection (m_selection);
+  mp_service->clear_highlights ();
 }
 
 void
@@ -198,21 +245,54 @@ PropertiesPage::readonly ()
   return false;
 }
 
-void 
-PropertiesPage::min_max_return_pressed ()
+void
+PropertiesPage::get_xmin_xmax (double &xmin, double &xmax, bool &has_error_out)
 {
-BEGIN_PROTECTED
+  bool has_error = false;
 
+  try {
+    tl::from_string_ext (tl::to_string (from_le->text ()), xmin);
+    lay::indicate_error (from_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (from_le, &ex);
+    has_error = true;
+  }
+
+  try {
+    tl::from_string_ext (tl::to_string (to_le->text ()), xmax);
+    lay::indicate_error (to_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (to_le, &ex);
+    has_error = true;
+  }
+
+  if (! has_error && xmin >= xmax) {
+    tl::Exception ex (tl::to_string (QObject::tr ("Invalid data value range (min. value must be less than max. value)")));
+    lay::indicate_error (from_le, &ex);
+    lay::indicate_error (to_le, &ex);
+    has_error = true;
+  }
+
+  if (has_error) {
+    has_error_out = true;
+  }
+}
+
+void 
+PropertiesPage::min_max_value_changed ()
+{
   value_le->setText (QString ());
   value_le->setEnabled (false);
 
-  color_pb->setEnabled (false_color_control->has_selection ());
+  colors->setEnabled (false_color_control->has_selection ());
+  colors->set_single_mode (false);
 
   double xmin, xmax;
-  tl::from_string (tl::to_string (from_le->text ()), xmin);
-  tl::from_string (tl::to_string (to_le->text ()), xmax);
-  if (xmin >= xmax) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Invalid data value range (min. value must be less than max. value)")));
+  bool has_error = false;
+  get_xmin_xmax (xmin, xmax, has_error);
+
+  if (has_error) {
+    return;
   }
 
   if (false_color_control->has_selection () && false_color_control->selected_node () > 0 && false_color_control->selected_node () < int (false_color_control->nodes ().size ()) - 1) {
@@ -223,13 +303,15 @@ BEGIN_PROTECTED
     value_le->setText (tl::to_qstring (tl::sprintf ("%.4g", xx)));
     value_le->setEnabled (true);
 
+  } else if (false_color_control->has_selection ()) {
+
+    colors->set_single_mode (true);
+
   }
 
   recompute_histogram ();
 
-  preview ();
-
-END_PROTECTED
+  emit edited ();
 }
 
 void
@@ -237,21 +319,20 @@ PropertiesPage::color_mapping_changed ()
 {
   if (! m_no_signals) {
 
+    bool has_error = false;
+
     value_le->setText (QString ());
     value_le->setEnabled (false);
 
-    color_pb->setEnabled (false_color_control->has_selection ());
+    colors->setEnabled (false_color_control->has_selection ());
+    colors->set_single_mode (false);
 
-    try {
+    if (false_color_control->has_selection () && false_color_control->selected_node () > 0 && false_color_control->selected_node () < int (false_color_control->nodes ().size ()) - 1) {
 
-      if (false_color_control->has_selection () && false_color_control->selected_node () > 0 && false_color_control->selected_node () < int (false_color_control->nodes ().size ()) - 1) {
+      double xmin, xmax;
+      get_xmin_xmax (xmin, xmax, has_error);
 
-        double xmin, xmax;
-        tl::from_string (tl::to_string (from_le->text ()), xmin);
-        tl::from_string (tl::to_string (to_le->text ()), xmax);
-        if (xmin >= xmax) {
-          throw tl::Exception ("");
-        }
+      if (! has_error) {
 
         double x = false_color_control->nodes () [false_color_control->selected_node ()].first;
         double xx = x * (xmax - xmin) + xmin;
@@ -261,40 +342,55 @@ PropertiesPage::color_mapping_changed ()
 
       }
 
-    } catch (...) { }
+    } else if (false_color_control->has_selection ()) {
 
-    preview ();
+      colors->set_single_mode (true);
+
+    }
+
+    if (! has_error) {
+      m_in_color_mapping_signal = true;
+      emit edited ();
+      m_in_color_mapping_signal = false;
+    }
 
   }
 }
 
 void
-PropertiesPage::value_return_pressed ()
+PropertiesPage::value_changed ()
 {
-BEGIN_PROTECTED
+  double xx = 0;
+  bool has_error = false;
 
   double xmin, xmax;
-  tl::from_string (tl::to_string (from_le->text ()), xmin);
-  tl::from_string (tl::to_string (to_le->text ()), xmax);
-  if (xmin >= xmax) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Invalid data value range (min. value must be less than max. value)")));
-  }
+  get_xmin_xmax (xmin, xmax, has_error);
 
   double x = 0.0;
-  tl::from_string (tl::to_string (value_le->text ()), x);
-
-  double xx = (x - xmin) / (xmax - xmin);
-  if (xx < 0 || xx > 1.0) {
-    throw tl::Exception (tl::to_string (QObject::tr ("The position entered (%g) must be between the minimum (%g) and maximum (%g) value")), x, xmin, xmax);
+  try {
+    tl::from_string_ext (tl::to_string (value_le->text ()), x);
+    lay::indicate_error (value_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (value_le, &ex);
+    has_error = true;
   }
 
-  m_no_signals = true;
-  false_color_control->set_current_position (xx);
-  m_no_signals = false;
+  xx = (x - xmin) / (xmax - xmin);
+  if (! has_error && (xx < 0 || xx > 1.0)) {
+    tl::Exception ex (tl::to_string (QObject::tr ("The position entered (%g) must be between the minimum (%g) and maximum (%g) value")), x, xmin, xmax);
+    lay::indicate_error (value_le, &ex);
+    has_error = true;
+  }
 
-  preview ();
+  if (! has_error) {
 
-END_PROTECTED
+    m_no_signals = true;
+    false_color_control->set_current_position (xx);
+    m_no_signals = false;
+
+    emit edited ();
+
+  }
 }
 
 inline double 
@@ -306,15 +402,19 @@ round_to_zero (double x)
 void 
 PropertiesPage::update ()
 {
+  if (m_in_color_mapping_signal) {
+    return;
+  }
+
   m_no_signals = true;
 
   if (mp_service) {
 
-    mp_service->highlight (std::distance (m_selection.begin (), m_pos));
+    mp_service->highlight (m_index);
 
     //  create a local copy in which we can apply modifications
     if (! mp_direct_image) {
-      const img::Object *image = dynamic_cast <const img::Object *> ((*m_pos)->ptr ());
+      const img::Object *image = dynamic_cast <const img::Object *> (m_selection [m_index]->ptr ());
       mp_direct_image = new img::Object (*image);
     }
 
@@ -380,6 +480,8 @@ PropertiesPage::update ()
   to_le->setText (tl::to_qstring (tl::to_string (mp_direct_image->max_value ())));
   to_le->setCursorPosition (0);
 
+  layer_binding_cbx->set_current_layer (mp_direct_image->layer_binding ());
+
   false_color_control->set_nodes (mp_direct_image->data_mapping ().false_color_nodes);
 
   brightness_slider->setValue (int (floor (mp_direct_image->data_mapping ().brightness * 100 + 0.5)));
@@ -418,8 +520,8 @@ PropertiesPage::recompute_histogram ()
   try {
 
     double xmin, xmax;
-    tl::from_string (tl::to_string (from_le->text ()), xmin);
-    tl::from_string (tl::to_string (to_le->text ()), xmax);
+    tl::from_string_ext (tl::to_string (from_le->text ()), xmin);
+    tl::from_string_ext (tl::to_string (to_le->text ()), xmax);
     if (xmin >= xmax) {
       throw tl::Exception ("");
     }
@@ -484,7 +586,7 @@ PropertiesPage::brightness_slider_changed (int value)
 
   m_no_signals = true;
   brightness_sb->setValue (value);
-  preview ();
+  emit edited ();
   m_no_signals = false;
 }
 
@@ -497,7 +599,7 @@ PropertiesPage::brightness_spinbox_changed (int value)
 
   m_no_signals = true;
   brightness_slider->setValue (value);
-  preview ();
+  emit edited ();
   m_no_signals = false;
 }
 
@@ -510,7 +612,7 @@ PropertiesPage::contrast_slider_changed (int value)
 
   m_no_signals = true;
   contrast_sb->setValue (value);
-  preview ();
+  emit edited ();
   m_no_signals = false;
 }
 
@@ -523,7 +625,7 @@ PropertiesPage::contrast_spinbox_changed (int value)
 
   m_no_signals = true;
   contrast_slider->setValue (value);
-  preview ();
+  emit edited ();
   m_no_signals = false;
 }
 
@@ -542,7 +644,7 @@ PropertiesPage::gamma_spinbox_changed (double value)
     gamma_slider->setValue (50 + int (0.5 + (value - 1.0) / (max_gamma - 1.0) * 50.0));
   }
 
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -564,7 +666,7 @@ PropertiesPage::gamma_slider_changed (int value)
   }
 
   gamma_sb->setValue (gamma);
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -581,7 +683,7 @@ PropertiesPage::red_slider_changed (int value)
   double gain = value * 0.02;
 
   r_sb->setValue (gain);
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -596,7 +698,7 @@ PropertiesPage::red_spinbox_changed (double value)
   m_no_signals = true;
 
   r_slider->setValue (int (0.5 + value * 50.0));
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -613,7 +715,7 @@ PropertiesPage::green_slider_changed (int value)
   double gain = value * 0.02;
 
   g_sb->setValue (gain);
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -628,7 +730,7 @@ PropertiesPage::green_spinbox_changed (double value)
   m_no_signals = true;
 
   g_slider->setValue (int (0.5 + value * 50.0));
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -645,7 +747,7 @@ PropertiesPage::blue_slider_changed (int value)
   double gain = value * 0.02;
 
   b_sb->setValue (gain);
-  preview ();
+  emit edited ();
 
   m_no_signals = false;
 }
@@ -660,61 +762,68 @@ PropertiesPage::blue_spinbox_changed (double value)
   m_no_signals = true;
 
   b_slider->setValue (int (0.5 + value * 50.0));
-  preview ();
+  emit edited ();
 
-  m_no_signals = false;
+  m_no_signals = false;  
 }
 
 void  
 PropertiesPage::black_to_white ()
 {
-  std::vector <std::pair <double, QColor> > nodes;
-  nodes.push_back (std::make_pair (0.0, QColor (0, 0, 0)));
-  nodes.push_back (std::make_pair (1.0, QColor (255, 255, 255)));
+  std::vector <std::pair <double, std::pair<tl::Color, tl::Color> > > nodes;
+  nodes.push_back (std::make_pair (0.0, std::make_pair (tl::Color (0, 0, 0), tl::Color (0, 0, 0))));
+  nodes.push_back (std::make_pair (1.0, std::make_pair (tl::Color (255, 255, 255), tl::Color (255, 255, 255))));
   false_color_control->set_nodes (nodes);
+  emit edited ();
 }
 
 void  
 PropertiesPage::white_to_black ()
 {
-  std::vector <std::pair <double, QColor> > nodes;
-  nodes.push_back (std::make_pair (0.0, QColor (255, 255, 255)));
-  nodes.push_back (std::make_pair (1.0, QColor (0, 0, 0)));
+  std::vector <std::pair <double, std::pair<tl::Color, tl::Color> > > nodes;
+  nodes.push_back (std::make_pair (0.0, std::make_pair (tl::Color (255, 255, 255), tl::Color (255, 255, 255))));
+  nodes.push_back (std::make_pair (1.0, std::make_pair (tl::Color (0, 0, 0), tl::Color (0, 0, 0))));
   false_color_control->set_nodes (nodes);
+  emit edited ();
 }
 
 void  
 PropertiesPage::red_to_blue ()
 {
-  std::vector <std::pair <double, QColor> > nodes;
-  nodes.push_back (std::make_pair (0.0, QColor (255, 0, 0)));
-  nodes.push_back (std::make_pair (1.0, QColor (0, 0, 255)));
+  std::vector <std::pair <double, std::pair<tl::Color, tl::Color> > > nodes;
+  nodes.push_back (std::make_pair (0.0, std::make_pair (tl::Color (255, 0, 0), tl::Color (255, 0, 0))));
+  nodes.push_back (std::make_pair (1.0, std::make_pair (tl::Color (0, 0, 255), tl::Color (0, 0, 255))));
   false_color_control->set_nodes (nodes);
-
+  emit edited ();
 }
 
 void  
 PropertiesPage::blue_to_red ()
 {
-  std::vector <std::pair <double, QColor> > nodes;
-  nodes.push_back (std::make_pair (0.0, QColor (0, 0, 255)));
-  nodes.push_back (std::make_pair (1.0, QColor (255, 0, 0)));
+  std::vector <std::pair <double, std::pair<tl::Color, tl::Color> > > nodes;
+  nodes.push_back (std::make_pair (0.0, std::make_pair (tl::Color (0, 0, 255), tl::Color (0, 0, 255))));
+  nodes.push_back (std::make_pair (1.0, std::make_pair (tl::Color (255, 0, 0), tl::Color (255, 0, 0))));
   false_color_control->set_nodes (nodes);
+  emit edited ();
 }
 
 void  
 PropertiesPage::reverse_color_order ()
 {
-  std::vector <std::pair <double, QColor> > nodes (false_color_control->nodes ());
+  std::vector <std::pair <double, std::pair<tl::Color, tl::Color> > > nodes (false_color_control->nodes ());
   for (size_t i = 0; i < nodes.size () / 2; ++i) {
-    std::swap (nodes [i].second, nodes [nodes.size () - 1 - i].second);
+    std::swap (nodes [i].second.second, nodes [nodes.size () - 1 - i].second.first);
+    std::swap (nodes [i].second.first, nodes [nodes.size () - 1 - i].second.second);
   }
   false_color_control->set_nodes (nodes);
+  emit edited ();
 }
 
 void 
-PropertiesPage::apply ()
+PropertiesPage::apply (bool /*commit*/)
 {
+  bool has_error = false;
+
   db::Matrix3d matrix = mp_direct_image->matrix ();
 
   //  The observer distance for perspective distortion is the average of the images width and height.
@@ -726,45 +835,89 @@ PropertiesPage::apply ()
   double w = matrix.mag_x (), h = matrix.mag_y (), x = matrix.disp ().x (), y = matrix.disp ().y (), 
          a = matrix.angle (), sa = matrix.shear_angle (), tx = matrix.perspective_tilt_x (z), ty = matrix.perspective_tilt_y (z);
 
-  bool mirror;
-
-  if (width_le->text () != tl::to_qstring (tl::micron_to_string (matrix.mag_x ()))) {
-    tl::from_string (tl::to_string (width_le->text ()), w);
-  }
-  if (height_le->text () != tl::to_qstring (tl::micron_to_string (matrix.mag_y ()))) {
-    tl::from_string (tl::to_string (height_le->text ()), h);
-  }
-  if (x_offset_le->text () != tl::to_qstring (tl::micron_to_string (round_to_zero (matrix.disp ().x ())))) {
-    tl::from_string (tl::to_string (x_offset_le->text ()), x);
-  }
-  if (y_offset_le->text () != tl::to_qstring (tl::micron_to_string (round_to_zero (matrix.disp ().y ())))) {
-    tl::from_string (tl::to_string (y_offset_le->text ()), y);
-  }
-  if (angle_le->text () != tl::to_qstring (tl::to_string (round_to_zero (matrix.angle ())))) {
-    tl::from_string (tl::to_string (angle_le->text ()), a);
-  }
-  if (shear_le->text () != tl::to_qstring (tl::to_string (round_to_zero (matrix.shear_angle ())))) {
-    tl::from_string (tl::to_string (shear_le->text ()), sa);
-  }
-  if (persp_tx_le->text () != tl::to_qstring (tl::to_string (round_to_zero (matrix.perspective_tilt_x (z))))) {
-    tl::from_string (tl::to_string (persp_tx_le->text ()), tx);
-  }
-  if (persp_ty_le->text () != tl::to_qstring (tl::to_string (round_to_zero (matrix.perspective_tilt_y (z))))) {
-    tl::from_string (tl::to_string (persp_ty_le->text ()), ty);
+  try {
+    tl::from_string_ext (tl::to_string (width_le->text ()), w);
+    if (w <= 0.0 || h <= 0.0) {
+      throw tl::Exception (tl::to_string (QObject::tr ("Pixel width or height must be positive, non-null values")));
+    }
+    lay::indicate_error (width_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (width_le, &ex);
+    has_error = true;
   }
 
-  mirror = mirror_cbx->isChecked ();
-
-  if (w <= 0.0 || h <= 0.0) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Pixel width or height must be positive, non-null values")));
+  try {
+    tl::from_string_ext (tl::to_string (height_le->text ()), h);
+    lay::indicate_error (height_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (height_le, &ex);
+    has_error = true;
   }
 
-  if (sa <= -45 || sa >= 45) {
-    throw tl::Exception (tl::to_string (QObject::tr ("The shear angle must be larger than -45 and less than 45 degree")));
+  try {
+    tl::from_string_ext (tl::to_string (x_offset_le->text ()), x);
+    lay::indicate_error (x_offset_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (x_offset_le, &ex);
+    has_error = true;
   }
 
-  if (tx <= -90 || tx >= 90 || ty <= -90 || ty >= 90) {
-    throw tl::Exception (tl::to_string (QObject::tr ("The perspective tilt angles must be larger than -90 and less than 90 degree")));
+  try {
+    tl::from_string_ext (tl::to_string (y_offset_le->text ()), y);
+    lay::indicate_error (y_offset_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (y_offset_le, &ex);
+    has_error = true;
+  }
+
+  try {
+    tl::from_string_ext (tl::to_string (angle_le->text ()), a);
+    lay::indicate_error (angle_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (angle_le, &ex);
+    has_error = true;
+  }
+
+  try {
+    tl::from_string_ext (tl::to_string (shear_le->text ()), sa);
+    if (sa <= -45 || sa >= 45) {
+      throw tl::Exception (tl::to_string (QObject::tr ("The shear angle must be larger than -45 and less than 45 degree")));
+    }
+    lay::indicate_error (shear_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (shear_le, &ex);
+    has_error = true;
+  }
+
+  try {
+    tl::from_string_ext (tl::to_string (persp_tx_le->text ()), tx);
+    if (tx <= -90 || tx >= 90) {
+      throw tl::Exception (tl::to_string (QObject::tr ("The perspective tilt angles must be larger than -90 and less than 90 degree")));
+    }
+    lay::indicate_error (persp_tx_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (persp_tx_le, &ex);
+    has_error = true;
+  }
+
+  try {
+    tl::from_string_ext (tl::to_string (persp_ty_le->text ()), ty);
+    if (ty <= -90 || ty >= 90) {
+      throw tl::Exception (tl::to_string (QObject::tr ("The perspective tilt angles must be larger than -90 and less than 90 degree")));
+    }
+    lay::indicate_error (persp_ty_le, (tl::Exception *) 0);
+  } catch (tl::Exception &ex) {
+    lay::indicate_error (persp_ty_le, &ex);
+    has_error = true;
+  }
+
+  bool mirror = mirror_cbx->isChecked ();
+
+  double xmin, xmax;
+  get_xmin_xmax (xmin, xmax, has_error);
+
+  if (has_error) {
+    throw tl::Exception (tl::to_string (tr ("At least one value is invalid - see highlighted entry fields")));
   }
 
   //  Compute the new observer distance
@@ -775,16 +928,11 @@ PropertiesPage::apply ()
   matrix = db::Matrix3d::disp (db::DVector (x, y)) * db::Matrix3d::perspective (tx, ty, z) * db::Matrix3d::rotation (a) * db::Matrix3d::shear (sa) * db::Matrix3d::mag (w, h) * db::Matrix3d::mirror (mirror);
   mp_direct_image->set_matrix (matrix);
 
-  double xmin, xmax;
-  tl::from_string (tl::to_string (from_le->text ()), xmin);
-  tl::from_string (tl::to_string (to_le->text ()), xmax);
-  if (xmin >= xmax) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Invalid data value range (min. value must be less than max. value)")));
-  }
-
   mp_direct_image->set_min_value (xmin);
   mp_direct_image->set_max_value (xmax);
 
+  mp_direct_image->set_layer_binding (layer_binding_cbx->current_layer_props ());
+
   img::DataMapping dm (mp_direct_image->data_mapping ());
   dm.brightness = brightness_sb->value () * 0.01;
   dm.contrast = contrast_sb->value () * 0.01;
@@ -796,7 +944,7 @@ PropertiesPage::apply ()
   mp_direct_image->set_data_mapping (dm);
 
   if (mp_service) {
-    mp_service->change_image (*m_pos, *mp_direct_image);
+    mp_service->change_image (m_selection [m_index], *mp_direct_image);
   }
 }
 
@@ -805,7 +953,7 @@ PropertiesPage::browse ()
 {
 BEGIN_PROTECTED
 
-  apply ();
+  apply (true);
 
   lay::FileDialog file_dialog (this, tl::to_string (QObject::tr ("Load Image File")), tl::to_string (QObject::tr ("All files (*)")));
 
@@ -826,6 +974,31 @@ BEGIN_PROTECTED
 END_PROTECTED
 }
 
+void
+PropertiesPage::save_pressed ()
+{
+BEGIN_PROTECTED
+
+  apply (true);
+
+  lay::FileDialog file_dialog (this, tl::to_string (QObject::tr ("Save As KLayout Image File")), tl::to_string (QObject::tr ("KLayout image files (*.lyimg);;All files (*)")));
+
+  std::string filename = mp_direct_image->filename ();
+  if (! filename.empty () && tl::extension (filename) != "lyimg") {
+    filename = tl::basename (filename) + ".lyimg";
+  }
+
+  if (file_dialog.get_save (filename)) {
+
+    tl::OutputFile file (filename);
+    tl::OutputStream stream (file);
+    img::ImageStreamer::write (stream, *mp_direct_image);
+
+  }
+
+END_PROTECTED
+}
+
 void
 PropertiesPage::reset_pressed ()
 {
@@ -851,30 +1024,7 @@ PropertiesPage::reset_pressed ()
 
   m_no_signals = false;
 
-  preview ();
-}
-
-void
-PropertiesPage::preview_checked ()
-{
-  preview ();
-}
-
-void
-PropertiesPage::preview ()
-{
-  if (preview_cbx->isChecked ()) {
-
-    BEGIN_PROTECTED_CLEANUP
-
-    apply (); // this is a HACK, because it changes the current object
-
-    END_PROTECTED_CLEANUP
-    {
-      preview_cbx->setChecked (false);
-    }
-
-  }
+  emit edited ();
 }
 
 void
@@ -882,9 +1032,12 @@ PropertiesPage::define_landmarks_pressed ()
 {
   if (mp_direct_image) {
     img::LandmarksDialog dialog (this, *mp_direct_image);
-    dialog.exec ();
+    if (dialog.exec ()) {
+      emit edited ();
+    }
   }
 }
 
 }
 
+#endif
