@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -38,14 +38,16 @@ class GSIHelpProvider
   : public HelpProvider
 {
 public:
-  virtual std::string folder () const;
-  virtual std::string title () const;
-  virtual void toc (std::vector<std::string> &t);
-  virtual QDomDocument get (const std::string &u) const;
+  GSIHelpProvider ();
+
+  virtual std::string folder (lay::HelpSource *src) const;
+  virtual std::string title (lay::HelpSource *src) const;
+  virtual void toc (lay::HelpSource *src, std::vector<std::string> &t);
+  virtual QDomDocument get (lay::HelpSource *src, const std::string &u) const;
 
 private:
   std::string produce_class_doc (const std::string &cls) const;
-  std::string produce_class_index () const;
+  std::string produce_class_index (HelpSource *src, const char *module_name) const;
 };
 
 }
