@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include "lymMacroInterpreter.h"
 #include "lymMacro.h"
 #include "gsiDecl.h"
+#include "tlFileUtils.h"
 
 #include <QDir>
 #include <QUrl>
@@ -46,55 +47,40 @@ MacroController::MacroController ()
     dm_sync_file_watcher (this, &MacroController::sync_file_watcher),
     dm_sync_files (this, &MacroController::sync_files)
 {
-  connect (&m_temp_macros, SIGNAL (menu_needs_update ()), this, SLOT (macro_collection_changed ()));
-  connect (&m_temp_macros, SIGNAL (macro_collection_changed (lym::MacroCollection *)), this, SLOT (macro_collection_changed ()));
+  //  .. nothing yet ..
 }
 
-static lay::MacroController::MacroCategory ruby_cat ()
+void
+MacroController::add_macro_category (const std::string &name, const std::string &description, const std::vector<std::string> &folders)
 {
   lay::MacroController::MacroCategory cat;
-  cat.name = "macros";
-  cat.description = tl::to_string (QObject::tr ("Ruby"));
-  cat.folders.push_back ("macros");
-  cat.folders.push_back ("ruby");
-  return cat;
-}
+  cat.name = name;
+  cat.description = description;
+  cat.folders = folders;
 
-static lay::MacroController::MacroCategory python_cat ()
-{
-  lay::MacroController::MacroCategory cat;
-  cat.name = "pymacros";
-  cat.description = tl::to_string (QObject::tr ("Python"));
-  cat.folders.push_back ("pymacros");
-  cat.folders.push_back ("python");
-  return cat;
-}
+  //  replace an existing category or add to back
+  for (auto c = m_macro_categories.begin (); c != m_macro_categories.end (); ++c) {
+    if (c->name == name) {
+      *c = cat;
+      return;
+    }
+  }
 
-static lay::MacroController::MacroCategory drc_cat ()
-{
-  lay::MacroController::MacroCategory cat;
-  cat.name = "drc";
-  cat.description = tl::to_string (QObject::tr ("DRC"));
-  cat.folders.push_back ("drc");
-  return cat;
+  m_macro_categories.push_back (cat);
 }
 
 void
 MacroController::finish ()
 {
+  lym::MacroCollection &root = lym::MacroCollection::root ();
+
+  root.clear ();
+
   //  Scan built-in macros
   //  These macros are always taken, even if there are no macros requested (they are required to
   //  fully form the API).
-  lym::MacroCollection::root ().add_folder (tl::to_string (QObject::tr ("Built-In")), ":/built-in-macros", "macros", true);
-  lym::MacroCollection::root ().add_folder (tl::to_string (QObject::tr ("Built-In")), ":/built-in-pymacros", "pymacros", true);
-
-  //  TODO: consider adding "drc" dynamically and allow more dynamic categories
-  //  We can do so if we first load the macros with the initial interpreters, then do autorun (which creates DSL interpreters) and then
-  //  register the remaining categories.
-
-  m_macro_categories.push_back (ruby_cat ());
-  m_macro_categories.push_back (python_cat ());
-  m_macro_categories.push_back (drc_cat ());
+  root.add_folder (tl::to_string (QObject::tr ("Built-In")), ":/built-in-macros", "macros", true);
+  root.add_folder (tl::to_string (QObject::tr ("Built-In")), ":/built-in-pymacros", "pymacros", true);
 
   //  scans the macros from techs and packages (this will allow autorun-early on them)
   //  and updates m_external_paths
@@ -118,13 +104,13 @@ MacroController::finish ()
               description += " - " + tl::to_string (tr ("%1 branch").arg (tl::to_qstring (*f)));
             }
 
-            lym::MacroCollection::root ().add_folder (description, mp, m_macro_categories[c].name, p->readonly);
+            root.add_folder (description, mp, m_macro_categories[c].name, p->readonly);
 
           }
 
         } else if (p->cat == m_macro_categories[c].name) {
 
-          lym::MacroCollection::root ().add_folder (p->description, p->path, m_macro_categories[c].name, p->readonly);
+          root.add_folder (p->description, p->path, m_macro_categories[c].name, p->readonly);
 
         }
 
@@ -147,7 +133,7 @@ MacroController::finish ()
 
   if (! m_no_implicit_macros) {
     for (std::vector <ExternalPathDescriptor>::const_iterator p = m_external_paths.begin (); p != m_external_paths.end (); ++p) {
-      lym::MacroCollection *mc = lym::MacroCollection::root ().add_folder (p->description, p->path, p->cat, p->readonly);
+      lym::MacroCollection *mc = root.add_folder (p->description, p->path, p->cat, p->readonly);
       if (mc) {
         mc->set_virtual_mode (p->type);
       }
@@ -160,8 +146,11 @@ MacroController::finish ()
 }
 
 void
-MacroController::initialized (lay::PluginRoot *root)
+MacroController::initialized (lay::Dispatcher *root)
 {
+  connect (&m_temp_macros, SIGNAL (menu_needs_update ()), this, SLOT (macro_collection_changed ()));
+  connect (&m_temp_macros, SIGNAL (macro_collection_changed (lym::MacroCollection *)), this, SLOT (macro_collection_changed ()));
+
   mp_mw = lay::MainWindow::instance ();
   if (mp_mw) {
     mp_macro_editor = new lay::MacroEditorDialog (root, &lym::MacroCollection::root ());
@@ -193,7 +182,7 @@ MacroController::initialized (lay::PluginRoot *root)
 }
 
 void
-MacroController::uninitialize (lay::PluginRoot * /*root*/)
+MacroController::uninitialize (lay::Dispatcher * /*root*/)
 {
   disconnect (&lym::MacroCollection::root (), SIGNAL (menu_needs_update ()), this, SLOT (macro_collection_changed ()));
   disconnect (&lym::MacroCollection::root (), SIGNAL (macro_collection_changed (lym::MacroCollection *)), this, SLOT (macro_collection_changed ()));
@@ -218,13 +207,8 @@ MacroController::uninitialize (lay::PluginRoot * /*root*/)
 }
 
 bool
-MacroController::configure (const std::string &key, const std::string &value)
+MacroController::configure (const std::string & /*key*/, const std::string & /*value*/)
 {
-  if (key == cfg_key_bindings) {
-    m_key_bindings = unpack_key_binding (value);
-  } else if (key == cfg_menu_items_hidden) {
-    m_menu_items_hidden = unpack_menu_items_hidden (value);
-  }
   return false;
 }
 
@@ -235,7 +219,7 @@ MacroController::config_finalize()
 }
 
 bool
-MacroController::can_exit (lay::PluginRoot * /*root*/) const
+MacroController::can_exit (lay::Dispatcher * /*root*/) const
 {
   if (mp_macro_editor) {
     return mp_macro_editor->can_exit ();
@@ -281,11 +265,11 @@ MacroController::drop_url (const std::string &path_or_url)
   }
 
   //  load and run macro
-  std::auto_ptr<lym::Macro> macro (new lym::Macro ());
+  std::unique_ptr<lym::Macro> macro (new lym::Macro ());
   macro->load_from (path);
   macro->set_file_path (path);
 
-  if (macro->is_autorun () || macro->show_in_menu ()) {
+  if ((macro->is_autorun () || macro->show_in_menu ()) && ! lay::ApplicationBase::instance ()->appdata_path ().empty ()) {
 
     //  install macro permanently
     if (QMessageBox::question (mp_mw,
@@ -351,6 +335,10 @@ MacroController::drop_url (const std::string &path_or_url)
 void
 MacroController::show_editor (const std::string &cat, bool force_add)
 {
+  if (macro_categories ().empty ()) {
+    throw tl::Exception (tl::to_string (QObject::tr ("Application has not been compiled with scripting support - no macro IDE available")));
+  }
+
   if (mp_macro_editor) {
     mp_macro_editor->show (cat, force_add);
   }
@@ -367,6 +355,8 @@ MacroController::sync_package_paths ()
 {
   std::vector<std::string> package_locations;
 
+  //  Add package locations for packages
+
   lay::SaltController *sc = lay::SaltController::instance ();
   if (sc) {
     lay::Salt &salt = sc->salt ();
@@ -375,6 +365,15 @@ MacroController::sync_package_paths ()
     }
   }
 
+  //  Add package locations for technologies which share the same structure
+
+  for (db::Technologies::const_iterator t = db::Technologies::instance ()->begin (); t != db::Technologies::instance ()->end (); ++t) {
+    std::string bd = t->base_path ();
+    if (! bd.empty () && tl::is_dir (bd)) {
+      package_locations.push_back (bd);
+    }
+  }
+
   //  refresh the package locations by first removing the package locations and then rebuilding
   //  TODO: maybe that is a performance bottleneck, but right now, remove_package_location doesn't do a lot.
 
@@ -396,22 +395,21 @@ MacroController::sync_package_paths ()
 void
 MacroController::sync_implicit_macros (bool ask_before_autorun)
 {
+  //  determine the paths currently in use
+  std::map<std::string, ExternalPathDescriptor> prev_folders_by_path;
+  for (std::vector<ExternalPathDescriptor>::const_iterator p = m_external_paths.begin (); p != m_external_paths.end (); ++p) {
+    prev_folders_by_path.insert (std::make_pair (p->path, *p));
+  }
+
+  //  gets the external paths (tech, packages) into m_external_paths
+  sync_macro_sources ();
+
   if (m_no_implicit_macros) {
 
-    sync_macro_sources ();
     sync_package_paths ();
 
   } else {
 
-    //  determine the paths currently in use
-    std::map<std::string, const ExternalPathDescriptor *> prev_folders_by_path;
-    for (std::vector<ExternalPathDescriptor>::const_iterator p = m_external_paths.begin (); p != m_external_paths.end (); ++p) {
-      prev_folders_by_path.insert (std::make_pair (p->path, p.operator-> ()));
-    }
-
-    //  gets the external paths (tech, packages) into m_external_paths
-    sync_macro_sources ();
-
     //  delete macro collections which are no longer required or update description
 
     std::vector<lym::MacroCollection *> folders_to_delete;
@@ -452,23 +450,39 @@ MacroController::sync_implicit_macros (bool ask_before_autorun)
 
     for (std::vector<ExternalPathDescriptor>::const_iterator p = m_external_paths.begin (); p != m_external_paths.end (); ++p) {
 
-      if (prev_folders_by_path.find (p->path) != prev_folders_by_path.end ()) {
-        continue;
-      }
+      auto pf = prev_folders_by_path.find (p->path);
+      if (pf != prev_folders_by_path.end ()) {
 
-      if (tl::verbosity () >= 20) {
-        tl::info << "Adding macro folder " << p->path << ", category '" << p->cat << "' for '" << p->description << "'";
-      }
+        if (pf->second.version != p->version) {
 
-      //  Add the folder. Note: it may happen that a macro folder for the tech specific macros already exists in
-      //  a non-tech context.
-      //  In that case, the add_folder method will return 0.
+          if (tl::verbosity () >= 20) {
+            tl::info << "New version (" << p->version << " vs. " << pf->second.version << ") of macro folder " << p->path << ", category '" << p->cat << "' for '" << p->description << "'";
+          }
+
+          lym::MacroCollection *mc = lym::MacroCollection::root ().folder_by_name (p->path);
+          if (mc) {
+            new_folders.push_back (mc);
+          }
+
+        }
+
+      } else {
+
+        if (tl::verbosity () >= 20) {
+          tl::info << "Adding macro folder " << p->path << ", category '" << p->cat << "' for '" << p->description << "'";
+        }
+
+        //  Add the folder. Note: it may happen that a macro folder for the tech specific macros already exists in
+        //  a non-tech context.
+        //  In that case, the add_folder method will return 0.
+
+        //  TODO: is it wise to make this writeable?
+        lym::MacroCollection *mc = lym::MacroCollection::root ().add_folder (p->description, p->path, p->cat, p->readonly);
+        if (mc) {
+          mc->set_virtual_mode (p->type);
+          new_folders.push_back (mc);
+        }
 
-      //  TODO: is it wise to make this writeable?
-      lym::MacroCollection *mc = lym::MacroCollection::root ().add_folder (p->description, p->path, p->cat, p->readonly);
-      if (mc) {
-        mc->set_virtual_mode (p->type);
-        new_folders.push_back (mc);
       }
 
     }
@@ -505,11 +519,11 @@ MacroController::sync_macro_sources ()
   std::map<std::string, std::vector<std::string> > grain_names_by_path;
   std::set<std::string> readonly_paths;
 
-  for (lay::Technologies::const_iterator t = lay::Technologies::instance ()->begin (); t != lay::Technologies::instance ()->end (); ++t) {
+  for (db::Technologies::const_iterator t = db::Technologies::instance ()->begin (); t != db::Technologies::instance ()->end (); ++t) {
     if (! t->base_path ().empty ()) {
       QDir base_dir (tl::to_qstring (t->base_path ()));
       if (base_dir.exists ()) {
-        std::string path = tl::to_string (base_dir.absolutePath ());
+        std::string path = tl::to_string (base_dir.canonicalPath ());
         tech_names_by_path [path].push_back (t->name ());
         if (t->is_readonly ()) {
           readonly_paths.insert (path);
@@ -552,7 +566,7 @@ MacroController::sync_macro_sources ()
             description += " - " + tl::to_string (tr ("%1 branch").arg (tl::to_qstring (*f)));
           }
 
-          external_paths.push_back (ExternalPathDescriptor (tl::to_string (macro_dir.path ()), description, macro_categories () [c].name, lym::MacroCollection::TechFolder, readonly_paths.find (t->first) != readonly_paths.end ()));
+          external_paths.push_back (ExternalPathDescriptor (tl::to_string (macro_dir.canonicalPath ()), description, macro_categories () [c].name, lym::MacroCollection::TechFolder, readonly_paths.find (t->first) != readonly_paths.end ()));
 
         }
 
@@ -584,7 +598,7 @@ MacroController::sync_macro_sources ()
             if (*f != macro_categories () [c].name) {
               description += " - " + tl::to_string (tr ("%1 branch").arg (tl::to_qstring (*f)));
             }
-            external_paths.push_back (ExternalPathDescriptor (tl::to_string (macro_dir.path ()), description, macro_categories () [c].name, lym::MacroCollection::SaltFolder, g->is_readonly ()));
+            external_paths.push_back (ExternalPathDescriptor (tl::to_string (macro_dir.canonicalPath ()), description, macro_categories () [c].name, lym::MacroCollection::SaltFolder, g->is_readonly (), g->version ()));
 
           }
 
@@ -654,8 +668,42 @@ static std::string menu_name (std::set<std::string> &used_names, const std::stri
   return name;
 }
 
+namespace {
+
+class RunMacroAction
+  : public lay::Action
+{
+public:
+  RunMacroAction (lym::Macro *lym)
+    : lay::Action (), mp_lym (lym)
+  {
+    if (lym->description ().empty ()) {
+      set_title (lym->path ());
+    } else {
+      set_title (lym->description ());
+    }
+  }
+
+  void triggered ()
+  {
+    if (mp_lym.get ()) {
+      mp_lym->run ();
+    }
+  }
+
+  lym::Macro *macro () const
+  {
+    return const_cast<lym::Macro *> (mp_lym.get ());
+  }
+
+private:
+  tl::weak_ptr<lym::Macro> mp_lym;
+};
+
+}
+
 void
-MacroController::add_macro_items_to_menu (lym::MacroCollection &collection, std::set<std::string> &used_names, std::set<std::string> &groups, const lay::Technology *tech)
+MacroController::add_macro_items_to_menu (lym::MacroCollection &collection, std::set<std::string> &used_names, std::set<std::string> &groups, const db::Technology *tech)
 {
   for (lym::MacroCollection::child_iterator c = collection.begin_children (); c != collection.end_children (); ++c) {
 
@@ -690,62 +738,32 @@ MacroController::add_macro_items_to_menu (lym::MacroCollection &collection, std:
       std::string gn = tl::trim (c->second->group_name ());
       if (! gn.empty () && groups.find (gn) == groups.end ()) {
         groups.insert (gn);
-        lay::Action as;
-        as.set_separator (true);
+        lay::Action *as = new lay::Action ();
+        as->set_separator (true);
         m_macro_actions.push_back (as);
         mp_mw->menu ()->insert_item (mp, menu_name (used_names, std::string ()), as);
       }
 
-      lay::Action a;
-      if (c->second->description ().empty ()) {
-        a.set_title (c->second->path ());
-      } else {
-        a.set_title (c->second->description ());
-      }
-      a.set_default_shortcut (sc);
+      lay::Action *a = new RunMacroAction (c->second);
+      a->set_default_shortcut (sc);
       m_macro_actions.push_back (a);
       mp_mw->menu ()->insert_item (mp, menu_name (used_names, c->second->name ()), a);
 
-      m_action_to_macro.insert (std::make_pair (a.qaction (), c->second));
-
-      lym::MacroSignalAdaptor *adaptor = new lym::MacroSignalAdaptor (a.qaction (), c->second);
-      QObject::connect (a.qaction (), SIGNAL (triggered ()), adaptor, SLOT (run ()));
-
     } else if (! sc.empty ()) {
 
       //  Create actions for shortcut-only actions too and add them to the main window
       //  to register their shortcut.
 
-      lay::Action a;
-      if (c->second->description ().empty ()) {
-        a.set_title (c->second->path ());
-      } else {
-        a.set_title (c->second->description ());
-      }
-      a.set_shortcut (sc);
+      lay::Action *a = new RunMacroAction (c->second);
+      a->set_shortcut (sc);
       m_macro_actions.push_back (a);
-
-      mp_mw->addAction (a.qaction ());
-      lym::MacroSignalAdaptor *adaptor = new lym::MacroSignalAdaptor (a.qaction (), c->second);
-      QObject::connect (a.qaction (), SIGNAL (triggered ()), adaptor, SLOT (run ()));
+      mp_mw->addAction (a->qaction ());
 
     }
 
   }
 }
 
-lym::Macro *
-MacroController::macro_for_action (const lay::Action *action)
-{
-  if (action) {
-    std::map<QAction *, lym::Macro *>::const_iterator a2m = m_action_to_macro.find (action->qaction ());
-    if (a2m != m_action_to_macro.end ()) {
-      return a2m->second;
-    }
-  }
-  return 0;
-}
-
 void
 MacroController::sync_with_external_sources ()
 {
@@ -766,7 +784,6 @@ void
 MacroController::macro_collection_changed ()
 {
   //  empty action to macro table now we know it's invalid
-  m_action_to_macro.clear ();
   dm_do_update_menu_with_macros ();
   dm_sync_file_watcher ();
 }
@@ -778,38 +795,31 @@ MacroController::do_update_menu_with_macros ()
     return;
   }
 
-  const lay::Technology *tech = 0;
+  const db::Technology *tech = 0;
   if (lay::TechnologyController::instance ()) {
     tech = lay::TechnologyController::instance ()->active_technology ();
   }
 
   //  delete all existing items
-  for (std::vector<lay::Action>::iterator a = m_macro_actions.begin (); a != m_macro_actions.end (); ++a) {
+  std::vector<lay::Action *> actions;
+  for (tl::weak_collection<lay::Action>::iterator a = m_macro_actions.begin (); a != m_macro_actions.end (); ++a) {
+    if (a.operator-> ()) {
+      actions.push_back (a.operator-> ());
+    }
+  }
+  for (std::vector<lay::Action *>::const_iterator a = actions.begin (); a != actions.end (); ++a) {
     mp_mw->menu ()->delete_items (*a);
   }
   m_macro_actions.clear ();
-  m_action_to_macro.clear ();
 
   std::set<std::string> groups;
   std::set<std::string> used_names;
   add_macro_items_to_menu (m_temp_macros, used_names, groups, tech);
   add_macro_items_to_menu (lym::MacroCollection::root (), used_names, groups, tech);
 
-  //  apply the custom keyboard shortcuts
-  for (std::vector<std::pair<std::string, std::string> >::const_iterator kb = m_key_bindings.begin (); kb != m_key_bindings.end (); ++kb) {
-    if (mp_mw->menu ()->is_valid (kb->first)) {
-      lay::Action a = mp_mw->menu ()->action (kb->first);
-      a.set_shortcut (kb->second);
-    }
-  }
-
-  //  apply the custom hidden flags
-  for (std::vector<std::pair<std::string, bool> >::const_iterator hf = m_menu_items_hidden.begin (); hf != m_menu_items_hidden.end (); ++hf) {
-    if (mp_mw->menu ()->is_valid (hf->first)) {
-      lay::Action a = mp_mw->menu ()->action (hf->first);
-      a.set_hidden (hf->second);
-    }
-  }
+  //  apply the custom keyboard shortcuts and hidden flags
+  mp_mw->apply_key_bindings ();
+  mp_mw->apply_hidden ();
 }
 
 void
@@ -832,10 +842,12 @@ add_collections_to_file_watcher (const lym::MacroCollection &collection, tl::Fil
 void
 MacroController::sync_file_watcher ()
 {
-  m_file_watcher->clear ();
-  m_file_watcher->enable (false);
-  add_collections_to_file_watcher (lym::MacroCollection::root (), m_file_watcher);
-  m_file_watcher->enable (true);
+  if (m_file_watcher) {
+    m_file_watcher->clear ();
+    m_file_watcher->enable (false);
+    add_collections_to_file_watcher (lym::MacroCollection::root (), m_file_watcher);
+    m_file_watcher->enable (true);
+  }
 }
 
 void
@@ -862,7 +874,8 @@ static tl::RegisteredClass<lay::PluginDeclaration> macro_controller_decl (new la
 
 static lym::Macro *macro_for_action (const lay::Action *action)
 {
-  return MacroController::instance () ? MacroController::instance ()->macro_for_action (action) : 0;
+  const RunMacroAction *rma = dynamic_cast<const RunMacroAction *> (action);
+  return rma ? rma->macro () : 0;
 }
 
 //  extend lay::Action with the ability to associate a macro with it
