@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -46,9 +46,7 @@ public:
    */
   TextProgress (int verbosity);
 
-  virtual void set_progress_can_cancel (bool f);
-  virtual void set_progress_text (const std::string &text);
-  virtual void set_progress_value (double v, const std::string &value);
+  virtual void update_progress (tl::Progress *progress);
   virtual void show_progress_bar (bool show);
 
 private:
