@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -30,20 +30,23 @@
 #include <vector>
 
 #include "tlObject.h"
-#include "tlFileSystemWatcher.h"
-#include "layTechnology.h"
 #include "dbLayout.h"
 #include "dbMetaInfo.h"
 #include "dbReader.h"
 #include "dbSaveLayoutOptions.h"
 #include "dbLoadLayoutOptions.h"
 #include "dbInstElement.h"
+#include "dbTechnology.h"
 #include "gsi.h"
 
+#if defined(HAVE_QT)
+#  include "tlFileSystemWatcher.h"
+#endif
+
 namespace lay 
 {
 
-class LayoutView;
+class LayoutViewBase;
 
 /**
  *  @brief A layout handle
@@ -83,7 +86,7 @@ public:
    *  If "force" is set to true, the layout will be given that name, irregardless if
    *  the name already is being used. If "force" is false, a new unique name is created.
    */
-  void rename (const std::string &name, bool force = false) throw (tl::Exception);
+  void rename (const std::string &name, bool force = false);
 
   /**
    *  @brief Gets the name of the handle
@@ -108,17 +111,14 @@ public:
   /**
    *  @brief Gets the technology attached to this layout
    */
-  const lay::Technology *technology () const;
+  const db::Technology *technology () const;
 
   /**
    *  @brief Gets the technology name for this layout
    *
    *  An empty name indicates the default technology should be used.
    */
-  const std::string &tech_name () const
-  {
-    return m_tech_name;
-  }
+  const std::string &tech_name () const;
 
   /**
    *  @brief Applies the given technology
@@ -144,6 +144,14 @@ public:
    */
   static LayoutHandle *find (const std::string &name);
 
+  /**
+   *  @brief Finds a handle by layout object
+   *
+   *  @param layout The Layout object bound to the handle
+   *  @return 0, if there is no layout object with this name. Otherwise a pointer to its handle
+   */
+  static LayoutHandle *find_layout (const db::Layout *layout);
+
   /**
    *  @brief Gets the names of all registered layout objects
    */
@@ -215,7 +223,7 @@ public:
    *  Save the layout under the given file name and with the given options.
    *  If update is true, this method updates the cell view's filename, title, save options and dirty flag.
    */
-  void save_as (const std::string &filename, tl::OutputStream::OutputStreamMode om, const db::SaveLayoutOptions &options, bool update = true);
+  void save_as (const std::string &filename, tl::OutputStream::OutputStreamMode om, const db::SaveLayoutOptions &options, bool update = true, int keep_backups = 0);
 
   /**
    *  @brief Sets the save options and a flag indicating whether they are valid
@@ -234,15 +242,6 @@ public:
     return m_save_options;
   }
 
-  /**
-   *  @brief Updates the given save options with attributes from this cell view
-   *
-   *  Some formats will initialize attributes from the cell view and the layout's
-   *  metadata (example: libname of GDS2). This method will update the options
-   *  if the layout provides attributes for initializing the latter.
-   */
-  void update_save_options (db::SaveLayoutOptions &options);
-
   /**
    *  @brief Gets a flag indicating whether the save options are valid
    *
@@ -290,24 +289,39 @@ public:
    */
   void layout_changed ();
 
+#if defined(HAVE_QT)
   /**
    *  @brief Gets the file system watcher that delivers events when one of the layouts gets updated
    */
   static tl::FileSystemWatcher &file_watcher ();
+#endif
+
+  /**
+   *  @brief Removes a file from the watcher
+   */
+  static void remove_file_from_watcher (const std::string &path);
+
+  /**
+   *  @brief Adds a file to the watcher
+   */
+  static void add_file_to_watcher (const std::string &path);
 
 private:
   db::Layout *mp_layout;
   int m_ref_count;
   std::string m_name;
   std::string m_filename;
-  std::string m_tech_name;
   bool m_dirty;
   db::SaveLayoutOptions m_save_options;
   bool m_save_options_valid;
   db::LoadLayoutOptions m_load_options;
 
+  void on_technology_changed ();
+
   static std::map <std::string, LayoutHandle *> ms_dict;
+#if defined(HAVE_QT)
   static tl::FileSystemWatcher *mp_file_watcher;
+#endif
 };
 
 /**
@@ -524,6 +538,11 @@ public:
    */
   db::ICplxTrans context_trans () const;
 
+  /**
+   *  @brief Retrive the accumulated transformation induced by the context part of the path as a micron-unit transformation
+   */
+  db::DCplxTrans context_dtrans () const;
+
   /**
    *  @brief Deep copy of the cellview
    *
@@ -571,7 +590,7 @@ public:
    *  @param cv The reference to the target cellview
    *  @param view The reference to the layout view
    */
-  CellViewRef (lay::CellView *cv, lay::LayoutView *view);
+  CellViewRef (lay::CellView *cv, lay::LayoutViewBase *view);
 
   /**
    *  @brief Gets the cellview index of this reference
@@ -580,9 +599,9 @@ public:
   int index () const;
 
   /**
-   *  @brief Gets the LayoutView the reference is pointing to
+   *  @brief Gets the LayoutViewBase the reference is pointing to
    */
-  lay::LayoutView *view ();
+  lay::LayoutViewBase *view ();
 
   /**
    *  @brief Equality: Gives true, if the cellviews are identical
@@ -732,9 +751,14 @@ public:
    */
   db::ICplxTrans context_trans () const;
 
+  /**
+   *  @brief Retrive the accumulated transformation induced by the context part of the path in micron units
+   */
+  db::DCplxTrans context_dtrans() const;
+
 private:
   tl::weak_ptr<lay::CellView> mp_cv;
-  tl::weak_ptr<lay::LayoutView> mp_view;
+  tl::weak_ptr<lay::LayoutViewBase> mp_view;
 };
 
 }
