@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -24,8 +24,10 @@
 
 #include "layDitherPattern.h"
 #include "tlAssert.h"
+#include "tlThreads.h"
 
 #include <ctype.h>
+#include <string.h>
 #include <algorithm>
 
 namespace lay
@@ -49,8 +51,8 @@ static const char *dither_strings [] = {
   "*.\n"
   ".*",
 
-  // 3: coarsly dotted
-  "coarsly dotted",
+  // 3: coarsely dotted
+  "coarsely dotted",
   "*...\n"
   "....\n"
   "..*.\n"
@@ -460,6 +462,8 @@ static const char *dither_strings [] = {
 // ---------------------------------------------------------------------
 //  DitherPatternInfo implementation
 
+static tl::Mutex s_mutex;
+
 DitherPatternInfo::DitherPatternInfo ()
   : m_width (1), m_height (1), m_order_index (0)
 {
@@ -480,23 +484,31 @@ DitherPatternInfo &
 DitherPatternInfo::operator= (const DitherPatternInfo &d)
 {
   if (&d != this) {
-
-    m_order_index = d.m_order_index;
-    m_name = d.m_name;
-    m_width = d.m_width;
-    m_pattern_stride = d.m_pattern_stride;
-    m_height = d.m_height;
-
-    for (size_t i = 0; i < sizeof (m_pattern) / sizeof (m_pattern [0]); ++i) {
-      m_pattern [i] = &m_buffer [0] + (d.m_pattern [i] - d.m_buffer);
-    }
-    memcpy (m_buffer, d.m_buffer, sizeof (m_buffer));
+    tl::MutexLocker locker (& s_mutex);
+    assign_no_lock (d);
   }
 
   return *this;
 }
 
-bool 
+void
+DitherPatternInfo::assign_no_lock (const DitherPatternInfo &d)
+{
+  m_scaled_pattern.reset ();
+
+  m_order_index = d.m_order_index;
+  m_name = d.m_name;
+  m_width = d.m_width;
+  m_pattern_stride = d.m_pattern_stride;
+  m_height = d.m_height;
+
+  for (size_t i = 0; i < sizeof (m_pattern) / sizeof (m_pattern [0]); ++i) {
+    m_pattern [i] = &m_buffer [0] + (d.m_pattern [i] - d.m_buffer);
+  }
+  memcpy (m_buffer, d.m_buffer, sizeof (m_buffer));
+}
+
+bool
 DitherPatternInfo::same_bitmap (const DitherPatternInfo &d) const
 {
   if (m_width != d.m_width || m_height != d.m_height) {
@@ -555,9 +567,11 @@ DitherPatternInfo::operator< (const DitherPatternInfo &d) const
   return m_order_index < d.m_order_index;
 }
 
+#if defined(HAVE_QT)
+
 // TODO including a scaling algorithm in this formula, or give more resolution to the dither
 QBitmap
-DitherPatternInfo::get_bitmap (int width, int height) const
+DitherPatternInfo::get_bitmap (int width, int height, int frame_width) const
 {
   if (height < 0) {
     height = 36;
@@ -565,6 +579,7 @@ DitherPatternInfo::get_bitmap (int width, int height) const
   if (width < 0) {
     width = 34;
   }
+  unsigned int fw = frame_width < 0 ? 1 : frame_width;
 
   const uint32_t * const *p = pattern ();
   unsigned int stride = (width + 7) / 8;
@@ -572,17 +587,14 @@ DitherPatternInfo::get_bitmap (int width, int height) const
   unsigned char *data = new unsigned char[stride * height];
   memset (data, 0x00, size_t (stride * height));
 
-  for (unsigned int i = 1; i < (unsigned int)(height - 1); ++i) {
-    for (unsigned int j = 0; j < stride; ++j) {
-      data [i * stride + j] = 0xff;
+  for (unsigned int i = 0; i < (unsigned int) height; ++i) {
+    uint32_t w = 0xffffffff;
+    if (i >= fw && i < (unsigned int) height - fw) {
+      w = *(p [(height - 1 - i) % m_height]);
     }
-  }
-
-  for (unsigned int i = 0; i < (unsigned int)(height - 4); ++i) {
-    uint32_t w = *(p [(height - 5 - i) % m_height]);
-    for (unsigned int j = 0; j < (unsigned int)(width - 2); ++j) {
-      if (! (w & (1 << (j % m_width)))) {
-        data [stride * (i + 2) + (j + 1) / 8] &= ~(1 << ((j + 1) % 8));
+    for (unsigned int j = 0; j < (unsigned int) width; ++j) {
+      if (j < fw || j >= (unsigned int) width - fw || (w & (1 << (j % m_width))) != 0) {
+        data [stride * i + j / 8] |= (1 << (j % 8));
       }
     }
   }
@@ -593,13 +605,24 @@ DitherPatternInfo::get_bitmap (int width, int height) const
   return bitmap;
 }
 
+#endif
+
 void 
 DitherPatternInfo::set_pattern (const uint32_t *pt, unsigned int w, unsigned int h) 
+{
+  tl::MutexLocker locker (& s_mutex);
+  m_scaled_pattern.reset (0);
+
+  set_pattern_impl (pt, w, h);
+}
+
+void
+DitherPatternInfo::set_pattern_impl (const uint32_t *pt, unsigned int w, unsigned int h)
 {
   //  pattern size must be 1x1 at least
   if (w == 0 || h == 0) {
     uint32_t zero = 0;
-    set_pattern (&zero, 1, 1);
+    set_pattern_impl (&zero, 1, 1);
     return;
   }
 
@@ -649,6 +672,192 @@ DitherPatternInfo::set_pattern (const uint32_t *pt, unsigned int w, unsigned int
   }
 }
 
+void
+DitherPatternInfo::set_pattern (const uint64_t *pt, unsigned int w, unsigned int h)
+{
+  tl::MutexLocker locker (& s_mutex);
+  m_scaled_pattern.reset (0);
+
+  set_pattern_impl (pt, w, h);
+}
+
+void
+DitherPatternInfo::set_pattern_impl (const uint64_t *pt, unsigned int w, unsigned int h)
+{
+  //  pattern size must be 1x1 at least
+  if (w == 0 || h == 0) {
+    uint32_t zero = 0;
+    set_pattern_impl (&zero, 1, 1);
+    return;
+  }
+
+  memset (m_buffer, 0, sizeof (m_buffer));
+
+  if (w >= 64) {
+    w = 64;
+  }
+  m_width = w;
+
+  if (h >= 64) {
+    h = 64;
+  }
+  m_height = h;
+
+  //  compute pattern stride
+  m_pattern_stride = 1;
+  while ((m_pattern_stride * 32) % w != 0) {
+    ++m_pattern_stride;
+  }
+
+  uint32_t *pp = &m_buffer[0];
+
+  for (unsigned int j = 0; j < sizeof (m_pattern) / sizeof (m_pattern [0]); ++j) {
+
+    m_pattern [j] = pp;
+
+    uint64_t din = pt[j % h];
+    uint64_t dd = din;
+
+    unsigned int b = 0;
+    for (unsigned int i = 0; i < m_pattern_stride; ++i) {
+      uint32_t dout = 0;
+      for (uint32_t m = 1; m != 0; m <<= 1) {
+        if ((dd & 1) != 0) {
+          dout |= m;
+        }
+        dd >>= 1;
+        if (++b == w) {
+          dd = din;
+          b = 0;
+        }
+      }
+      *pp++ = dout;
+    }
+
+  }
+}
+
+const DitherPatternInfo &
+DitherPatternInfo::scaled (unsigned int n) const
+{
+  if (n <= 1) {
+    return *this;
+  }
+
+  tl::MutexLocker locker (& s_mutex);
+
+  if (! m_scaled_pattern.get ()) {
+    m_scaled_pattern.reset (new std::map<unsigned int, DitherPatternInfo> ());
+  }
+
+  auto i = m_scaled_pattern->find (n);
+  if (i != m_scaled_pattern->end ()) {
+    return i->second;
+  }
+
+  DitherPatternInfo &sp = (*m_scaled_pattern) [n];
+  sp.assign_no_lock (*this);
+  sp.scale_pattern (n);
+  return sp;
+}
+
+void
+DitherPatternInfo::scale_pattern (unsigned int n)
+{
+  //  limit scale factor such that the width and height do not get larger than 64
+  while (n * m_width > 64 || n * m_height > 64) {
+    --n;
+  }
+
+  if (n <= 1) {
+    return;
+  }
+
+  std::vector<uint64_t> new_pattern;
+  new_pattern.resize (n * m_height, (uint64_t) 0);
+
+  for (unsigned int r = 0; r < m_height; ++r) {
+
+    const uint32_t *p = pattern () [r];
+    const uint32_t *pb = pattern () [(r + m_height - 1) % m_height];
+    const uint32_t *pt = pattern () [(r + 1) % m_height];
+
+    for (unsigned int l = 0; l < n; ++l) {
+
+      const uint32_t *py1 = (l < n / 2) ? pb : pt;
+      const uint32_t *py2 = (l < n / 2) ? pt : pb;
+
+      uint64_t d = 0;
+      uint64_t mm = 1;
+      uint32_t m = 1;
+      uint32_t mmax = 1 << m_width;
+      uint32_t ml = m_width > 1 ? (1 << (m_width - 1)) : 1;
+      uint32_t mr = m_width > 1 ? 2 : 1;
+
+      for (unsigned int c = 0; c < m_width; ++c) {
+        for (unsigned int b = 0; b < n; ++b) {
+          if ((*p & m) != 0) {
+            d |= mm;
+          } else {
+            //  Try interpolation.
+            //  In the following cases, the center's pixel lower-right quadrant fill be filled:
+            //
+            //  (A1)     (A2)     (A3)
+            //  x 0 0    x 0 0    x 0 1
+            //  0 0 1    0 0 1    0 0 1
+            //  0 1 x    1 1 x    0 1 x
+            //
+            //  (B1)     (B2)
+            //  0 1 x    0 0 0
+            //  0 0 1    1 0 1
+            //  0 1 x    x 1 x
+            //
+            //  For easy implementation, we encode the pattern into a byte k with the following significant bits
+            //  (for lower-right subpixel, mirrored accordingly for the other subpixels)
+            //
+            //  k bits:
+            //  0 1 2
+            //  3 - 4
+            //  5 6 7
+            //
+            uint32_t mx1 = (b < n / 2) ? ml : mr;
+            uint32_t mx2 = (b < n / 2) ? mr : ml;
+            uint8_t k = ((*py2 & mx2) != 0 ? 1   : 0) |
+                        ((*py2 & m)   != 0 ? 2   : 0) |
+                        ((*py2 & mx1) != 0 ? 4   : 0) |
+                        ((*p & mx2)   != 0 ? 8   : 0) |
+                        ((*p & mx1)   != 0 ? 16  : 0) |
+                        ((*py1 & mx2) != 0 ? 32  : 0) |
+                        ((*py1 & m)   != 0 ? 64  : 0) |
+                        ((*py1 & mx1) != 0 ? 128 : 0);
+            if ((k & 0x7e) == 0x50 /*(A1)*/ ||
+                (k & 0x7e) == 0x70 /*(A2)*/ ||
+                (k & 0x7e) == 0x54 /*(A3)*/ ||
+                (k & 0x7b) == 0x52 /*(B1)*/ ||
+                (k & 0x5f) == 0x58 /*(B2)*/) {
+              d |= mm;
+            }
+          }
+          mm <<= 1;
+        }
+        m <<= 1;
+        if ((ml <<= 1) == mmax) {
+          ml = 1;
+        }
+        if ((mr <<= 1) == mmax) {
+          mr = 1;
+        }
+      }
+
+      new_pattern [r * n + l] = d;
+
+    }
+
+  }
+
+  set_pattern_impl (new_pattern.begin ().operator-> (), n * m_width, n * m_height);
+}
+
 std::string
 DitherPatternInfo::to_string () const
 {
@@ -771,8 +980,8 @@ struct ReplaceDitherPatternOp
   DitherPatternInfo m_old, m_new;
 };
 
-DitherPattern::DitherPattern ()
-  : QObject (), db::Object (0)
+DitherPattern::DitherPattern () :
+    db::Object (0)
 {
   for (unsigned int d = 0; d < sizeof (dither_strings) / sizeof (dither_strings [0]); d += 2) {
     m_pattern.push_back (DitherPatternInfo ());
@@ -786,8 +995,8 @@ DitherPattern::~DitherPattern ()
   //  .. nothing yet ..
 }
 
-DitherPattern::DitherPattern (const DitherPattern &p)
-  : QObject (), db::Object (0)
+DitherPattern::DitherPattern (const DitherPattern &p) :
+    db::Object (0)
 {
   m_pattern = p.m_pattern;
 }
@@ -807,16 +1016,6 @@ DitherPattern::operator= (const DitherPattern &p)
   return *this;
 }
 
-QBitmap
-DitherPattern::get_bitmap (unsigned int i, int width, int height) const
-{
-  if (i < count ()) {
-    return m_pattern [i].get_bitmap (width, height);
-  } else {
-    return m_pattern [1].get_bitmap (width, height);
-  }
-}
-
 const DitherPatternInfo &
 DitherPattern::pattern (unsigned int i) const
 {
@@ -831,11 +1030,8 @@ DitherPattern::pattern (unsigned int i) const
 void 
 DitherPattern::replace_pattern (unsigned int i, const DitherPatternInfo &p)
 {
-  bool chg = false;
-
   while (i >= count ()) {
     m_pattern.push_back (DitherPatternInfo ());
-    chg = true;
   }
 
   if (m_pattern [i] != p) {
@@ -843,12 +1039,6 @@ DitherPattern::replace_pattern (unsigned int i, const DitherPatternInfo &p)
       manager ()->queue (this, new ReplaceDitherPatternOp (i, m_pattern [i], p));
     }
     m_pattern [i] = p;
-    chg = true;
-  }
-
-  //  if something has changed emit the signal
-  if (chg) {
-    emit changed ();
   }
 }
 
@@ -876,6 +1066,14 @@ DitherPattern::add_pattern (const DitherPatternInfo &p)
   return index;
 }
 
+void
+DitherPattern::scale_pattern (unsigned int n)
+{
+  for (auto i = m_pattern.begin (); i != m_pattern.end (); ++i) {
+    i->scale_pattern (n);
+  }
+}
+
 namespace {
   struct display_order
   {
@@ -950,13 +1148,13 @@ DitherPattern::merge (const DitherPattern &other, std::map<unsigned int, unsigne
 {
   //  insert the standard pattern into the map (for completeness)
   for (iterator c = begin (); c != begin_custom (); ++c) {
-    index_map.insert (std::make_pair (std::distance (begin (), c), std::distance (begin (), c)));
+    index_map.insert (std::make_pair ((unsigned int) std::distance (begin (), c), (unsigned int) std::distance (begin (), c)));
   }
 
   //  build an index of present pattern
   std::map <DitherPatternInfo, unsigned int, pattern_less_f> patterns;
   for (iterator c = begin_custom (); c != end (); ++c) {
-    patterns.insert (std::make_pair (*c, std::distance (begin (), c)));
+    patterns.insert (std::make_pair (*c, (unsigned int) std::distance (begin (), c)));
   }
 
   //  map the pattern of other into *this, possibly creating new ones
@@ -969,7 +1167,7 @@ DitherPattern::merge (const DitherPattern &other, std::map<unsigned int, unsigne
     } else {
       new_index = p->second;
     }
-    index_map.insert (std::make_pair (std::distance (other.begin (), c), new_index));
+    index_map.insert (std::make_pair ((unsigned int) std::distance (other.begin (), c), new_index));
   }
 }
 
