@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 
 #include "tlProgress.h"
 #include "layFinder.h"
+#include "layTextInfo.h"
 
 namespace lay
 {
@@ -62,7 +63,7 @@ static int inst_point_sel_tests = 10000;
 
 Finder::Finder (bool point_mode, bool top_level_sel)
   : m_min_level (0), m_max_level (0),
-    mp_layout (0), mp_view (0), m_cv_index (0), m_point_mode (point_mode), m_top_level_sel (top_level_sel)
+    mp_layout (0), mp_view (0), m_cv_index (0), m_point_mode (point_mode), m_catch_all (false), m_consider_viewport (true), m_top_level_sel (top_level_sel)
 {
   m_distance = std::numeric_limits<double>::max ();
 }
@@ -86,10 +87,13 @@ Finder::closer (double d)
 }
 
 void 
-Finder::start (lay::LayoutView *view, const lay::CellView &cv, unsigned int cv_index, const std::vector<db::ICplxTrans> &trans, const db::Box &region, int min_level, int max_level, const std::vector<int> &layers)
+Finder::start (lay::LayoutViewBase *view, unsigned int cv_index, const std::vector<db::DCplxTrans> &trans, const db::DBox &region, const db::DBox &scan_region, int min_level, int max_level, const std::vector<unsigned int> &layers)
 {
+  const lay::CellView &cv = view->cellview (cv_index);
+
+  reset_counter ();
+
   m_layers = layers;
-  m_region = region; 
   mp_layout = &cv->layout ();
   mp_view = view;
   m_cv_index = cv_index;
@@ -97,94 +101,148 @@ Finder::start (lay::LayoutView *view, const lay::CellView &cv, unsigned int cv_i
   m_max_level = std::max (m_min_level, std::min (max_level, m_top_level_sel ? ((int) cv.specific_path ().size () + 1) : max_level));
 
   if (layers.size () == 1) {
-    m_box_convert = db::box_convert <db::CellInst> (*mp_layout, (unsigned int) layers [0]);
-    m_cell_box_convert = db::box_convert <db::Cell> ((unsigned int) layers [0]);
+
+    m_box_convert = db::box_convert <db::CellInst, false> (*mp_layout, layers [0]);
+    m_cell_box_convert = db::box_convert <db::Cell, false> (layers [0]);
+
   } else {
-    m_box_convert = db::box_convert <db::CellInst> (*mp_layout);
-    m_cell_box_convert = db::box_convert <db::Cell> ();
+
+    m_box_convert = db::box_convert <db::CellInst, false> (*mp_layout);
+    m_cell_box_convert = db::box_convert <db::Cell, false> ();
+
   }
 
   m_path.erase (m_path.begin (), m_path.end ());
 
-  for (std::vector<db::ICplxTrans>::const_iterator t = trans.begin (); t != trans.end (); ++t) {
-    do_find (*cv.cell (), cv.specific_path ().size (), *t * cv.context_trans ());
+  for (std::vector<db::DCplxTrans>::const_iterator t = trans.begin (); t != trans.end (); ++t) {
+
+    db::VCplxTrans it = (*t * db::CplxTrans (mp_layout->dbu ())).inverted ();
+    m_region = it * region;
+    m_scan_region = it * scan_region;
+
+    do_find (*cv.cell (), int (cv.specific_path ().size ()), view->viewport ().trans () * *t, cv.context_trans ());
+
+  }
+}
+
+void
+Finder::test_edge (const db::ICplxTrans &trans, const db::Edge &edge, double &distance, bool &match)
+{
+  if (test_edge (trans, edge, true, distance, match) == 0) {
+    test_edge (trans, edge, false, distance, match);
   }
 }
 
 unsigned int
-Finder::test_edge (const db::Edge &edg, double &distance, bool &match)
+Finder::test_edge (const db::ICplxTrans &trans, const db::Edge &edg, bool points, double &distance, bool &match)
 {
+  db::Point p1 = trans * edg.p1 ();
+  db::Point p2 = trans * edg.p2 ();
+
   unsigned int ret = 0;
 
-  //  we hit the region with the edge end points - take the closest vertex
-  if (m_region.contains (edg.p1 ()) || m_region.contains (edg.p2 ())) {
+  if (points) {
 
-    double d1 = edg.p1 ().double_distance (m_region.center ());
-    double d2 = edg.p2 ().double_distance (m_region.center ());
+    //  we hit the region with the edge end points - take the closest vertex
+    if (m_region.contains (p1) || m_region.contains (p2)) {
+
+      double d1 = p1.double_distance (m_region.center ());
+      double d2 = p2.double_distance (m_region.center ());
+      if (d1 < d2) {
+        ret = 1;
+      } else {
+        ret = 2;
+      }
+
+      double d = std::min (d1, d2);
+      //  add a penalty of 1 DBU for being on the wrong
+      //  side of the edge - this favors the right edge
+      //  in case of butting corners
+      if (ret == 1) {
+        if (db::sprod_sign (m_region.center () - p1, p2 - p1) < 0) {
+          d += trans.ctrans (1);
+        }
+      } else {
+        if (db::sprod_sign (m_region.center () - p2, p1 - p2) < 0) {
+          d += trans.ctrans (1);
+        }
+      }
+
+      if (! match || d < distance) {
+        distance = d;
+      }
+
+      match = true;
 
-    //  snap to the point - nothing can get closer
-    distance = 0.0;
-    if (d1 < d2) {
-      ret = 1;
-    } else {
-      ret = 2;
     }
 
-    match = true;
-
-  }
+  } else {
   
-  //  if the edge cuts through the active region: test the
-  //  edge as a whole
-  if (ret == 0 && edg.clipped (m_region).first) {
-    double d = edg.distance_abs (m_region.center ());
-    if (! match || d < distance) {
-      distance = d;
+    //  if the edge cuts through the active region: test the
+    //  edge as a whole
+    db::Edge edg_trans (p1, p2);
+    if (edg_trans.clipped (m_region).first) {
+
+      double d = edg_trans.distance_abs (m_region.center ());
+      if (! match || d < distance) {
+        distance = d;
+      }
+
       ret = 3;
+      match = true;
+
     }
-    match = true;
+
   }
 
   return ret;
 }
 
 void
-Finder::do_find (const db::Cell &cell, int level, const db::ICplxTrans &t)
+Finder::do_find (const db::Cell &cell, int level, const db::DCplxTrans &vp, const db::ICplxTrans &t)
 {
   if (level <= m_max_level /*take level of cell itself*/ 
       && cell.is_proxy () 
       && m_layers.size () == 1 
-      && (unsigned int) m_layers [0] == mp_layout->guiding_shape_layer ()) {
+      && m_layers [0] == mp_layout->guiding_shape_layer ()) {
 
     //  when looking at the guiding shape layer, we can visit this cell as well allowing to find the guiding shapes
 
-    db::Box touch_box (t.inverted () * m_region);
+    db::ICplxTrans it = t.inverted ();
+    db::Box scan_box (it * m_scan_region);
+    db::Box hit_box (it * m_region);
 
     if (level >= m_min_level) {
-      visit_cell (cell, touch_box, t, level);
+      visit_cell (cell, hit_box, scan_box, vp, t, level);
     }
 
   } else if (level < m_max_level 
-      && (t * m_cell_box_convert (cell)).touches (m_region) 
+      && (t * m_cell_box_convert (cell)).touches (m_scan_region)
       && (mp_view->select_inside_pcells_mode () || !cell.is_proxy ()) 
-      && !mp_view->is_cell_hidden (cell.cell_index (), m_cv_index)) {
+      && !mp_view->is_cell_hidden (cell.cell_index (), m_cv_index)
+      && !cell.is_ghost_cell ()) {
 
-    db::Box touch_box (t.inverted () * m_region);
+    db::ICplxTrans it = t.inverted ();
+    db::Box scan_box (it * m_scan_region);
+    db::Box hit_box (it * m_region);
 
     if (level >= m_min_level) {
-      visit_cell (cell, touch_box, t, level);
+      visit_cell (cell, hit_box, scan_box, vp, t, level);
     }
 
-    db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
+    db::Cell::touching_iterator inst = cell.begin_touching (scan_box);
     while (! inst.at_end ()) {
 
       const db::CellInstArray &cell_inst = inst->cell_inst ();
-      for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, m_box_convert); ! p.at_end (); ++p) {
+      for (db::CellInstArray::iterator p = cell_inst.begin_touching (scan_box, m_box_convert); ! p.at_end (); ++p) {
 
         m_path.push_back (db::InstElement (*inst, p));
 
+        checkpoint ();
+
         do_find (mp_layout->cell (cell_inst.object ().cell_index ()), 
                  level + 1,
+                 vp,
                  t * cell_inst.complex_trans (*p));
 
         m_path.pop_back ();
@@ -201,13 +259,15 @@ Finder::do_find (const db::Cell &cell, int level, const db::ICplxTrans &t)
 // -------------------------------------------------------------
 //  ShapeFinder implementation
 
-ShapeFinder::ShapeFinder (bool point_mode, bool top_level_sel, db::ShapeIterator::flags_type flags, const std::set<lay::ObjectInstPath> *excludes)
+ShapeFinder::ShapeFinder (bool point_mode, bool top_level_sel, db::ShapeIterator::flags_type flags, const std::set<lay::ObjectInstPath> *excludes, bool capture_all_shapes)
   : Finder (point_mode, top_level_sel), 
     mp_excludes ((excludes && !excludes->empty ()) ? excludes : 0),
     m_flags (flags), m_cv_index (0), m_topcell (0), 
-    mp_prop_sel (0), m_inv_prop_sel (false), mp_progress (0)
+    mp_text_info (0),
+    mp_prop_sel (0), m_inv_prop_sel (false), mp_progress (0),
+    m_capture_all_shapes (capture_all_shapes)
 {
-  m_tries = point_sel_tests;
+  m_try_counter = m_tries = point_sel_tests;
 }
 
 struct LPContextEqualOp
@@ -259,9 +319,9 @@ struct LPContextCompareOp
 };
 
 bool
-ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
+ShapeFinder::find (LayoutViewBase *view, const db::DBox &region_mu)
 {
-  tl::AbsoluteProgress progress (tl::to_string (QObject::tr ("Selecting ...")));
+  tl::AbsoluteProgress progress (tl::to_string (tr ("Selecting ...")));
   progress.set_unit (1000);
   progress.set_format ("");
   mp_progress = &progress;
@@ -269,6 +329,9 @@ ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
   m_context_layers.clear ();
   m_cells_with_context.clear ();
 
+  lay::TextInfo text_info (view);
+  mp_text_info = (m_flags & db::ShapeIterator::Texts) != 0 && point_mode () ? &text_info : 0;
+
   std::vector<lay::LayerPropertiesConstIterator> lprops;
   for (lay::LayerPropertiesConstIterator lp = view->begin_layers (); ! lp.at_end (); ++lp) {
     if (lp->is_visual ()) {
@@ -278,7 +341,7 @@ ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
 
   std::sort (lprops.begin (), lprops.end (), LPContextCompareOp ());
 
-  std::vector<int> layers;
+  std::vector<unsigned int> layers;
   for (std::vector<lay::LayerPropertiesConstIterator>::const_iterator llp = lprops.begin (); llp != lprops.end (); ) {
 
     layers.clear ();
@@ -286,7 +349,10 @@ ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
     lay::LayerPropertiesConstIterator lp0 = *llp;
     LPContextEqualOp eq;
     do {
-      layers.push_back ((*llp)->layer_index ());
+      int li = (*llp)->layer_index ();
+      if (li >= 0) {
+        layers.push_back ((unsigned int) li);
+      }
       ++llp;
     } while (llp != lprops.end () && eq(lp0, *llp));
 
@@ -324,9 +390,9 @@ ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
 }
 
 bool 
-ShapeFinder::find (lay::LayoutView *view, const lay::LayerProperties &lprops, const db::DBox &region_mu)
+ShapeFinder::find (lay::LayoutViewBase *view, const lay::LayerProperties &lprops, const db::DBox &region_mu)
 {
-  tl::AbsoluteProgress progress (tl::to_string (QObject::tr ("Selecting ...")));
+  tl::AbsoluteProgress progress (tl::to_string (tr ("Selecting ...")));
   progress.set_unit (1000);
   progress.set_format ("");
   mp_progress = &progress;
@@ -334,16 +400,22 @@ ShapeFinder::find (lay::LayoutView *view, const lay::LayerProperties &lprops, co
   m_cells_with_context.clear ();
   m_context_layers.clear ();
 
-  std::vector<int> layers;
-  layers.push_back (lprops.layer_index ());
+  lay::TextInfo text_info (view);
+  mp_text_info = (m_flags & db::ShapeIterator::Texts) != 0 ? &text_info : 0;
+
+  std::vector<unsigned int> layers;
+  int li = lprops.layer_index ();
+  if (li >= 0) {
+    layers.push_back ((unsigned int) li);
+  }
   bool result = find_internal (view, lprops.cellview_index (), &lprops.prop_sel (), lprops.inverse_prop_sel (), lprops.hier_levels (), lprops.trans (), layers, region_mu);
 
   mp_progress = 0;
   return result;
 }
 
-bool 
-ShapeFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const std::set<db::properties_id_type> *prop_sel, bool inv_prop_sel, const lay::HierarchyLevelSelection &hier_sel, const std::vector<db::DCplxTrans> &trans_mu, const std::vector<int> &layers, const db::DBox &region_mu)
+bool
+ShapeFinder::find_internal (lay::LayoutViewBase *view, unsigned int cv_index, const std::set<db::properties_id_type> *prop_sel, bool inv_prop_sel, const lay::HierarchyLevelSelection &hier_sel, const std::vector<db::DCplxTrans> &trans_mu, const std::vector<unsigned int> &layers, const db::DBox &region_mu)
 {
   m_cv_index = cv_index;
 
@@ -354,14 +426,6 @@ ShapeFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const
 
   m_topcell = cv.cell_index ();
 
-  double dbu = cv->layout ().dbu ();
-  db::Box region = db::VCplxTrans (1.0 / dbu) * region_mu;
-  std::vector<db::ICplxTrans> trans;
-  trans.reserve(trans_mu.size());
-  for (std::vector<db::DCplxTrans>::const_iterator t = trans_mu.begin(); t != trans_mu.end(); ++t) {
-    trans.push_back(db::VCplxTrans(1.0 / dbu) * *t * db::CplxTrans(dbu));
-  }
-
   mp_prop_sel = prop_sel;
   m_inv_prop_sel = inv_prop_sel;
 
@@ -376,13 +440,43 @@ ShapeFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const
     max_level = hier_sel.to_level (ctx_path_length, max_level);
   }
 
-  //  actually find
-  try {
-    start (view, cv, m_cv_index, trans, region, min_level, max_level, layers);
-  } catch (StopException) {
-    // ..
+  auto flags_saved = m_flags;
+
+  if ((m_flags & db::ShapeIterator::Texts) != 0 && mp_text_info && ! mp_text_info->point_mode ()) {
+
+    m_flags = db::ShapeIterator::Texts;
+
+    try {
+
+      //  for catching all labels we search the whole view area
+      db::DBox scan_region_mu = view->viewport ().box ();
+      start (view, m_cv_index, trans_mu, region_mu, scan_region_mu, min_level, max_level, layers);
+
+    } catch (StopException) {
+      //  ...
+    } catch (...) {
+      m_flags = flags_saved;
+      throw;
+    }
+
+    m_flags = db::ShapeIterator::flags_type (flags_saved - db::ShapeIterator::Texts);
+
   }
 
+  try {
+
+    //  another pass with tight search box and without texts
+    start (view, m_cv_index, trans_mu, region_mu, region_mu, min_level, max_level, layers);
+
+  } catch (StopException) {
+    //  ...
+  } catch (...) {
+    m_flags = flags_saved;
+    throw;
+  }
+
+  m_flags = flags_saved;
+
   //  return true if anything was found
   return ! m_founds.empty ();
 }
@@ -393,15 +487,29 @@ ShapeFinder::checkpoint ()
   if (! point_mode ()) {
     ++*mp_progress;
   } else {
-    if (--m_tries < 0) {
+    if (--m_try_counter < 0) {
       throw StopException ();
     }
   }
 }
 
-void 
-ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::ICplxTrans &t, int /*level*/)
+void
+ShapeFinder::reset_counter ()
 {
+  m_try_counter = m_tries;
+}
+
+void 
+ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &hit_box, const db::Box &scan_box, const db::DCplxTrans &vp, const db::ICplxTrans &t, int /*level*/)
+{
+  checkpoint ();
+
+  //  Viewport in current cell coordinate space (DBU)
+  db::Box viewport_box;
+  if (consider_viewport ()) {
+    viewport_box = (vp * db::CplxTrans (layout ().dbu ()) * t).inverted () * db::DBox (0, 0, view ()->viewport ().width (), view ()->viewport ().height ());
+  }
+
   if (! m_context_layers.empty ()) {
 
     std::map<db::cell_index_type, bool>::const_iterator ctx = m_cells_with_context.find (cell.cell_index ());
@@ -426,21 +534,33 @@ ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const
 
   if (! point_mode ()) {
 
-    checkpoint ();
+    for (std::vector<unsigned int>::const_iterator l = layers ().begin (); l != layers ().end (); ++l) {
 
-    for (std::vector<int>::const_iterator l = layers ().begin (); l != layers ().end (); ++l) {
+      if (layers ().size () == 1 || (layers ().size () > 1 && cell.bbox (*l).touches (scan_box))) {
 
-      if (layers ().size () == 1 || (layers ().size () > 1 && cell.bbox ((unsigned int) *l).touches (search_box))) {
+        checkpoint ();
 
-        const db::Shapes &shapes = cell.shapes ((unsigned int) *l);
+        const db::Shapes &shapes = cell.shapes (*l);
 
-        db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
+        db::ShapeIterator shape = shapes.begin_touching (scan_box, m_flags, mp_prop_sel, m_inv_prop_sel);
         while (! shape.at_end ()) {
 
           checkpoint ();
 
+          db::Box bbox;
+          if (! view ()->text_visible () && shape->is_text ()) {
+            //  ignore texts if the view does not show them
+          } else if (text_info () && shape->is_text ()) {
+            db::CplxTrans t_dbu = db::CplxTrans (layout ().dbu ()) * t;
+            db::Text text;
+            shape->text (text);
+            bbox = t_dbu.inverted () * text_info ()->bbox (t_dbu * text, vp);
+          } else {
+            bbox = shape->bbox ();
+          }
+
           //  in box mode, just test the boxes
-          if (shape->bbox ().inside (search_box)) {
+          if (bbox.inside (hit_box)) {
 
             m_founds.push_back (lay::ObjectInstPath ());
             m_founds.back ().set_cv_index (m_cv_index);
@@ -466,46 +586,56 @@ ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const
 
   } else {
 
-    for (std::vector<int>::const_iterator l = layers ().begin (); l != layers ().end (); ++l) {
+    for (std::vector<unsigned int>::const_iterator l = layers ().begin (); l != layers ().end (); ++l) {
 
-      if (layers ().size () == 1 || (layers ().size () > 1 && cell.bbox ((unsigned int) *l).touches (search_box))) {
+      if (layers ().size () == 1 || (layers ().size () > 1 && cell.bbox (*l).touches (scan_box))) {
 
         checkpoint ();
 
         const db::Shapes &shapes = cell.shapes (*l);
 
-        db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
+        db::ShapeIterator shape = shapes.begin_touching (scan_box, m_flags, mp_prop_sel, m_inv_prop_sel);
         while (! shape.at_end ()) {
 
+          checkpoint ();
+
           bool match = false;
           double d = std::numeric_limits<double>::max ();
 
-          checkpoint ();
-
-          db::Point point (search_box.center ());
+          db::Point point (hit_box.center ());
 
           //  in point mode, test the edges and use a "closest" criterion
           if (shape->is_polygon ()) {
 
+            bool any_valid_edge = m_capture_all_shapes;
             for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); ++e) {
-              test_edge (t * *e, d, match);
+              if (viewport_box.empty () || (*e).clipped (viewport_box).first) {
+                any_valid_edge = true;
+                test_edge (t, *e, d, match);
+              }
             }
 
             //  test if inside the polygon
-            if (! match && inside_poly (shape->begin_edge (), point) >= 0) { 
+            if (! match && any_valid_edge && inside_poly (shape->begin_edge (), point) >= 0) {
               d = t.ctrans (poly_dist (shape->begin_edge (), point)); 
               match = true;
             }
 
           } else if (shape->is_path ()) {
 
+            bool any_valid_edge = m_capture_all_shapes;
+
             //  test the "spine"
-            db::Shape::point_iterator pt = shape->begin_point (); 
+            db::Shape::point_iterator pt = shape->begin_point ();
             if (pt != shape->end_point ()) {
               db::Point p (*pt);
               ++pt;
               for (; pt != shape->end_point (); ++pt) {
-                test_edge (t * db::Edge (p, *pt), d, match);
+                db::Edge e (p, *pt);
+                if (viewport_box.empty () || e.clipped (viewport_box).first) {
+                  any_valid_edge = true;
+                  test_edge (t, e, d, match);
+                }
                 p = *pt;
               }
             }
@@ -514,18 +644,27 @@ ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const
             db::Polygon poly;
             shape->polygon (poly);
             for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
-              test_edge (t * *e, d, match);
+              if (viewport_box.empty () || (*e).clipped (viewport_box).first) {
+                any_valid_edge = true;
+                test_edge (t, *e, d, match);
+              }
             }
 
             //  test if inside the polygon
-            if (! match && inside_poly (poly.begin_edge (), point) >= 0) { 
+            if (! match && any_valid_edge && inside_poly (poly.begin_edge (), point) >= 0) {
               d = t.ctrans (poly_dist (poly.begin_edge (), point));
               match = true;
             }
 
-          } else if (shape->is_box ()) {
+          } else if (shape->is_box () || shape->is_point () || (shape->is_text () && view ()->text_visible ())) {
 
-            const db::Box &box = shape->box ();
+            db::Box box = shape->bbox ();
+            if (text_info () && shape->is_text ()) {
+              db::CplxTrans t_dbu = db::CplxTrans (layout ().dbu ()) * t;
+              db::Text text;
+              shape->text (text);
+              box = t_dbu.inverted () * text_info ()->bbox (t_dbu * text, vp);
+            }
 
             //  point-like boxes are handles which attract the finder
             if (box.width () == 0 && box.height () == 0) {
@@ -533,66 +672,54 @@ ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const
               match = true;
             } else {
 
+              bool any_valid_edge = m_capture_all_shapes;
+
               //  convert to polygon and test those edges
               db::Polygon poly (box);
               for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
-                test_edge (t * *e, d, match);
+                if (viewport_box.empty () || (*e).clipped (viewport_box).first) {
+                  any_valid_edge = true;
+                  test_edge (t, *e, d, match);
+                }
               }
 
-              if (! match && box.contains (search_box.center ())) {
-                d = t.ctrans (poly_dist (poly.begin_edge (), point)); 
+              if (! match && any_valid_edge && box.contains (hit_box.center ())) {
+                d = t.ctrans (poly_dist (poly.begin_edge (), point));
                 match = true;
               }
 
             }
 
-          } else if (shape->is_text ()) {
-
-            db::Point tp (shape->text_trans () * db::Point ());
-            if (search_box.contains (tp)) {
-              d = t.ctrans (tp.distance (search_box.center ()));
-              match = true;
-            }
-
           }
 
           if (match) {
 
+            lay::ObjectInstPath found;
+            found.set_cv_index (m_cv_index);
+            found.set_topcell (m_topcell);
+            found.assign_path (path ().begin (), path ().end ());
+            found.set_layer (*l);
+            found.set_shape (*shape);
+
             if (mp_excludes) {
 
               //  with an exclude list first create the selection item so we can check
               //  if it's part of the exclude set.
 
-              lay::ObjectInstPath found;
-              found.set_cv_index (m_cv_index);
-              found.set_topcell (m_topcell);
-              found.assign_path (path ().begin (), path ().end ());
-              found.set_layer (*l);
-              found.set_shape (*shape);
-
               //  in point mode just store the found object that has the least "distance" and is
               //  not in the exclude set
-              if (mp_excludes->find (found) == mp_excludes->end () && closer (d)) {
+              match = (mp_excludes->find (found) == mp_excludes->end ());
 
-                if (m_founds.empty ()) {
-                  m_founds.push_back (found);
-                } else {
-                  m_founds.front () = found;
-                }
-              }
+            }
 
-            } else if (closer (d)) {
+            if (match && (catch_all () || closer (d))) {
 
               //  in point mode just store that found that has the least "distance"
-              if (m_founds.empty ()) {
-                m_founds.push_back (lay::ObjectInstPath ());
+              if (m_founds.empty () || catch_all ()) {
+                m_founds.push_back (found);
               }
 
-              m_founds.back ().set_cv_index (m_cv_index);
-              m_founds.back ().set_topcell (m_topcell);
-              m_founds.back ().assign_path (path ().begin (), path ().end ());
-              m_founds.back ().set_layer (*l);
-              m_founds.back ().set_shape (*shape);
+              m_founds.back () = found;
 
             }
 
@@ -620,6 +747,8 @@ InstFinder::InstFinder (bool point_mode, bool top_level_sel, bool full_arrays, b
     m_full_arrays (full_arrays), 
     m_enclose_insts (enclose_inst), 
     m_visible_layers (visible_layers),
+    m_consider_ghost_cells (true),
+    m_consider_normal_cells (true),
     mp_view (0), 
     mp_progress (0)
 {
@@ -627,14 +756,14 @@ InstFinder::InstFinder (bool point_mode, bool top_level_sel, bool full_arrays, b
 }
 
 bool
-InstFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
+InstFinder::find (lay::LayoutViewBase *view, const db::DBox &region_mu)
 {
-  tl::AbsoluteProgress progress (tl::to_string (QObject::tr ("Selecting ...")));
+  tl::AbsoluteProgress progress (tl::to_string (tr ("Selecting ...")));
   progress.set_unit (1000);
   progress.set_format ("");
   mp_progress = &progress;
 
-  std::set< std::pair<db::DCplxTrans, int> > variants = view->cv_transform_variants ();
+  std::set< std::pair<db::DCplxTrans, int> > variants = view->cv_transform_variants_with_empty ();
   for (std::set< std::pair<db::DCplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {
     find (view, v->second, v->first, region_mu);
   }
@@ -644,9 +773,9 @@ InstFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
 }
 
 bool 
-InstFinder::find (lay::LayoutView *view, unsigned int cv_index, const db::DCplxTrans &trans_mu, const db::DBox &region_mu)
+InstFinder::find (LayoutViewBase *view, unsigned int cv_index, const db::DCplxTrans &trans_mu, const db::DBox &region_mu)
 {
-  tl::AbsoluteProgress progress (tl::to_string (QObject::tr ("Selecting ...")));
+  tl::AbsoluteProgress progress (tl::to_string (tr ("Selecting ...")));
   progress.set_unit (1000);
   progress.set_format ("");
   mp_progress = &progress;
@@ -658,7 +787,7 @@ InstFinder::find (lay::LayoutView *view, unsigned int cv_index, const db::DCplxT
 }
 
 bool 
-InstFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const db::DCplxTrans &trans_mu, const db::DBox &region_mu)
+InstFinder::find_internal (LayoutViewBase *view, unsigned int cv_index, const db::DCplxTrans &trans_mu, const db::DBox &region_mu)
 {
   const lay::CellView &cv = view->cellview (cv_index);
   if (! cv.is_valid ()) {
@@ -672,24 +801,22 @@ InstFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const d
         m_visible_layer_indexes.push_back (l->layer_index ());
       }
     }
-  }
-
-  if (!m_visible_layers || view->guiding_shapes_visible ()) {
-    m_visible_layer_indexes.push_back (cv->layout ().guiding_shape_layer ());
+    //  add guiding shape and error layers so we can select cells by error markers or guiding shapes
+    if (view->guiding_shapes_visible ()) {
+      m_visible_layer_indexes.push_back (cv->layout ().guiding_shape_layer ());
+    }
+    m_visible_layer_indexes.push_back (cv->layout ().error_layer ());
   }
 
   m_cv_index = cv_index;
   m_topcell = cv.cell ()->cell_index ();
   mp_view = view;
 
-  double dbu = cv->layout ().dbu ();
-  db::Box region = db::VCplxTrans (1.0 / dbu) * region_mu;
-
   //  actually find
   try {
-    std::vector<db::ICplxTrans> tv;
-    tv.push_back (db::VCplxTrans (1.0 / dbu) * trans_mu * db::CplxTrans (dbu));
-    start (view, cv, cv_index, tv, region, view->get_min_hier_levels (), view->get_max_hier_levels (), std::vector<int> ());
+    std::vector<db::DCplxTrans> tv;
+    tv.push_back (trans_mu);
+    start (view, cv_index, tv, region_mu, region_mu, view->get_min_hier_levels (), view->get_max_hier_levels (), std::vector<unsigned int> ());
   } catch (StopException) {
     // ..
   }
@@ -698,37 +825,71 @@ InstFinder::find_internal (lay::LayoutView *view, unsigned int cv_index, const d
   return ! m_founds.empty ();
 }
 
-void 
-InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::ICplxTrans &t, int level)
+void
+InstFinder::checkpoint ()
 {
   if (! point_mode ()) {
-
     ++*mp_progress;
+  } else {
+    if (--m_try_counter < 0) {
+      throw StopException ();
+    }
+  }
+}
+
+void
+InstFinder::reset_counter ()
+{
+  m_try_counter = m_tries;
+}
+
+bool
+InstFinder::consider_cell (const db::Cell &cell) const
+{
+  return cell.is_ghost_cell () ? m_consider_ghost_cells : m_consider_normal_cells;
+}
+
+void 
+InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::Box & /*scan_box*/, const db::DCplxTrans &vp, const db::ICplxTrans &t, int level)
+{
+  checkpoint ();
+
+  //  Viewport in current cell coordinate space (DBU)
+  db::Box viewport_box;
+  if (consider_viewport ()) {
+    viewport_box = (vp * db::CplxTrans (layout ().dbu ()) * t).inverted () * db::DBox (0, 0, view ()->viewport ().width (), view ()->viewport ().height ());
+  }
+
+  if (! point_mode ()) {
 
     //  look for instances to check here ..
-    db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
-    while (! inst.at_end ()) {
+    for (db::Cell::touching_iterator inst = cell.begin_touching (search_box); ! inst.at_end (); ++inst) {
 
       const db::CellInstArray &cell_inst = inst->cell_inst ();
       const db::Cell &inst_cell = layout ().cell (cell_inst.object ().cell_index ());
 
-      ++*mp_progress;
+      checkpoint ();
 
-      //  just consider the instances exactly at the last level of 
+      if (! consider_cell (inst_cell)) {
+        continue;
+      }
+
+      //  just consider the instances exactly at the last level of
       //  hierarchy (this is where the boxes are drawn) or of cells that
       //  are hidden.
-      if (level == max_level () - 1 || inst_cell.is_proxy () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
+      if (level == max_level () - 1 || inst_cell.is_proxy () || inst_cell.is_ghost_cell () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
 
         db::box_convert <db::CellInst, false> bc (layout ());
         for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
         
-          ++*mp_progress;
+          checkpoint ();
 
           db::Box ibox;
-          if (inst_cell.bbox ().empty ()) {
-            ibox = db::Box (db::Point (0, 0), db::Point (0, 0));
-          } else if (! m_visible_layers || level == mp_view->get_max_hier_levels () - 1 || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
-            ibox = inst_cell.bbox ();
+          if (! m_visible_layers || level == mp_view->get_max_hier_levels () - 1 || inst_cell.is_ghost_cell () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
+            ibox = inst_cell.bbox_with_empty ();
+          } else if (inst_cell.bbox ().empty ()) {
+            //  empty cells cannot be found by visible layers, so we always select them here
+            ibox = inst_cell.bbox_with_empty ();
           } else {
             for (std::vector<int>::const_iterator l = m_visible_layer_indexes.begin (); l != m_visible_layer_indexes.end (); ++l) {
               ibox += inst_cell.bbox (*l);
@@ -774,47 +935,41 @@ InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const d
 
       }
 
-      ++inst;
-
     }
 
   } else {
 
-    if (--m_tries < 0) {
-      throw StopException ();
-    }
-
     //  look for instances to check here ..
-    db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
-    while (! inst.at_end ()) {
+    for (db::Cell::touching_iterator inst = cell.begin_touching (search_box); ! inst.at_end (); ++inst) {
 
-      if (--m_tries < 0) {
-        throw StopException ();
-      }
+      checkpoint ();
 
       const db::CellInstArray &cell_inst = inst->cell_inst ();
       const db::Cell &inst_cell = layout ().cell (cell_inst.object ().cell_index ());
 
+      if (! consider_cell (inst_cell)) {
+        continue;
+      }
+
       //  just consider the instances exactly at the last level of 
       //  hierarchy (this is where the boxes are drawn) or if of cells that
       //  are hidden.
-      if (level == max_level () - 1 || inst_cell.is_proxy () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
+      if (level == max_level () - 1 || inst_cell.is_proxy () || inst_cell.is_ghost_cell () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
 
         db::box_convert <db::CellInst, false> bc (layout ());
         for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
-        
-          if (--m_tries < 0) {
-            throw StopException ();
-          }
+
+          checkpoint ();
 
           bool match = false;
           double d = std::numeric_limits<double>::max ();
 
           db::Box ibox;
-          if (inst_cell.bbox ().empty ()) {
-            ibox = db::Box (db::Point (0, 0), db::Point (0, 0));
-          } else if (! m_visible_layers || level == mp_view->get_max_hier_levels () - 1 || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
-            ibox = inst_cell.bbox ();
+          if (! m_visible_layers || level == mp_view->get_max_hier_levels () - 1 || inst_cell.is_ghost_cell () || mp_view->is_cell_hidden (inst_cell.cell_index (), m_cv_index)) {
+            ibox = inst_cell.bbox_with_empty ();
+          } else if (inst_cell.bbox ().empty ()) {
+            //  empty cells cannot be found by visible layers, so we always select them here
+            ibox = inst_cell.bbox_with_empty ();
           } else {
             for (std::vector<int>::const_iterator l = m_visible_layer_indexes.begin (); l != m_visible_layer_indexes.end (); ++l) {
               ibox += inst_cell.bbox (*l);
@@ -831,11 +986,16 @@ InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const d
               //  convert to polygon and test those edges
               db::Polygon poly (cell_inst.complex_trans (*p) * db::Polygon (ibox));
 
+              bool any_valid_edge = false;
               for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
-                test_edge (t * *e, d, match);
+                //  only consider edges that cut through the viewport
+                if (viewport_box.empty () || (*e).clipped (viewport_box).first) {
+                  any_valid_edge = true;
+                  test_edge (t, *e, d, match);
+                }
               }
 
-              if (! match && db::inside_poly (poly.begin_edge (), search_box.center ())) {
+              if (! match && any_valid_edge && db::inside_poly (poly.begin_edge (), search_box.center ())) {
                 d = t.ctrans (poly_dist (poly.begin_edge (), search_box.center ()));
                 match = true;
               }
@@ -848,52 +1008,38 @@ InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const d
 
           if (match) {
 
+            lay::ObjectInstPath found;
+            found.set_cv_index (m_cv_index);
+            found.set_topcell (m_topcell);
+            found.assign_path (path ().begin (), path ().end ());
+
+            //  add the selected instance as the last element of the path
+            db::InstElement el;
+            el.inst_ptr = *inst;
+            if (! m_full_arrays) {
+              el.array_inst = p;
+            }
+            found.add_path (el);
+
             if (mp_excludes) {
 
               //  with an exclude list first create the selection item so we can check
               //  if it's part of the exclude set.
 
-              lay::ObjectInstPath found;
-              found.set_cv_index (m_cv_index);
-              found.set_topcell (m_topcell);
-              found.assign_path (path ().begin (), path ().end ());
-
-              //  add the selected instance as the last element of the path
-              db::InstElement el;
-              el.inst_ptr = *inst;
-              if (! m_full_arrays) {
-                el.array_inst = p;
-              }
-              found.add_path (el);
-
               //  in point mode just store the found object that has the least "distance" and is
               //  not in the exclude set
-              if (mp_excludes->find (found) == mp_excludes->end () && closer (d)) {
-                if (m_founds.empty ()) {
-                  m_founds.push_back (found);
-                } else {
-                  m_founds.front () = found;
-                }
-              }
+              match = (mp_excludes->find (found) == mp_excludes->end ());
 
-            } else if (closer (d)) {
+            }
+
+            if (match && (catch_all () || closer (d))) {
 
               //  in point mode just store that found that has the least "distance"
-              if (m_founds.empty ()) {
+              if (m_founds.empty () || catch_all ()) {
                 m_founds.push_back (lay::ObjectInstPath ());
               }
 
-              m_founds.back ().set_cv_index (m_cv_index);
-              m_founds.back ().set_topcell (m_topcell);
-              m_founds.back ().assign_path (path ().begin (), path ().end ());
-
-              //  add the selected instance as the last element of the path
-              db::InstElement el;
-              el.inst_ptr = *inst;
-              if (! m_full_arrays) {
-                el.array_inst = p;
-              }
-              m_founds.back ().add_path (el);
+              m_founds.back () = found;
 
             }
 
@@ -903,13 +1049,11 @@ InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const d
 
       }
 
-      ++inst;
-
     }
 
   }
 
 }
 
-} // namespace edt
+} // namespace lay
 
