@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -48,7 +48,7 @@ namespace tl {
 
 namespace lay {
 
-class LayoutView;
+class LayoutViewBase;
 class LayerPropertiesList;
 class LayerPropertiesNode;
 
@@ -101,6 +101,18 @@ public:
    */
   LayerProperties &operator= (const LayerProperties &d);
 
+  /**
+   *  @brief Gets the generation number
+   *
+   *  The generation number changes whenever something is changed
+   *  with this layer properties object. "0" is reserved for the
+   *  "uninitialized" state.
+   */
+  size_t gen_id () const
+  {
+    return m_gen_id;
+  }
+
   /**
    *  @brief Assignment alias for GSI binding
    */
@@ -125,7 +137,7 @@ public:
   /**
    *  @brief Utility: compute the effective color from a color with brightness correction
    */
-  static color_t brighter (color_t in, int b);
+  static tl::color_t brighter (tl::color_t in, int b);
 
   /**
    *  @brief render the effective frame color 
@@ -133,7 +145,7 @@ public:
    *  The effective frame color is computed from the frame color brightness and the
    *  frame color.
    */
-  color_t eff_frame_color (bool real) const;
+  tl::color_t eff_frame_color (bool real) const;
 
   /**
    *  @brief render the effective fill color
@@ -141,7 +153,7 @@ public:
    *  The effective fill color is computed from the frame color brightness and the
    *  frame color.
    */
-  color_t eff_fill_color (bool real) const;
+  tl::color_t eff_fill_color (bool real) const;
 
   /**
    *  @brief render the effective frame color plus an additional brightness adjustment
@@ -149,7 +161,7 @@ public:
    *  This method returns the effective frame color with an additional brightness adjustment 
    *  applied.
    */
-  color_t eff_frame_color_brighter (bool real, int plus_brightness) const;
+  tl::color_t eff_frame_color_brighter (bool real, int plus_brightness) const;
 
   /**
    *  @brief render the effective frame color plus an additional brightness adjustment
@@ -157,19 +169,20 @@ public:
    *  This method returns the effective fill color with an additional brightness adjustment 
    *  applied.
    */
-  color_t eff_fill_color_brighter (bool real, int plus_brightness) const;
+  tl::color_t eff_fill_color_brighter (bool real, int plus_brightness) const;
 
   /**
    *  @brief Get the frame color
    *
    *  This method may return an invalid color if the color is not set.
    */
-  color_t frame_color (bool real) const
+  tl::color_t frame_color (bool real) const
   {
     if (real) {
       ensure_visual_realized ();
       return m_frame_color_real;
     } else {
+      refresh ();
       return m_frame_color;
     }
   }
@@ -180,8 +193,9 @@ public:
    *  To clear the frame color, pass 0 to this method. Valid colors have the
    *  upper 8 bits set.
    */
-  void set_frame_color_code (color_t c)
+  void set_frame_color_code (tl::color_t c)
   {
+    refresh ();
     if (m_frame_color != c) {
       m_frame_color = c;
       need_realize (nr_visual);
@@ -191,7 +205,7 @@ public:
   /**
    *  @brief Set the frame color to the given value
    */
-  void set_frame_color (color_t c)
+  void set_frame_color (tl::color_t c)
   {
     set_frame_color_code (c | 0xff000000);
   }
@@ -217,12 +231,13 @@ public:
    *
    *  This method may return an invalid color if the color is not set.
    */
-  color_t fill_color (bool real) const
+  tl::color_t fill_color (bool real) const
   {
     if (real) {
       ensure_visual_realized ();
       return m_fill_color_real;
     } else {
+      refresh ();
       return m_fill_color;
     }
   }
@@ -233,8 +248,9 @@ public:
    *  To clear the fill color, pass 0 to this method. Valid colors have the
    *  upper 8 bits set.
    */
-  void set_fill_color_code (color_t c)
+  void set_fill_color_code (tl::color_t c)
   {
+    refresh ();
     if (m_fill_color != c) {
       m_fill_color = c;
       need_realize (nr_visual);
@@ -244,7 +260,7 @@ public:
   /**
    *  @brief Set the fill color to the given value
    */
-  void set_fill_color (color_t c)
+  void set_fill_color (tl::color_t c)
   {
     set_fill_color_code (c | 0xff000000);
   }
@@ -272,6 +288,7 @@ public:
    */
   void set_frame_brightness (int b)
   {
+    refresh ();
     if (m_frame_brightness != b) {
       m_frame_brightness = b;
       need_realize (nr_visual);
@@ -287,6 +304,7 @@ public:
       ensure_visual_realized ();
       return m_frame_brightness_real;
     } else {
+      refresh ();
       return m_frame_brightness;
     }
   }
@@ -298,6 +316,7 @@ public:
    */
   void set_fill_brightness (int b)
   {
+    refresh ();
     if (m_fill_brightness != b) {
       m_fill_brightness = b;
       need_realize (nr_visual);
@@ -315,6 +334,7 @@ public:
       ensure_visual_realized ();
       return m_fill_brightness_real;
     } else {
+      refresh ();
       return m_fill_brightness;
     }
   }
@@ -324,6 +344,7 @@ public:
    */
   void set_dither_pattern (int index)
   {
+    refresh ();
     if (m_dither_pattern != index) {
       m_dither_pattern = index;
       need_realize (nr_visual);
@@ -353,6 +374,7 @@ public:
       ensure_visual_realized ();
       return m_dither_pattern_real;
     } else {
+      refresh ();
       return m_dither_pattern;
     }
   }
@@ -378,6 +400,7 @@ public:
    */
   void set_line_style (int index)
   {
+    refresh ();
     if (m_line_style != index) {
       m_line_style = index;
       need_realize (nr_visual);
@@ -407,6 +430,7 @@ public:
       ensure_visual_realized ();
       return m_line_style_real;
     } else {
+      refresh ();
       return m_line_style;
     }
   }
@@ -432,6 +456,7 @@ public:
    */
   void set_valid (bool v)
   {
+    refresh ();
     if (m_valid != v) {
       m_valid = v;
       need_realize (nr_visual);
@@ -447,6 +472,7 @@ public:
       ensure_visual_realized ();
       return m_valid_real;
     } else {
+      refresh ();
       return m_valid;
     }
   }
@@ -456,6 +482,7 @@ public:
    */
   void set_visible (bool v)
   {
+    refresh ();
     if (m_visible != v) {
       m_visible = v;
       need_realize (nr_visual);
@@ -471,12 +498,13 @@ public:
       ensure_visual_realized ();
       return m_visible_real;
     } else {
+      refresh ();
       return m_visible;
     }
   }
 
   /**
-   *  @brief Return true, if the layer is showing "something"
+   *  @brief Returns true, if the layer is showing "something"
    *
    *  A layer "shows something" if it is visible and it displays some information,
    *  either shapes or cell boxes. Invalid layers, i.e. such that have a layer selection
@@ -523,6 +551,7 @@ public:
    */
   void set_transparent (bool t)
   {
+    refresh ();
     if (m_transparent != t) {
       m_transparent = t;
       need_realize (nr_visual);
@@ -538,6 +567,7 @@ public:
       ensure_visual_realized ();
       return m_transparent_real;
     } else {
+      refresh ();
       return m_transparent;
     }
   }
@@ -547,6 +577,7 @@ public:
    */
   void set_width (int w)
   {
+    refresh ();
     if (m_width != w) {
       m_width = w;
       need_realize (nr_visual);
@@ -562,6 +593,7 @@ public:
       ensure_visual_realized ();
       return m_width_real;
     } else {
+      refresh ();
       return m_width;
     }
   }
@@ -571,6 +603,7 @@ public:
    */
   void set_marked (bool t)
   {
+    refresh ();
     if (m_marked != t) {
       m_marked = t;
       need_realize (nr_visual);
@@ -586,6 +619,7 @@ public:
       ensure_visual_realized ();
       return m_marked_real;
     } else {
+      refresh ();
       return m_marked;
     }
   }
@@ -595,6 +629,7 @@ public:
    */
   void set_animation (int a)
   {
+    refresh ();
     if (m_animation != a) {
       m_animation = a;
       need_realize (nr_visual);
@@ -610,6 +645,7 @@ public:
       ensure_visual_realized ();
       return m_animation_real;
     } else {
+      refresh ();
       return m_animation;
     }
   }
@@ -623,6 +659,7 @@ public:
       ensure_visual_realized ();
       return m_xfill_real;
     } else {
+      refresh ();
       return m_xfill;
     }
   }
@@ -637,7 +674,11 @@ public:
    */
   void set_name (const std::string &n)
   {
-    m_name = n;
+    refresh ();
+    if (m_name != n) {
+      m_name = n;
+      need_realize (nr_meta);
+    }
   }
   
   /**
@@ -645,6 +686,7 @@ public:
    */
   const std::string &name () const
   {
+    refresh ();
     return m_name;
   }
   
@@ -658,7 +700,7 @@ public:
    *  If it is set to false, the view's always_show_source attribute with determine whether the source is 
    *  shown.
    */
-  std::string display_string (const lay::LayoutView *view, bool real, bool always_with_source = false) const;
+  std::string display_string (const lay::LayoutViewBase *view, bool real, bool always_with_source = false) const;
   
   /**
    *  @brief The source specification 
@@ -675,7 +717,7 @@ public:
    *
    *  This method may throw an exception if the specification
    *  is not valid. In order to make the layer usable, the properties
-   *  object must be "realized" with respect to a LayoutView object.
+   *  object must be "realized" with respect to a LayoutViewBase object.
    */
   void set_source (const std::string &s);
   
@@ -683,7 +725,7 @@ public:
    *  @brief Load the source specification 
    *
    *  In order to make the layer usable, the properties
-   *  object must be "realized" with respect to a LayoutView object.
+   *  object must be "realized" with respect to a LayoutViewBase object.
    */
   void set_source (const lay::ParsedLayerSource &s);
   
@@ -696,6 +738,7 @@ public:
       ensure_source_realized ();
       return m_source_real;
     } else {
+      refresh ();
       return m_source;
     }
   }
@@ -801,8 +844,8 @@ public:
   /**
    *  @brief Adaptors required for the XML reader
    */
-  color_t frame_color_loc () const            { return frame_color (false);      }
-  color_t fill_color_loc () const             { return fill_color (false);       }
+  tl::color_t frame_color_loc () const        { return frame_color (false);      }
+  tl::color_t fill_color_loc () const         { return fill_color (false);       }
   int frame_brightness_loc () const           { return frame_brightness (false); }
   int fill_brightness_loc () const            { return fill_brightness (false);  }
   int dither_pattern_loc () const             { return dither_pattern (false);   }
@@ -844,12 +887,7 @@ protected:
    *
    *  @param view The view the properties refer to or 0 if there is no view.
    */
-  void do_realize (const LayoutView *view) const;
-
-  /** 
-   *  @brief Tell the children that a realize of the visual properties is needed
-   */
-  virtual void need_realize (unsigned int flags, bool force = false);
+  void do_realize (const LayoutViewBase *view) const;
 
   /** 
    *  @brief Tell, if a realize of the visual properties is needed
@@ -867,12 +905,38 @@ protected:
     return m_realize_needed_source;
   }
 
+  /**
+   *  @brief Marks the properties object as modified
+   *
+   *  This will basically increment the generation count.
+   *  This method is implied when calling "needs_realize".
+   */
+  void touch ();
+
+  /**
+   *  @brief Tells the children that a realize of the visual properties is needed
+   *  This is also the "forward sync" for the LayerPropertiesNodeRef.
+   */
+  virtual void need_realize (unsigned int flags, bool force = false);
+
+  /**
+   *  @brief indicates a change of the collapsed/expanded state
+   */
+  virtual void expanded_state_changed ();
+
+  /**
+   *  @brief Fetches the current status from the original properties for the LayerPropertiesNodeRef implementation
+   */
+  virtual void refresh () const { }
+
 private:
+  //  the generation number
+  size_t m_gen_id;
   //  display styles
-  color_t m_frame_color;
-  mutable color_t m_frame_color_real;
-  color_t m_fill_color;
-  mutable color_t m_fill_color_real;
+  tl::color_t m_frame_color;
+  mutable tl::color_t m_frame_color_real;
+  tl::color_t m_fill_color;
+  mutable tl::color_t m_fill_color_real;
   int m_frame_brightness;
   mutable int m_frame_brightness_real;
   int m_fill_brightness;
@@ -916,7 +980,8 @@ protected:
   enum {
     nr_visual = 1,
     nr_source = 2,
-    nr_hierarchy = 4
+    nr_meta = 4,
+    nr_hierarchy = 8
   };
 
   mutable bool m_realize_needed_source : 1;
@@ -1009,11 +1074,26 @@ public:
     return r;
   }
 
+  /**
+   *  @brief Sets the expanded state of the layer properties tree node
+   */
+  void set_expanded (bool ex);
+
+  /**
+   *  @brief Gets the expanded state of the layer properties node
+   */
+  bool expanded () const
+  {
+    refresh ();
+    return m_expanded;
+  }
+
   /**
    *  @brief Child layers: begin iterator
    */
   const_iterator begin_children () const
   {
+    refresh ();
     return m_children.begin ();
   }
   
@@ -1022,6 +1102,7 @@ public:
    */
   const_iterator end_children () const
   {
+    refresh ();
     return m_children.end ();
   }
   
@@ -1030,6 +1111,7 @@ public:
    */
   iterator begin_children ()
   {
+    refresh ();
     return m_children.begin ();
   }
   
@@ -1038,6 +1120,7 @@ public:
    */
   iterator end_children () 
   {
+    refresh ();
     return m_children.end ();
   }
 
@@ -1099,21 +1182,30 @@ public:
    *  @return A bbox in micron units
    */
   db::DBox bbox () const;
-  
+
+  /**
+   *  @brief Computes the overall box of this layer
+   *
+   *  This is not a layer specific box, but an all-layer box,
+   *  including transformations (if specified).
+   *  This box is equivalent to the box delivered by
+   *  a cell frame layer.
+   */
+  db::DBox overall_bbox () const;
+
   /**
    *  @brief Attach to a view
    *
    *  This method attaches the properties node and it's children to a view.
    *  This enables the node to realize itself against the view, i.e.
    *  compute the actual property selection set.
-   *  This method is supposed to be 
    */
-  void attach_view (LayoutView *view, unsigned int list_index);
+  void attach_view (LayoutViewBase *view, unsigned int list_index);
 
   /**
    *  @brief Gets the layout view the node lives in
    */
-  LayoutView *view () const;
+  LayoutViewBase *view () const;
 
   /**
    *  @brief Gets the index of the layer properties list that node lives in
@@ -1144,14 +1236,21 @@ public:
   virtual void realize_source () const;
   virtual void realize_visual () const;
 
+  void set_expanded_silent (bool ex)
+  {
+    m_expanded = ex;
+  }
+
 protected: 
   virtual void need_realize (unsigned int flags, bool force);
+  virtual void expanded_state_changed ();
   void set_parent (const LayerPropertiesNode *);
 
 private:
   //  A reference to the view 
-  tl::weak_ptr<lay::LayoutView> mp_view;
+  tl::weak_ptr<lay::LayoutViewBase> mp_view;
   unsigned int m_list_index;
+  bool m_expanded;
   //  the parent node
   tl::weak_ptr<LayerPropertiesNode> mp_parent;
   //  the list of children
@@ -1407,7 +1506,7 @@ public:
     if (! mp_obj) {
       set_obj ();
     }
-    return mp_obj;
+    return mp_obj.get ();
   }
 
   /**
@@ -1423,7 +1522,7 @@ private:
 
   size_t m_uint;
   tl::weak_ptr<LayerPropertiesList> m_list;
-  mutable const LayerPropertiesNode *mp_obj;
+  mutable tl::weak_ptr<LayerPropertiesNode> mp_obj;
 
   void inc (unsigned int d);
   std::pair <size_t, size_t> factor () const;
@@ -1446,7 +1545,7 @@ struct LAYBASIC_PUBLIC CompareLayerIteratorBottomUp
  *  @brief Flat, non-const layer iterator
  *
  *  This iterator provides a flat view for the layers in the layer tree
- *  and allows to change the item pointed to.
+ *  and allows changing the item pointed to.
  */
 class LAYBASIC_PUBLIC LayerPropertiesIterator
   : public LayerPropertiesConstIterator
@@ -1848,12 +1947,12 @@ public:
    *  "load" automatically attaches the view.
    *  This method has the side effect of recomputing the layer source parameters.
    */
-  void attach_view (lay::LayoutView *view, unsigned int list_index);
+  void attach_view (LayoutViewBase *view, unsigned int list_index);
 
   /**
    *  @brief Gets the layout view this list is attached to
    */
-  lay::LayoutView *view () const;
+  lay::LayoutViewBase *view () const;
 
   /**
    *  @brief Gets the layout list
@@ -1861,7 +1960,7 @@ public:
   unsigned int list_index () const;
 
 private:
-  tl::weak_ptr<lay::LayoutView> mp_view;
+  tl::weak_ptr<lay::LayoutViewBase> mp_view;
   unsigned int m_list_index;
   layer_list m_layer_properties;
   lay::DitherPattern m_dither_pattern;
@@ -1878,6 +1977,10 @@ private:
  *  a node's property will update the properties in the view as well.
  *  Second, changes in the node's hierarchy will be reflected in the
  *  view's layer hierarchy too.
+ *
+ *  The implementation is based on a synchronized mirror copy of the
+ *  original node. This is not quite efficient and cumbersome. In the
+ *  future, this may be replaced by a simple reference.
  */
 class LAYBASIC_PUBLIC LayerPropertiesNodeRef
   : public LayerPropertiesNode
@@ -1888,14 +1991,6 @@ public:
    */
   LayerPropertiesNodeRef (LayerPropertiesNode *node);
 
-  /**
-   *  @brief Constructor from an iterator
-   *  The iterator is a pointer since the erase implementation requires us to
-   *  modify the iterator. Hence with this version, the original iterator will
-   *  follow up after the erase.
-   */
-  LayerPropertiesNodeRef (LayerPropertiesConstIterator *iter);
-
   /**
    *  @brief Constructor from an iterator with an iterator copy
    */
@@ -1919,7 +2014,7 @@ public:
 
   /**
    *  @brief Deletes the current node
-   *  After this operation, the reference will point to the
+   *  After this operation, the reference will point to the next element.
    */
   void erase ();
 
@@ -1947,10 +2042,12 @@ public:
 
 private:
   LayerPropertiesConstIterator m_iter;
-  tl::weak_ptr<LayerPropertiesConstIterator> mp_iter;
   tl::weak_ptr<LayerPropertiesNode> mp_node;
+  size_t m_synched_gen_id;
 
-  void need_realize (unsigned int flags, bool force);
+  virtual void need_realize (unsigned int flags, bool force);
+  virtual void expanded_state_changed ();
+  virtual void refresh () const;
 };
 
 }
