@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -20,32 +20,19 @@
 
 */
 
+#include "tlDefs.h"
 
 #if !defined(HDR_pyaCommon_h)
 # define HDR_pyaCommon_h
 
-# if defined _WIN32 || defined __CYGWIN__
-
-#   ifdef MAKE_PYA_LIBRARY
-#     define PYA_PUBLIC __declspec(dllexport)
-#   else
-#     define PYA_PUBLIC __declspec(dllimport)
-#   endif
-#   define PYA_LOCAL
-#   define PYA_PUBLIC_TEMPLATE
-
+# ifdef MAKE_PYA_LIBRARY
+#   define PYA_PUBLIC           DEF_INSIDE_PUBLIC
+#   define PYA_PUBLIC_TEMPLATE  DEF_INSIDE_PUBLIC_TEMPLATE
+#   define PYA_LOCAL            DEF_INSIDE_LOCAL
 # else
-
-#   if __GNUC__ >= 4 || defined(__clang__)
-#     define PYA_PUBLIC __attribute__ ((visibility ("default")))
-#     define PYA_PUBLIC_TEMPLATE __attribute__ ((visibility ("default")))
-#     define PYA_LOCAL  __attribute__ ((visibility ("hidden")))
-#   else
-#     define PYA_PUBLIC
-#     define PYA_PUBLIC_TEMPLATE
-#     define PYA_LOCAL
-#   endif
-
+#   define PYA_PUBLIC           DEF_OUTSIDE_PUBLIC
+#   define PYA_PUBLIC_TEMPLATE  DEF_OUTSIDE_PUBLIC_TEMPLATE
+#   define PYA_LOCAL            DEF_OUTSIDE_LOCAL
 # endif
 
 #endif
