@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -23,154 +23,161 @@
 
 #include "pyaConvert.h"
 #include "pyaObject.h"
-
-#include "pya.h"
+#include "pyaModule.h"
+#include "pyaStatusChangedListener.h"
 #include "pyaUtils.h"
 
+#include "gsiClassBase.h"
+
 #include <string>
 
 namespace pya
 {
 
+bool is_derived_from (const gsi::ClassBase *cls, const std::type_info &ti)
+{
+  return cls->is_derived_from (gsi::class_by_typeinfo_no_assert (ti));
+}
+
 template <>
-long python2c<long> (PyObject *rval, tl::Heap *)
+long python2c_func<long>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsLong (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsLong (rval);
   } else if (PyFloat_Check (rval)) {
     return (long) (PyFloat_AsDouble (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to an integer")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to an integer")));
   }
 }
 
 template <>
-bool python2c<bool> (PyObject *rval, tl::Heap *)
+bool python2c_func<bool>::operator() (PyObject *rval)
 {
   return PyObject_IsTrue (rval);
 }
 
 template <>
-char python2c<char> (PyObject *rval, tl::Heap *)
+char python2c_func<char>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return char (PyInt_AsLong (rval));
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return char (PyLong_AsLong (rval));
   } else if (PyFloat_Check (rval)) {
     return char (PyFloat_AsDouble (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to a character")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to a character")));
   }
 }
 
 template <>
-unsigned long python2c<unsigned long> (PyObject *rval, tl::Heap *)
+unsigned long python2c_func<unsigned long>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsUnsignedLongMask (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsUnsignedLongMask (rval);
   } else if (PyFloat_Check (rval)) {
     return (unsigned long) (PyFloat_AsDouble (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to an integer")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to an integer")));
   }
 }
 
 template <>
-long long python2c<long long> (PyObject *rval, tl::Heap *)
+long long python2c_func<long long>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsLong (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsLongLong (rval);
   } else if (PyFloat_Check (rval)) {
     return (long long) (PyFloat_AsDouble (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to an integer")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to an integer")));
   }
 }
 
 template <>
-unsigned long long python2c<unsigned long long> (PyObject *rval, tl::Heap *)
+unsigned long long python2c_func<unsigned long long>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsUnsignedLongMask (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsUnsignedLongLongMask (rval);
   } else if (PyFloat_Check (rval)) {
     return (unsigned long long) (PyFloat_AsDouble (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to an integer")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to an integer")));
   }
 }
 
 #if defined(HAVE_64BIT_COORD)
 template <>
-__int128 python2c<__int128> (PyObject *rval, tl::Heap *)
+__int128 python2c_func<__int128>::operator() (PyObject *rval)
 {
-  // TOOD: this is pretty simplistic
+  // TODO: this is pretty simplistic
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsLong (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsLongLong (rval);
   } else if (PyFloat_Check (rval)) {
     return PyFloat_AsDouble (rval);
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to an integer")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to an integer")));
   }
 }
 #endif
 
 template <>
-double python2c<double> (PyObject *rval, tl::Heap *)
+double python2c_func<double>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyInt_Check (rval)) {
     return PyInt_AsLong (rval);
-  } else 
+  } else
 #endif
   if (PyLong_Check (rval)) {
     return PyLong_AsLongLong (rval);
   } else if (PyFloat_Check (rval)) {
     return PyFloat_AsDouble (rval);
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to a floating-point value")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to a floating-point value")));
   }
 }
 
 template <>
-std::string python2c<std::string> (PyObject *rval, tl::Heap *)
+std::string python2c_func<std::string>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyString_Check (rval)) {
     return std::string (PyString_AsString (rval), PyString_Size (rval));
-  } else 
+  } else
 #else
   if (PyBytes_Check (rval)) {
     return std::string (PyBytes_AsString (rval), PyBytes_Size (rval));
-  } else 
+  } else
 #endif
   if (PyUnicode_Check (rval)) {
     PythonRef ba (PyUnicode_AsUTF8String (rval));
@@ -181,21 +188,58 @@ std::string python2c<std::string> (PyObject *rval, tl::Heap *)
   } else if (PyByteArray_Check (rval)) {
     return std::string (PyByteArray_AsString (rval), PyByteArray_Size (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to a string")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to a string")));
   }
 }
 
 template <>
-QByteArray python2c<QByteArray> (PyObject *rval, tl::Heap *)
+std::vector<char> python2c_func<std::vector<char> >::operator() (PyObject *rval)
+{
+#if PY_MAJOR_VERSION < 3
+  if (PyString_Check (rval)) {
+    const char *cp = PyString_AsString (rval);
+    return std::vector<char> (cp, cp + PyString_Size (rval));
+  } else
+#else
+  if (PyBytes_Check (rval)) {
+    char *cp = 0;
+    Py_ssize_t sz = 0;
+    PyBytes_AsStringAndSize (rval, &cp, &sz);
+    tl_assert (cp != 0);
+    return std::vector<char> (cp, cp + sz);
+  } else
+#endif
+  if (PyUnicode_Check (rval)) {
+    PythonRef ba (PyUnicode_AsUTF8String (rval));
+    if (! ba) {
+      check_error ();
+    }
+    char *cp = 0;
+    Py_ssize_t sz = 0;
+    PyBytes_AsStringAndSize (ba.get (), &cp, &sz);
+    tl_assert (cp != 0);
+    return std::vector<char> (cp, cp + sz);
+  } else if (PyByteArray_Check (rval)) {
+    char *cp = PyByteArray_AsString (rval);
+    Py_ssize_t sz = PyByteArray_Size (rval);
+    return std::vector<char> (cp, cp + sz);
+  } else {
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to a byte array")));
+  }
+}
+
+#if defined(HAVE_QT)
+template <>
+QByteArray python2c_func<QByteArray>::operator() (PyObject *rval)
 {
 #if PY_MAJOR_VERSION < 3
   if (PyString_Check (rval)) {
     return QByteArray (PyString_AsString (rval), PyString_Size (rval));
-  } else 
+  } else
 #else
   if (PyBytes_Check (rval)) {
     return QByteArray (PyBytes_AsString (rval), PyBytes_Size (rval));
-  } else 
+  } else
 #endif
   if (PyUnicode_Check (rval)) {
     PythonRef ba (PyUnicode_AsUTF8String (rval));
@@ -206,71 +250,44 @@ QByteArray python2c<QByteArray> (PyObject *rval, tl::Heap *)
   } else if (PyByteArray_Check (rval)) {
     return QByteArray (PyByteArray_AsString (rval), PyByteArray_Size (rval));
   } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to a byte array")));
+    throw tl::TypeError (tl::to_string (tr ("Value cannot be converted to a byte array")));
   }
 }
 
 template <>
-const char *python2c<const char *> (PyObject *rval, tl::Heap *heap)
-{
-#if PY_MAJOR_VERSION < 3
-  if (PyString_Check (rval)) {
-    return PyString_AsString (rval);
-  } else 
-#else
-  if (PyBytes_Check (rval)) {
-    return PyBytes_AsString (rval);
-  } else 
-#endif
-  if (PyUnicode_Check (rval)) {
-
-    tl_assert (heap);
-
-    //  for creating a const char * we need a temporary object. Otherwise we cannot keep that
-    //  value.
-    std::string *converted_value = new std::string (python2c<std::string> (rval, heap));
-    heap->push (converted_value);
-
-    return converted_value->c_str ();
-
-  } else if (PyByteArray_Check (rval)) {
-    return PyByteArray_AsString (rval);
-  } else {
-    throw tl::Exception (tl::to_string (QObject::tr ("Argument cannot be converted to a string")));
-  }
-}
-
-template <>
-QString python2c<QString> (PyObject *rval, tl::Heap *heap)
+QString python2c_func<QString>::operator() (PyObject *rval)
 {
   //  TODO: directly convert Unicode strings to QString if possible
-  return tl::to_qstring (python2c<std::string> (rval, heap));
+  return tl::to_qstring (python2c<std::string> (rval));
 }
+#endif
 
 template <>
-tl::Variant python2c<tl::Variant> (PyObject *rval, tl::Heap *heap)
+tl::Variant python2c_func<tl::Variant>::operator() (PyObject *rval)
 {
   if (rval == NULL || rval == Py_None) {
     return tl::Variant ();
   } else if (PyBool_Check (rval)) {
-    return tl::Variant (python2c<bool> (rval, heap));
+    return tl::Variant (python2c<bool> (rval));
   } else if (PyLong_Check (rval)) {
-    return tl::Variant (python2c<long long> (rval, heap));
+    return tl::Variant (python2c<long long> (rval));
 #if PY_MAJOR_VERSION < 3
   } else if (PyInt_Check (rval)) {
-    return tl::Variant (python2c<int> (rval, heap));
+    return tl::Variant (python2c<int> (rval));
 #endif
   } else if (PyFloat_Check (rval)) {
-    return tl::Variant (python2c<double> (rval, heap));
+    return tl::Variant (python2c<double> (rval));
 #if PY_MAJOR_VERSION < 3
   } else if (PyString_Check (rval)) {
-    return tl::Variant (python2c<std::string> (rval, heap));
+    return tl::Variant (python2c<std::string> (rval));
 #else
   } else if (PyBytes_Check (rval)) {
-    return tl::Variant (python2c<std::string> (rval, heap));
+    return tl::Variant (python2c<std::vector<char> > (rval));
 #endif
-  } else if (PyUnicode_Check (rval) || PyByteArray_Check (rval)) {
-    return tl::Variant (python2c<std::string> (rval, heap));
+  } else if (PyUnicode_Check (rval)) {
+    return tl::Variant (python2c<std::string> (rval));
+  } else if (PyByteArray_Check (rval)) {
+    return tl::Variant (python2c<std::vector<char> > (rval));
   } else if (PyList_Check (rval)) {
 
     size_t len = PyList_Size (rval);
@@ -279,7 +296,7 @@ tl::Variant python2c<tl::Variant> (PyObject *rval, tl::Heap *heap)
     tl::Variant r (empty.begin (), empty.end ());
     r.get_list ().reserve (len);
     for (size_t i = 0; i < len; ++i) {
-      r.get_list ().push_back (python2c<tl::Variant> (PyList_GetItem (rval, i), heap));
+      r.get_list ().push_back (python2c<tl::Variant> (PyList_GetItem (rval, i)));
     }
     return r;
 
@@ -291,7 +308,7 @@ tl::Variant python2c<tl::Variant> (PyObject *rval, tl::Heap *heap)
     tl::Variant r (empty.begin (), empty.end ());
     r.get_list ().reserve (len);
     for (size_t i = 0; i < len; ++i) {
-      r.get_list ().push_back (python2c<tl::Variant> (PyTuple_GetItem (rval, i), heap));
+      r.get_list ().push_back (python2c<tl::Variant> (PyTuple_GetItem (rval, i)));
     }
     return r;
 
@@ -308,13 +325,13 @@ tl::Variant python2c<tl::Variant> (PyObject *rval, tl::Heap *heap)
     return r;
 
   } else {
-    
-    const gsi::ClassBase *cls = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (rval));
+
+    const gsi::ClassBase *cls = PythonModule::cls_for_type (Py_TYPE (rval));
     if (cls) {
 
-      PYAObjectBase *p = (PYAObjectBase *) rval;
+      PYAObjectBase *p = PYAObjectBase::from_pyobject (rval);
 
-      //  employ the tl::Variant binding capabilities of the Expression binding to derive the 
+      //  employ the tl::Variant binding capabilities of the Expression binding to derive the
       //  variant value.
 
       void *obj = p->obj ();
@@ -349,7 +366,7 @@ tl::Variant python2c<tl::Variant> (PyObject *rval, tl::Heap *heap)
       std::string m ("<unknown type>");
       PythonRef msg_str (PyObject_Str (rval));
       if (msg_str) {
-        m = python2c<std::string> (msg_str.get (), heap);
+        m = python2c<std::string> (msg_str.get ());
       }
       return tl::Variant (m);
 
@@ -366,7 +383,7 @@ object_to_python (void *obj, PYAObjectBase *self, const gsi::ArgType &atype)
   bool is_direct   = !(atype.is_ptr () || atype.is_ref () || atype.is_cptr () || atype.is_cref ());
   bool pass_obj    = atype.pass_obj () || is_direct;
   bool is_const    = atype.is_cptr () || atype.is_cref ();
-  bool prefer_copy = atype.is_cref ();
+  bool prefer_copy = atype.prefer_copy ();
   bool can_destroy = prefer_copy || atype.is_ptr ();
 
   return object_to_python (obj, self, cls, pass_obj, is_const, prefer_copy, can_destroy);
@@ -376,13 +393,12 @@ object_to_python (void *obj, PYAObjectBase *self, const gsi::ArgType &atype)
  *  @brief Correct constness if a reference is const and a non-const reference is required
  *  HINT: this is a workaround for the fact that unlike C++, Python does not have const or non-const
  *  references. Since a reference is identical with the object it points to, there are only const or non-const
- *  objects. We deliver const objects first, but if a non-const version is requestes, the
+ *  objects. We deliver const objects first, but if a non-const version is requested, the
  *  object turns into a non-const one. This may be confusing but provides a certain level
  *  of "constness", at least until there is another non-const reference for that object.
  */
-void correct_constness (PyObject *obj, bool const_required)
+void correct_constness (PYAObjectBase *p, bool const_required)
 {
-  PYAObjectBase *p = (PYAObjectBase *) obj;
   if (p->const_ref () && ! const_required) {
     //  promote to non-const object
     p->set_const_ref (false);
@@ -424,12 +440,12 @@ object_to_python (void *obj, PYAObjectBase *self, const gsi::ClassBase *cls, boo
       obj = clsact->create_from_adapted (obj);
     }
 
-    //  we wil own the new object
+    //  we will own the new object
     pass_obj = true;
 
   }
 
-  if (! pass_obj && prefer_copy && ! clsact->adapted_type_info () && ! clsact->is_managed () && clsact->can_copy ()) {
+  if (! pass_obj && prefer_copy && ! clsact->adapted_type_info () && ! clsact->is_managed () && clsact->can_copy () && clsact->can_default_create ()) {
 
     //  We copy objects passed by const reference if they are not managed
     //  (derived from gsi::ObjectBase, in that case, client_data is not 0).
@@ -438,70 +454,79 @@ object_to_python (void *obj, PYAObjectBase *self, const gsi::ClassBase *cls, boo
     //  can't guarantee the lifetime of the container will exceed that
     //  of the exposed property. Hence copying is safer.
 
-    PyTypeObject *type = PythonInterpreter::instance ()->type_for_cls (clsact);
-    tl_assert (type != NULL);
+    PyTypeObject *type = PythonModule::type_for_cls (clsact);
+    if (!type) {
+      throw tl::Exception (tl::sprintf (tl::to_string (tr ("Requested type %s.%s is not bound to a Python class (did you load the '%s' module?)")), clsact->module (), clsact->name (), clsact->module ()));
+    }
 
     //  create a instance and copy the value
-    PYAObjectBase *new_object = (PYAObjectBase *) type->tp_alloc (type, 0);
-    new (new_object) PYAObjectBase (clsact);
+    PyObject *new_pyobject = type->tp_alloc (type, 0);
+    PYAObjectBase *new_object = PYAObjectBase::from_pyobject_unsafe (new_pyobject);
+    new (new_object) PYAObjectBase (clsact, new_pyobject);
     clsact->assign (new_object->obj (), obj);
-    return new_object;
+    return new_pyobject;
 
   } else if (pya_object) {
 
     //  we have a that is located in C++ space but is supposed to get attached
     //  a Python object. If it already has, we simply return a reference to this
-    Py_INCREF (pya_object);
+    Py_INCREF (pya_object->py_object ());
 
     correct_constness (pya_object, is_const);
 
-    return pya_object;
+    return pya_object->py_object ();
 
   } else {
 
-    PyTypeObject *type = PythonInterpreter::instance ()->type_for_cls (clsact);
-    tl_assert (type != NULL);
+    PyTypeObject *type = PythonModule::type_for_cls (clsact);
+    if (!type) {
+      throw tl::Exception (tl::sprintf (tl::to_string (tr ("Requested type %s.%s is not bound to a Python class (did you load the '%s' module?)")), clsact->module (), clsact->name (), clsact->module ()));
+    }
 
     //  create a instance and copy the value
-    PYAObjectBase *new_object = (PYAObjectBase *) type->tp_alloc (type, 0);
-    new (new_object) PYAObjectBase (clsact);
+    PyObject *new_pyobject = type->tp_alloc (type, 0);
+    PYAObjectBase *new_object = PYAObjectBase::from_pyobject_unsafe (new_pyobject);
+    new (new_object) PYAObjectBase (clsact, new_pyobject);
     new_object->set (obj, pass_obj, is_const, can_destroy);
-    return new_object;
+
+    return new_pyobject;
 
   }
 }
 
 template <>
-PyObject *c2python<tl::Variant> (const tl::Variant &c)
+PyObject *c2python_func<const tl::Variant &>::operator() (const tl::Variant &c)
 {
   if (c.is_double ()) {
-    return c2python<double> (c.to_double ());
+    return c2python (c.to_double ());
   } else if (c.is_bool ()) {
-    return c2python<bool> (c.to_bool ());
+    return c2python (c.to_bool ());
   } else if (c.is_a_string ()) {
-    return c2python<std::string> (c.to_string ());
+    return c2python (c.to_stdstring ());
+  } else if (c.is_a_bytearray ()) {
+    return c2python (c.to_bytearray ());
   } else if (c.is_long ()) {
-    return c2python<long> (c.to_long ());
+    return c2python (c.to_long ());
   } else if (c.is_ulong ()) {
-    return c2python<unsigned long> (c.to_ulong ());
+    return c2python (c.to_ulong ());
   } else if (c.is_longlong ()) {
-    return c2python<long long> (c.to_longlong ());
+    return c2python (c.to_longlong ());
   } else if (c.is_ulonglong ()) {
-    return c2python<unsigned long long> (c.to_ulonglong ());
+    return c2python (c.to_ulonglong ());
   } else if (c.is_array ()) {
 
     PyObject *ret = PyDict_New ();
     for (tl::Variant::const_array_iterator i = c.begin_array (); i != c.end_array (); ++i) {
-      PyDict_SetItem (ret, c2python<tl::Variant> (i->first), c2python<tl::Variant> (i->second));
+      PyDict_SetItem (ret, c2python (i->first), c2python (i->second));
     }
     return ret;
-    
+
   } else if (c.is_list ()) {
 
     PyObject *ret = PyList_New (c.get_list ().size ());
     size_t index = 0;
     for (tl::Variant::const_iterator i = c.begin (); i != c.end (); ++i, ++index) {
-      PyList_SetItem (ret, index, c2python<tl::Variant> (*i));
+      PyList_SetItem (ret, index, c2python (*i));
     }
     return ret;
 
@@ -509,8 +534,13 @@ PyObject *c2python<tl::Variant> (const tl::Variant &c)
 
     const gsi::ClassBase *cls = c.gsi_cls ();
     if (cls) {
-      void *obj = const_cast<void *> (c.to_user ());
-      return object_to_python (obj, 0, c.user_cls ()->gsi_cls (), false, false, true, false);
+      if (! c.user_is_ref () && cls->is_managed ()) {
+        void *obj = c.user_unshare ();
+        return object_to_python (obj, 0, c.user_cls ()->gsi_cls (), true, c.user_is_const (), false, false);
+      } else {
+        void *obj = const_cast<void *> (c.to_user ());
+        return object_to_python (obj, 0, c.user_cls ()->gsi_cls (), false, false, true, false);
+      }
     } else {
       //  not a known type -> return nil
       Py_RETURN_NONE;
@@ -522,31 +552,7 @@ PyObject *c2python<tl::Variant> (const tl::Variant &c)
 }
 
 template <>
-PyObject *c2python<gsi::ObjectType> (const gsi::ObjectType &)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Internal error: abstract object type read")));
-}
-
-template <>
-PyObject *c2python<gsi::VectorType> (const gsi::VectorType &)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Internal error: abstract vector type read")));
-}
-
-template <>
-PyObject *c2python<gsi::MapType> (const gsi::MapType &)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Internal error: abstract map type read")));
-}
-
-template <>
-PyObject *c2python<gsi::VoidType> (const gsi::VoidType &)
-{
-  throw tl::Exception (tl::to_string (QObject::tr ("Internal error: abstract void type read")));
-}
-
-template <>
-PyObject *c2python<std::string> (const std::string &c)
+PyObject *c2python_func<const std::string &>::operator() (const std::string &c)
 {
 #if PY_MAJOR_VERSION < 3
   return PyString_FromStringAndSize (c.c_str (), Py_ssize_t (c.size ()));
@@ -560,7 +566,17 @@ PyObject *c2python<std::string> (const std::string &c)
 }
 
 template <>
-PyObject *c2python<const char *> (const char * const &p)
+PyObject *c2python_func<const std::vector<char> &>::operator() (const std::vector<char> &c)
+{
+#if PY_MAJOR_VERSION < 3
+  return PyByteArray_FromStringAndSize (&c.front (), Py_ssize_t (c.size ()));
+#else
+  return PyBytes_FromStringAndSize (&c.front (), Py_ssize_t (c.size ()));
+#endif
+}
+
+template <>
+PyObject *c2python_func<const char *>::operator() (const char *p)
 {
   const char *s = p;
   if (! s) {
@@ -570,7 +586,7 @@ PyObject *c2python<const char *> (const char * const &p)
 #if PY_MAJOR_VERSION < 3
   return PyString_FromString (s);
 #else
-  PyObject *ret = PyUnicode_DecodeUTF8 (p, strlen (p), NULL);
+  PyObject *ret = PyUnicode_DecodeUTF8 (s, strlen (s), NULL);
   if (ret == NULL) {
     check_error ();
   }
@@ -578,31 +594,34 @@ PyObject *c2python<const char *> (const char * const &p)
 #endif
 }
 
+#if defined(HAVE_QT)
 template <>
-PyObject *c2python<QByteArray> (const QByteArray &qba)
+PyObject *c2python_func<const QByteArray &>::operator() (const QByteArray &qba)
 {
   if (qba.isNull ()) {
     Py_RETURN_NONE;
   } else {
 #if PY_MAJOR_VERSION < 3
-    return PyString_FromStringAndSize (qba.constData (), Py_ssize_t (qba.size ()));
-#else
     return PyByteArray_FromStringAndSize (qba.constData (), Py_ssize_t (qba.size ()));
+#else
+    return PyBytes_FromStringAndSize (qba.constData (), Py_ssize_t (qba.size ()));
 #endif
   }
 }
+#endif
 
+#if defined(HAVE_QT)
 template <>
-PyObject *c2python<QString> (const QString &qs)
+PyObject *c2python_func<const QString &>::operator() (const QString &qs)
 {
   if (qs.isNull ()) {
     Py_RETURN_NONE;
   } else {
-    //  TODO: can be done more efficently
+    //  TODO: can be done more efficiently
     std::string c (tl::to_string (qs));
-    return c2python<std::string> (c);
+    return c2python (c);
   }
 }
+#endif
 
 }
-
