@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -26,6 +26,7 @@
 #include "pyaMarshal.h"
 #include "pyaObject.h"
 #include "pyaConvert.h"
+#include "pyaSignalHandler.h"
 #include "pya.h"
 
 namespace pya
@@ -48,7 +49,7 @@ pya_channel_write (PyObject *self, PyObject *args)
   }
 
   PYAChannelObject *channel = (PYAChannelObject *) self;
-  if (PythonInterpreter::instance ()->current_console ()) {
+  if (PythonInterpreter::instance () && PythonInterpreter::instance ()->current_console ()) {
     PythonInterpreter::instance ()->current_console ()->write_str (msg, channel->channel);
   }
 
@@ -65,7 +66,7 @@ pya_channel_flush (PyObject * /*self*/, PyObject *args)
     return NULL;
   }
 
-  if (PythonInterpreter::instance ()->current_console ()) {
+  if (PythonInterpreter::instance () && PythonInterpreter::instance ()->current_console ()) {
     PythonInterpreter::instance ()->current_console ()->flush ();
   }
 
@@ -82,7 +83,7 @@ pya_channel_isatty (PyObject * /*self*/, PyObject *args)
     return NULL;
   }
 
-  if (PythonInterpreter::instance ()->current_console () && PythonInterpreter::instance ()->current_console ()->is_tty ()) {
+  if (PythonInterpreter::instance () && PythonInterpreter::instance ()->current_console () && PythonInterpreter::instance ()->current_console ()->is_tty ()) {
     Py_RETURN_TRUE;
   } else {
     Py_RETURN_FALSE;
@@ -101,11 +102,11 @@ pya_channel_init (PyObject *self, PyObject *, PyObject *)
 }
 
 void
-PYAChannelObject::make_class (PyObject *module)
+PYAChannelObject::make_class ()
 {
   static PyTypeObject channel_type = {
     PyVarObject_HEAD_INIT (&PyType_Type, 0)
-    "pya._Channel",             // tp_name
+    "__PYA_Channel",            // tp_name
     sizeof (PYAChannelObject)   // tp_size
   };
 
@@ -113,7 +114,7 @@ PYAChannelObject::make_class (PyObject *module)
       {"write", (PyCFunction) &pya_channel_write, METH_VARARGS, "internal stdout/stderr redirection object: write method" },
       {"flush", (PyCFunction) &pya_channel_flush, METH_VARARGS, "internal stdout/stderr redirection object: flush method" },
       {"isatty", (PyCFunction) &pya_channel_isatty, METH_VARARGS, "internal stdout/stderr redirection object: isatty method" },
-      {NULL,	NULL},
+      {NULL,  NULL},
   };
 
   channel_type.tp_flags = Py_TPFLAGS_DEFAULT;
@@ -123,7 +124,8 @@ PYAChannelObject::make_class (PyObject *module)
   PyType_Ready (&channel_type);
   Py_INCREF (&channel_type);
 
-  PyModule_AddObject (module, "_Channel", (PyObject *) &channel_type);
+  PyObject *module = PyImport_AddModule("__main__");
+  PyModule_AddObject (module, "__PYA_Channel", (PyObject *) &channel_type);
 
   cls = &channel_type;
 }
@@ -157,7 +159,7 @@ pya_static_attribute_descriptor_get (PyObject *self, PyObject * /*obj*/, PyObjec
     return (*(attr->getter)) ((PyObject *) attr->type, NULL);
   } else {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Attribute not readable"));
+    msg += tl::to_string (tr ("Attribute not readable"));
     msg += ": ";
     msg += attr->type->tp_name;
     msg += ".";
@@ -185,7 +187,7 @@ pya_static_attribute_descriptor_set (PyObject *self, PyObject * /*obj*/, PyObjec
     }
   } else {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Attribute cannot be changed"));
+    msg += tl::to_string (tr ("Attribute cannot be changed"));
     msg += ": ";
     msg += attr->type->tp_name;
     msg += ".";
@@ -356,7 +358,18 @@ pya_plain_iterator_next (PyObject *self)
 {
   PYAIteratorObject *iter = (PYAIteratorObject *) self;
 
-  if (! iter->iter || iter->iter->at_end ()) {
+  if (! iter->iter) {
+    PyErr_SetNone (PyExc_StopIteration);
+    return NULL;
+  }
+
+  //  increment except on first visit
+  if (! iter->first) {
+    iter->iter->inc ();
+  }
+  iter->first = false;
+
+  if (iter->iter->at_end ()) {
     PyErr_SetNone (PyExc_StopIteration);
     return NULL;
   }
@@ -366,9 +379,7 @@ pya_plain_iterator_next (PyObject *self)
 
   gsi::SerialArgs args (iter->iter->serial_size ());
   iter->iter->get (args);
-  PythonRef obj = pop_arg (*iter->value_type, args, 0, heap);
-
-  iter->iter->inc ();
+  PythonRef obj = pull_arg (*iter->value_type, args, 0, heap);
 
   return obj.release ();
 }
@@ -425,6 +436,7 @@ PYAIteratorObject::create (PyObject *origin, gsi::IterAdaptorAbstractBase *iter,
     iter_obj->origin = origin;
     iter_obj->iter = iter;
     iter_obj->value_type = value_type;
+    iter_obj->first = true;
   }
   return iter_obj;
 }
@@ -447,7 +459,7 @@ pya_signal_add (PyObject *self, PyObject *args)
 
   if (! PyCallable_Check (callable)) {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Signal's += operator needs a callable object"));
+    msg += tl::to_string (tr ("Signal's += operator needs a callable object"));
     PyErr_SetString (PyExc_AttributeError, msg.c_str ());
     return NULL;
   }
@@ -465,7 +477,7 @@ pya_signal_inplace_add (PyObject *self, PyObject *callable)
 {
   if (! PyCallable_Check (callable)) {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Signal's += operator needs a callable object"));
+    msg += tl::to_string (tr ("Signal's += operator needs a callable object"));
     PyErr_SetString (PyExc_AttributeError, msg.c_str ());
     return NULL;
   }
@@ -492,7 +504,7 @@ pya_signal_remove (PyObject *self, PyObject *args)
 
   if (! PyCallable_Check (callable)) {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Signal's -= operator needs a callable object"));
+    msg += tl::to_string (tr ("Signal's -= operator needs a callable object"));
     PyErr_SetString (PyExc_AttributeError, msg.c_str ());
     return NULL;
   }
@@ -513,7 +525,7 @@ pya_signal_inplace_remove (PyObject *self, PyObject *callable)
 {
   if (! PyCallable_Check (callable)) {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Signal's -= operator needs a callable object"));
+    msg += tl::to_string (tr ("Signal's -= operator needs a callable object"));
     PyErr_SetString (PyExc_AttributeError, msg.c_str ());
     return NULL;
   }
@@ -540,7 +552,7 @@ pya_signal_set (PyObject *self, PyObject *args)
 
   if (! PyCallable_Check (callable)) {
     std::string msg;
-    msg += tl::to_string (QObject::tr ("Signal's 'set' method needs a callable object"));
+    msg += tl::to_string (tr ("Signal's 'set' method needs a callable object"));
     PyErr_SetString (PyExc_AttributeError, msg.c_str ());
     return NULL;
   }
@@ -616,10 +628,12 @@ PYASignal::make_class (PyObject *module)
 
   static PyMethodDef signal_methods[] = {
       {"add", (PyCFunction) &pya_signal_add, METH_VARARGS, "internal signal proxy object: += operator" },
+      {"connect", (PyCFunction) &pya_signal_add, METH_VARARGS, "synonym to 'add' or '+='" },
       {"remove", (PyCFunction) &pya_signal_remove, METH_VARARGS, "internal signal proxy object: -= operator" },
+      {"disconnect", (PyCFunction) &pya_signal_remove, METH_VARARGS, "synonym to 'remove' or '-='" },
       {"set", (PyCFunction) &pya_signal_set, METH_VARARGS, "internal signal proxy object: assignment" },
       {"clear", (PyCFunction) &pya_signal_clear, METH_NOARGS, "internal signal proxy object: clears all receivers" },
-      {NULL,	NULL},
+      {NULL,  NULL},
   };
 
   static PyNumberMethods nm = { };
