@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
 #include "pyaMarshal.h"
 #include "pyaObject.h"
 #include "pyaConvert.h"
-#include "pya.h"
+#include "pyaModule.h"
 
 #include "gsiTypes.h"
 #include "gsiObjectHolder.h"
@@ -32,20 +32,39 @@
 namespace pya
 {
 
+void push_args (gsi::SerialArgs &arglist, const gsi::MethodBase *meth, PyObject *args, PyObject *kwargs, tl::Heap &heap);
+
 // -------------------------------------------------------------------
 //  Serialization adaptors for strings, variants, vectors and maps
 
 /**
- *  @brief An adaptor for a string from ruby objects 
+ *  @brief An adaptor for a string from ruby objects
  */
-class PythonBasedStringAdaptor   
+class PythonBasedStringAdaptor
   : public gsi::StringAdaptor
 {
 public:
   PythonBasedStringAdaptor (const PythonPtr &string)
-    : m_stdstr (python2c<std::string> (string.get ())), m_string (string)
+    : m_string (string)
   {
-    //  .. nothing yet ..
+#if PY_MAJOR_VERSION < 3
+    if (PyString_Check (string.get ())) {
+      m_stdstr = python2c<std::string> (string.get ());
+    } else
+#else
+    if (PyBytes_Check (string.get ())) {
+      m_stdstr = python2c<std::string> (string.get ());
+    } else
+#endif
+    if (PyUnicode_Check (string.get ()) || PyByteArray_Check (string.get ())) {
+      m_stdstr = python2c<std::string> (string.get ());
+    } else {
+      //  use object protocol to get the string through str(...)
+      PythonRef as_str (PyObject_Str (string.get ()));
+      if (as_str) {
+        m_stdstr = python2c<std::string> (as_str.get ());
+      }
+    }
   }
 
   virtual const char *c_str () const
@@ -61,6 +80,7 @@ public:
   virtual void set (const char * /*c_str*/, size_t /*s*/, tl::Heap & /*heap*/)
   {
     //  TODO: is there a setter for a string?
+    //  So far it's not possible to have string OUT parameter
   }
 
 private:
@@ -69,16 +89,51 @@ private:
 };
 
 /**
- *  @brief An adaptor for a variant from ruby objects 
+ *  @brief An adaptor for a byte array from ruby objects
  */
-class PythonBasedVariantAdaptor   
+class PythonBasedByteArrayAdaptor
+  : public gsi::ByteArrayAdaptor
+{
+public:
+  PythonBasedByteArrayAdaptor (const PythonPtr &ba)
+    : m_bytearray (python2c<std::vector<char> > (ba.get ())), m_bytes (ba)
+  {
+    //  .. nothing yet ..
+  }
+
+  virtual const char *c_str () const
+  {
+    return &m_bytearray.front ();
+  }
+
+  virtual size_t size () const
+  {
+    return m_bytearray.size ();
+  }
+
+  virtual void set (const char * /*c_str*/, size_t /*s*/, tl::Heap & /*heap*/)
+  {
+    //  TODO: is there a setter for a byte array?
+    //  So far it's not possible to have bytes OUT parameter
+  }
+
+private:
+  std::vector<char> m_bytearray;
+  PythonPtr m_bytes;
+};
+
+/**
+ *  @brief An adaptor for a variant from ruby objects
+ */
+class PythonBasedVariantAdaptor
   : public gsi::VariantAdaptor
 {
 public:
   PythonBasedVariantAdaptor (const PythonPtr &var);
 
   virtual tl::Variant var () const;
-  virtual void set (const tl::Variant &v);
+  virtual void set (const tl::Variant &v, tl::Heap & /*heap*/);
+  const PythonPtr &ptr () const { return m_var; }
 
 private:
   PythonPtr m_var;
@@ -106,7 +161,7 @@ private:
 /**
  *  @brief An adaptor for a vector from Python objects
  */
-class PythonBasedVectorAdaptor   
+class PythonBasedVectorAdaptor
   : public gsi::VectorAdaptor
 {
 public:
@@ -132,7 +187,7 @@ class PythonBasedMapAdaptorIterator
 public:
   PythonBasedMapAdaptorIterator (const PythonPtr &hash, const gsi::ArgType *ainner, const gsi::ArgType *ainner_k);
 
-  virtual void get (gsi::SerialArgs &w, tl::Heap &heap) const; 
+  virtual void get (gsi::SerialArgs &w, tl::Heap &heap) const;
   virtual bool at_end () const;
   virtual void inc ();
 
@@ -147,7 +202,7 @@ private:
 /**
  *  @brief An adaptor for a map from Python objects
  */
-class PythonBasedMapAdaptor   
+class PythonBasedMapAdaptor
   : public gsi::MapAdaptor
 {
 public:
@@ -171,11 +226,11 @@ template <class R>
 struct get_boxed_value_func
 {
   void operator() (void **ret, PyObject *arg, tl::Heap *heap)
-  { 
-    const gsi::ClassBase *cls_decl = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (arg));
+  {
+    const gsi::ClassBase *cls_decl = PythonModule::cls_for_type (Py_TYPE (arg));
     if (! cls_decl) {
 
-      R *v = new R (python2c<R> (arg, heap));
+      R *v = new R (python2c<R> (arg));
       heap->push (v);
       *ret = v;
 
@@ -184,10 +239,10 @@ struct get_boxed_value_func
       const gsi::ClassBase *bt = gsi::cls_decl <gsi::Value> ();
 
       if (!cls_decl->is_derived_from (bt)) {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Passing an object to pointer or reference requires a boxed type (pya.%s)")), bt->name ()));
+        throw tl::Exception (tl::sprintf (tl::to_string (tr ("Passing an object to pointer or reference requires a boxed type (pya.%s)")), bt->name ()));
       }
 
-      PYAObjectBase *p = (PYAObjectBase *) arg; 
+      PYAObjectBase *p = PYAObjectBase::from_pyobject (arg);
       gsi::Value *bo = reinterpret_cast<gsi::Value *> (p->obj ());
       if (bo) {
         *ret = bo->value ().template morph<R> ().native_ptr ();
@@ -219,7 +274,7 @@ struct writer
     if (arg == Py_None || arg == NULL) {
 
       if (atype.is_ref () || atype.is_cref ()) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference type cannot be passed None")));
+        throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference type cannot be passed None")));
       } else if (atype.is_ptr ()) {
         aa->write<R *> ((R *)0);
       } else if (atype.is_cptr ()) {
@@ -234,28 +289,28 @@ struct writer
         // references or pointers require a boxed object. Pointers also allow None.
         void *vc = boxed_value_ptr (atype.type (), arg, *heap);
         if (! vc && atype.is_ref ()) {
-          throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference or direct type cannot be passed None or an empty boxed value object")));
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed None or an empty boxed value object")));
         }
         aa->write<void *> (vc);
       } else if (atype.is_cref ()) {
         //  Note: POD's are written as copies for const refs, so we can pass a temporary here:
         //  (avoids having to create a temp object)
-        aa->write<const R &> (python2c<R> (arg, heap));
+        aa->write<const R &> (python2c<R> (arg));
       } else if (atype.is_cptr ()) {
         //  Note: POD's are written as copies for const ptrs, so we can pass a temporary here:
         //  (avoids having to create a temp object)
-        R r = python2c<R> (arg, heap);
+        R r = python2c<R> (arg);
         aa->write<const R *> (&r);
       } else {
-        aa->write<R> (python2c<R> (arg, heap));
+        aa->write<R> (python2c<R> (arg));
       }
 
     }
   }
 };
 
-/** 
- *  @brief Serialization for strings 
+/**
+ *  @brief Serialization for strings
  */
 template <>
 struct writer<gsi::StringType>
@@ -282,7 +337,7 @@ struct writer<gsi::StringType>
         void *vc = 0;
         get_boxed_value_func<std::string> () (&vc, arg, heap);
         if (! vc && atype.is_ref ()) {
-          throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
         }
 
         //  NOTE: by convention we pass the ownership to the receiver for adaptors.
@@ -303,6 +358,55 @@ struct writer<gsi::StringType>
   }
 };
 
+/**
+ *  @brief Serialization for strings
+ */
+template <>
+struct writer<gsi::ByteArrayType>
+{
+  void operator() (gsi::SerialArgs *aa, PyObject *arg, const gsi::ArgType &atype, tl::Heap *heap)
+  {
+    //  Cannot pass ownership currently
+    tl_assert (!atype.pass_obj ());
+
+    if (arg == Py_None || arg == NULL) {
+
+      if (! (atype.is_ptr () || atype.is_cptr ())) {
+        //  nil is treated as an empty string for references
+        aa->write<void *> ((void *)new gsi::ByteArrayAdaptorImpl<std::vector<char> > (std::vector<char> ()));
+      } else {
+        aa->write<void *> ((void *)0);
+      }
+
+    } else {
+
+      if (atype.is_ref () || atype.is_ptr ()) {
+
+        // references or pointers require a boxed object. Pointers also allow nil.
+        void *vc = 0;
+        get_boxed_value_func<std::vector<char> > () (&vc, arg, heap);
+        if (! vc && atype.is_ref ()) {
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
+        }
+
+        //  NOTE: by convention we pass the ownership to the receiver for adaptors.
+        if (! vc) {
+          aa->write<void *> (0);
+        } else {
+          aa->write<void *> ((void *)new gsi::ByteArrayAdaptorImpl<std::vector<char> > ((std::vector<char> *) vc));
+        }
+
+      } else {
+
+        //  NOTE: by convention we pass the ownership to the receiver for adaptors.
+        aa->write<void *> ((void *)new PythonBasedByteArrayAdaptor (arg));
+
+      }
+
+    }
+  }
+};
+
 /**
  *  @brief Specialization for Variant
  */
@@ -327,7 +431,7 @@ struct writer<gsi::VectorType>
   {
     if (arg == Py_None || arg == NULL) {
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed nil")));
       } else {
         aa->write<void *> ((void *)0);
       }
@@ -348,7 +452,7 @@ struct writer<gsi::MapType>
   {
     if (arg == Py_None || arg == NULL) {
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed nil")));
       } else {
         aa->write<void *> ((void *)0);
       }
@@ -362,17 +466,19 @@ struct writer<gsi::MapType>
 
 /**
  *  @brief A serialization wrapper (write mode)
- *  Specialisation for objects 
+ *  Specialization for objects
  */
 template <>
 struct writer<gsi::ObjectType>
 {
   void operator() (gsi::SerialArgs *aa, PyObject *arg, const gsi::ArgType &atype, tl::Heap *heap)
   {
+    const gsi::ClassBase *acls = atype.cls ();
+
     if (arg == Py_None || arg == NULL) {
 
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed null")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed null")));
       } else {
         aa->write<void *> ((void *) 0);
         return;
@@ -380,16 +486,52 @@ struct writer<gsi::ObjectType>
 
     }
 
-    if (atype.is_ptr () || atype.is_cptr () || atype.is_ref () || atype.is_cref ()) {
+    if (PyTuple_Check (arg) || PyList_Check (arg)) {
 
-      const gsi::ClassBase *cls_decl = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (arg));
-      if (! cls_decl) {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), Py_TYPE (arg)->tp_name));
+      //  we may implicitly convert a tuple into a constructor call of a target object -
+      //  for now we only check whether the number of arguments is compatible with the list given.
+
+      int n = PyTuple_Check (arg) ? int (PyTuple_Size (arg)) : int (PyList_Size (arg));
+      const gsi::MethodBase *meth = 0;
+      for (gsi::ClassBase::method_iterator c = acls->begin_constructors (); c != acls->end_constructors (); ++c) {
+        if ((*c)->compatible_with_num_args (n)) {
+          meth = *c;
+          break;
+        }
       }
 
-      if (cls_decl->is_derived_from (atype.cls ())) {
+      if (!meth) {
+        throw tl::Exception (tl::to_string (tr ("No constructor of %s available that takes %d arguments (implicit call from tuple)")), acls->name (), n);
+      }
 
-        PYAObjectBase *p = (PYAObjectBase *) (arg); 
+      //  implicit call of constructor
+      gsi::SerialArgs retlist (meth->retsize ());
+      gsi::SerialArgs arglist (meth->argsize ());
+
+      push_args (arglist, meth, arg, NULL, *heap);
+
+      meth->call (0, arglist, retlist);
+
+      void *new_obj = retlist.read<void *> (*heap);
+      if (new_obj && (atype.is_ptr () || atype.is_cptr () || atype.is_ref () || atype.is_cref ())) {
+        //  For pointers or refs, ownership over these objects is not transferred.
+        //  Hence we have to keep them on the heap.
+        //  TODO: what if the called method takes ownership using keep()?
+        heap->push (new gsi::ObjectHolder (acls, new_obj));
+      }
+
+      aa->write<void *> (new_obj);
+
+    } else if (atype.is_ptr () || atype.is_cptr () || atype.is_ref () || atype.is_cref ()) {
+
+      const gsi::ClassBase *cls_decl = PythonModule::cls_for_type (Py_TYPE (arg));
+      if (! cls_decl) {
+        throw tl::TypeError (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), Py_TYPE (arg)->tp_name));
+      }
+
+      if (cls_decl->is_derived_from (acls)) {
+
+        PYAObjectBase *p = PYAObjectBase::from_pyobject (arg);
 
         if (cls_decl->adapted_type_info ()) {
           //  resolved adapted type
@@ -398,30 +540,31 @@ struct writer<gsi::ObjectType>
           aa->write<void *> (p->obj ());
         }
 
-      } else if (cls_decl->can_convert_to (atype.cls ())) {
+      } else if (cls_decl->can_convert_to (acls)) {
 
-        PYAObjectBase *p = (PYAObjectBase *) (arg); 
+        PYAObjectBase *p = PYAObjectBase::from_pyobject (arg);
 
         //  We can convert objects for cref and cptr, but ownership over these objects is not transferred.
         //  Hence we have to keep them on the heap.
-        void *new_obj = atype.cls ()->create_obj_from (p->cls_decl (), p->obj ());
-        heap->push (new gsi::ObjectHolder (atype.cls (), new_obj));
+        //  TODO: what if the called method takes ownership using keep()?
+        void *new_obj = acls->create_obj_from (p->cls_decl (), p->obj ());
+        heap->push (new gsi::ObjectHolder (acls, new_obj));
         aa->write<void *> (new_obj);
 
       } else {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), cls_decl->name ()));
+        throw tl::TypeError (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), cls_decl->name ()));
       }
 
     } else {
 
-      const gsi::ClassBase *cls_decl = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (arg));
+      const gsi::ClassBase *cls_decl = PythonModule::cls_for_type (Py_TYPE (arg));
       if (! cls_decl) {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), Py_TYPE (arg)->tp_name));
+        throw tl::TypeError (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), Py_TYPE (arg)->tp_name));
       }
 
-      if (cls_decl->is_derived_from (atype.cls ())) {
+      if (cls_decl->is_derived_from (acls)) {
 
-        PYAObjectBase *p = (PYAObjectBase *) (arg); 
+        PYAObjectBase *p = PYAObjectBase::from_pyobject (arg);
 
         if (cls_decl->adapted_type_info ()) {
           //  resolved adapted type
@@ -430,13 +573,13 @@ struct writer<gsi::ObjectType>
           aa->write<void *> (atype.cls ()->clone (p->obj ()));
         }
 
-      } else if (cls_decl->can_convert_to (atype.cls ())) {
+      } else if (cls_decl->can_convert_to (acls)) {
 
-        PYAObjectBase *p = (PYAObjectBase *) (arg); 
+        PYAObjectBase *p = PYAObjectBase::from_pyobject (arg);
         aa->write<void *> (atype.cls ()->create_obj_from (cls_decl, p->obj ()));
 
       } else {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), cls_decl->name ()));
+        throw tl::TypeError (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), cls_decl->name ()));
       }
 
     }
@@ -446,7 +589,7 @@ struct writer<gsi::ObjectType>
 
 /**
  *  @brief A serialization wrapper (write mode)
- *  Specialisation for void 
+ *  Specialization for void
  */
 template <>
 struct writer<gsi::VoidType>
@@ -464,81 +607,105 @@ push_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, PyObject *arg, tl
 }
 
 /**
- *  @brief Deseralisation wrapper 
+ *  @brief Deserialization wrapper
  *
  *  The default implementation is for POD types, strings and variants
  */
-template <class R> 
+template <class R>
 struct reader
 {
-  void operator() (gsi::SerialArgs *rr, PythonRef *ret, PyObject * /*self*/, const gsi::ArgType &arg, tl::Heap *heap)
+  void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &arg, tl::Heap *heap)
   {
     if (arg.is_ref ()) {
-      *ret = c2python<R> (rr->template read<R &> (*heap));
+      *ret = c2python (rr->template read<R &> (*heap));
     } else if (arg.is_cref ()) {
-      *ret = c2python<R> (rr->template read<const R &> (*heap));
+      *ret = c2python (rr->template read<const R &> (*heap));
     } else if (arg.is_ptr ()) {
       R *p = rr->template read<R *> (*heap);
       if (p) {
-        *ret = c2python<R> (*p);
+        *ret = c2python (*p);
       } else {
         *ret = PythonRef (Py_None, false /*borrowed*/);
       }
     } else if (arg.is_cptr ()) {
       const R *p = rr->template read<const R *> (*heap);
       if (p) {
-        *ret = c2python<R> (*p);
+        *ret = c2python (*p);
       } else {
         *ret = PythonRef (Py_None, false /*borrowed*/);
       }
     } else {
-      *ret = c2python<R> (rr->template read<R> (*heap));
+      *ret = c2python (rr->template read<R> (*heap));
     }
   }
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for const char *
+ *  @brief Deserialization wrapper: specialization for const char *
  *
  *  Without that would would have to handle void *&, void * const &, ...
  *  TODO: right now these types are not supported.
  */
-template <> 
+template <>
 struct reader<void *>
 {
-  void operator() (gsi::SerialArgs *rr, PythonRef *ret, PyObject * /*self*/, const gsi::ArgType &arg, tl::Heap *heap)
+  void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &arg, tl::Heap *heap)
   {
     tl_assert (! arg.is_cref ());
     tl_assert (! arg.is_ref ());
     tl_assert (! arg.is_cptr ());
     tl_assert (! arg.is_ptr ());
-    *ret = c2python<void *> (rr->read<void *> (*heap));
+    *ret = c2python (rr->read<void *> (*heap));
   }
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for strings
+ *  @brief Deserialization wrapper: specialization for strings
  */
-template <> 
+template <>
 struct reader<gsi::StringType>
 {
   void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::StringAdaptor> a ((gsi::StringAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::StringAdaptor> a ((gsi::StringAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = PythonRef (Py_None, false /*borrowed*/);
     } else {
-      *ret = c2python<std::string> (std::string (a->c_str (), a->size ()));
+      *ret = c2python (std::string (a->c_str (), a->size ()));
+    }
+  }
+};
+
+/**
+ *  @brief Deserialization wrapper: specialization for byte arrays
+ */
+template <>
+struct reader<gsi::ByteArrayType>
+{
+  void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &, tl::Heap *heap)
+  {
+    std::unique_ptr<gsi::ByteArrayAdaptor> a ((gsi::ByteArrayAdaptor *) rr->read<void *>(*heap));
+    if (!a.get ()) {
+      *ret = PythonRef (Py_None, false /*borrowed*/);
+    } else {
+      const char *cp = a->c_str ();
+      size_t sz = a->size ();
+#if PY_MAJOR_VERSION < 3
+      *ret = PyByteArray_FromStringAndSize (cp, sz);
+#else
+      *ret = PyBytes_FromStringAndSize (cp, sz);
+#endif
     }
   }
 };
 
 static
-PyObject *object_from_variant (const tl::Variant &var, PYAObjectBase *self, const gsi::ArgType &atype)
+PyObject *object_from_variant (tl::Variant &var, PYAObjectBase *self, const gsi::ArgType &atype, bool transfer = false)
 {
   if (var.is_user()) {
 
-    bool pass_obj = atype.pass_obj() || (! atype.is_cptr() && ! atype.is_ptr () && ! atype.is_cref () && ! atype.is_ref ());
+    bool is_direct = (! atype.is_cptr() && ! atype.is_ptr () && ! atype.is_cref () && ! atype.is_ref ());
+    bool pass_obj = atype.pass_obj() || is_direct;
     bool is_const = atype.is_cptr() || atype.is_cref();
     bool prefer_copy = false;
     bool can_destroy = false;
@@ -546,7 +713,10 @@ PyObject *object_from_variant (const tl::Variant &var, PYAObjectBase *self, cons
     //  TODO: ugly const_cast, but there is no "const shared reference" ...
     gsi::Proxy *holder = dynamic_cast<gsi::Proxy *>(const_cast<tl::Object *>(var.to_object ()));
 
-    if (pass_obj) {
+    void *obj = var.to_user ();
+    const gsi::ClassBase *cls = var.user_cls ()->gsi_cls ();
+
+    if (pass_obj || transfer) {
 
       if (holder) {
 
@@ -567,77 +737,89 @@ PyObject *object_from_variant (const tl::Variant &var, PYAObjectBase *self, cons
         //  If the object was not owned before, it is not owned after (bears risk of invalid
         //  pointers, but it's probably rarely the case. Non-managed objects are usually copied
         //  between the ownership spaces.
-        if (var.user_is_ref()) {
+        //  If the variant holds the user object, we can take it from it and claim ownership.
+        if (var.user_is_ref ()) {
           prefer_copy = false;   // unsafe
+          pass_obj = false;
         } else {
-          prefer_copy = true;   // safe
+          obj = var.user_take ();
+          can_destroy = true;
         }
 
-        pass_obj = false;
-
       }
 
+    } else {
+
+      //  This is the case for return values that prefer to be copied (e.g. from const &)
+      prefer_copy = atype.prefer_copy ();
+
     }
 
-    return object_to_python ((void *) var.to_user (), self, var.user_cls ()->gsi_cls (), pass_obj, is_const, prefer_copy, can_destroy);
+    return object_to_python (obj, self, cls, pass_obj, is_const, prefer_copy, can_destroy);
 
   } else {
-    return c2python<tl::Variant> (var);
+    return c2python (var);
   }
 }
 
 /**
- *  @brief Deseralisation wrapper: specialization for variants
+ *  @brief Deserialization wrapper: specialization for variants
  */
-template <> 
+template <>
 struct reader<gsi::VariantType>
 {
   void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase *self, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::VariantAdaptor> a ((gsi::VariantAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::VariantAdaptor> a ((gsi::VariantAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = PythonRef (Py_None, false /*borrowed*/);
     } else {
       gsi::VariantAdaptorImpl<tl::Variant> *aa = dynamic_cast<gsi::VariantAdaptorImpl<tl::Variant> *> (a.get ());
+      PythonBasedVariantAdaptor *pa = dynamic_cast<PythonBasedVariantAdaptor *> (a.get ());
       if (aa) {
         //  A small optimization that saves one variant copy
-        *ret = object_from_variant (aa->var_ref (), self, atype);
+        *ret = object_from_variant (aa->var_ref_nc (), self, atype);
+      } else if (pa) {
+        //  Optimization for Python to Python transfer
+        *ret = pa->ptr ();
       } else {
-        *ret = object_from_variant (a->var (), self, atype);
+        tl::Variant v = a->var ();
+        //  NOTE: as v may hold the object, we need to transfer ownership
+        *ret = object_from_variant (v, self, atype, true);
       }
     }
   }
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for vectors
+ *  @brief Deserialization wrapper: specialization for vectors
  */
-template <> 
+template <>
 struct reader<gsi::VectorType>
 {
   void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::VectorAdaptor> a ((gsi::VectorAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::VectorAdaptor> a ((gsi::VectorAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = PythonRef (Py_None, false /*borrowed*/);
     } else {
       *ret = PyList_New (0);
       tl_assert (atype.inner () != 0);
       PythonBasedVectorAdaptor t (*ret, atype.inner ());
-      a->copy_to (&t, *heap); 
+      a->copy_to (&t, *heap);
     }
   }
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for maps
+ *  @brief Deserialization wrapper: specialization for maps
  */
-template <> 
+template <>
 struct reader<gsi::MapType>
 {
   void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase * /*self*/, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::MapAdaptor> a ((gsi::MapAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::MapAdaptor> a ((gsi::MapAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = PythonRef (Py_None, false /*borrowed*/);
     } else {
@@ -651,9 +833,9 @@ struct reader<gsi::MapType>
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for object 
+ *  @brief Deserialization wrapper: specialization for object
  */
-template <> 
+template <>
 struct reader<gsi::ObjectType>
 {
   void operator() (gsi::SerialArgs *rr, PythonRef *ret, PYAObjectBase *self, const gsi::ArgType &atype, tl::Heap *heap)
@@ -668,9 +850,9 @@ struct reader<gsi::ObjectType>
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for void
+ *  @brief Deserialization wrapper: specialization for void
  */
-template <> 
+template <>
 struct reader<gsi::VoidType>
 {
   void operator() (gsi::SerialArgs *, PythonRef *, PYAObjectBase *, const gsi::ArgType &, tl::Heap *)
@@ -679,8 +861,8 @@ struct reader<gsi::VoidType>
   }
 };
 
-PythonRef 
-pop_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, PYAObjectBase *self, tl::Heap &heap)
+PythonRef
+pull_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, PYAObjectBase *self, tl::Heap &heap)
 {
   PythonRef ret;
   gsi::do_on_type<reader> () (atype.type (), &aserial, &ret, self, atype, &heap);
@@ -701,7 +883,7 @@ tl::Variant PythonBasedVariantAdaptor::var () const
   return python2c<tl::Variant> (m_var.get ());
 }
 
-void PythonBasedVariantAdaptor::set (const tl::Variant & /*v*/) 
+void PythonBasedVariantAdaptor::set (const tl::Variant & /*v*/, tl::Heap & /*heap*/)
 {
   //  TODO: is there a setter for a string?
 }
@@ -715,7 +897,7 @@ PythonBasedVectorAdaptorIterator::PythonBasedVectorAdaptorIterator (const Python
   //  .. nothing yet ..
 }
 
-void PythonBasedVectorAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap) const 
+void PythonBasedVectorAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap) const
 {
   PyObject *member = NULL;
   if (PyTuple_Check (m_array.get ())) {
@@ -726,12 +908,12 @@ void PythonBasedVectorAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap)
   gsi::do_on_type<writer> () (mp_ainner->type (), &w, member, *mp_ainner, &heap);
 }
 
-bool PythonBasedVectorAdaptorIterator::at_end () const 
+bool PythonBasedVectorAdaptorIterator::at_end () const
 {
   return m_i == m_len;
 }
 
-void PythonBasedVectorAdaptorIterator::inc () 
+void PythonBasedVectorAdaptorIterator::inc ()
 {
   ++m_i;
 }
@@ -750,21 +932,23 @@ gsi::VectorAdaptorIterator *PythonBasedVectorAdaptor::create_iterator () const
   return new PythonBasedVectorAdaptorIterator (m_array, size (), mp_ainner);
 }
 
-void PythonBasedVectorAdaptor::push (gsi::SerialArgs &r, tl::Heap &heap) 
+void PythonBasedVectorAdaptor::push (gsi::SerialArgs &r, tl::Heap &heap)
 {
   if (PyList_Check (m_array.get ())) {
     PythonRef member;
     gsi::do_on_type<reader> () (mp_ainner->type (), &r, &member, (PYAObjectBase *) 0, *mp_ainner, &heap);
     PyList_Append (m_array.get (), member.get ());
   } else if (PyTuple_Check (m_array.get ())) {
-    throw tl::Exception (tl::to_string (QObject::tr ("Tuples cannot be modified and cannot be used as out parameters")));
+    throw tl::Exception (tl::to_string (tr ("Tuples cannot be modified and cannot be used as out parameters")));
   }
 }
 
-void PythonBasedVectorAdaptor::clear () 
+void PythonBasedVectorAdaptor::clear ()
 {
-  if (PySequence_Check (m_array.get ())) {
-    PySequence_DelSlice (m_array.get (), 0, PySequence_Length (m_array.get ()));
+  if (PyList_Check (m_array.get ())) {
+    PyList_SetSlice (m_array.get (), 0, PyList_Size (m_array.get ()), NULL);
+  } else if (PyTuple_Check (m_array.get ())) {
+    throw tl::Exception (tl::to_string (tr ("Tuples cannot be modified and cannot be used as out parameters")));
   }
 }
 
@@ -777,7 +961,7 @@ size_t PythonBasedVectorAdaptor::size () const
   }
 }
 
-size_t PythonBasedVectorAdaptor::serial_size () const 
+size_t PythonBasedVectorAdaptor::serial_size () const
 {
   return mp_ainner->size ();
 }
@@ -793,18 +977,18 @@ PythonBasedMapAdaptorIterator::PythonBasedMapAdaptorIterator (const PythonPtr &h
   inc ();
 }
 
-void PythonBasedMapAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap) const 
+void PythonBasedMapAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap) const
 {
   gsi::do_on_type<writer> () (mp_ainner_k->type (), &w, m_key, *mp_ainner_k, &heap);
   gsi::do_on_type<writer> () (mp_ainner->type (), &w, m_value, *mp_ainner, &heap);
 }
 
-bool PythonBasedMapAdaptorIterator::at_end () const 
+bool PythonBasedMapAdaptorIterator::at_end () const
 {
   return ! m_has_items;
 }
 
-void PythonBasedMapAdaptorIterator::inc () 
+void PythonBasedMapAdaptorIterator::inc ()
 {
   m_has_items = PyDict_Next(m_hash.get (), &m_pos, &m_key, &m_value);
 }
@@ -813,7 +997,7 @@ void PythonBasedMapAdaptorIterator::inc ()
 //  PythonBasedMapAdaptor implementation
 
 PythonBasedMapAdaptor::PythonBasedMapAdaptor (const PythonPtr &hash, const gsi::ArgType *ainner, const gsi::ArgType *ainner_k)
-  : mp_ainner (ainner), mp_ainner_k (ainner_k), m_hash (hash) 
+  : mp_ainner (ainner), mp_ainner_k (ainner_k), m_hash (hash)
 {
 }
 
@@ -822,7 +1006,7 @@ gsi::MapAdaptorIterator *PythonBasedMapAdaptor::create_iterator () const
   return new PythonBasedMapAdaptorIterator (m_hash, mp_ainner, mp_ainner_k);
 }
 
-void PythonBasedMapAdaptor::insert (gsi::SerialArgs &r, tl::Heap &heap) 
+void PythonBasedMapAdaptor::insert (gsi::SerialArgs &r, tl::Heap &heap)
 {
   PythonRef k, v;
   gsi::do_on_type<reader> () (mp_ainner_k->type (), &r, &k, (PYAObjectBase *) 0, *mp_ainner_k, &heap);
@@ -830,7 +1014,7 @@ void PythonBasedMapAdaptor::insert (gsi::SerialArgs &r, tl::Heap &heap)
   PyDict_SetItem (m_hash.get (), k.get (), v.get ());
 }
 
-void PythonBasedMapAdaptor::clear () 
+void PythonBasedMapAdaptor::clear ()
 {
   PyDict_Clear (m_hash.get ());
 }
@@ -840,7 +1024,7 @@ size_t PythonBasedMapAdaptor::size () const
   return PyDict_Size (m_hash.get ());
 }
 
-size_t PythonBasedMapAdaptor::serial_size () const 
+size_t PythonBasedMapAdaptor::serial_size () const
 {
   return mp_ainner_k->size () + mp_ainner->size ();
 }
@@ -862,7 +1046,7 @@ size_t PythonBasedMapAdaptor::serial_size () const
 template <class R>
 struct test_arg_func
 {
-  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose)
+  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Py_None) {
 
@@ -876,7 +1060,7 @@ struct test_arg_func
       if (atype.is_ptr () || atype.is_ref ()) {
 
         //  check if we have a boxed type
-        const gsi::ClassBase *cls_decl = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (arg));
+        const gsi::ClassBase *cls_decl = PythonModule::cls_for_type (Py_TYPE (arg));
         if (cls_decl) {
           const gsi::ClassBase *bc = gsi::cls_decl <gsi::Value> ();
           if (cls_decl->is_derived_from (bc)) {
@@ -899,7 +1083,7 @@ struct test_arg_func
 template <>
 struct test_arg_func<gsi::VariantType>
 {
-  void operator() (bool *ret, PyObject *, const gsi::ArgType &, bool)
+  void operator() (bool *ret, PyObject *, const gsi::ArgType &, bool, bool)
   {
     //  we assume we can convert everything into a variant
     *ret = true;
@@ -909,16 +1093,16 @@ struct test_arg_func<gsi::VariantType>
 template <>
 struct test_arg_func<gsi::StringType>
 {
-  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &, bool)
+  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &, bool, bool)
   {
 #if PY_MAJOR_VERSION < 3
     if (PyString_Check (arg)) {
       *ret = true;
-    } else 
+    } else
 #else
     if (PyBytes_Check (arg)) {
       *ret = true;
-    } else 
+    } else
 #endif
     if (PyUnicode_Check (arg)) {
       *ret = true;
@@ -933,8 +1117,14 @@ struct test_arg_func<gsi::StringType>
 template <>
 struct test_arg_func<gsi::VectorType>
 {
-  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose)
+  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
+    if ((atype.is_cptr () || atype.is_ptr ()) && arg == Py_None) {
+      //  for ptr or cptr, null is an allowed value
+      *ret = true;
+      return;
+    }
+
     if (! PyTuple_Check (arg) && ! PyList_Check (arg)) {
       *ret = false;
       return;
@@ -948,7 +1138,7 @@ struct test_arg_func<gsi::VectorType>
 
       size_t n = PyTuple_Size (arg);
       for (size_t i = 0; i < n && *ret; ++i) {
-        if (! test_arg (ainner, PyTuple_GetItem (arg, i), loose)) {
+        if (! test_arg (ainner, PyTuple_GetItem (arg, i), loose, true /*issue-1651*/)) {
           *ret = false;
         }
       }
@@ -957,7 +1147,7 @@ struct test_arg_func<gsi::VectorType>
 
       size_t n = PyList_Size (arg);
       for (size_t i = 0; i < n && *ret; ++i) {
-        if (! test_arg (ainner, PyList_GetItem (arg, i), loose)) {
+        if (! test_arg (ainner, PyList_GetItem (arg, i), loose, true /*issue-1651*/)) {
           *ret = false;
         }
       }
@@ -969,8 +1159,14 @@ struct test_arg_func<gsi::VectorType>
 template <>
 struct test_arg_func<gsi::MapType>
 {
-  void operator () (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose)
+  void operator () (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
+    if ((atype.is_cptr () || atype.is_ptr ()) && arg == Py_None) {
+      //  for ptr or cptr, null is an allowed value
+      *ret = true;
+      return;
+    }
+
     if (! PyDict_Check (arg)) {
       *ret = false;
       return;
@@ -981,42 +1177,62 @@ struct test_arg_func<gsi::MapType>
     const gsi::ArgType &ainner = *atype.inner ();
     const gsi::ArgType &ainner_k = *atype.inner ();
 
-    //  Note: we test key and value separately. That way we don't need to 
+    //  Note: we test key and value separately. That way we don't need to
     //  instantiate a 2d template with do_on_type2.
     *ret = true;
 
     PyObject *key, *value;
     Py_ssize_t pos = 0;
     while (PyDict_Next(arg, &pos, &key, &value)) {
-      if (! test_arg (ainner_k, key, loose)) {
+      if (! test_arg (ainner_k, key, loose, true /*issue-1651*/)) {
         *ret = false;
         break;
       }
-      if (! test_arg (ainner, value, loose)) {
+      if (! test_arg (ainner, value, loose, true /*issue-1651*/)) {
         *ret = false;
         break;
       }
     }
   }
 };
+
 template <>
 struct test_arg_func<gsi::ObjectType>
 {
-  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose)
+  void operator() (bool *ret, PyObject *arg, const gsi::ArgType &atype, bool loose, bool object_substitution)
   {
+    const gsi::ClassBase *acls = atype.cls ();
+
     //  for const X * or X *, null is an allowed value
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Py_None) {
       *ret = true;
       return;
     }
 
-    const gsi::ClassBase *cls_decl = PythonInterpreter::instance ()->cls_for_type (Py_TYPE (arg));
+    if (object_substitution && (PyTuple_Check (arg) || PyList_Check (arg))) {
+
+      //  we may implicitly convert a tuple into a constructor call of a target object -
+      //  for now we only check whether the number of arguments is compatible with the list given.
+
+      int n = PyTuple_Check (arg) ? int (PyTuple_Size (arg)) : int (PyList_Size (arg));
+      *ret = false;
+      for (gsi::ClassBase::method_iterator c = acls->begin_constructors (); c != acls->end_constructors (); ++c) {
+        if ((*c)->compatible_with_num_args (n)) {
+          *ret = true;
+          break;
+        }
+      }
+      return;
+
+    }
+
+    const gsi::ClassBase *cls_decl = PythonModule::cls_for_type (Py_TYPE (arg));
     if (! cls_decl) {
       *ret = false;
       return;
     }
 
-    if (! (cls_decl == atype.cls () || (loose && (cls_decl->is_derived_from (atype.cls ()) || cls_decl->can_convert_to(atype.cls ()))))) {
+    if (! (cls_decl == acls || (loose && (cls_decl->is_derived_from (atype.cls ()) || (object_substitution && cls_decl->can_convert_to (atype.cls ())))))) {
       *ret = false;
       return;
     }
@@ -1031,12 +1247,11 @@ struct test_arg_func<gsi::ObjectType>
 };
 
 bool
-test_arg (const gsi::ArgType &atype, PyObject *arg, bool loose)
+test_arg (const gsi::ArgType &atype, PyObject *arg, bool loose, bool object_substitution)
 {
   bool ret = false;
-  gsi::do_on_type<test_arg_func> () (atype.type (), &ret, arg, atype, loose);
+  gsi::do_on_type<test_arg_func> () (atype.type (), &ret, arg, atype, loose, object_substitution);
   return ret;
 }
-      
-}
 
+}
