@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -34,6 +34,8 @@
 namespace rba
 {
 
+void push_args (gsi::SerialArgs &arglist, const gsi::MethodBase *meth, VALUE *argv, int argc, VALUE kwargs, tl::Heap &heap);
+
 // -------------------------------------------------------------------
 //  Serialization adaptors for strings, variants, vectors and maps
 
@@ -46,7 +48,7 @@ class RubyBasedStringAdaptor
 public:
   RubyBasedStringAdaptor (VALUE value)
   {
-    m_string = rba_safe_string_value (value);
+    m_string = rba_safe_obj_as_string (value);
     gc_lock_object (m_string);
   }
 
@@ -68,12 +70,51 @@ public:
   virtual void set (const char * /*c_str*/, size_t /*s*/, tl::Heap & /*heap*/)
   {
     //  TODO: is there a setter for a string?
+    //  -> so far, string OUT parameters are not supported
   }
 
 private:
   VALUE m_string;
 };
 
+/**
+ *  @brief An adaptor for a byte array from ruby objects
+ */
+class RubyBasedByteArrayAdaptor
+  : public gsi::ByteArrayAdaptor
+{
+public:
+  RubyBasedByteArrayAdaptor (VALUE value)
+  {
+    m_bytes = rba_safe_string_value (value);
+    gc_lock_object (m_bytes);
+  }
+
+  ~RubyBasedByteArrayAdaptor ()
+  {
+    gc_unlock_object (m_bytes);
+  }
+
+  virtual const char *c_str () const
+  {
+    return RSTRING_PTR (m_bytes);
+  }
+
+  virtual size_t size () const
+  {
+    return RSTRING_LEN (m_bytes);
+  }
+
+  virtual void set (const char * /*c_str*/, size_t /*s*/, tl::Heap & /*heap*/)
+  {
+    //  TODO: is there a setter for a string?
+    //  -> so far, byte array OUT parameters are not supported
+  }
+
+private:
+  VALUE m_bytes;
+};
+
 /**
  *  @brief An adaptor for a variant from ruby objects
  */
@@ -85,7 +126,8 @@ public:
   ~RubyBasedVariantAdaptor ();
 
   virtual tl::Variant var () const;
-  virtual void set (const tl::Variant &v);
+  virtual void set (const tl::Variant &v, tl::Heap &heap);
+  VALUE value () const { return m_var; }
 
 private:
   VALUE m_var;
@@ -192,7 +234,7 @@ struct get_boxed_value_func
       Proxy *p = 0;
       Data_Get_Struct (arg, Proxy, p);
       if (!p->cls_decl ()->is_derived_from (bt)) {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Passing an object to pointer or reference requires a boxed type (RBA::%s)")), bt->name ()));
+        throw tl::Exception (tl::sprintf (tl::to_string (tr ("Passing an object to pointer or reference requires a boxed type (RBA::%s)")), bt->name ()));
       }
 
       gsi::Value *bo = reinterpret_cast<gsi::Value *> (p->obj ());
@@ -214,6 +256,7 @@ struct get_boxed_value_func_error
 
 template <> struct get_boxed_value_func<gsi::VariantType> : get_boxed_value_func_error { };
 template <> struct get_boxed_value_func<gsi::StringType> : get_boxed_value_func_error { };
+template <> struct get_boxed_value_func<gsi::ByteArrayType> : get_boxed_value_func_error { };
 template <> struct get_boxed_value_func<gsi::ObjectType> : get_boxed_value_func_error { };
 template <> struct get_boxed_value_func<gsi::VectorType> : get_boxed_value_func_error { };
 template <> struct get_boxed_value_func<gsi::MapType> : get_boxed_value_func_error { };
@@ -241,7 +284,7 @@ struct writer
     if (arg == Qnil) {
 
       if (atype.is_ref () || atype.is_cref ()) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference type cannot be passed nil")));
       } else if (atype.is_ptr ()) {
         aa->write<R *> ((R *)0);
       } else if (atype.is_cptr ()) {
@@ -256,7 +299,7 @@ struct writer
         // references or pointers require a boxed object. Pointers also allow nil.
         void *vc = boxed_value_ptr (atype.type (), arg, *heap);
         if (! vc && atype.is_ref ()) {
-          throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
         }
         aa->write<void *> (vc);
       } else if (atype.is_cref ()) {
@@ -300,7 +343,7 @@ struct writer<gsi::StringType>
         void *vc = 0;
         get_boxed_value_func<std::string> () (&vc, arg, heap);
         if (! vc && atype.is_ref ()) {
-          throw tl::Exception (tl::to_string (QObject::tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
         }
 
         //  NOTE: by convention we pass the ownership to the receiver for adaptors.
@@ -321,6 +364,55 @@ struct writer<gsi::StringType>
   }
 };
 
+/**
+ *  @brief Serialization for strings
+ */
+template <>
+struct writer<gsi::ByteArrayType>
+{
+  void operator() (gsi::SerialArgs *aa, VALUE arg, const gsi::ArgType &atype, tl::Heap *heap)
+  {
+    //  Cannot pass ownership currently
+    tl_assert (!atype.pass_obj ());
+
+    if (arg == Qnil) {
+
+      if (! (atype.is_ptr () || atype.is_cptr ())) {
+        //  nil is treated as an empty string for references
+        aa->write<void *> ((void *)new gsi::ByteArrayAdaptorImpl<std::vector<char> > (std::vector<char> ()));
+      } else {
+        aa->write<void *> ((void *)0);
+      }
+
+    } else {
+
+      if (atype.is_ref () || atype.is_ptr ()) {
+
+        // references or pointers require a boxed object. Pointers also allow nil.
+        void *vc = 0;
+        get_boxed_value_func<std::vector<char> > () (&vc, arg, heap);
+        if (! vc && atype.is_ref ()) {
+          throw tl::Exception (tl::to_string (tr ("Arguments or return values of reference or direct type cannot be passed nil or an empty boxed value object")));
+        }
+
+        //  NOTE: by convention we pass the ownership to the receiver for adaptors.
+        if (! vc) {
+          aa->write<void *> (0);
+        } else {
+          aa->write<void *> ((void *)new gsi::ByteArrayAdaptorImpl<std::vector<char> > ((std::vector<char> *) vc));
+        }
+
+      } else {
+
+        //  NOTE: by convention we pass the ownership to the receiver for adaptors.
+        aa->write<void *> ((void *)new RubyBasedByteArrayAdaptor (arg));
+
+      }
+
+    }
+  }
+};
+
 /**
  *  @brief Specialization for Variant
  */
@@ -345,13 +437,19 @@ struct writer<gsi::VectorType>
   {
     if (arg == Qnil) {
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed nil")));
       } else {
         aa->write<void *> ((void *)0);
       }
     } else {
+
+      if (TYPE (arg) != T_ARRAY) {
+        throw tl::Exception (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected array, got %s)")), rba_class_name (arg).c_str ()));
+      }
+
       tl_assert (atype.inner () != 0);
       aa->write<void *> ((void *)new RubyBasedVectorAdaptor (arg, atype.inner ()));
+
     }
   }
 };
@@ -366,20 +464,27 @@ struct writer<gsi::MapType>
   {
     if (arg == Qnil) {
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed nil")));
       } else {
         aa->write<void *> ((void *)0);
       }
+
     } else {
+
+      if (TYPE (arg) != T_HASH) {
+        throw tl::Exception (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected hash, got %s)")), rba_class_name (arg).c_str ()));
+      }
+
       tl_assert (atype.inner () != 0);
       tl_assert (atype.inner_k () != 0);
       aa->write<void *> ((void *)new RubyBasedMapAdaptor (arg, atype.inner (), atype.inner_k ()));
+
     }
   }
 };
 
 /**
- *  @brief A specialisation of the write function for vector types
+ *  @brief A specialization of the write function for vector types
  */
 template <>
 struct writer <gsi::ObjectType>
@@ -389,15 +494,51 @@ struct writer <gsi::ObjectType>
     if (arg == Qnil) {
 
       if (! (atype.is_ptr () || atype.is_cptr ())) {
-        throw tl::Exception (tl::to_string (QObject::tr ("Arguments of reference or direct type cannot be passed nil")));
+        throw tl::Exception (tl::to_string (tr ("Arguments of reference or direct type cannot be passed nil")));
       } else {
         aa->write<void *> ((void *) 0);
       }
 
+    } else if (TYPE (arg) == T_ARRAY) {
+
+      //  we may implicitly convert an array into a constructor call of a target object -
+      //  for now we only check whether the number of arguments is compatible with the array given.
+
+      int n = RARRAY_LEN (arg);
+      const gsi::MethodBase *meth = 0;
+      for (gsi::ClassBase::method_iterator c = atype.cls ()->begin_constructors (); c != atype.cls ()->end_constructors (); ++c) {
+        if ((*c)->compatible_with_num_args (n)) {
+          meth = *c;
+          break;
+        }
+      }
+
+      if (!meth) {
+        throw tl::Exception (tl::to_string (tr ("No constructor of %s available that takes %d arguments (implicit call from tuple)")), atype.cls ()->name (), n);
+      }
+
+      //  implicit call of constructor
+      gsi::SerialArgs retlist (meth->retsize ());
+      gsi::SerialArgs arglist (meth->argsize ());
+
+      push_args (arglist, meth, RARRAY_PTR (arg), n, Qnil, *heap);
+
+      meth->call (0, arglist, retlist);
+
+      void *new_obj = retlist.read<void *> (*heap);
+      if (new_obj && (atype.is_ptr () || atype.is_cptr () || atype.is_ref () || atype.is_cref ())) {
+        //  For pointers or refs, ownership over these objects is not transferred.
+        //  Hence we have to keep them on the heap.
+        //  TODO: what if the called method takes ownership using keep()?
+        heap->push (new gsi::ObjectHolder (atype.cls (), new_obj));
+      }
+
+      aa->write<void *> (new_obj);
+
     } else {
 
       if (TYPE (arg) != T_DATA) {
-        throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
+        throw tl::Exception (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
       }
 
       Proxy *p = 0;
@@ -423,7 +564,7 @@ struct writer <gsi::ObjectType>
           aa->write<void *> (new_obj);
 
         } else {
-          throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
+          throw tl::Exception (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
         }
 
       } else {
@@ -442,7 +583,7 @@ struct writer <gsi::ObjectType>
           aa->write<void *> (atype.cls ()->create_obj_from (p->cls_decl (), p->obj ()));
 
         } else {
-          throw tl::Exception (tl::sprintf (tl::to_string (QObject::tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
+          throw tl::Exception (tl::sprintf (tl::to_string (tr ("Unexpected object type (expected argument of class %s, got %s)")), atype.cls ()->name (), rba_class_name (arg).c_str ()));
         }
 
       }
@@ -474,7 +615,7 @@ push_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, VALUE arg, tl::He
 }
 
 /**
- *  @brief Deseralisation wrapper
+ *  @brief Deserialization wrapper
  *
  *  The default implementation is for POD types, strings and variants
  */
@@ -508,7 +649,7 @@ struct reader
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for void *
+ *  @brief Deserialization wrapper: specialization for void *
  *
  *  Without that would would have to handle void *&, void * const &, ...
  *  TODO: right now these types are not supported.
@@ -527,27 +668,45 @@ struct reader<void *>
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for strings
+ *  @brief Deserialization wrapper: specialization for strings
  */
 template <>
 struct reader<gsi::StringType>
 {
   void operator() (gsi::SerialArgs *rr, VALUE *ret, Proxy * /*self*/, const gsi::ArgType &, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::StringAdaptor> a ((gsi::StringAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::StringAdaptor> a ((gsi::StringAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = Qnil;
     } else {
-      *ret = rb_str_new (a->c_str (), a->size ());
+      *ret = rb_utf8_str_new (a->c_str (), long (a->size ()));
     }
   }
 };
 
-static VALUE object_from_variant (const tl::Variant &var, Proxy *self, const gsi::ArgType &atype)
+/**
+ *  @brief Deserialization wrapper: specialization for strings
+ */
+template <>
+struct reader<gsi::ByteArrayType>
+{
+  void operator() (gsi::SerialArgs *rr, VALUE *ret, Proxy * /*self*/, const gsi::ArgType &, tl::Heap *heap)
+  {
+    std::unique_ptr<gsi::ByteArrayAdaptor> a ((gsi::ByteArrayAdaptor *) rr->read<void *>(*heap));
+    if (!a.get ()) {
+      *ret = Qnil;
+    } else {
+      *ret = rb_str_new (a->c_str (), long (a->size ()));
+    }
+  }
+};
+
+static VALUE object_from_variant (tl::Variant &var, Proxy *self, const gsi::ArgType &atype, bool transfer = false)
 {
   if (var.is_user()) {
 
-    bool pass_obj = atype.pass_obj() || (! atype.is_cptr() && ! atype.is_ptr () && ! atype.is_cref () && ! atype.is_ref ());
+    bool is_direct = (! atype.is_cptr() && ! atype.is_ptr () && ! atype.is_cref () && ! atype.is_ref ());
+    bool pass_obj = atype.pass_obj() || is_direct;
     bool is_const = atype.is_cptr() || atype.is_cref();
     bool prefer_copy = false;
     bool can_destroy = false;
@@ -555,7 +714,10 @@ static VALUE object_from_variant (const tl::Variant &var, Proxy *self, const gsi
     //  TODO: ugly const_cast, but there is no "const shared reference" ...
     gsi::Proxy *holder = dynamic_cast<gsi::Proxy *>(const_cast<tl::Object *>(var.to_object ()));
 
-    if (pass_obj) {
+    void *obj = var.to_user ();
+    const gsi::ClassBase *cls = var.user_cls ()->gsi_cls ();
+
+    if (pass_obj || transfer) {
 
       if (holder) {
 
@@ -576,19 +738,25 @@ static VALUE object_from_variant (const tl::Variant &var, Proxy *self, const gsi
         //  If the object was not owned before, it is not owned after (bears risk of invalid
         //  pointers, but it's probably rarely the case. Non-managed objects are usually copied
         //  between the ownership spaces.
-        if (var.user_is_ref()) {
+        //  If the variant holds the user object, we can take it from it and claim ownership.
+        if (var.user_is_ref ()) {
           prefer_copy = false;   // unsafe
+          pass_obj = false;
         } else {
-          prefer_copy = true;   // safe
+          obj = var.user_take ();
+          can_destroy = true;
         }
 
-        pass_obj = false;
-
       }
 
+    } else {
+
+      //  This is the case for return values that prefer to be copied (e.g. from const &)
+      prefer_copy = atype.prefer_copy ();
+
     }
 
-    return object_to_ruby ((void *) var.to_user (), self, var.user_cls ()->gsi_cls (), pass_obj, is_const, prefer_copy, can_destroy);
+    return object_to_ruby (obj, self, cls, pass_obj, is_const, prefer_copy, can_destroy);
 
   } else {
     return c2ruby<tl::Variant> (var);
@@ -596,37 +764,43 @@ static VALUE object_from_variant (const tl::Variant &var, Proxy *self, const gsi
 }
 
 /**
- *  @brief Deseralisation wrapper: specialization for variants
+ *  @brief Deserialization wrapper: specialization for variants
  */
 template <>
 struct reader<gsi::VariantType>
 {
   void operator() (gsi::SerialArgs *rr, VALUE *ret, Proxy *self, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::VariantAdaptor> a ((gsi::VariantAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::VariantAdaptor> a ((gsi::VariantAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = Qnil;
     } else {
       gsi::VariantAdaptorImpl<tl::Variant> *aa = dynamic_cast<gsi::VariantAdaptorImpl<tl::Variant> *> (a.get ());
+      RubyBasedVariantAdaptor *pa = dynamic_cast<RubyBasedVariantAdaptor *> (a.get ());
       if (aa) {
         //  A small optimization that saves one variant copy
-        *ret = object_from_variant (aa->var_ref (), self, atype);
+        *ret = object_from_variant (aa->var_ref_nc (), self, atype);
+      } else if (pa) {
+        //  Optimization for Ruby to Ruby transfer
+        *ret = pa->value ();
       } else {
-        *ret = object_from_variant (a->var (), self, atype);
+        tl::Variant v = a->var ();
+        //  NOTE: as v may hold the object, we need to transfer ownership
+        *ret = object_from_variant (v, self, atype, true);
       }
     }
   }
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for vectors
+ *  @brief Deserialization wrapper: specialization for vectors
  */
 template <>
 struct reader<gsi::VectorType>
 {
   void operator() (gsi::SerialArgs *rr, VALUE *ret, Proxy * /*self*/, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::VectorAdaptor> a ((gsi::VectorAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::VectorAdaptor> a ((gsi::VectorAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = Qnil;
     } else {
@@ -639,14 +813,14 @@ struct reader<gsi::VectorType>
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for maps
+ *  @brief Deserialization wrapper: specialization for maps
  */
 template <>
 struct reader<gsi::MapType>
 {
   void operator() (gsi::SerialArgs *rr, VALUE *ret, Proxy * /*self*/, const gsi::ArgType &atype, tl::Heap *heap)
   {
-    std::auto_ptr<gsi::MapAdaptor> a ((gsi::MapAdaptor *) rr->read<void *>(*heap));
+    std::unique_ptr<gsi::MapAdaptor> a ((gsi::MapAdaptor *) rr->read<void *>(*heap));
     if (!a.get ()) {
       *ret = Qnil;
     } else {
@@ -660,7 +834,7 @@ struct reader<gsi::MapType>
 };
 
 /**
- *  @brief Deseralisation wrapper: specialization for objects
+ *  @brief Deserialization wrapper: specialization for objects
  */
 template <>
 struct reader<gsi::ObjectType>
@@ -704,7 +878,7 @@ tl::Variant RubyBasedVariantAdaptor::var () const
   return ruby2c<tl::Variant> (m_var);
 }
 
-void RubyBasedVariantAdaptor::set (const tl::Variant & /*v*/)
+void RubyBasedVariantAdaptor::set (const tl::Variant & /*v*/, tl::Heap & /*heap*/)
 {
   //  TODO: is there a setter for a string?
 }
@@ -720,7 +894,7 @@ RubyBasedVectorAdaptorIterator::RubyBasedVectorAdaptorIterator (VALUE array, con
 
 void RubyBasedVectorAdaptorIterator::get (gsi::SerialArgs &w, tl::Heap &heap) const
 {
-  gsi::do_on_type<writer> () (mp_ainner->type (), &w, rb_ary_entry (m_array, m_i), *mp_ainner, &heap);
+  gsi::do_on_type<writer> () (mp_ainner->type (), &w, rb_ary_entry (m_array, long (m_i)), *mp_ainner, &heap);
 }
 
 bool RubyBasedVectorAdaptorIterator::at_end () const
@@ -852,10 +1026,10 @@ size_t RubyBasedMapAdaptor::serial_size () const
 }
 
 // -------------------------------------------------------------------
-//  Pops an argument from the call or return stack
+//  Pulls an argument from the front of an argument queue
 
 VALUE
-pop_arg (const gsi::ArgType &atype, Proxy *self, gsi::SerialArgs &aserial, tl::Heap &heap)
+pull_arg (const gsi::ArgType &atype, Proxy *self, gsi::SerialArgs &aserial, tl::Heap &heap)
 {
   VALUE ret = Qnil;
   gsi::do_on_type<reader> () (atype.type (), &aserial, &ret, self, atype, &heap);
@@ -879,7 +1053,7 @@ pop_arg (const gsi::ArgType &atype, Proxy *self, gsi::SerialArgs &aserial, tl::H
 template <class R>
 struct test_arg_func
 {
-  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose)
+  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Qnil) {
 
@@ -927,7 +1101,7 @@ struct test_vector
     unsigned int len = RARRAY_LEN(arr);
     VALUE *el = RARRAY_PTR(arr);
     while (len-- > 0) {
-      if (! test_arg (ainner, *el++, loose)) {
+      if (! test_arg (ainner, *el++, loose, true /*issue-1651*/)) {
         *ret = false;
         break;
       }
@@ -938,12 +1112,12 @@ struct test_vector
 template <>
 struct test_arg_func<gsi::VectorType>
 {
-  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose)
+  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Qnil) {
       //  for pointers to vectors, nil is a valid value
       *ret = true;
-    } if (TYPE (arg) != T_ARRAY) {
+    } else if (TYPE (arg) != T_ARRAY) {
       *ret = false;
     } else {
 
@@ -967,11 +1141,11 @@ struct HashTestKeyValueData
 static int hash_test_value_key (VALUE key, VALUE value, VALUE a)
 {
   HashTestKeyValueData *args = (HashTestKeyValueData *)a;
-  if (! test_arg (*args->ainner_k, key, args->loose)) {
+  if (! test_arg (*args->ainner_k, key, args->loose, true /*issue-1651*/)) {
     *(args->ret) = false;
     return ST_STOP;
   }
-  if (! test_arg (*args->ainner, value, args->loose)) {
+  if (! test_arg (*args->ainner, value, args->loose, true /*issue-1651*/)) {
     *(args->ret) = false;
     return ST_STOP;
   }
@@ -981,7 +1155,7 @@ static int hash_test_value_key (VALUE key, VALUE value, VALUE a)
 template <>
 struct test_arg_func<gsi::MapType>
 {
-  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose)
+  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose, bool /*object_substitution*/)
   {
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Qnil) {
       //  for pointers to maps, nil is a valid value
@@ -1009,13 +1183,28 @@ struct test_arg_func<gsi::MapType>
 template <>
 struct test_arg_func<gsi::ObjectType>
 {
-  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose)
+  void operator () (bool *ret, VALUE arg, const gsi::ArgType &atype, bool loose, bool object_substitution)
   {
     if ((atype.is_cptr () || atype.is_ptr ()) && arg == Qnil) {
 
       //  for const X * or X *, nil is an allowed value
       *ret = true;
 
+    } else if (object_substitution && TYPE (arg) == T_ARRAY) {
+
+      //  we may implicitly convert an array into a constructor call of a target object -
+      //  for now we only check whether the number of arguments is compatible with the array given.
+
+      int n = RARRAY_LEN (arg);
+
+      *ret = false;
+      for (gsi::ClassBase::method_iterator c = atype.cls ()->begin_constructors (); c != atype.cls ()->end_constructors (); ++c) {
+        if ((*c)->compatible_with_num_args (n)) {
+          *ret = true;
+          break;
+        }
+      }
+
     } else {
 
       *ret = (TYPE (arg) == T_DATA);
@@ -1028,7 +1217,7 @@ struct test_arg_func<gsi::ObjectType>
 
         //  in loose mode (second pass) try to match the types via implicit constructors,
         //  in strict mode (first pass) require direct type match
-        if (p->cls_decl () == atype.cls () || (loose && (p->cls_decl ()->is_derived_from (atype.cls ()) || p->cls_decl ()->can_convert_to (atype.cls ())))) {
+        if (p->cls_decl () == atype.cls () || (loose && (p->cls_decl ()->is_derived_from (atype.cls ()) || (object_substitution && p->cls_decl ()->can_convert_to (atype.cls ()))))) {
           //  type matches: check constness
           if ((atype.is_ref () || atype.is_ptr ()) && p->const_ref ()) {
             *ret = false;
@@ -1045,10 +1234,10 @@ struct test_arg_func<gsi::ObjectType>
 };
 
 bool
-test_arg (const gsi::ArgType &atype, VALUE arg, bool loose)
+test_arg (const gsi::ArgType &atype, VALUE arg, bool loose, bool object_substitution)
 {
   bool ret = false;
-  gsi::do_on_type<test_arg_func> () (atype.type (), &ret, arg, atype, loose);
+  gsi::do_on_type<test_arg_func> () (atype.type (), &ret, arg, atype, loose, object_substitution);
   return ret;
 }
 
