@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -27,9 +27,11 @@
 #include "rdbCommon.h"
 
 #include "dbTrans.h"
+#include "dbObjectWithProperties.h"
 #include "gsi.h"
 #include "tlObject.h"
 #include "tlObjectCollection.h"
+#include "tlPixelBuffer.h"
 
 #include <string>
 #include <list>
@@ -37,7 +39,9 @@
 #include <set>
 #include <vector>
 
+#if defined(HAVE_QT)
 class QImage;
+#endif
 
 namespace tl
 {
@@ -47,6 +51,7 @@ namespace tl
 namespace db
 {
   class Shape;
+  class Layout;
 }
 
 namespace rdb
@@ -67,7 +72,7 @@ class Items;
  *  An item is member of exactly one category. This can be a check for example. 
  *  A category is described by a name and a description string. An Id is provided
  *  to reference this category from actual report items.
- *  Categories can be organised hierarchically for which a category collection 
+ *  Categories can be organized hierarchically for which a category collection 
  *  is provided and member of the individual category.
  *
  *  A category can only be created by the database object, since the
@@ -409,8 +414,8 @@ public:
   }
 
   void add_category (Category *cath);
-
   void set_database (Database *database);
+  Category *category_by_raw_name (const std::string &name);
 };
 
 /**
@@ -435,6 +440,8 @@ public:
 
   virtual std::string to_display_string () const = 0;
 
+  virtual bool is_shape () const = 0;
+
   virtual ValueBase *clone () const = 0;
 
   virtual int type_index () const = 0;
@@ -464,7 +471,7 @@ int type_index_of ();
  *  Using this class, any value can be stored inside the collection of Values.
  */
 template <class C>
-class RDB_PUBLIC Value 
+class RDB_PUBLIC_TEMPLATE Value
   : public ValueBase
 {
 public:
@@ -500,10 +507,9 @@ public:
     return type_index_of<C> (); 
   }
 
-  bool compare (const ValueBase *other) const 
-  {
-    return m_value < static_cast<const Value<C> *> (other)->m_value;
-  }
+  bool compare (const ValueBase *other) const;
+
+  bool is_shape () const;
 
   std::string to_string () const;
 
@@ -518,6 +524,24 @@ private:
   C m_value;
 };
 
+/**
+ *  @brief Type bindings
+ */
+template <class T>
+RDB_PUBLIC_TEMPLATE ValueBase *make_value (const T &value)
+{
+  return new Value<T> (value);
+}
+
+/**
+ *  @brief Type bindings
+ */
+template <class T>
+RDB_PUBLIC_TEMPLATE ValueBase *make_value (const db::object_with_properties<T> &value)
+{
+  return new Value<T> (value);
+}
+
 /**
  *  @brief A class encapsulating ValueBase pointer
  */
@@ -615,7 +639,7 @@ public:
   /**
    *  @brief Convert the values collection to a string 
    */
-  std::string to_string (const Database *rdb) const;
+  std::string to_string (const Database *rdb = 0) const;
 
   /**
    *  @brief Fill the values collection from the string
@@ -662,6 +686,19 @@ public:
    */
   Values &operator= (const Values &d);
 
+  /**
+   *  @brief Compare two value sets (less operator)
+   *
+   *  This compare function will use the tag mapping provided by tag map ("this" tag id to "other" tag id).
+   *  Values with tags not listed in the tag map will not be compared.
+   *  Untagged values (tag_id 0) will be compared always.
+   *
+   *  "rev_tag_map" needs to be the reverse of "tag_map".
+   *
+   *  The order of the values matters.
+   */
+  bool compare (const Values &other, const std::map<id_type, id_type> &tag_map, const std::map<id_type, id_type> &rev_tag_map) const;
+
   /**
    *  @brief The const iterator (begin)
    */
@@ -722,6 +759,14 @@ public:
     m_values.swap (other.m_values);
   }
 
+  /**
+   *  @brief Clears the values
+   */
+  void clear ()
+  {
+    m_values.clear ();
+  }
+
   /**
    *  @brief Convert the values collection to a string 
    */
@@ -848,9 +893,11 @@ public:
    *  @brief Add a value in a generic way
    */
   template <class V>
-  void add_value (const V &v)
+  ValueBase *add_value (const V &v, id_type tag_id = 0)
   {
-    values ().add (new Value<V> (v));
+    ValueBase *value = new Value<V> (v);
+    values ().add (value, tag_id);
+    return value;
   }
 
   /**
@@ -928,36 +975,72 @@ public:
    */
   void set_tag_str (const std::string &tags);
 
+#if defined(HAVE_QT)
   /**
-   *  @brief Get the image object attached to this item 
+   *  @brief Gets the image object attached to this item
    *
-   *  @return The image object or 0 if no object is attached
+   *  @return The image object or an empty image if no image is attached
    */
-  const QImage *image () const
-  {
-    return mp_image;
-  }
+  QImage image () const;
 
   /**
-   *  @brief Set an image for this item
-   *
-   *  The image object will become owned by the item
+   *  @brief Set the image for this item
    */
-  void set_image (QImage *image);
+  void set_image (const QImage &image);
+#endif
+
+#if defined(HAVE_PNG)
+  /**
+   *  @brief Gets the image as a tl::PixelBuffer object
+   */
+  tl::PixelBuffer image_pixels () const;
 
   /**
-   *  @brief Get an image as a string (base64 coded)
+   *  @brief Sets the image from a tl::PixelBuffer object
+   */
+  void set_image (const tl::PixelBuffer &image);
+#endif
+
+  /**
+   *  @brief Gets a value indicating whether the item has an image
+   */
+  bool has_image () const;
+
+  /**
+   *  @brief Gets the image as a string (PNG, base64 coded)
+   *
+   *  Note: if neither PNG support for Qt are compiled in, this string will be empty
    */
   std::string image_str () const;
 
   /**
-   *  @brief Get an image from a string (base64 coded)
+   *  @brief Gets the image from a string (PNG, base64 coded)
    *
    *  If the image string is empty, the image will be cleared.
    *  If the image string is not valid, the image will also be cleared.
    */
   void set_image_str (const std::string &s);
 
+  /**
+   *  @brief Gets the item comment
+   *
+   *  The comment string is an arbitrary string attached to the item.
+   */
+  const std::string &comment () const
+  {
+    return m_comment;
+  }
+
+  /**
+   *  @brief Sets the item comment
+   *
+   *  The comment string is an arbitrary string attached to the item.
+   */
+  void set_comment (const std::string &s)
+  {
+    m_comment = s;
+  }
+
   /**
    *  @brief Get the database reference
    */
@@ -982,10 +1065,11 @@ private:
   id_type m_cell_id;
   id_type m_category_id;
   size_t m_multiplicity;
+  std::string m_comment;
   bool m_visited;
   std::vector <bool> m_tag_ids;
   Database *mp_database;
-  QImage *mp_image;
+  std::string m_image_str;
 
   Item ();
 
@@ -1406,7 +1490,7 @@ public:
    *
    *  This method is provided for persistency application only. It should not be used otherwise.
    */
-  Cell (id_type id, const std::string &name, const std::string &variant);
+  Cell (id_type id, const std::string &name, const std::string &variant, const std::string &layout_name);
 
   /**
    *  @brief Cell destructor
@@ -1432,7 +1516,7 @@ public:
   }
 
   /**
-   *  @brief Get the cell name
+   *  @brief Gets the cell name
    */
   const std::string &name () const
   {
@@ -1440,7 +1524,7 @@ public:
   }
 
   /**
-   *  @brief Set the name string (setter)
+   *  @brief Sets the name string
    *
    *  This method must not be used for items in the database to keep the database consistent.
    */
@@ -1450,7 +1534,7 @@ public:
   }
 
   /**
-   *  @brief Get the variant id
+   *  @brief Gets the variant id
    */
   const std::string &variant () const
   {
@@ -1458,7 +1542,7 @@ public:
   }
 
   /**
-   *  @brief Set the variant string (setter)
+   *  @brief Sets the variant string
    *
    *  This method must not be used for items in the database to keep the database consistent.
    */
@@ -1467,6 +1551,24 @@ public:
     m_variant = v;
   }
 
+  /**
+   *  @brief Gets the layout cell name
+   */
+  const std::string &layout_name () const
+  {
+    return m_layout_name;
+  }
+
+  /**
+   *  @brief Sets the layout cell string
+   *
+   *  This method must not be used for items in the database to keep the database consistent.
+   */
+  void set_layout_name (const std::string &s)
+  {
+    m_layout_name = s;
+  }
+
   /**
    *  @brief Get the qualified name (name plus optionally the variant id separated by a colon)
    */
@@ -1544,6 +1646,7 @@ private:
   id_type m_id;
   std::string m_name;
   std::string m_variant;
+  std::string m_layout_name;
   size_t m_num_items;
   size_t m_num_items_visited;
   References m_references;
@@ -2048,6 +2151,14 @@ public:
     return *mp_categories;
   }
 
+  /**
+   *  @brief Get the reference to the categories collection (non-const version)
+   */
+  Categories &categories_non_const ()
+  {
+    return *mp_categories;
+  }
+
   /**
    *  @brief Import categories
    *
@@ -2095,6 +2206,20 @@ public:
     return const_cast<Database *> (this)->category_by_id_non_const (id);
   }
 
+  /**
+   *  @brief Get the category pointer for a category name (non-const version)
+   *
+   *  This method returns 0 if the category name is invalid.
+   */
+  Category *category_by_name_non_const (const std::string &name);
+
+  /**
+   *  @brief Get the category pointer for a category id (non-const version)
+   *
+   *  This method returns 0 if the category is invalid.
+   */
+  Category *category_by_id_non_const (id_type id);
+
   /**
    *  @brief Access to the cell collection (const)
    */
@@ -2103,6 +2228,14 @@ public:
     return m_cells;
   }
 
+  /**
+   *  @brief Access to the cell collection
+   */
+  Cells &cells_non_const ()
+  {
+    return m_cells;
+  }
+
   /**
    *  @brief Import cells
    *
@@ -2117,7 +2250,7 @@ public:
    */
   Cell *create_cell (const std::string &name)
   {
-    return create_cell (name, std::string ());
+    return create_cell (name, std::string (), std::string ());
   }
 
   /**
@@ -2126,8 +2259,11 @@ public:
    *  A cell with name name/variant combination must not exist already.
    *  If the variant string is empty, this method behaves the same as the 
    *  method without variant.
+   *
+   *  "layout_name" is the name of the cell in the layout. If empty, the layout
+   *  cell is assumed to be identical to "name".
    */
-  Cell *create_cell (const std::string &name, const std::string &variant);
+  Cell *create_cell (const std::string &name, const std::string &variant, const std::string &layout_name);
 
   /**
    *  @brief Get all variants registered for a given cell name (not qname!)
@@ -2156,6 +2292,20 @@ public:
     return const_cast<Database *> (this)->cell_by_id_non_const (id);
   }
 
+  /**
+   *  @brief Get the cell pointer for a cell name or name:variant combination (non-const version)
+   *
+   *  This method returns 0 if the cell name or name:variant combination is invalid.
+   */
+  Cell *cell_by_qname_non_const (const std::string &qname);
+
+  /**
+   *  @brief Get the cell pointer for a cell id (non-const version)
+   *
+   *  This method returns 0 if the cell id is invalid.
+   */
+  Cell *cell_by_id_non_const (id_type id);
+
   /**
    *  @brief Report the number of items in total
    */
@@ -2207,10 +2357,22 @@ public:
    */
   void remove_item_tag (const Item *item, id_type tag);
 
+  /**
+   *  @brief Sets the comment string of the item
+   */
+  void set_item_comment (const Item *item, const std::string &comment);
+
+#if defined(HAVE_QT)
   /**
    *  @brief Set the image of an item
    */
-  void set_item_image (const Item *item, QImage *image);
+  void set_item_image (const Item *item, const QImage &image);
+#endif
+
+  /**
+   *  @brief Set the image string of an item
+   */
+  void set_item_image_str (const Item *item, const std::string &image_str);
 
   /**
    *  @brief Set the multiplicity of an item
@@ -2225,6 +2387,14 @@ public:
     return *mp_items;
   }
 
+  /**
+   *  @brief Get the items collection (non-const version)
+   */
+  Items &items_non_const ()
+  {
+    return *mp_items;
+  }
+
   /**
    *  @brief Set the items collection
    *
@@ -2238,16 +2408,31 @@ public:
    */
   std::pair<const_item_ref_iterator, const_item_ref_iterator> items_by_cell (id_type cell_id) const; 
 
+  /**
+   *  @brief Get an iterator pair that delivers the non-const items (ItemRef) for a given cell
+   */
+  std::pair<item_ref_iterator, item_ref_iterator> items_by_cell (id_type cell_id);
+
   /**
    *  @brief Get an iterator that delivers the const items (ItemRef) for a given category
    */
   std::pair<const_item_ref_iterator, const_item_ref_iterator> items_by_category (id_type category_id) const; 
 
+  /**
+   *  @brief Get an iterator that delivers the non-const items (ItemRef) for a given category
+   */
+  std::pair<item_ref_iterator, item_ref_iterator> items_by_category (id_type category_id);
+
   /**
    *  @brief Get an iterator that delivers the const items (ItemRef) for a given cell and category
    */
   std::pair<const_item_ref_iterator, const_item_ref_iterator> items_by_cell_and_category (id_type cell_id, id_type category_id) const; 
 
+  /**
+   *  @brief Get an iterator that delivers the non-const items (ItemRef) for a given cell and category
+   */
+  std::pair<item_ref_iterator, item_ref_iterator> items_by_cell_and_category (id_type cell_id, id_type category_id);
+
   /**
    *  @brief Returns true, if the database was modified
    */
@@ -2269,12 +2454,39 @@ public:
    */
   void save (const std::string &filename);
 
+  /**
+   *  @brief Write the database to a file
+   *
+   *  This function is like "save", but does not update the file name attribute.
+   */
+  void write (const std::string &filename);
+
   /**
    *  @brief Load the database from a file
    *
-   *  @brief This clears the existing database.
+   *  Note: This clears the existing database.
+   *  The argument intentionally is a copy, so we can call
+   *  "load (this->filename ())" for reloading.
    */
-  void load (const std::string &filename);
+  void load (std::string filename);
+
+  /**
+   *  @brief Applies the attributes from a different database
+   *
+   *  Attributes are waived flags, images etc.
+   *  The attributes are applied to markers with identical value(s), category and cell context.
+   */
+  void apply (const rdb::Database &other);
+
+  /**
+   *  @brief Scans a layout into this RDB
+   *
+   *  @param layout The layout to scan
+   *  @param cell_index The top cell to scan
+   *  @param layers_and_descriptions The layers and (optional) descriptions/names of the layer to scan
+   *  @param flat True, to perform a flat scan
+   */
+  void scan_layout (const db::Layout &layout, db::cell_index_type cell_index, const std::vector<std::pair<unsigned int, std::string> > &layers_and_descriptions, bool flat);
 
 private:
   std::string m_generator;
@@ -2308,14 +2520,6 @@ private:
     m_modified = true;
   }
 
-  /**
-   *  @brief Get the items collection (non-const version)
-   */
-  Items &items_non_const () 
-  {
-    return *mp_items;
-  }
-
   /**
    *  @brief Get the reference to the tags collection (non-const version)
    */
@@ -2323,96 +2527,8 @@ private:
   {
     return m_tags;
   }
-
-  /**
-   *  @brief Get the reference to the categories collection (non-const version)
-   */
-  Categories &categories_non_const ()
-  {
-    return *mp_categories;
-  }
-
-  /**
-   *  @brief Get the category pointer for a category name
-   *
-   *  This method returns 0 if the category name is invalid.
-   */
-  Category *category_by_name_non_const (const std::string &name);
-
-  /**
-   *  @brief Get the category pointer for a category id
-   *
-   *  This method returns 0 if the category is invalid.
-   */
-  Category *category_by_id_non_const (id_type id);
-
-  /**
-   *  @brief Access to the cell collection
-   */
-  Cells &cells_non_const ()
-  {
-    return m_cells;
-  }
-
-  /**
-   *  @brief Get the cell pointer for a cell name or name:variant combination (non-const version)
-   *
-   *  This method returns 0 if the cell name or name:variant combination is invalid.
-   */
-  Cell *cell_by_qname_non_const (const std::string &qname);
-
-  /**
-   *  @brief Get the cell pointer for a cell id (non-const version)
-   *
-   *  This method returns 0 if the cell id is invalid.
-   */
-  Cell *cell_by_id_non_const (id_type id);
 };
 
 }
 
-namespace tl
-{
-  /**
-   *  @brief Type traits for Cell
-   */
-  template <> struct type_traits<rdb::Cell> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-
-  /**
-   *  @brief Type traits for Category
-   */
-  template <> struct type_traits<rdb::Category> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-
-  /**
-   *  @brief Type traits for Categories
-   */
-  template <> struct type_traits<rdb::Categories> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-
-  /**
-   *  @brief Type traits for Item
-   */
-  template <> struct type_traits<rdb::Item> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-
-  /**
-   *  @brief Type traits for Database
-   */
-  template <> struct type_traits<rdb::Database> : public type_traits<void> {
-    typedef tl::false_tag has_copy_constructor;
-    typedef tl::false_tag has_default_constructor;
-  };
-}
-
 #endif
-
