@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -28,16 +28,16 @@
 namespace tl
 {
 
-TL_PUBLIC void handle_event_exception (tl::Exception &ex)
+void handle_event_exception (tl::Exception &ex)
 {
-  //  Exceptions have unpredicable side effects for events so we just log
+  //  Exceptions have unpredictable side effects for events so we just log
   //  them here.
   tl::warn << ex.msg ();
 }
 
-TL_PUBLIC void handle_event_exception (std::exception &ex)
+void handle_event_exception (std::exception &ex)
 {
-  //  Exceptions have unpredicable side effects for events so we just log
+  //  Exceptions have unpredictable side effects for events so we just log
   //  them here.
   tl::warn << ex.what ();
 }
