@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -22,6 +22,8 @@
 
 #if defined(HAVE_CURL)
 
+#define NOMINMAX   //  for windows.h -> min/max not defined
+
 #include "tlHttpStream.h"
 #include "tlHttpStreamCurl.h"
 #include "tlLog.h"
@@ -31,14 +33,16 @@
 #include "tlAssert.h"
 #include "tlStaticObjects.h"
 #include "tlProgress.h"
-#include "tlDeferredExecution.h"
+#include "tlFileUtils.h"
+#include "tlUri.h"
+#include "tlString.h"
 
-#include <QUrl>
-#include <QFileInfo>
-#include <QCoreApplication>
-
-#include <sys/time.h>
-#include <unistd.h>
+#if !defined(_MSC_VER)
+# include <sys/time.h>
+# include <unistd.h>
+#else
+# include <WinSock2.h>
+#endif
 
 #include <curl/curl.h>
 
@@ -48,6 +52,7 @@
 #include <cassert>
 #include <iostream>
 #include <memory>
+#include <algorithm>
 
 // #define DEBUG_CURL 1
 
@@ -60,8 +65,8 @@ namespace tl
 
 std::string server_from_url (const std::string &url)
 {
-  QUrl qurl (tl::to_qstring (url));
-  return tl::to_string (qurl.scheme ()) + ":" + tl::to_string (qurl.host ());
+  tl::URI uri (url);
+  return uri.scheme () + "://" + uri.authority ();
 }
 
 std::string parse_realm (const std::string &header)
@@ -256,6 +261,7 @@ private:
 
 public:
   ChunkedBuffer ()
+    : m_current_chunk (m_chunks.end ())
   {
     //  .. nothing yet ..
   }
@@ -263,6 +269,7 @@ public:
   void clear ()
   {
     m_chunks.clear ();
+    m_current_chunk = m_chunks.end ();
   }
 
   void push (const char *data, size_t bytes)
@@ -270,6 +277,9 @@ public:
     if (bytes > 0) {
       m_chunks.push_back (ChunkInfo ());
       m_chunks.back ().set (data, bytes);
+      if (m_current_chunk == m_chunks.end ()) {
+        --m_current_chunk;
+      }
     }
   }
 
@@ -277,13 +287,13 @@ public:
   {
     char *start = data;
 
-    while (bytes > 0 && ! m_chunks.empty ()) {
+    while (bytes > 0 && m_current_chunk != m_chunks.end ()) {
 
-      size_t n = m_chunks.front ().fetch (data, bytes);
+      size_t n = m_current_chunk->fetch (data, bytes);
       data += n;
       bytes -= n;
-      if (m_chunks.front ().empty ()) {
-        m_chunks.pop_front ();
+      if (m_current_chunk->empty ()) {
+        ++m_current_chunk;
       }
 
     }
@@ -320,7 +330,38 @@ public:
     return true;
   }
 
+  size_t pos () const
+  {
+    size_t p = 0;
+    for (std::list<ChunkInfo>::const_iterator c = m_chunks.begin (); c != m_current_chunk; ++c) {
+      p += c->size;
+    }
+    if (m_current_chunk != m_chunks.end ()) {
+      p += (m_current_chunk->pos - m_current_chunk->start);
+    }
+    return p;
+  }
+
+  void seek (size_t pos)
+  {
+    for (std::list<ChunkInfo>::iterator c = m_chunks.begin (); c != m_chunks.end (); ++c) {
+      c->pos = c->start;
+    }
+
+    m_current_chunk = m_chunks.end ();
+    for (std::list<ChunkInfo>::iterator c = m_chunks.begin (); c != m_chunks.end (); ++c) {
+      if (pos < c->size) {
+        m_current_chunk = c;
+        c->pos = c->start + pos;
+        break;
+      } else {
+        pos -= c->size;
+      }
+    }
+  }
+
   std::list<ChunkInfo> m_chunks;
+  std::list<ChunkInfo>::iterator m_current_chunk;
 };
 
 // ---------------------------------------------------------------
@@ -334,7 +375,7 @@ public:
  *
  *  @code
  *    CurlNetworkManager mgr;
- *    std::auto_ptr<CurlConnection> conn (mgr.create_connection ());
+ *    std::unique_ptr<CurlConnection> conn (mgr.create_connection ());
  *    conn->set_url ("http://www.example.com");
  *    conn->send ();
  *    while (mgr.tick () > 0) {
@@ -429,7 +470,7 @@ public:
 
   /**
    *  @brief Checks the response once finished is true
-   *  This method will throw an exception if an error occured
+   *  This method will throw an exception if an error occurred
    */
   void check () const;
 
@@ -450,12 +491,14 @@ private:
 
   friend class CurlNetworkManager;
   friend size_t read_func (char *buffer, size_t size, size_t nitems, void *userdata);
+  friend size_t seek_func (void *userdata, curl_off_t offset, int origin);
   friend size_t write_func (char *ptr, size_t size, size_t nmemb, void *userdata);
   friend size_t write_header_func (char *ptr, size_t size, size_t nmemb, void *userdata);
 
   void add_read_data (const char *data, size_t n);
   void add_header_data (const char *data, size_t n);
   size_t fetch_data (char *buffer, size_t nbytes);
+  int seek (curl_off_t offset, int origin);
 
   void finished (int status);
   void init ();
@@ -522,7 +565,12 @@ public:
    *  @brief Must be called in regular intervals to update the status
    *  Returns the number of open connections.
    */
-  int tick ();
+  void tick ();
+
+  /**
+   *  @brief Returns true if a reply has arrived
+   */
+  bool has_reply () const;
 
   /**
    *  @brief The singleton instance
@@ -548,10 +596,139 @@ private:
   static CurlNetworkManager *ms_instance;
 };
 
+// ---------------------------------------------------------------
+//  InputHttpStream implementation
+
+InputHttpStream::InputHttpStream (const std::string &url)
+{
+  mp_data = new InputHttpStreamPrivateData (this, url);
+  mp_callback = 0;
+}
+
+InputHttpStream::~InputHttpStream ()
+{
+  delete mp_data;
+  mp_data = 0;
+}
+
+void
+InputHttpStream::set_credential_provider (HttpCredentialProvider *cp)
+{
+  CurlNetworkManager::instance ()->credentials ().set_provider (cp);
+  CurlNetworkManager::instance ()->proxy_credentials ().set_provider (cp);
+}
+
+void
+InputHttpStream::send ()
+{
+  mp_data->send ();
+}
+
+void
+InputHttpStream::close ()
+{
+  mp_data->close ();
+}
+
+void
+InputHttpStream::set_request (const char *r)
+{
+  mp_data->set_request (r);
+}
+
+void
+InputHttpStream::set_data (const char *data)
+{
+  mp_data->set_data (data);
+}
+
+void
+InputHttpStream::set_data (const char *data, size_t n)
+{
+  mp_data->set_data (data, n);
+}
+
+void
+InputHttpStream::add_header (const std::string &name, const std::string &value)
+{
+  mp_data->add_header (name, value);
+}
+
+tl::Event &
+InputHttpStream::ready ()
+{
+  return mp_data->ready ();
+}
+
+bool
+InputHttpStream::data_available ()
+{
+  return mp_data->data_available ();
+}
+
+size_t
+InputHttpStream::read (char *b, size_t n)
+{
+  return mp_data->read (b, n);
+}
+
+void
+InputHttpStream::reset ()
+{
+  mp_data->reset ();
+}
+
+std::string
+InputHttpStream::source () const
+{
+  return mp_data->source ();
+}
+
+std::string
+InputHttpStream::absolute_path () const
+{
+  return mp_data->absolute_path ();
+}
+
+std::string
+InputHttpStream::filename () const
+{
+  return mp_data->filename ();
+}
+
+bool
+InputHttpStream::is_available ()
+{
+  return true;
+}
+
+void
+InputHttpStream::tick ()
+{
+  if (mp_callback) {
+    mp_callback->wait_for_input ();
+  }
+  CurlNetworkManager::instance ()->tick ();
+}
+
+void
+InputHttpStream::set_timeout (double to)
+{
+  mp_data->set_timeout (to);
+}
+
+double
+InputHttpStream::timeout () const
+{
+  return mp_data->timeout ();
+}
+
+
 // ----------------------------------------------------------------------
 //  CurlConnection implementation
 
 size_t read_func (char *buffer, size_t size, size_t nitems, void *userdata);
+size_t seek_func (void *userdata, curl_off_t offset, int origin);
 size_t write_func (char *ptr, size_t size, size_t nmemb, void *userdata);
 size_t write_header_func (char *ptr, size_t size, size_t nmemb, void *userdata);
 
@@ -666,6 +843,18 @@ size_t CurlConnection::fetch_data (char *buffer, size_t nbytes)
   return m_data.fetch (buffer, nbytes);
 }
 
+int CurlConnection::seek (curl_off_t offset, int origin)
+{
+  if (origin == SEEK_CUR) {
+    m_data.seek (size_t (curl_off_t (m_data.pos ()) + offset));
+  } else if (origin == SEEK_END) {
+    m_data.seek (size_t (curl_off_t (m_data.size ()) + offset));
+  } else {
+    m_data.seek (size_t (offset));
+  }
+  return CURL_SEEKFUNC_OK;
+}
+
 void CurlConnection::send ()
 {
   tl_assert (mp_handle != 0);
@@ -702,6 +891,8 @@ void CurlConnection::send ()
 
   curl_easy_setopt (mp_handle, CURLOPT_READFUNCTION, &read_func);
   curl_easy_setopt (mp_handle, CURLOPT_READDATA, (void *) this);
+  curl_easy_setopt (mp_handle, CURLOPT_SEEKFUNCTION, &seek_func);
+  curl_easy_setopt (mp_handle, CURLOPT_SEEKDATA, (void *) this);
   curl_easy_setopt (mp_handle, CURLOPT_WRITEFUNCTION, &write_func);
   curl_easy_setopt (mp_handle, CURLOPT_WRITEDATA, (void *) this);
   curl_easy_setopt (mp_handle, CURLOPT_HEADERFUNCTION, &write_header_func);
@@ -752,7 +943,7 @@ void CurlConnection::check () const
 
   } else if (m_status != 0) {
 
-    throw tl::HttpErrorException (tl::to_string (QObject::tr ("Connection error (%1)").arg (QString::fromLatin1 (m_error_msg))), m_status, m_url);
+    throw tl::HttpErrorException (tl::sprintf (tl::to_string (tr ("Connection error (%s)")), m_error_msg), m_status, m_url);
 
   } else if (m_http_status < 200 || m_http_status >= 300) {
 
@@ -778,7 +969,7 @@ void CurlConnection::check () const
     if (error_text) {
       throw tl::HttpErrorException (error_text, m_http_status, m_url);
     } else {
-      throw tl::HttpErrorException (tl::to_string (QObject::tr ("HTTP error")), m_http_status, m_url);
+      throw tl::HttpErrorException (tl::to_string (tr ("HTTP error")), m_http_status, m_url);
     }
 
   }
@@ -863,20 +1054,26 @@ void CurlConnection::finished (int status)
   finished_event ();
 }
 
-size_t read_func(char *buffer, size_t size, size_t nitems, void *userdata)
+size_t read_func (char *buffer, size_t size, size_t nitems, void *userdata)
 {
   CurlConnection *connection = (CurlConnection *) userdata;
   return connection->fetch_data (buffer, size * nitems);
 }
 
-size_t write_func(char *buffer, size_t size, size_t nitems, void *userdata)
+size_t seek_func (void *userdata, curl_off_t offset, int origin)
+{
+  CurlConnection *connection = (CurlConnection *) userdata;
+  return connection->seek (offset, origin);
+}
+
+size_t write_func (char *buffer, size_t size, size_t nitems, void *userdata)
 {
   CurlConnection *connection = (CurlConnection *) userdata;
   connection->add_read_data (buffer, size * nitems);
   return size * nitems;
 }
 
-size_t write_header_func(char *buffer, size_t size, size_t nitems, void *userdata)
+size_t write_header_func (char *buffer, size_t size, size_t nitems, void *userdata)
 {
   CurlConnection *connection = (CurlConnection *) userdata;
   connection->add_header_data (buffer, size * nitems);
@@ -968,18 +1165,28 @@ void CurlNetworkManager::release_connection (CurlConnection *connection)
 
 void CurlNetworkManager::on_tick ()
 {
-  if (tick ()) {
-    dm_tick ();
+  tick ();
+  if (! has_reply ()) {
+    //  NOTE: don't reschedule if there is no DM scheduler. This will cause deep
+    //  recursion.
+    if (tl::DeferredMethodScheduler::instance ()) {
+      dm_tick ();
+    }
   }
 }
 
-int CurlNetworkManager::tick ()
+bool CurlNetworkManager::has_reply () const
+{
+  return m_still_running <= 0;
+}
+
+void CurlNetworkManager::tick ()
 {
 #if defined(DEBUG_CURL)
   std::cerr << "CurlNetworkManager::tick()" << std::endl;
 #endif
   if (m_still_running <= 0) {
-    return 0;
+    return;
   }
 
   struct timeval timeout;
@@ -1015,7 +1222,7 @@ int CurlNetworkManager::tick ()
   mc = curl_multi_fdset (mp_multi_handle, &fdread, &fdwrite, &fdexcep, &maxfd);
 
   if (mc != CURLM_OK) {
-    throw tl::HttpErrorException (tl::to_string (QObject::tr ("Connection error (curl_multi_fdset() failed)")), mc, std::string ());
+    throw tl::HttpErrorException (tl::to_string (tr ("Connection error (curl_multi_fdset() failed)")), mc, std::string ());
   }
 
   /* On success the value of maxfd is guaranteed to be >= -1. We call
@@ -1072,68 +1279,79 @@ int CurlNetworkManager::tick ()
     }
 
   }
-
-  return m_still_running;
 }
 
 // ---------------------------------------------------------------
-//  InputHttpFileCurl implementation
+//  InputHttpStreamPrivateData implementation
 
-InputHttpStream::InputHttpStream (const std::string &url)
+InputHttpStreamPrivateData::InputHttpStreamPrivateData (InputHttpStream *stream, const std::string &url)
+  : m_timeout (InputHttpStream::get_default_timeout ()), mp_stream (stream)
 {
   m_sent = false;
   m_ready = false;
 
   m_connection.reset (CurlNetworkManager::instance ()->create_connection ());
   m_connection->set_url (url.c_str ());
-  m_connection->data_available_event.add (this, &InputHttpStream::on_data_available);
-  m_connection->finished_event.add (this, &InputHttpStream::on_finished);
+  m_connection->data_available_event.add (this, &InputHttpStreamPrivateData::on_data_available);
+  m_connection->finished_event.add (this, &InputHttpStreamPrivateData::on_finished);
 }
 
-InputHttpStream::~InputHttpStream ()
+InputHttpStreamPrivateData::~InputHttpStreamPrivateData ()
 {
   // .. nothing yet ..
 }
 
+void
+InputHttpStreamPrivateData::set_timeout (double to)
+{
+  m_timeout = to;
+}
+
+double
+InputHttpStreamPrivateData::timeout () const
+{
+  return m_timeout;
+}
+
 bool
-InputHttpStream::data_available ()
+InputHttpStreamPrivateData::data_available ()
 {
   return m_connection->read_available () > 0;
 }
 
 void
-InputHttpStream::set_request (const char *r)
+InputHttpStreamPrivateData::set_request (const char *r)
 {
   m_connection->set_request (r);
 }
 
 void
-InputHttpStream::set_data (const char *data)
+InputHttpStreamPrivateData::set_data (const char *data)
 {
   m_connection->set_data (data);
 }
 
 void
-InputHttpStream::set_data (const char *data, size_t n)
+InputHttpStreamPrivateData::set_data (const char *data, size_t n)
 {
   m_connection->set_data (data, n);
 }
 
 void
-InputHttpStream::add_header (const std::string &name, const std::string &value)
+InputHttpStreamPrivateData::add_header (const std::string &name, const std::string &value)
 {
   m_connection->add_header (name.c_str (), value.c_str ());
 }
 
 void
-InputHttpStream::on_finished ()
+InputHttpStreamPrivateData::on_finished ()
 {
   m_progress.reset (0);
   m_ready_event ();
 }
 
 void
-InputHttpStream::on_data_available ()
+InputHttpStreamPrivateData::on_data_available ()
 {
   //  send the ready event just once
   if (! m_ready) {
@@ -1143,7 +1361,7 @@ InputHttpStream::on_data_available ()
 }
 
 void
-InputHttpStream::send ()
+InputHttpStreamPrivateData::send ()
 {
   m_ready = false;
   m_progress.reset (0);
@@ -1152,7 +1370,7 @@ InputHttpStream::send ()
 }
 
 void
-InputHttpStream::check ()
+InputHttpStreamPrivateData::check ()
 {
   if (m_connection->finished ()) {
     m_connection->check ();
@@ -1160,7 +1378,7 @@ InputHttpStream::check ()
 }
 
 size_t
-InputHttpStream::read (char *b, size_t n)
+InputHttpStreamPrivateData::read (char *b, size_t n)
 {
   if (! m_sent) {
     send ();
@@ -1172,25 +1390,35 @@ InputHttpStream::read (char *b, size_t n)
     tl::NoDeferredMethods silent;
 
     if (! m_progress.get ()) {
-      m_progress.reset (new tl::AbsoluteProgress (tl::to_string (QObject::tr ("Downloading")) + " " + m_connection->url (), 1));
+      m_progress.reset (new tl::AbsoluteProgress (tl::to_string (tr ("Downloading")) + " " + m_connection->url (), 1));
     }
 
-    while (n > m_connection->read_available () && ! m_connection->finished () && CurlNetworkManager::instance ()->tick ()) {
-      ++*m_progress;
+    tl::Clock start_time = tl::Clock::current ();
+    while (n > m_connection->read_available () && ! m_connection->finished () && ! tl::CurlNetworkManager::instance ()->has_reply ()) {
+
+      //  Check for timeout
+      if (m_timeout > 0.0 && (tl::Clock::current() - start_time).seconds () >= m_timeout) {
+        throw tl::HttpErrorException (tl::sprintf (tl::to_string (tr ("Connection timed out (timeout is %.1fs)")), m_timeout), 0, m_connection->url ());
+      }
+
+      mp_stream->tick ();
+      if (m_progress.get ()) {  //  might have been reset by tick()
+        ++*m_progress;
+      }
     }
   }
 
   if (m_connection->finished ()) {
     m_connection->check ();
   } else if (tl::verbosity() >= 40) {
-    tl::info << "HTTP reponse data read: " << m_connection->read_data_to_string ();
+    tl::info << "HTTP response data read: " << m_connection->read_data_to_string ();
   }
 
   return m_connection->fetch_read_data (b, n);
 }
 
 void
-InputHttpStream::close ()
+InputHttpStreamPrivateData::close ()
 {
   m_progress.reset (0);
   if (m_connection.get ()) {
@@ -1200,36 +1428,30 @@ InputHttpStream::close ()
 }
 
 void
-InputHttpStream::reset ()
+InputHttpStreamPrivateData::reset ()
 {
-  throw tl::Exception (tl::to_string (QObject::tr ("'reset' is not supported on HTTP input streams")));
+  throw tl::Exception (tl::to_string (tr ("'reset' is not supported on HTTP input streams")));
 }
 
 std::string
-InputHttpStream::filename () const
+InputHttpStreamPrivateData::filename () const
 {
-  return tl::to_string (QFileInfo (QUrl (tl::to_qstring (m_connection->url ())).path ()).fileName ());
+  return tl::filename (tl::URI (m_connection->url ()).path ());
 }
 
 std::string
-InputHttpStream::source () const
+InputHttpStreamPrivateData::source () const
 {
   return m_connection->url ();
 }
 
 std::string
-InputHttpStream::absolute_path () const
+InputHttpStreamPrivateData::absolute_path () const
 {
   return m_connection->url ();
 }
 
-void
-InputHttpStream::set_credential_provider (HttpCredentialProvider *cp)
-{
-  CurlNetworkManager::instance ()->credentials ().set_provider (cp);
-  CurlNetworkManager::instance ()->proxy_credentials ().set_provider (cp);
-}
-
 }
 
 #endif
+
