@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -29,6 +29,8 @@
 #include <string>
 #include <vector>
 #include <map>
+#include <set>
+#include <list>
 #include <stdexcept>
 #include <typeinfo>
 
@@ -36,9 +38,11 @@
 #include "tlAssert.h"
 #include "tlObject.h"
 
-#include <QString>
-#include <QByteArray>
-#include <QVariant>
+#if defined(HAVE_QT)
+# include <QString>
+# include <QByteArray>
+# include <QVariant>
+#endif
 
 namespace gsi
 {
@@ -51,8 +55,11 @@ namespace gsi
 namespace tl
 {
 
+class Variant;
 class Extractor;
 class EvalClass;
+template <class T> void extractor_impl (tl::Extractor &, T &);
+template <class T> bool test_extractor_impl (tl::Extractor &, T &);
 
 /**
  *  @brief A base class which describes a class, i.e. an object capable of converting and handling void *
@@ -72,6 +79,9 @@ public:
   virtual bool less (const void *, const void *) const = 0;
   virtual void *clone (const void *) const = 0;
   virtual std::string to_string (const void *) const = 0;
+  virtual int to_int (const void *) const = 0;
+  virtual double to_double (const void *) const = 0;
+  virtual void to_variant (const void *, tl::Variant &var) const = 0;
   virtual void read (void *, tl::Extractor &ex) const = 0;
   virtual const char *name () const = 0;
   virtual bool is_const () const = 0;
@@ -79,6 +89,7 @@ public:
   virtual const gsi::ClassBase *gsi_cls () const = 0;
   virtual const tl::EvalClass *eval_cls () const = 0;
   virtual void *deref_proxy (tl::Object *proxy) const = 0;
+  virtual const tl::VariantUserClassBase *change_constness (bool constness) const = 0;
 
   const void *deref_proxy_const (const tl::Object *proxy) const
   {
@@ -116,6 +127,11 @@ public:
     return VariantUserClassBase::instance (typeid (T), is_const);
   }
 
+  const tl::VariantUserClassBase *change_constness (bool constness) const
+  {
+    return instance (constness);
+  }
+
 private:
   static const tl::VariantUserClassBase *ms_instances[4];
 
@@ -162,8 +178,11 @@ public:
     t_double, 
     t_string, 
     t_stdstring, 
-    t_qstring, 
+    t_bytearray,
+#if defined(HAVE_QT)
+    t_qstring,
     t_qbytearray, 
+#endif
     t_list, 
     t_array, 
     t_user,
@@ -186,6 +205,12 @@ public:
    */
   Variant (const tl::Variant &d);
 
+  /**
+   *  @brief Initialize the Variant with a std::vector<char>
+   */
+  Variant (const std::vector<char> &s);
+
+#if defined(HAVE_QT)
   /**
    *  @brief Initialize the Variant with a QByteArray
    */
@@ -196,6 +221,14 @@ public:
    */
   Variant (const QString &s);
 
+  /**
+   *  @brief Create from a QVariant
+   *
+   *  This constructor will convert a QVariant into a tl::Variant as far as possible.
+   */
+  explicit Variant (const QVariant &v);
+#endif
+
   /**
    *  @brief Initialize the Variant with "string"
    */
@@ -291,13 +324,6 @@ public:
    */
   Variant (size_t l, bool /*dummy*/);
 
-  /**
-   *  @brief Create from a QVariant
-   *
-   *  This constructor will convert a QVariant into a tl::Variant as far as possible.
-   */
-  explicit Variant (const QVariant &v);
-
   /**
    *  @brief Initialize with a user type based on void *
    *
@@ -339,6 +365,116 @@ public:
     m_var.mp_user.cls = c;
   }
 
+  /**
+   *  @brief Initialize with a const user type (will always create a reference and NOT own the object)
+   */
+  template <class T>
+  Variant (const T *obj)
+    : m_type (t_nil), m_string (0)
+  {
+    if (obj) {
+      *this = make_variant_ref (obj);
+    }
+  }
+
+  /**
+   *  @brief Initialize with a non-const user type (will always create a reference and NOT own the object)
+   */
+  template <class T>
+  Variant (T *obj)
+    : m_type (t_nil), m_string (0)
+  {
+    if (obj) {
+      *this = make_variant_ref (obj);
+    }
+  }
+
+  /**
+   *  @brief Initialize the Variant with a STL vector
+   */
+  template <class T>
+  Variant (const std::vector<T> &list)
+    : m_type (t_list), m_string (0)
+  {
+    m_var.m_list = new std::vector<tl::Variant> ();
+    m_var.m_list->reserve (list.size ());
+    for (auto i = list.begin (); i != list.end (); ++i) {
+      m_var.m_list->push_back (tl::Variant (*i));
+    }
+  }
+
+  /**
+   *  @brief Initialize the Variant with a STL list
+   */
+  template <class T>
+  Variant (const std::list<T> &list)
+    : m_type (t_list), m_string (0)
+  {
+    m_var.m_list = new std::vector<tl::Variant> ();
+    m_var.m_list->reserve (list.size ());
+    for (auto i = list.begin (); i != list.end (); ++i) {
+      m_var.m_list->push_back (tl::Variant (*i));
+    }
+  }
+
+  /**
+   *  @brief Initialize the Variant with a STL set (not maintaining the set character)
+   */
+  template <class T>
+  Variant (const std::set<T> &list)
+    : m_type (t_list), m_string (0)
+  {
+    m_var.m_list = new std::vector<tl::Variant> ();
+    m_var.m_list->reserve (list.size ());
+    for (auto i = list.begin (); i != list.end (); ++i) {
+      m_var.m_list->push_back (tl::Variant (*i));
+    }
+  }
+
+  /**
+   *  @brief Initialize the Variant with a STL pair
+   */
+  template <class A, class B>
+  Variant (const std::pair<A, B> &pair)
+    : m_type (t_list), m_string (0)
+  {
+    m_var.m_list = new std::vector<tl::Variant> ();
+    m_var.m_list->reserve (2);
+    m_var.m_list->push_back (tl::Variant (pair.first));
+    m_var.m_list->push_back (tl::Variant (pair.second));
+  }
+
+  /**
+   *  @brief Initialize the Variant with a STL map
+   */
+  template <class K, class V>
+  Variant (const std::map<K, V> &map)
+    : m_type (t_array), m_string (0)
+  {
+    m_var.m_array = new std::map<tl::Variant, tl::Variant> ();
+    for (auto i = map.begin (); i != map.end (); ++i) {
+      m_var.m_array->insert (std::make_pair (tl::Variant (i->first), tl::Variant (i->second)));
+    }
+  }
+
+  /**
+   *  @brief Initialize the Variant with an explicit vector of variants
+   */
+  Variant (const std::vector<tl::Variant> &list)
+    : m_type (t_list), m_string (0)
+  {
+    m_var.m_list = new std::vector<tl::Variant> (list);
+  }
+
+  /**
+   *  @brief Initialize the Variant with an explicit map of variants
+   */
+  Variant (const std::map<tl::Variant, tl::Variant> &map)
+    : m_type (t_array), m_string (0)
+  {
+    m_var.m_array = new std::map<tl::Variant, tl::Variant> (map);
+  }
+
   /**
    *  @brief Initialize the Variant with a list
    */
@@ -398,10 +534,12 @@ public:
     return tl::Variant ((void *) new T(t), c, true);
   }
 
+#if defined(HAVE_QT)
   /**
    *  @brief Convert to a QVariant
    */
   QVariant to_qvariant () const;
+#endif
 
   /**
    *  @brief Assignment
@@ -413,6 +551,7 @@ public:
    */
   Variant &operator= (const char *v);
 
+#if defined(HAVE_QT)
   /**
    *  @brief Assignment of a QByteArray
    */
@@ -422,12 +561,18 @@ public:
    *  @brief Assignment of a QString
    */
   Variant &operator= (const QString &v);
+#endif
 
   /**
    *  @brief Assignment of a string
    */
   Variant &operator= (const std::string &v);
 
+  /**
+   *  @brief Assignment of a STL byte array
+   */
+  Variant &operator= (const std::vector<char> &v);
+
   /**
    *  @brief Assignment of a double
    */
@@ -580,6 +725,22 @@ public:
    */
   bool operator< (const Variant &d) const;
 
+  /**
+   *  @brief An exact compare
+   *
+   *  In constrast to operator==, this method compares the two variants exactly -
+   *  i.e. also the types have to be identical, not only the (normalized type) values.
+   */
+  bool equal (const Variant &d) const;
+
+  /**
+   *  @brief An exact less operator
+   *
+   *  This is the companion for "equal". It is also an exact compare - i.e.
+   *  the type codes are compared as well.
+   */
+  bool less (const Variant &d) const;
+
   /**
    *  @brief Conversion to a string
    *
@@ -588,6 +749,7 @@ public:
    */
   const char *to_string () const;
 
+#if defined(HAVE_QT)
   /**
    *  @brief Conversion to a QByteArray
    *
@@ -603,6 +765,15 @@ public:
    *  No conversion is provided to user types currently.
    */
   QString to_qstring () const;
+#endif
+
+  /**
+   *  @brief Conversion to a STL byte array
+   *
+   *  This performs the conversion to a std::vector<char> as far as possible.
+   *  No conversion is provided to user types currently.
+   */
+  std::vector<char> to_bytearray () const;
 
   /**
    *  @brief Conversion to a std::string
@@ -840,6 +1011,23 @@ public:
    */
   void user_destroy ();
 
+  /**
+   *  @brief Takes the user object and releases it from the variant
+   */
+  void *user_take ();
+
+  /**
+   *  @brief Takes the user object and releases ownership by the variant
+   *  This method is const as it does not change the value, but the ownership of
+   *  the contained object. The object must not be "user_is_ref".
+   */
+  void *user_unshare () const;
+
+  /**
+   *  @brief Changes the constness of the user object
+   */
+  void user_change_constness (bool constness);
+
   /**
    *  @brief Assigns the object stored in other to self
    *
@@ -1305,6 +1493,16 @@ public:
     return m_type == t_id;
   }
 
+  /**
+   *  @brief Test, if it is a std::vector<char> byte array
+   */
+  bool is_bytearray () const
+  {
+    return m_type == t_bytearray;
+  }
+
+#if defined(HAVE_QT)
+
   /**
    *  @brief Test, if it is a QByteArray
    */
@@ -1321,6 +1519,8 @@ public:
     return m_type == t_qstring;
   }
 
+#endif
+
   /**
    *  @brief Test, if it is a std::string
    */
@@ -1342,7 +1542,23 @@ public:
    */
   bool is_a_string () const
   {
-    return m_type == t_string || m_type == t_stdstring || m_type == t_qstring || m_type == t_qbytearray;
+#if defined(HAVE_QT)
+    return m_type == t_string || m_type == t_stdstring || m_type == t_qstring;
+#else
+    return m_type == t_string || m_type == t_stdstring;
+#endif
+  }
+
+  /**
+   *  @brief Test, if it is a byte array
+   */
+  bool is_a_bytearray () const
+  {
+#if defined(HAVE_QT)
+    return m_type == t_bytearray || m_type == t_qbytearray;
+#else
+    return m_type == t_bytearray;
+#endif
   }
 
   /**
@@ -1459,6 +1675,11 @@ public:
    */
   std::string to_parsable_string () const;
 
+  /**
+   *  @brief Gets a hash value for the variant
+   */
+  size_t hash () const;
+
 private:
   type m_type;
 
@@ -1492,8 +1713,11 @@ private:
       char ptr [sizeof (WeakOrSharedPtr)];
       const VariantUserClassBase *cls;
     } mp_user_ref;
+#if defined(HAVE_QT)
     QString *m_qstring;
     QByteArray *m_qbytearray;
+#endif
+    std::vector<char> *m_bytearray;
     std::string *m_stdstring;
   } m_var;
 
@@ -1501,6 +1725,8 @@ private:
   mutable char *m_string;
 
   void set_user_object (void *obj, bool shared);
+  bool equal_core (const Variant &d, type t) const;
+  bool less_core (const Variant &d, type t) const;
 };
 
 //  specializations of the to ... methods
@@ -1522,8 +1748,11 @@ template<> inline __int128 Variant::to<__int128> () const
 template<> inline double Variant::to<double> () const                             { return to_double (); }
 template<> inline float Variant::to<float> () const                               { return to_float (); }
 template<> inline std::string Variant::to<std::string> () const                   { return to_stdstring (); }
+template<> inline std::vector<char> Variant::to<std::vector<char> > () const      { return to_bytearray (); }
+#if defined(HAVE_QT)
 template<> inline QString Variant::to<QString> () const                           { return to_qstring (); }
 template<> inline QByteArray Variant::to<QByteArray> () const                     { return to_qbytearray (); }
+#endif
 template<> inline const char *Variant::to<const char *> () const                  { return to_string (); }
 
 //  specializations if the is.. methods
@@ -1545,8 +1774,11 @@ template<> inline bool Variant::is<__int128> () const             { return m_typ
 template<> inline bool Variant::is<double> () const               { return m_type == t_double; }
 template<> inline bool Variant::is<float> () const                { return m_type == t_float; }
 template<> inline bool Variant::is<std::string> () const          { return m_type == t_stdstring; }
+template<> inline bool Variant::is<std::vector<char> > () const   { return m_type == t_bytearray; }
+#if defined(HAVE_QT)
 template<> inline bool Variant::is<QString> () const              { return m_type == t_qstring; }
 template<> inline bool Variant::is<QByteArray> () const           { return m_type == t_qbytearray; }
+#endif
 template<> inline bool Variant::is<const char *> () const         { return m_type == t_string; }
 
 //  specializations of the can_convert.. methods
@@ -1568,8 +1800,11 @@ template<> inline bool Variant::can_convert_to<__int128> () const             {
 template<> inline bool Variant::can_convert_to<double> () const               { return can_convert_to_double (); }
 template<> inline bool Variant::can_convert_to<float> () const                { return can_convert_to_float (); }
 template<> inline bool Variant::can_convert_to<std::string> () const          { return true; }
+template<> inline bool Variant::can_convert_to<std::vector<char> > () const   { return true; }
+#if defined(HAVE_QT)
 template<> inline bool Variant::can_convert_to<QString> () const              { return true; }
 template<> inline bool Variant::can_convert_to<QByteArray> () const           { return true; }
+#endif
 template<> inline bool Variant::can_convert_to<const char *> () const         { return true; }
 
 /**
@@ -1577,7 +1812,28 @@ template<> inline bool Variant::can_convert_to<const char *> () const         {
  */
 void initialize_variant_class_table ();
 
+/**
+ *  @brief Special extractors for Variant
+ */
+template<> TL_PUBLIC void extractor_impl<Variant> (tl::Extractor &ex, Variant &v);
+template<> TL_PUBLIC bool test_extractor_impl<Variant> (tl::Extractor &ex, Variant &v);
+
 } // namespace tl
 
+namespace std
+{
+  /**
+   *  @brief That hash function for tl::Variant
+   */
+  template <>
+  struct hash <tl::Variant>
+  {
+    size_t operator() (const tl::Variant &o) const
+    {
+      return o.hash ();
+    }
+  };
+}
+
 #endif
 
