@@ -2,7 +2,7 @@
 /*
 
   KLayout Layout Viewer
-  Copyright (C) 2006-2018 Matthias Koefferlein
+  Copyright (C) 2006-2026 Matthias Koefferlein
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -26,10 +26,6 @@
 
 #include "tlCommon.h"
 
-#include <QXmlInputSource>
-#include <QXmlDefaultHandler>
-#include <QXmlLocator>
-
 #include <list>
 #include <vector>
 
@@ -51,7 +47,7 @@ class TL_PUBLIC XMLException : public tl::Exception
 {
 public: 
   XMLException (const char *msg)
-    : Exception (tl::to_string (QObject::tr ("XML parser error: %s")).c_str ()),
+    : Exception (tl::to_string (tr ("XML parser error: %s")).c_str ()),
       m_msg (msg)
   {
     //  .. nothing yet ..
@@ -87,9 +83,9 @@ private:
   static std::string fmt (int line, int /*column*/)
   {
     if (line < 0) {
-      return tl::to_string (QObject::tr ("XML parser error: %s")).c_str ();
+      return tl::to_string (tr ("XML parser error: %s")).c_str ();
     } else {
-      return tl::to_string (QObject::tr ("XML parser error: %s in line %d, column %d")).c_str ();
+      return tl::to_string (tr ("XML parser error: %s in line %d, column %d")).c_str ();
     }
   }
 };
@@ -301,6 +297,9 @@ private:
   std::vector <XMLReaderProxyBase *> m_objects;
 };
 
+//  The opaque source type
+class XMLSourcePrivateData;
+
 /**
  *  @brief A generic XML text source class
  *
@@ -312,17 +311,24 @@ private:
 class TL_PUBLIC XMLSource 
 {
 public:
-  XMLSource ()
+  XMLSource ();
+  ~XMLSource ();
+
+  XMLSourcePrivateData *source ()
   {
-    //  ..  nothing yet ..
-  }
-  
-  virtual ~XMLSource ()
-  {
-    //  ..  nothing yet ..
+    return mp_source;
   }
 
-  virtual QXmlInputSource *source () = 0;
+  void reset ();
+
+protected:
+  void set_source (XMLSourcePrivateData *source)
+  {
+    mp_source = source;
+  }
+
+private:
+  XMLSourcePrivateData *mp_source;
 };
 
 /**
@@ -333,15 +339,12 @@ class TL_PUBLIC XMLStringSource : public XMLSource
 {
 public:
   XMLStringSource (const std::string &string);
-  virtual ~XMLStringSource ();
-
-  virtual QXmlInputSource *source ()  
-  {
-    return mp_source;
-  }
+  XMLStringSource (const char *cp);
+  XMLStringSource (const char *cp, size_t len);
+  ~XMLStringSource ();
 
 private:
-  QXmlInputSource *mp_source;
+  std::string m_copy;
 };
 
 /**
@@ -353,17 +356,7 @@ class TL_PUBLIC XMLFileSource : public XMLSource
 public:
   XMLFileSource (const std::string &path);
   XMLFileSource (const std::string &path, const std::string &progress_message);
-  virtual ~XMLFileSource ();
-
-  virtual QXmlInputSource *source ()  
-  {
-    return mp_source;
-  }
-
-private:
-  QXmlInputSource *mp_source;
-  QIODevice *mp_io;
-  tl::InputStream m_stream;
+  ~XMLFileSource ();
 };
 
 /**
@@ -377,38 +370,10 @@ class TL_PUBLIC XMLStreamSource : public XMLSource
 public:
   XMLStreamSource (tl::InputStream &stream);
   XMLStreamSource (tl::InputStream &stream, const std::string &progress_message);
-  virtual ~XMLStreamSource ();
-
-  virtual QXmlInputSource *source ()  
-  {
-    return mp_source;
-  }
-
-private:
-  QXmlInputSource *mp_source;
-  QIODevice *mp_io;
+  ~XMLStreamSource ();
 };
 
 
-/**
- *  @brief The XML parser class
- *
- *  The class is the basic wrapper around the Qt XML parser.
- *  It is provided mainly for compatibility with the "libparsifal branch".
- */
-
-class TL_PUBLIC XMLParser
-{
-public:
-  XMLParser ();
-  ~XMLParser ();
-
-  void parse (XMLSource &source, QXmlDefaultHandler &handler);
-
-private:
-  QXmlSimpleReader *mp_reader;
-};
-
 // -----------------------------------------------------------------
 //  The C++ structure definition interface (for use cases see tlXMLParser.ut)
 
@@ -429,28 +394,45 @@ struct pass_by_ref_tag {
  *  elements to implement the search algorithm.
  */
 
-class TL_PUBLIC XMLStructureHandler 
-  : public QXmlDefaultHandler
+class TL_PUBLIC XMLStructureHandler
 {
 public:
   XMLStructureHandler (const XMLElementBase *root, XMLReaderState *reader_state);
 
-  bool characters (const QString &ch);
-  bool endElement (const QString &namespaceURI, const QString &localName, const QString &qName);
-  bool startElement (const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
-  bool error (const QXmlParseException &exception);
-  bool fatalError (const QXmlParseException &exception);
-  bool warning (const QXmlParseException &exception);
-
-  void setDocumentLocator (QXmlLocator *locator);
+  void characters (const std::string &ch);
+  void end_element (const std::string &uri, const std::string &lname, const std::string &qname);
+  void start_element (const std::string &uri, const std::string &lname, const std::string &qname);
 
 private:
-  std::vector <const XMLElementBase *> m_stack;  
+  std::vector <const XMLElementBase *> m_stack;
   const XMLElementBase *mp_root;
-  QXmlLocator *mp_locator;
   XMLReaderState *mp_state;
 };
 
+class XMLParserPrivateData;
+
+/**
+ *  @brief The XML parser class
+ *  This is the main entry point. It will take a structure handler and
+ *  parse the given source.
+ */
+class TL_PUBLIC XMLParser
+{
+public:
+  XMLParser ();
+  ~XMLParser ();
+
+  void parse (XMLSource &source, XMLStructureHandler &handler);
+
+  /**
+   *  @brief Returns true, if XML support is compiled in
+   */
+  static bool is_available ();
+
+private:
+  XMLParserPrivateData *mp_data;
+};
+
 /**
  *  @brief XMLElementProxy element of the XML structure definition
  *
@@ -1840,12 +1822,12 @@ make_member (void (Parent::*setter) (const Value &), const std::string &name)
  *  @brief Utility: create a XMLMember object
  */
 template <class Value, class Parent>
-XMLMember<Value, Parent, XMLMemberAccReadAdaptor <Value, Parent>, XMLMemberDummyWriteAdaptor <Value, Parent>, XMLStdConverter <Value> > 
+XMLMember<Value, Parent, XMLMemberDummyReadAdaptor <Value, Parent>, XMLMemberAccWriteAdaptor <Value, Parent>, XMLStdConverter <Value> >
 make_member (void (Parent::*setter) (Value), const std::string &name)
 {
-  return XMLMember<Value, Parent, XMLMemberAccReadAdaptor <Value, Parent>, XMLMemberDummyWriteAdaptor <Value, Parent>, XMLStdConverter <Value> > ( 
-          XMLMemberAccReadAdaptor <Value, Parent> (setter), 
-          XMLMemberDummyWriteAdaptor <Value, Parent> (), name); 
+  return XMLMember<Value, Parent, XMLMemberDummyReadAdaptor <Value, Parent>, XMLMemberAccWriteAdaptor <Value, Parent>, XMLStdConverter <Value> > (
+          XMLMemberDummyReadAdaptor <Value, Parent> (),
+          XMLMemberAccWriteAdaptor <Value, Parent> (setter), name);
 }
 
 /**
