@@ -1,7 +1,7 @@
 # encoding: UTF-8
 
 # KLayout Layout Viewer
-# Copyright (C) 2006-2018 Matthias Koefferlein
+# Copyright (C) 2006-2026 Matthias Koefferlein
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -80,8 +80,12 @@ class DBTrans_TestClass < TestBase
     assert_equal( RBA::Trans::new(RBA::Trans::R180, RBA::DVector::new(5,-7)).to_s, "r180 5,-7" )
     assert_equal( RBA::Trans::new(RBA::Trans::R180).to_s, "r180 0,0" )
 
-    assert_equal( e.trans( RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
-    assert_equal( ( e * RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
+    assert_equal( e.ctrans( 2.0 ), 2.0 )
+    assert_equal( e.ctrans( -2.0 ), -2.0 )
+    assert_equal( e * 2.0, 2.0 )
+    assert_equal( e * -2.0, -2.0 )
+    assert_equal( e.trans( RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( ( e * RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( e.trans( RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( ( e * RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( e.trans( RBA::Text::new("text", RBA::Vector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
@@ -140,7 +144,13 @@ class DBTrans_TestClass < TestBase
 
     assert_equal( mb.trans( RBA::DPoint::new( 1, 0 )).to_s, "17,3" )
     assert_equal( mb.ctrans(2).to_s, "4.0" )
+    assert_equal( mb.ctrans(-2).to_s, "-4.0" )
+    assert_equal( (mb * 2).to_s, "4.0" )
+    assert_equal( (mb * -2).to_s, "-4.0" )
     assert_equal( i.ctrans(2).to_s, "1.0" )
+    assert_equal( i.ctrans(-2).to_s, "-1.0" )
+    assert_equal( (i * 2).to_s, "1.0" )
+    assert_equal( (i * -2).to_s, "-1.0" )
 
   end
 
@@ -155,21 +165,22 @@ class DBTrans_TestClass < TestBase
     assert_equal( c.is_unity?, false )
     assert_equal( c.is_ortho?, true )
     assert_equal( c.is_mag?, false )
+    assert_equal( c.is_complex?, false )
     assert_equal( c.is_mirror?, true )
     assert_equal( c.rot, RBA::DCplxTrans::M135.rot )
     assert_equal( c.s_trans.to_s, "m135 0,0" )
     assert_equal( c.angle, 270 )
 
-    assert_equal( c.trans( RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
-    assert_equal( ( c * RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
-    assert_equal( c.trans( RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
-    assert_equal( ( c * RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
-    assert_equal( c.trans( RBA::Text::new("text", RBA::Vector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
-    assert_equal( ( c * RBA::Text::new("text", RBA::Vector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
-    assert_equal( c.trans( RBA::Polygon::new( [ RBA::Point::new(0, 1), RBA::Point::new(2, -3), RBA::Point::new(4, 5) ] ) ).to_s, "(-5,-4;-1,0;3,-2)" )
-    assert_equal( ( c * RBA::Polygon::new( [ RBA::Point::new(0, 1), RBA::Point::new(2, -3), RBA::Point::new(4, 5) ] ) ).to_s, "(-5,-4;-1,0;3,-2)" )
-    assert_equal( c.trans( RBA::Path::new( [ RBA::Point::new(0, 1), RBA::Point::new(2, 3) ], 10 ) ).to_s, "(-1,0;-3,-2) w=10 bx=0 ex=0 r=false" )
-    assert_equal( ( c * RBA::Path::new( [ RBA::Point::new(0, 1), RBA::Point::new(2, 3) ], 10 ) ).to_s, "(-1,0;-3,-2) w=10 bx=0 ex=0 r=false" )
+    assert_equal( c.trans( RBA::DEdge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( ( c * RBA::DEdge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( c.trans( RBA::DBox::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( ( c * RBA::DBox::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( c.trans( RBA::DText::new("text", RBA::DVector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
+    assert_equal( ( c * RBA::DText::new("text", RBA::DVector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
+    assert_equal( c.trans( RBA::DPolygon::new( [ RBA::DPoint::new(0, 1), RBA::DPoint::new(2, -3), RBA::DPoint::new(4, 5) ] ) ).to_s, "(-5,-4;-1,0;3,-2)" )
+    assert_equal( ( c * RBA::DPolygon::new( [ RBA::DPoint::new(0, 1), RBA::DPoint::new(2, -3), RBA::DPoint::new(4, 5) ] ) ).to_s, "(-5,-4;-1,0;3,-2)" )
+    assert_equal( c.trans( RBA::DPath::new( [ RBA::DPoint::new(0, 1), RBA::DPoint::new(2, 3) ], 10 ) ).to_s, "(-1,0;-3,-2) w=10 bx=0 ex=0 r=false" )
+    assert_equal( ( c * RBA::DPath::new( [ RBA::DPoint::new(0, 1), RBA::DPoint::new(2, 3) ], 10 ) ).to_s, "(-1,0;-3,-2) w=10 bx=0 ex=0 r=false" )
 
     c = RBA::DCplxTrans::from_itrans( RBA::CplxTrans::M135 )
     assert_equal( c.to_s, "m135 *1 0,0" )
@@ -179,6 +190,7 @@ class DBTrans_TestClass < TestBase
     assert_equal( c.is_unity?, false )
     assert_equal( c.is_ortho?, true )
     assert_equal( c.is_mag?, true )
+    assert_equal( c.is_complex?, true )
     assert_equal( c.is_mirror?, false )
     assert_equal( c.rot, RBA::DCplxTrans::R0.rot )
     assert_equal( c.s_trans.to_s, "r0 0,0" )
@@ -190,12 +202,16 @@ class DBTrans_TestClass < TestBase
     assert_equal( c.to_s, "m22.5 *0.75 2.5,-12.5" )
     assert_equal( c.is_unity?, false )
     assert_equal( c.is_ortho?, false )
+    assert_equal( c.is_complex?, true )
     assert_equal( c.is_mag?, true )
     assert_equal( c.rot, RBA::DCplxTrans::M0.rot )
     assert_equal( c.s_trans.to_s, "m0 2.5,-12.5" )
     assert_equal( (c.angle - 45).abs < 1e-10, true )
 
     assert_equal( c.ctrans( 5 ).to_s, "3.75" )
+    assert_equal( c.ctrans( -5 ).to_s, "-3.75" )
+    assert_equal( (c * 5).to_s, "3.75" )
+    assert_equal( (c * -5).to_s, "-3.75" )
     assert_equal( c.trans( RBA::DPoint::new( 12, 16 ) ).to_s, "17.3492424049,-14.6213203436" )
 
     assert_equal( RBA::DCplxTrans::new.to_s, "r0 *1 0,0" )
@@ -262,8 +278,13 @@ class DBTrans_TestClass < TestBase
     c.mirror = true
     assert_equal( c.to_s, "m135 1,7" )
 
-    assert_equal( e.trans( RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
-    assert_equal( ( e * RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-1,0;-3,-2)" )
+    assert_equal( e.ctrans( 2 ), 2 )
+    assert_equal( e.ctrans( -2 ), -2 )
+    assert_equal( e * 2, 2 )
+    assert_equal( e * -2, -2 )
+
+    assert_equal( e.trans( RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
+    assert_equal( ( e * RBA::Edge::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( e.trans( RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( ( e * RBA::Box::new(0, 1, 2, 3) ).to_s, "(-3,-2;-1,0)" )
     assert_equal( e.trans( RBA::Text::new("text", RBA::Vector::new(0, 1)) ).to_s, "('text',m135 -1,0)" )
@@ -303,7 +324,13 @@ class DBTrans_TestClass < TestBase
 
     assert_equal( mb.trans( RBA::Point::new( 1, 0 )).to_s, "17,3" )
     assert_equal( mb.ctrans(2).to_s, "4.0" )
+    assert_equal( mb.ctrans(-2).to_s, "-4.0" )
+    assert_equal( (mb * 2).to_s, "4.0" )
+    assert_equal( (mb * -2).to_s, "-4.0" )
     assert_equal( i.ctrans(2).to_s, "1.0" )
+    assert_equal( i.ctrans(-2).to_s, "-1.0" )
+    assert_equal( (i * 2).to_s, "1.0" )
+    assert_equal( (i * -2).to_s, "-1.0" )
 
   end
 
@@ -350,6 +377,9 @@ class DBTrans_TestClass < TestBase
     assert_equal( (c.angle - 45).abs < 1e-10, true )
 
     assert_equal( c.ctrans( 5 ).to_s, "3.75" )
+    assert_equal( c.ctrans( -5 ).to_s, "-3.75" )
+    assert_equal( (c * 5).to_s, "3.75" )
+    assert_equal( (c * -5).to_s, "-3.75" )
     assert_equal( c.trans( RBA::Point::new( 12, 16 ) ).to_s, "17.3492424049,-14.6213203436" )
 
     assert_equal( RBA::CplxTrans::new.to_s, "r0 *1 0,0" )
@@ -615,6 +645,37 @@ class DBTrans_TestClass < TestBase
 
   end
 
+  # Complex trans conversions, issue #1586
+  def test_6_CplxTransConversions
+
+    itrans = RBA::ICplxTrans::new(1.0, 0.0, false, RBA::Vector::new(1, 2))
+    vtrans = RBA::VCplxTrans::new(1000.0, 0.0, false, RBA::Vector::new(1, 2))
+    dtrans = RBA::DCplxTrans::new(1.0, 0.0, false, RBA::DVector::new(1, 2))
+    ctrans = RBA::CplxTrans::new(0.001, 0.0, false, RBA::DVector::new(1, 2))
+
+    assert_equal(RBA::ICplxTrans::new(vtrans, 0.001).to_s, "r0 *1 1,2")
+    assert_equal(RBA::ICplxTrans::new(dtrans, 0.001).to_s, "r0 *1 1000,2000")
+    assert_equal(RBA::ICplxTrans::new(ctrans, 0.001).to_s, "r0 *1 1000,2000")
+
+    assert_equal(RBA::VCplxTrans::new(itrans, 0.001).to_s, "r0 *1000 1,2")
+    assert_equal(RBA::VCplxTrans::new(dtrans, 0.001).to_s, "r0 *1000 1000,2000")
+    assert_equal(RBA::VCplxTrans::new(ctrans, 0.001).to_s, "r0 *1000 1000,2000")
+
+    assert_equal(RBA::DCplxTrans::new(itrans, 0.001).to_s, "r0 *1 0.001,0.002")
+    assert_equal(RBA::DCplxTrans::new(vtrans, 0.001).to_s, "r0 *1 0.001,0.002")
+    assert_equal(RBA::DCplxTrans::new(ctrans, 0.001).to_s, "r0 *1 1,2")
+
+    assert_equal(RBA::CplxTrans::new(itrans, 0.001).to_s, "r0 *0.001 0.001,0.002")
+    assert_equal(RBA::CplxTrans::new(vtrans, 0.001).to_s, "r0 *0.001 0.001,0.002")
+    assert_equal(RBA::CplxTrans::new(dtrans, 0.001).to_s, "r0 *0.001 1,2")
+
+    # issue #1586 (NOTE: to_itrans is deprecated)
+    t = RBA::DCplxTrans::new(1.0, 45.0, false, 12.345678, 20.000000)
+    assert_equal(t.to_itrans(0.001).to_s, "r45 *1 12345.678,20000")
+    assert_equal(RBA::ICplxTrans::new(t, 0.001).to_s, "r45 *1 12345.678,20000")
+
+  end
+
 end
 
 load("test_epilogue.rb")
