@@ -1,7 +1,7 @@
 # encoding: UTF-8
 
 # KLayout Layout Viewer
-# Copyright (C) 2006-2018 Matthias Koefferlein
+# Copyright (C) 2006-2026 Matthias Koefferlein
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,20 +25,20 @@ load("test_prologue.rb")
 
 class LAYLayers_TestClass < TestBase
 
-  def lnode_str( space, l )
-    return space + l.current.source( true ) + "\n";
+  def lnode_str(space, l)
+    return space + l.current.source(true) + "\n";
   end
 
-  def lnodes_str( space, l )
+  def lnodes_str(space, l)
     res = ""
     while ! l.at_end?  
-      res += lnode_str( space, l )
+      res += lnode_str(space, l)
       if (l.current.has_children?) 
         l.down_first_child
-        res += lnodes_str( "  #{space}", l )
+        res += lnodes_str("  #{space}", l)
         l.up
       end
-      l.next_sibling( 1 )
+      l.next_sibling(1)
     end
     return res
   end
@@ -61,6 +61,10 @@ class LAYLayers_TestClass < TestBase
 
   def test_1
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -148,7 +152,7 @@ class LAYLayers_TestClass < TestBase
     l12_new.source = "@* (0,0 *0.5) (0,5 r45 *2.5)"
     cv.set_layer_properties( p12, l12_new )
     trans = p12.current.trans( true )
-    assert_equal( trans.to_s, p12.current.trans.to_s )
+    assert_equal( trans.map(&:to_s), p12.current.trans.map(&:to_s) )
     assert_equal( trans.size, 2 )
     assert_equal( trans [0].to_s, "r0 *0.5 0,0" )
     assert_equal( trans [1].to_s, "r45 *2.5 0,5" )
@@ -213,6 +217,10 @@ class LAYLayers_TestClass < TestBase
 
   def test_1a
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -226,6 +234,8 @@ class LAYLayers_TestClass < TestBase
 
     cv.insert_layer_list(1)
     cv.rename_layer_list(1, "x")
+    assert_equal(cv.layer_list_name(1), "x")
+    assert_equal(cv.layer_list_name(10000), "")  # must not crash
     assert_equal(cv.current_layer_list, 1)
     cv.set_current_layer_list(0)
     assert_equal(cv.current_layer_list, 0)
@@ -390,7 +400,7 @@ class LAYLayers_TestClass < TestBase
     new_p.clear_children
     assert_equal(new_p.has_children?, false)
 
-    assert_equal(ll.has_children?, true)
+    assert_equal(ll.has_children?, false)
 
     cv.transaction("Delete")
     li = cv.begin_layers(0)
@@ -424,6 +434,10 @@ class LAYLayers_TestClass < TestBase
 
   def test_2
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     p = RBA::LayerPropertiesNode::new
 
     assert_equal( p.source( false ), "*/*@*" )
@@ -655,18 +669,28 @@ class LAYLayers_TestClass < TestBase
     p.clear_line_style
     assert_equal( p.has_line_style?, false )
 
+    assert_equal( p.is_expanded?, false )
+    p.expanded = true
+    assert_equal( p.is_expanded?, true )
+
     pp = RBA::LayerPropertiesNode::new
     assert_equal( pp == p, false )
     assert_equal( pp != p, true )
+    assert_equal( pp.is_expanded?, false )
     pp = p.dup
     assert_equal( pp == p, true )
     assert_equal( pp != p, false )
+    assert_equal( pp.is_expanded?, true )
 
   end
 
   # direct replacement of objects and attributes
   def test_3
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -687,6 +711,44 @@ class LAYLayers_TestClass < TestBase
     assert_equal( lnodes_str( "", cv.begin_layers ), "*/*@*\n" )
     assert_equal( lnodes_str2(cv), "*/*@*" )
 
+    assert_equal( cv.begin_layers.current.name, "" )
+    assert_equal( cv.begin_layers.current.visible?, true )
+    assert_equal( cv.begin_layers.current.dither_pattern, -1 )
+    assert_equal( cv.begin_layers.current.line_style, -1 )
+    assert_equal( cv.begin_layers.current.valid?, true )
+    assert_equal( cv.begin_layers.current.transparent?, false )
+
+    # test LayerPropertiesNodeRef
+    pos.current.name = "NAME"
+    pos.current.visible = false
+    pos.current.fill_color = 0xff012345
+    pos.current.frame_color = 0xff123456
+    pos.current.fill_brightness = 42
+    pos.current.frame_brightness = 17
+    pos.current.dither_pattern = 4
+    pos.current.line_style = 3
+    pos.current.valid = false
+    pos.current.transparent = true
+    pos.current.marked = false
+    pos.current.xfill = false
+    pos.current.width = 2
+    pos.current.animation = 2
+
+    assert_equal( cv.begin_layers.current.name, "NAME" )
+    assert_equal( cv.begin_layers.current.visible?, false )
+    assert_equal( cv.begin_layers.current.fill_color, 0xff012345 )
+    assert_equal( cv.begin_layers.current.frame_color, 0xff123456 )
+    assert_equal( cv.begin_layers.current.fill_brightness, 42 )
+    assert_equal( cv.begin_layers.current.frame_brightness, 17 )
+    assert_equal( cv.begin_layers.current.dither_pattern, 4 )
+    assert_equal( cv.begin_layers.current.line_style, 3 )
+    assert_equal( cv.begin_layers.current.valid?, false )
+    assert_equal( cv.begin_layers.current.transparent?, true )
+    assert_equal( cv.begin_layers.current.marked?, false )
+    assert_equal( cv.begin_layers.current.xfill?, false )
+    assert_equal( cv.begin_layers.current.width, 2 )
+    assert_equal( cv.begin_layers.current.animation, 2 )
+
     new_p = RBA::LayerProperties::new
     new_p.source = "1/0@1"
     assert_equal( new_p.flat.source, "1/0@1" )
@@ -749,6 +811,10 @@ class LAYLayers_TestClass < TestBase
   # propagation of "real" attributes through the hierarchy
   def test_4
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -786,6 +852,10 @@ class LAYLayers_TestClass < TestBase
   # delete method of iterator
   def test_5
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -830,6 +900,10 @@ class LAYLayers_TestClass < TestBase
   # custom stipples and line styles
   def test_6
 
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
     app = RBA::Application.instance
     mw = app.main_window
     mw.close_all
@@ -874,6 +948,135 @@ class LAYLayers_TestClass < TestBase
 
   end
 
+  # dynamic update of tree with LayerPropertiesNodeRef
+  def test_7
+
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
+    app = RBA::Application.instance
+    mw = app.main_window
+    mw.close_all
+
+    mw.load_layout( ENV["TESTSRC"] + "/testdata/gds/t11.gds", 1 ) 
+
+    lv = mw.current_view
+    lv.clear_layers
+
+    lp = RBA::LayerProperties::new
+    lp.source = "A@*"
+     
+    lp = lv.insert_layer(lv.begin_layers, lp)
+     
+    lpp = RBA::LayerProperties::new
+    lpp = lp.add_child(lpp)
+    lpp.source = "A.1"
+     
+    lppp = RBA::LayerProperties::new
+    lppp = lpp.add_child(lppp)
+    lppp.source = "A.1.1@*"
+     
+    lppp = RBA::LayerProperties::new
+    lppp = lpp.add_child(lppp)
+    li = lv.begin_layers # lp
+    li.down_first_child # lpp
+    li.down_first_child # before lppp
+    li.next # lppp
+    li.current.source = "X"
+    assert_equal(lppp.source, "X@1")
+    lppp.source = "A.1.2"
+    assert_equal(li.current.source, "A.1.2@1")
+     
+    lpp = RBA::LayerProperties::new
+    lpp = lp.add_child(lpp)
+    lpp.source = "A.2@*"
+     
+    lppp = RBA::LayerProperties::new
+    lppp = lpp.add_child(lppp)
+    lppp.source = "A.2.1"
+    lppp_saved = lppp
+
+    lppp = RBA::LayerProperties::new
+    lppp = lpp.add_child(lppp)
+    lppp.source = "A.2.2@*"
+
+    assert_equal(lnodes_str("", lv.begin_layers), 
+      "A@*\n" +
+      "  A.1@1\n" +
+      "    A.1.1@1\n" +
+      "    A.1.2@1\n" +
+      "  A.2@*\n" +
+      "    A.2.1@1\n" +
+      "    A.2.2@*\n"
+    )
+
+    lp.source = "B@2"
+
+    assert_equal(lnodes_str("", lv.begin_layers), 
+      "B@2\n" +
+      "  A.1@1\n" +
+      "    A.1.1@1\n" +
+      "    A.1.2@1\n" +
+      "  A.2@2\n" +
+      "    A.2.1@1\n" +
+      "    A.2.2@2\n"
+    )
+
+    lppp_saved.delete
+
+    assert_equal(lnodes_str("", lv.begin_layers), 
+      "B@2\n" +
+      "  A.1@1\n" +
+      "    A.1.1@1\n" +
+      "    A.1.2@1\n" +
+      "  A.2@2\n" +
+      "    A.2.2@2\n"
+    )
+
+    assert_equal(lppp.source, "A.2.2@*")
+
+  end
+
+  # clear_layers with index and layer list with name
+  def test_8
+
+    if !RBA.constants.member?(:Application)
+      return
+    end
+
+    app = RBA::Application.instance
+    mw = app.main_window
+    mw.close_all
+
+    mw.load_layout( ENV["TESTSRC"] + "/testdata/gds/t11.gds", 1 ) 
+
+    cv = mw.current_view
+
+    cv.clear_layers
+    assert_equal(lnodes_str("", cv.begin_layers(0)), "")
+
+    cv.rename_layer_list(0, "x")
+    assert_equal(cv.layer_list_name(0), "x")
+
+    cv.insert_layer(0, cv.end_layers(0), RBA::LayerProperties::new)
+    assert_equal(lnodes_str("", cv.begin_layers(0)), "*/*@*\n")
+
+    cv.clear_layers(0)
+    assert_equal(lnodes_str("", cv.begin_layers(0)), "")
+    assert_equal(cv.layer_list_name(0), "x")
+
+    cv.rename_layer_list(cv.current_layer_list, "y")
+    assert_equal(cv.layer_list_name(0), "y")
+    cv.insert_layer(cv.end_layers, RBA::LayerProperties::new)
+    assert_equal(lnodes_str("", cv.begin_layers), "*/*@*\n")
+
+    cv.clear_layers
+    assert_equal(lnodes_str("", cv.begin_layers), "")
+    assert_equal(cv.layer_list_name(cv.current_layer_list), "y")
+
+  end
+
 end
 
 load("test_epilogue.rb")
